// Written by Christopher E. Miller
// See the included license.txt for copyright and license details.


module dfl.application;

private import std.path, std.file, std.thread, std.c.stdlib;
private import std.string, std.c.stdlib;

private import dfl.base, dfl.form, dfl.winapi, dfl.event;
private import dfl.control, dfl.menu, dfl.drawing, dfl.label;
private import dfl.button, dfl.textbox, dfl.wincom, dfl.environment;
private import dfl.utf;


//debug = APP_PRINT;
//debug = SHOW_MESSAGE_INFO; // Slow.

debug(APP_PRINT)
{
	pragma(msg, "DFL: debug app print.");
	
	version(DFL_LIB)
		static assert(0);
}


private extern(C) void abort();


class ApplicationContext // docmain
{
	this()
	{
	}
	
	
	// If onMainFormClose isn't overridden, the message
	// loop terminates when the main form is destroyed.
	this(Form mainForm)
	{
		mform = mainForm;
		mainForm.closed ~= &onMainFormClosed;
	}
	
	
	final void mainForm(Form mainForm) // setter
	{
		if(mform)
			mform.closed.removeHandler(&onMainFormClosed);
		
		mform = mainForm;
		
		if(mainForm)
			mainForm.closed ~= &onMainFormClosed;
	}
	
	
	final Form mainForm() // getter
	{
		return mform;
	}
	
	
	EventHandler threadExit;
	
	
	final void exitThread()
	{
		exitThreadCore();
	}
	
	
	protected:
	void exitThreadCore()
	{
		threadExit(this, EventArgs.empty);
		//ExitThread(0);
	}
	
	
	void onMainFormClosed(Object sender, EventArgs args)
	{
		exitThreadCore();
	}
	
	
	private:
	Form mform; // The context form.
}


private extern(Windows)
{
	alias UINT function(LPCWSTR lpPathName, LPCWSTR lpPrefixString, UINT uUnique,
		LPWSTR lpTempFileName) GetTempFileNameWProc;
	alias DWORD function(DWORD nBufferLength, LPWSTR lpBuffer) GetTempPathWProc;
	alias HANDLE function(PACTCTXW pActCtx) CreateActCtxWProc;
	alias BOOL function(HANDLE hActCtx, ULONG** lpCookie) ActivateActCtxProc;
}


class Application // docmain
{
	private this() {}
	
	
	static:
	// Should be called before creating any controls.
	// This is typically the first function called in main().
	// Does nothing if not supported.
	void enableVisualStyles()
	{
		const char[] MANIFEST = `<?xml version="1.0" encoding="UTF-8" standalone="yes"?>` \r\n
			`<assembly xmlns="urn:schemas-microsoft-com:asm.v1" manifestVersion="1.0">` \r\n
				`<dependency>` \r\n
					`<dependentAssembly>` \r\n
						`<assemblyIdentity` \r\n
							`type="win32"` \r\n
							`name="Microsoft.Windows.Common-Controls"` \r\n
							`version="6.0.0.0"` \r\n
							`processorArchitecture="X86"` \r\n
							`publicKeyToken="6595b64144ccf1df"` \r\n
							`language="*"` \r\n
						`/>` \r\n
					`</dependentAssembly>` \r\n
				`</dependency>` \r\n
			`</assembly>` \r\n;
		
		HMODULE kernel32;
		kernel32 = GetModuleHandleA("kernel32.dll");
		//if(kernel32)
		assert(kernel32 !is null);
		{
			CreateActCtxWProc createActCtxW;
			createActCtxW = cast(CreateActCtxWProc)GetProcAddress(kernel32, "CreateActCtxW");
			if(createActCtxW)
			{
				GetTempPathWProc getTempPathW;
				GetTempFileNameWProc getTempFileNameW;
				ActivateActCtxProc activateActCtx;
				
				getTempPathW = cast(GetTempPathWProc)GetProcAddress(kernel32, "GetTempPathW");
				assert(getTempPathW !is null);
				getTempFileNameW = cast(GetTempFileNameWProc)GetProcAddress(kernel32, "GetTempFileNameW");
				assert(getTempFileNameW !is null);
				activateActCtx = cast(ActivateActCtxProc)GetProcAddress(kernel32, "ActivateActCtx");
				assert(activateActCtx !is null);
				
				DWORD pathlen = MAX_PATH;
				wchar* pathbuf = cast(wchar*)alloca(pathlen * wchar.sizeof);
				if(pathbuf)
				{
					pathlen = getTempPathW(pathlen, pathbuf);
					if(pathlen)
					{
						DWORD manifestlen = MAX_PATH;
						wchar* manifestbuf = cast(wchar*)alloca(manifestlen * wchar.sizeof);
						if(manifestbuf)
						{
							manifestlen = getTempFileNameW(pathbuf, "dmf", 0, manifestbuf);
							if(manifestlen)
							{
								HANDLE hf;
								hf = CreateFileW(manifestbuf, GENERIC_WRITE, 0, null, CREATE_ALWAYS, FILE_ATTRIBUTE_NORMAL | FILE_FLAG_SEQUENTIAL_SCAN, null);
								if(hf)
								{
									DWORD written;
									if(WriteFile(hf, MANIFEST, MANIFEST.length, &written, null))
									{
										CloseHandle(hf);
										
										ACTCTXW ac;
										HANDLE hac;
										
										ac.cbSize = ACTCTXW.sizeof;
										ac.dwFlags = 0;
										ac.lpSource = manifestbuf;
										
										hac = createActCtxW(&ac);
										if(hac != INVALID_HANDLE_VALUE)
										{
											ULONG* ulp;
											activateActCtx(hac, &ulp);
											
											_initCommonControls(ICC_STANDARD_CLASSES);
										}
										else
										{
											debug(APP_PRINT)
												printf("CreateActCtxW failed.\n");
										}
									}
									else
									{
										CloseHandle(hf);
									}
								}
								
								DeleteFileW(manifestbuf);
							}
						}
					}
				}
			}
		}
	}
	
	
	char[] executablePath() // getter
	{
		char[] pathbuf = new char[MAX_PATH];
		DWORD len;
		
		len = GetModuleFileNameA(null, pathbuf, pathbuf.length);
		if(!len)
			throw new DflException("Unable to obtain executable path.");
		return pathbuf[0 .. len];
	}
	
	
	bool messageLoop() // getter
	{
		return running != 0;
	}
	
	
	char[] startupPath() // getter
	{
		char[] pathbuf = new char[MAX_PATH];
		DWORD len;
		
		len = GetModuleFileNameA(null, pathbuf, pathbuf.length);
		if(!len)
			throw new DflException("Unable to obtain startup path.");
		return getDirName(pathbuf[0 .. len]);
	}
	
	
	void addMessageFilter(IMessageFilter mf)
	{
		//filters ~= mf;
		
		IMessageFilter[] fs = filters;
		fs ~= mf;
		filters = fs;
	}
	
	
	void removeMessageFilter(IMessageFilter mf)
	{
		uint i;
		for(i = 0; i != filters.length; i++)
		{
			if(mf is filters[i])
			{
				if(!i)
					filters = filters[1 .. filters.length];
				else if(i == filters.length - 1)
					filters = filters[0 .. i];
				else
					filters = filters[0 .. i] ~ filters[i + 1 .. filters.length];
				break;
			}
		}
	}
	
	
	// Process all messages in the queue.
	void doEvents()
	{
		try
		{
			Message msg;
			
			while(PeekMessageA(&msg._winMsg, null, 0, 0, PM_REMOVE))
			{
				gotMessage(msg);
				
				if(msg.msg == WM_QUIT)
				{
					running = false;
					return;
				}
			}
			
			// Execution continues after this so it's not idle.
		}
		catch(Object e)
		{
			onThreadException(e);
		}
	}
	
	
	void run()
	{
		run(new ApplicationContext);
	}
	
	
	void run(void delegate() whileIdle)
	{
		run(new ApplicationContext, whileIdle);
	}
	
	
	void run(ApplicationContext appcon)
	{
		void whileIdle()
		{
			waitMsg();
		}
		
		
		run(appcon, &whileIdle);
	}
	
	
	// -whileIdle- is called repeatedly while there are no messages in the queue.
	// Application.idle events are suppressed; however, the -whileIdle- handler
	// may manually fire the Application.idle event.
	void run(ApplicationContext appcon, void delegate() whileIdle)
	{
		if(running)
		{
			//throw new DflException("Cannot have more than one message loop per thread.");
			assert(0); // Cannot have more than one message loop per thread.
			return;
		}
		running = true;
		
		
		void threadJustExited(Object sender, EventArgs ea)
		{
			exitThread();
		}
		
		
		appcon.threadExit ~= &threadJustExited;
		
		if(appcon.mainForm)
			appcon.mainForm.createControl();
		
		version(CUSTOM_MSG_HOOK)
		{
			HHOOK _msghook = SetWindowsHookExA(WH_CALLWNDPROCRET, &globalMsgHook, null, GetCurrentThreadId());
			if(!_msghook)
				throw new DflException("Unable to get window messages.");
			msghook = _msghook;
		}
		
		try
		{
			for(;;)
			{
				try
				{
					still_running:
					while(running)
					{
						Message msg;
						
						while(PeekMessageA(&msg._winMsg, null, 0, 0, PM_REMOVE))
						{
							gotMessage(msg);
							
							if(!running)
								break still_running;
							
							if(msg.msg == WM_QUIT)
							{
								running = false;
								break still_running;
							}
						}
						
						whileIdle();
					}
					
					// Stopped running.
					threadExit(Thread.getThis(), EventArgs.empty);
					assert(!running);
					return;
				}
				catch(Object e)
				{
					onThreadException(e);
				}
			}
		}
		finally
		{
			version(CUSTOM_MSG_HOOK)
				UnhookWindowsHookEx(msghook);
			appcon.threadExit.removeHandler(&threadJustExited);
		}
	}
	
	
	// Makes the form -mainForm- visible.
	void run(Form mainForm, void delegate() whileIdle)
	{
		ApplicationContext appcon = new ApplicationContext(mainForm);
		//mainForm.show(); // Interferes with -running-.
		run(appcon, whileIdle);
	}
	
	
	void run(Form mainForm)
	{
		ApplicationContext appcon = new ApplicationContext(mainForm);
		//mainForm.show(); // Interferes with -running-.
		run(appcon);
	}
	
	
	void exit()
	{
		PostQuitMessage(0);
	}
	
	
	// Actually only stops the current run() loop.
	void exitThread()
	{
		running = false;
	}
	
	
	// ApartmentState oleRequired() ...
	
	
	private static class ErrForm: Form
	{
		protected override void onLoad(EventArgs ea)
		{
			okBtn.focus();
		}
		
		
		protected override void onClosing(CancelEventArgs cea)
		{
			cea.cancel = !errdone;
		}
		
		
		const int PADDING = 10;
		
		
		void onOkClick(Object sender, EventArgs ea)
		{
			errdone = true;
			ctnu = true;
			close();
		}
		
		
		void onCancelClick(Object sender, EventArgs ea)
		{
			errdone = true;
			ctnu = false;
			close();
		}
		
		
		this(char[] errmsg)
		{
			text = "Error";
			clientSize = Size(340, 150);
			startPosition = FormStartPosition.CENTER_SCREEN;
			formBorderStyle = FormBorderStyle.FIXED_DIALOG;
			minimizeBox = false;
			maximizeBox = false;
			controlBox = false;
			
			with(new Label)
			{
				bounds = Rect(PADDING, PADDING, this.clientSize.width - PADDING * 2, 40);
				text = "An application exception has occured. Click Continue to allow "
					"the application to ignore this error and attempt to continue.";
				parent = this;
			}
			
			
			with(errBox = new TextBox)
			{
				text = errmsg;
				bounds = Rect(PADDING, 40 + PADDING, this.clientSize.width - PADDING * 2, 50);
				backColor = this.backColor;
				readOnly = true;
				multiline = true;
				parent = this;
			}
			
			with(okBtn = new Button)
			{
				width = 100;
				location = Point(this.clientSize.width - width - PADDING - width - PADDING,
					this.clientSize.height - height - PADDING);
				text = "&Continue";
				parent = this;
				click ~= &onOkClick;
			}
			acceptButton = okBtn;
			
			with(new Button)
			{
				width = 100;
				location = Point(this.clientSize.width - width - PADDING,
					this.clientSize.height - height - PADDING);
				text = "&Quit";
				parent = this;
				click ~= &onCancelClick;
			}
		}
		
		
		private int inThread2()
		{
			try
			{
				// Create in this thread so that it owns the handle.
				assert(!isHandleCreated);
				show();
				SetForegroundWindow(handle);
				
				MSG msg;
				assert(isHandleCreated);
				while(GetMessageA(&msg, null, 0, 0))
				{
					if(!IsDialogMessageA(handle, &msg))
					{
						TranslateMessage(&msg);
						DispatchMessageA(&msg);
					}
					
					if(!created)
						break;
				}
			}
			finally
			{
				dispose();
				assert(!isHandleCreated);
				
				thread1.resume();
			}
			
			return 0;
		}
		
		
		private Thread thread1;
		
		bit doContinue()
		{
			assert(!isHandleCreated);
			
			// Need to use a separate thread so that all the main thread's messages
			// will be there still when the exception is recovered from.
			// This is very important for some messages, such as socket events.
			thread1 = Thread.getThis();
			Thread thd;
			thd = new Thread(&inThread2);
			thd.start();
			SuspendThread(GetCurrentThread()); // Thread2 will resume me.
			
			return ctnu;
		}
		
		
		override char[] toString()
		{
			return errBox.text;
		}
		
		
		private:
		bit errdone = false;
		bit ctnu = false;
		Button okBtn;
		TextBox errBox;
	}
	
	
	void onThreadException(Object e)
	{
		static bit except = false;
		
		if(except)
		{
			printf("Error: ");
			e.print();
			
			abort();
			return;
		}
		
		except = true;
		if(threadException.handlers.length)
		{
			threadException(Thread.getThis(), new ThreadExceptionEventArgs(e));
			except = false;
			return;
		}
		else
		{
			// No thread exception handlers, display a dialog.
			
			/+
			if(IDYES == MessageBoxA(null,
				"An application exception has occured. Click Yes to allow\r\n"
				"the application to ignore this error and attempt to continue.\r\n"
				"Click No to quit the application.\r\n\r\n"~
				e.toString(),
				null, MB_ICONWARNING | MB_TASKMODAL | MB_YESNO))
			{
				except = false;
				return;
			}
			+/
			
			//try
			{
				if((new ErrForm(e.toString())).doContinue())
				{
					except = false;
					return;
				}
			}
			/+
			catch
			{
				MessageBoxA(null, "Error displaying error message.", "DFL", MB_ICONERROR | MB_TASKMODAL);
			}
			+/
		}
		//except = false;
		
		//throw e;
		printf("Error: ");
		e.print();
		//exitThread();
		Environment.exit(35); // ?
	}
	
	
	// EventHandler applicationExit; // Just use a static destructor.
	EventHandler idle; // Finished processing and is now idle.
	ThreadExceptionEventHandler threadException;
	EventHandler threadExit;
	
	
	// Returns null if not found.
	package Control lookupHwnd(HWND hwnd)
	{
		if(hwnd in controls)
			return controls[hwnd];
		return null;
	}
	
	
	package void removeHwnd(HWND hwnd)
	{
		//delete controls[hwnd];
		controls.remove(hwnd);
	}
	
	
	// Returns its new unique menu ID.
	package int addMenuItem(MenuItem menu)
	{
		if(nmenus == LAST_MENU_ID - FIRST_MENU_ID)
			throw new DflException("Out of menus.");
		
		int result;
		typeof(menus) tempmenus;
		
		result = nmenus;
		nmenus++;
		tempmenus = cast(Menu*)realloc(menus, Menu.sizeof * nmenus);
		if(!tempmenus)
		{
			nmenus--;
			//throw new OutOfMemory;
			throw new DflException("Out of memory.");
		}
		menus = tempmenus;
		
		menus[result] = menu;
		
		return result + FIRST_MENU_ID;
	}
	
	
	package void addContextMenu(ContextMenu menu)
	{
		if(nmenus == LAST_MENU_ID - FIRST_MENU_ID)
			throw new DflException("Out of menus.");
		
		typeof(menus) tempmenus;
		int idx;
		
		idx = nmenus;
		nmenus++;
		tempmenus = cast(Menu*)realloc(menus, Menu.sizeof * nmenus);
		if(!tempmenus)
		{
			nmenus--;
			//throw new OutOfMemory;
			throw new DflException("Out of memory.");
		}
		menus = tempmenus;
		
		menus[idx] = menu;
	}
	
	
	package void removeMenu(Menu menu)
	{
		uint idx;
		
		for(idx = 0; idx != nmenus; idx++)
		{
			if(menus[idx] is menu)
			{
				goto found;
			}
		}
		return;
		
		found:
		if(nmenus == 1)
		{
			free(menus);
			menus = null;
			nmenus--;
		}
		else
		{
			if(idx != nmenus - 1)
				menus[idx] = menus[nmenus - 1]; // Move last one in its place
			
			nmenus--;
			menus = cast(Menu*)realloc(menus, Menu.sizeof * nmenus);
			assert(menus != null); // Memory shrink shouldn't be a problem.
		}
	}
	
	
	package MenuItem lookupMenuID(int menuID)
	{
		uint idx;
		MenuItem mi;
		
		for(idx = 0; idx != nmenus; idx++)
		{
			mi = cast(MenuItem)menus[idx];
			if(mi && mi._menuID == menuID)
				return mi;
		}
		return null;
	}
	
	
	package Menu lookupMenu(HMENU hmenu)
	{
		uint idx;
		
		for(idx = 0; idx != nmenus; idx++)
		{
			if(menus[idx].handle == hmenu)
				return menus[idx];
		}
		return null;
	}
	
	
	package void creatingControl(Control ctrl)
	{
		TlsSetValue(tlsControl, cast(Control*)ctrl);
	}
	
	
	package void waitMsg()
	{
		idle(Thread.getThis(), EventArgs.empty);
		WaitMessage();
	}
	
	
	// Because waiting for an event enters an idle state,
	// this function fires the -idle- event.
	void waitForEvent()
	{
		waitMsg();
	}
	
	
	package void refCountInc(void* p)
	{
		if(p in _refs)
			_refs[p] = _refs[p] + 1;
		else
			_refs[p] = 1;
	}
	
	
	// Returns the new ref count.
	package size_t refCountDec(void* p)
	{
		if(p in _refs)
		{
			size_t count;
			count = _refs[p];
			if(!--count)
			//	delete _refs[p];
				_refs.remove(p);
			else
				_refs[p] = count;
			return count;
		}
		return 0;
	}
	
	
	package void ppin(void* p)
	{
		//std.gc.pin(p);
	}
	
	
	package void punpin(void* p)
	{
		//std.gc.unpin(p);
	}
	
	
	private:
	static:
	size_t[void*] _refs;
	IMessageFilter[] filters;
	DWORD tlsThreadFlags;
	DWORD tlsControl;
	DWORD tlsFilter; // IMessageFilter[]*.
	version(CUSTOM_MSG_HOOK)
		DWORD tlsHook; // HHOOK.
	Control[HWND] controls;
	ushort lastMenuID = 200;
	
	// malloc() is needed so the menus can be garbage collected.
	uint nmenus = 0; // Number of -menus-.
	Menu* menus = null; // WARNING: malloc()'d memory!
	
	// Menus.
	const ushort FIRST_MENU_ID = 200;
	const ushort LAST_MENU_ID = 10000;
	
	// Controls.
	const ushort FIRST_CTRL_ID = LAST_MENU_ID + 1;
	const ushort LAST_CTRL_ID = 65500;
	
	
	// Destroy all menu handles at program exit because Windows will not
	// unless it is assigned to a window.
	private void sdtorFreeAllMenus()
	{
		foreach(Menu m; menus[0 .. nmenus])
		{
			DestroyMenu(m.handle);
		}
		nmenus = 0;
		free(menus);
		menus = null;
	}
	
	
	private struct TlsFilterValue
	{
		IMessageFilter[] filters;
	}
	
	
	/+
	void filters(IMessageFilter[] filters) // setter
	{
		// The TlsFilterValue is being garbage collected!
		
		TlsFilterValue* val = cast(TlsFilterValue*)TlsGetValue(tlsFilter);
		if(!val)
			val = new TlsFilterValue;
		val.filters = filters;
		TlsSetValue(tlsFilter, cast(LPVOID)val);
	}
	
	
	IMessageFilter[] filters() // getter
	{
		TlsFilterValue* val = cast(TlsFilterValue*)TlsGetValue(tlsFilter);
		if(!val)
			return null;
		return val.filters;
	}
	+/
	
	
	version(CUSTOM_MSG_HOOK)
	{
		void msghook(HHOOK hhook) // setter
		{
			TlsSetValue(tlsHook, cast(LPVOID)hhook);
		}
		
		
		HHOOK msghook() // getter
		{
			return cast(HHOOK)TlsGetValue(tlsHook);
		}
	}
	
	
	Control getCreatingControl()
	{
		return cast(Control)cast(Control*)TlsGetValue(tlsControl);
	}
	
	
	// Thread flags.
	enum TF: DWORD
	{
		RUNNING = 1,
	}
	
	
	TF threadFlags() // getter
	{
		return cast(TF)cast(DWORD)TlsGetValue(tlsThreadFlags);
	}
	
	
	void threadFlags(TF flags) // setter
	{
		if(!TlsSetValue(tlsThreadFlags, cast(LPVOID)cast(DWORD)flags))
			assert(0);
	}
	
	
	DWORD running() // getter
	{
		return threadFlags & TF.RUNNING;
	}
	
	
	void running(DWORD byes) // setter
	{
		if(byes)
			threadFlags = threadFlags | TF.RUNNING;
		else
			threadFlags = threadFlags & ~TF.RUNNING;
	}
	
	
	void gotMessage(inout Message msg)
	{
		// Don't bother with this extra stuff if there aren't any filters.
		if(filters.length)
		{
			try
			{
				// Keep a local reference so that handlers
				// may be added and removed during filtering.
				IMessageFilter[] local = filters;
				
				foreach(IMessageFilter mf; local)
				{
					// Returning true prevents dispatching.
					if(mf.preFilterMessage(msg))
					{
						Control ctrl;
						ctrl = lookupHwnd(msg.hWnd);
						if(ctrl)
							ctrl.mustWndProc(msg);
						return;
					}
				}
			}
			catch(Object o)
			{
				Control ctrl;
				ctrl = lookupHwnd(msg.hWnd);
				if(ctrl)
					ctrl.mustWndProc(msg);
				throw o;
			}
		}
		
		TranslateMessage(&msg._winMsg);
		DispatchMessageA(&msg._winMsg);
	}
}


package:


debug(SHOW_MESSAGE_INFO)
{
	private import std.stdio, std.string;
	
	
	void showMessageInfo(inout Message m)
	{
		void writeWm(char[] wmName)
		{
			writef("Message %s=%d(0x%X)\r\n", wmName, m.msg, m.msg);
		}
		
		
		switch(m.msg)
		{
			case WM_NULL: writeWm("WM_NULL"); break;
			case WM_CREATE: writeWm("WM_CREATE"); break;
			case WM_DESTROY: writeWm("WM_DESTROY"); break;
			case WM_MOVE: writeWm("WM_MOVE"); break;
			case WM_SIZE: writeWm("WM_SIZE"); break;
			case WM_ACTIVATE: writeWm("WM_ACTIVATE"); break;
			case WM_SETFOCUS: writeWm("WM_SETFOCUS"); break;
			case WM_KILLFOCUS: writeWm("WM_KILLFOCUS"); break;
			case WM_ENABLE: writeWm("WM_ENABLE"); break;
			case WM_SETREDRAW: writeWm("WM_SETREDRAW"); break;
			case WM_SETTEXT: writeWm("WM_SETTEXT"); break;
			case WM_GETTEXT: writeWm("WM_GETTEXT"); break;
			case WM_GETTEXTLENGTH: writeWm("WM_GETTEXTLENGTH"); break;
			case WM_PAINT: writeWm("WM_PAINT"); break;
			case WM_CLOSE: writeWm("WM_CLOSE"); break;
			case WM_QUERYENDSESSION: writeWm("WM_QUERYENDSESSION"); break;
			case WM_QUIT: writeWm("WM_QUIT"); break;
			case WM_QUERYOPEN: writeWm("WM_QUERYOPEN"); break;
			case WM_ERASEBKGND: writeWm("WM_ERASEBKGND"); break;
			case WM_SYSCOLORCHANGE: writeWm("WM_SYSCOLORCHANGE"); break;
			case WM_ENDSESSION: writeWm("WM_ENDSESSION"); break;
			case WM_SHOWWINDOW: writeWm("WM_SHOWWINDOW"); break;
			//case WM_WININICHANGE: writeWm("WM_WININICHANGE"); break;
			case WM_SETTINGCHANGE: writeWm("WM_SETTINGCHANGE"); break;
			case WM_DEVMODECHANGE: writeWm("WM_DEVMODECHANGE"); break;
			case WM_ACTIVATEAPP: writeWm("WM_ACTIVATEAPP"); break;
			case WM_FONTCHANGE: writeWm("WM_FONTCHANGE"); break;
			case WM_TIMECHANGE: writeWm("WM_TIMECHANGE"); break;
			case WM_CANCELMODE: writeWm("WM_CANCELMODE"); break;
			case WM_SETCURSOR: writeWm("WM_SETCURSOR"); break;
			case WM_MOUSEACTIVATE: writeWm("WM_MOUSEACTIVATE"); break;
			case WM_CHILDACTIVATE: writeWm("WM_CHILDACTIVATE"); break;
			case WM_QUEUESYNC: writeWm("WM_QUEUESYNC"); break;
			case WM_GETMINMAXINFO: writeWm("WM_GETMINMAXINFO"); break;
			case WM_NOTIFY: writeWm("WM_NOTIFY"); break;
			case WM_INPUTLANGCHANGEREQUEST: writeWm("WM_INPUTLANGCHANGEREQUEST"); break;
			case WM_INPUTLANGCHANGE: writeWm("WM_INPUTLANGCHANGE"); break;
			case WM_TCARD: writeWm("WM_TCARD"); break;
			case WM_HELP: writeWm("WM_HELP"); break;
			case WM_USERCHANGED: writeWm("WM_USERCHANGED"); break;
			case WM_NOTIFYFORMAT: writeWm("WM_NOTIFYFORMAT"); break;
			case WM_CONTEXTMENU: writeWm("WM_CONTEXTMENU"); break;
			case WM_STYLECHANGING: writeWm("WM_STYLECHANGING"); break;
			case WM_STYLECHANGED: writeWm("WM_STYLECHANGED"); break;
			case WM_DISPLAYCHANGE: writeWm("WM_DISPLAYCHANGE"); break;
			case WM_GETICON: writeWm("WM_GETICON"); break;
			case WM_SETICON: writeWm("WM_SETICON"); break;
			case WM_NCCREATE: writeWm("WM_NCCREATE"); break;
			case WM_NCDESTROY: writeWm("WM_NCDESTROY"); break;
			case WM_NCCALCSIZE: writeWm("WM_NCCALCSIZE"); break;
			case WM_NCHITTEST: writeWm("WM_NCHITTEST"); break;
			case WM_NCPAINT: writeWm("WM_NCPAINT"); break;
			case WM_NCACTIVATE: writeWm("WM_NCACTIVATE"); break;
			case WM_GETDLGCODE: writeWm("WM_GETDLGCODE"); break;
			case WM_NCMOUSEMOVE: writeWm("WM_NCMOUSEMOVE"); break;
			case WM_NCLBUTTONDOWN: writeWm("WM_NCLBUTTONDOWN"); break;
			case WM_NCLBUTTONUP: writeWm("WM_NCLBUTTONUP"); break;
			case WM_NCLBUTTONDBLCLK: writeWm("WM_NCLBUTTONDBLCLK"); break;
			case WM_NCRBUTTONDOWN: writeWm("WM_NCRBUTTONDOWN"); break;
			case WM_NCRBUTTONUP: writeWm("WM_NCRBUTTONUP"); break;
			case WM_NCRBUTTONDBLCLK: writeWm("WM_NCRBUTTONDBLCLK"); break;
			case WM_NCMBUTTONDOWN: writeWm("WM_NCMBUTTONDOWN"); break;
			case WM_NCMBUTTONUP: writeWm("WM_NCMBUTTONUP"); break;
			case WM_NCMBUTTONDBLCLK: writeWm("WM_NCMBUTTONDBLCLK"); break;
			case WM_KEYDOWN: writeWm("WM_KEYDOWN"); break;
			case WM_KEYUP: writeWm("WM_KEYUP"); break;
			case WM_CHAR: writeWm("WM_CHAR"); break;
			case WM_DEADCHAR: writeWm("WM_DEADCHAR"); break;
			case WM_SYSKEYDOWN: writeWm("WM_SYSKEYDOWN"); break;
			case WM_SYSKEYUP: writeWm("WM_SYSKEYUP"); break;
			case WM_SYSCHAR: writeWm("WM_SYSCHAR"); break;
			case WM_SYSDEADCHAR: writeWm("WM_SYSDEADCHAR"); break;
			case WM_IME_STARTCOMPOSITION: writeWm("WM_IME_STARTCOMPOSITION"); break;
			case WM_IME_ENDCOMPOSITION: writeWm("WM_IME_ENDCOMPOSITION"); break;
			case WM_IME_COMPOSITION: writeWm("WM_IME_COMPOSITION"); break;
			case WM_INITDIALOG: writeWm("WM_INITDIALOG"); break;
			case WM_COMMAND: writeWm("WM_COMMAND"); break;
			case WM_SYSCOMMAND: writeWm("WM_SYSCOMMAND"); break;
			case WM_TIMER: writeWm("WM_TIMER"); break;
			case WM_HSCROLL: writeWm("WM_HSCROLL"); break;
			case WM_VSCROLL: writeWm("WM_VSCROLL"); break;
			case WM_INITMENU: writeWm("WM_INITMENU"); break;
			case WM_INITMENUPOPUP: writeWm("WM_INITMENUPOPUP"); break;
			case WM_MENUSELECT: writeWm("WM_MENUSELECT"); break;
			case WM_MENUCHAR: writeWm("WM_MENUCHAR"); break;
			case WM_ENTERIDLE: writeWm("WM_ENTERIDLE"); break;
			case WM_CTLCOLORMSGBOX: writeWm("WM_CTLCOLORMSGBOX"); break;
			case WM_CTLCOLOREDIT: writeWm("WM_CTLCOLOREDIT"); break;
			case WM_CTLCOLORLISTBOX: writeWm("WM_CTLCOLORLISTBOX"); break;
			case WM_CTLCOLORBTN: writeWm("WM_CTLCOLORBTN"); break;
			case WM_CTLCOLORDLG: writeWm("WM_CTLCOLORDLG"); break;
			case WM_CTLCOLORSCROLLBAR: writeWm("WM_CTLCOLORSCROLLBAR"); break;
			case WM_CTLCOLORSTATIC: writeWm("WM_CTLCOLORSTATIC"); break;
			case WM_MOUSEMOVE: writeWm("WM_MOUSEMOVE"); break;
			case WM_LBUTTONDOWN: writeWm("WM_LBUTTONDOWN"); break;
			case WM_LBUTTONUP: writeWm("WM_LBUTTONUP"); break;
			case WM_LBUTTONDBLCLK: writeWm("WM_LBUTTONDBLCLK"); break;
			case WM_RBUTTONDOWN: writeWm("WM_RBUTTONDOWN"); break;
			case WM_RBUTTONUP: writeWm("WM_RBUTTONUP"); break;
			case WM_RBUTTONDBLCLK: writeWm("WM_RBUTTONDBLCLK"); break;
			case WM_MBUTTONDOWN: writeWm("WM_MBUTTONDOWN"); break;
			case WM_MBUTTONUP: writeWm("WM_MBUTTONUP"); break;
			case WM_MBUTTONDBLCLK: writeWm("WM_MBUTTONDBLCLK"); break;
			case WM_PARENTNOTIFY: writeWm("WM_PARENTNOTIFY"); break;
			case WM_ENTERMENULOOP: writeWm("WM_ENTERMENULOOP"); break;
			case WM_EXITMENULOOP: writeWm("WM_EXITMENULOOP"); break;
			case WM_NEXTMENU: writeWm("WM_NEXTMENU"); break;
			case WM_SETFONT: writeWm("WM_SETFONT"); break;
			case WM_GETFONT: writeWm("WM_GETFONT"); break;
			case WM_USER: writeWm("WM_USER"); break;
			case WM_NEXTDLGCTL: writeWm("WM_NEXTDLGCTL"); break;
			case WM_CAPTURECHANGED: writeWm("WM_CAPTURECHANGED"); break;
			case WM_WINDOWPOSCHANGING: writeWm("WM_WINDOWPOSCHANGING"); break;
			case WM_WINDOWPOSCHANGED: writeWm("WM_WINDOWPOSCHANGED"); break;
			case WM_DRAWITEM: writeWm("WM_DRAWITEM"); break;
			case WM_CLEAR: writeWm("WM_CLEAR"); break;
			case WM_CUT: writeWm("WM_CUT"); break;
			case WM_COPY: writeWm("WM_COPY"); break;
			case WM_PASTE: writeWm("WM_PASTE"); break;
			case WM_MDITILE: writeWm("WM_MDITILE"); break;
			case WM_MDICASCADE: writeWm("WM_MDICASCADE"); break;
			case WM_MDIICONARRANGE: writeWm("WM_MDIICONARRANGE"); break;
			case WM_MDIGETACTIVE: writeWm("WM_MDIGETACTIVE"); break;
			case WM_MOUSEWHEEL: writeWm("WM_MOUSEWHEEL"); break;
			case WM_MOUSEHOVER: writeWm("WM_MOUSEHOVER"); break;
			case WM_MOUSELEAVE: writeWm("WM_MOUSELEAVE"); break;
			case WM_PRINT: writeWm("WM_PRINT"); break;
			case WM_PRINTCLIENT: writeWm("WM_PRINTCLIENT"); break;
			case WM_MEASUREITEM: writeWm("WM_MEASUREITEM"); break;
			
			default:
				if(m.msg >= WM_USER && m.msg <= 0x7FFF)
				{
					writeWm("WM_USER+" ~ std.string.toString(m.msg - WM_USER));
				}
				else if(m.msg >=0xC000 && m.msg <= 0xFFFF)
				{
					writeWm("RegisterWindowMessage");
				}
				else
				{
					writeWm("?");
				}
		}
		
		writef("HWND=%d(0x%X) WPARAM=%d(0x%X) LPARAM=%d(0x%X)\r\n\r\n",
			cast(size_t)m.hWnd, cast(size_t)m.hWnd,
			m.wParam, m.wParam,
			m.lParam, m.lParam);
	}
}


extern(Windows) LRESULT dflWndProc(HWND hwnd, UINT msg, WPARAM wparam, LPARAM lparam)
{
	//printf("HWND %p; WM %d(0x%X); WPARAM %d(0x%X); LPARAM %d(0x%X);\n", hwnd, msg, msg, wparam, wparam, lparam, lparam);
	
	if(msg == wmDfl)
	{
		switch(wparam)
		{
			case WPARAM_DFL_INVOKE:
				{
					InvokeData* pinv;
					pinv = cast(InvokeData*)lparam;
					try
					{
						pinv.result = pinv.dg(pinv.args);
					}
					catch(Object e)
					{
						//Application.onThreadException(e);
						try
						{
							if(e)
								pinv.exception = e;
							else
								pinv.exception = new Object;
						}
						catch(Object e2)
						{
							Application.onThreadException(e2);
						}
					}
					return LRESULT_DFL_INVOKE;
				}
				break;
			
			case WPARAM_DFL_DELAY_INVOKE:
				try
				{
					(cast(void function())lparam)();
				}
				catch(Object e)
				{
					Application.onThreadException(e);
				}
				break;
			
			default: ;
		}
	}
	
	Message dm = Message(hwnd, msg, wparam, lparam);
	Control ctrl;
	
	debug(SHOW_MESSAGE_INFO)
		showMessageInfo(dm);
	
	if(msg == WM_NCCREATE)
	{
		ctrl = Application.getCreatingControl();
		if(!ctrl)
		{
			debug(APP_PRINT)
				printf("Unable to add window 0x%X.\n", hwnd);
			return dm.result;
		}
		Application.creatingControl(null); // Reset.
		
		Application.controls[hwnd] = ctrl;
		ctrl.hwnd = hwnd;
		debug(APP_PRINT)
			printf("Added window 0x%X.\n", hwnd);
		
		//ctrl.finishCreating(hwnd);
		goto do_msg;
	}
	else
	{
		ctrl = Application.lookupHwnd(hwnd);
	}
	
	if(ctrl)
	{
		do_msg:
		ctrl.mustWndProc(dm);
		if(!ctrl.preProcessMessage(dm))
			ctrl._wndProc(dm);
	}
	return dm.result;
}


version(CUSTOM_MSG_HOOK)
{
	typedef CWPRETSTRUCT CustomMsg;
	
	
	// Needs to be re-entrant.
	extern(Windows) LRESULT globalMsgHook(int code, WPARAM wparam, LPARAM lparam)
	{
		if(code == HC_ACTION)
		{
			CustomMsg* msg = cast(CustomMsg*)lparam;
			Control ctrl;
			
			switch(msg.message)
			{
				// ...
			}
		}
		
		return CallNextHookEx(Application.msghook, code, wparam, lparam);
	}
}
else
{
	struct CustomMsg
	{
		HWND hwnd;
		UINT message;
		WPARAM wParam;
		LPARAM lParam;
	}
}


const LRESULT LRESULT_DFL_INVOKE = 0x95FADF; // Magic number.


struct InvokeData
{
	Object delegate(Object[]) dg;
	Object[] args;
	Object result;
	Object exception = null;
}


UINT wmDfl;

enum: WPARAM
{
	WPARAM_DFL_INVOKE = 78,
	WPARAM_DFL_DELAY_INVOKE = 79,
}


extern(Windows)
{
	alias BOOL function(LPTRACKMOUSEEVENT lpEventTrack) TrackMouseEventProc;
	alias BOOL function(HWND, COLORREF, BYTE, DWORD) SetLayeredWindowAttributesProc;
	
	alias HTHEME function(HWND) GetWindowThemeProc;
	alias BOOL function(HTHEME hTheme, int iPartId, int iStateId) IsThemeBackgroundPartiallyTransparentProc;
	alias HRESULT function(HWND hwnd, HDC hdc, RECT* prc) DrawThemeParentBackgroundProc;
	alias void function(DWORD dwFlags) SetThemeAppPropertiesProc;
}


// Set version = SUPPORTS_MOUSE_TRACKING if it is guaranteed to be supported.
TrackMouseEventProc trackMouseEvent;

// Set version = SUPPORTS_OPACITY if it is guaranteed to be supported.
SetLayeredWindowAttributesProc setLayeredWindowAttributes;

/+
GetWindowThemeProc getWindowTheme;
IsThemeBackgroundPartiallyTransparentProc isThemeBackgroundPartiallyTransparent;
DrawThemeParentBackgroundProc drawThemeParentBackground;
SetThemeAppPropertiesProc setThemeAppProperties;
+/


const char[] CONTROL_CLASSNAME = "DFL_Control";
const char[] FORM_CLASSNAME = "DFL_Form";
const char[] TEXTBOX_CLASSNAME = "DFL_TextBox";
const char[] LISTBOX_CLASSNAME = "DFL_ListBox";
const char[] LABEL_CLASSNAME = "DFL_Label";
const char[] BUTTON_CLASSNAME = "DFL_Button";
const char[] MDICLIENT_CLASSNAME = "DFL_MdiClient";
const char[] RICHTEXTBOX_CLASSNAME = "DFL_RichTextBox";
const char[] COMBOBOX_CLASSNAME = "DFL_ComboBox";
const char[] TREEVIEW_CLASSNAME = "DFL_TreeView";
const char[] TABCONTROL_CLASSNAME = "DFL_TabControl";
const char[] LISTVIEW_CLASSNAME = "DFL_ListView";

WNDPROC textBoxPrevWndProc;
WNDPROC listboxPrevWndProc;
WNDPROC labelPrevWndProc;
WNDPROC buttonPrevWndProc;
WNDPROC mdiclientPrevWndProc;
WNDPROC richtextboxPrevWndProc;
WNDPROC comboboxPrevWndProc;
WNDPROC treeviewPrevWndProc;
WNDPROC tabcontrolPrevWndProc;
WNDPROC listviewPrevWndProc;

LONG textBoxClassStyle;
LONG listboxClassStyle;
LONG labelClassStyle;
LONG buttonClassStyle;
LONG mdiclientClassStyle;
LONG richtextboxClassStyle;
LONG comboboxClassStyle;
LONG treeviewClassStyle;
LONG tabcontrolClassStyle;
LONG listviewClassStyle;

HMODULE hmodRichtextbox = null;

// DMD 0.93: CS_HREDRAW | CS_VREDRAW | CS_DBLCLKS is not an expression
//const UINT WNDCLASS_STYLE = CS_HREDRAW | CS_VREDRAW | CS_DBLCLKS;
//const UINT WNDCLASS_STYLE = 11;

//const UINT WNDCLASS_STYLE = CS_DBLCLKS;
// DMD 0.106: CS_DBLCLKS is not an expression
const UINT WNDCLASS_STYLE = 0x0008;


extern(Windows)
{
	alias BOOL function(LPINITCOMMONCONTROLSEX lpInitCtrls) InitCommonControlsExProc;
}


// For this to work properly on Windows 95, Internet Explorer 4.0 must be installed.
void _initCommonControls(DWORD dwControls)
{
	version(SUPPORTS_COMMON_CONTROLS_EX)
	{
		pragma(msg, "DFL: extended common controls supported at compile time.");
		
		alias InitCommonControlsEx initProc;
	}
	else
	{
		// Make sure InitCommonControlsEx() is in comctl32.dll,
		// otherwise use the old InitCommonControls().
		
		HMODULE hmodCommonControls;
		InitCommonControlsExProc initProc;
		
		hmodCommonControls = LoadLibraryA("comctl32.dll");
		if(!hmodCommonControls)
		//	throw new DflException("Unable to load 'comctl32.dll'.");
			goto no_comctl32;
		
		initProc = cast(InitCommonControlsExProc)GetProcAddress(hmodCommonControls, "InitCommonControlsEx");
		if(!initProc)
		{
			//FreeLibrary(hmodCommonControls);
			no_comctl32:
			InitCommonControls();
			return;
		}
	}
	
	INITCOMMONCONTROLSEX icce;
	icce.dwSize = INITCOMMONCONTROLSEX.sizeof;
	icce.dwICC = dwControls;
	initProc(&icce);
}


static this()
{
	dfl.utf._init();
	
	Application.tlsThreadFlags = TlsAlloc();
	Application.tlsControl = TlsAlloc();
	Application.tlsFilter = TlsAlloc();
	version(CUSTOM_MSG_HOOK)
		Application.tlsHook = TlsAlloc();
	
	wmDfl = RegisterWindowMessageA("WM_DFL");
	if(!wmDfl)
		wmDfl = WM_USER + 0x7CD;
	
	//InitCommonControls(); // Done later. Needs to be linked with comctl32.lib.
	OleInitialize(null); // Needs to be linked with ole32.lib.
	
	HMODULE user32 = GetModuleHandleA("user32.dll");
	
	version(SUPPORTS_MOUSE_TRACKING)
	{
		pragma(msg, "DFL: mouse tracking supported at compile time.");
		
		trackMouseEvent = &TrackMouseEvent;
	}
	else
	{
		trackMouseEvent = cast(TrackMouseEventProc)GetProcAddress(user32, "TrackMouseEvent");
		if(!trackMouseEvent) // Must be Windows 95; check if common controls has it (IE 5.5).
			trackMouseEvent = cast(TrackMouseEventProc)GetProcAddress(GetModuleHandleA("comctl32.dll"), "_TrackMouseEvent");
	}
	
	version(SUPPORTS_OPACITY)
	{
		pragma(msg, "DFL: opacity supported at compile time.");
		
		setLayeredWindowAttributes = &SetLayeredWindowAttributes;
	}
	else
	{
		setLayeredWindowAttributes = cast(SetLayeredWindowAttributesProc)GetProcAddress(user32, "SetLayeredWindowAttributes");
	}
	
	WNDCLASSA wc;
	(cast(ubyte*)&wc)[0 .. wc.sizeof] = 0;
	wc.style = WNDCLASS_STYLE;
	wc.hInstance = GetModuleHandleA(null);
	wc.lpfnWndProc = &dflWndProc;
	
	// Control wndclass.
	wc.lpszClassName = CONTROL_CLASSNAME;
	if(!RegisterClassA(&wc))
		throw new DflException("Unable to initialize Control.");
	
	// Form wndclass.
	wc.cbWndExtra = DLGWINDOWEXTRA;
	wc.lpszClassName = FORM_CLASSNAME;
	if(!RegisterClassA(&wc))
		throw new DflException("Unable to initialize Form.");
	
	/+
	WndClass wc;
	wc.wc.style = WNDCLASS_STYLE;
	wc.wc.hInstance = GetModuleHandleA(null);
	wc.wc.lpfnWndProc = &dflWndProc;
	
	// Control wndclass.
	wc.className = CONTROL_CLASSNAME;
	if(!dfl.utf.registerClass(wc))
		throw new DflException("Unable to initialize Control.");
	
	// Form wndclass.
	wc.wc.cbWndExtra = DLGWINDOWEXTRA;
	wc.className = FORM_CLASSNAME;
	if(!dfl.utf.registerClass(wc))
		throw new DflException("Unable to initialize Form.");
	+/
}


static ~this()
{
	Application.sdtorFreeAllMenus();
	
	if(hmodRichtextbox)
		FreeLibrary(hmodRichtextbox);
}


extern(Windows)
{
	void _initTextBox()
	{
		if(!textBoxPrevWndProc)
		{
			WNDCLASS info;
			textBoxPrevWndProc = superClass(null, "EDIT", TEXTBOX_CLASSNAME, info);
			if(!textBoxPrevWndProc)
				throw new DflException("Unable to initialize TextBoxBase.");
			textBoxClassStyle = info.style;
		}
	}
	
	
	void _initListbox()
	{
		if(!listboxPrevWndProc)
		{
			WNDCLASS info;
			listboxPrevWndProc = superClass(null, "LISTBOX", LISTBOX_CLASSNAME, info);
			if(!listboxPrevWndProc)
				throw new DflException("Unable to initialize ListBox.");
			listboxClassStyle = info.style;
		}
	}
	
	
	void _initLabel()
	{
		if(!labelPrevWndProc)
		{
			WNDCLASS info;
			labelPrevWndProc = superClass(null, "STATIC", LABEL_CLASSNAME, info);
			if(!labelPrevWndProc)
				throw new DflException("Unable to initialize Label.");
			labelClassStyle = info.style;
		}
	}
	
	
	void _initButton()
	{
		if(!buttonPrevWndProc)
		{
			WNDCLASS info;
			buttonPrevWndProc = superClass(null, "BUTTON", BUTTON_CLASSNAME, info);
			if(!buttonPrevWndProc)
				throw new DflException("Unable to initialize ButtonBase.");
			buttonClassStyle = info.style;
		}
	}
	
	
	void _initMdiclient()
	{
		if(!mdiclientPrevWndProc)
		{
			WNDCLASS info;
			mdiclientPrevWndProc = superClass(null, "MDICLIENT", MDICLIENT_CLASSNAME, info);
			if(!mdiclientPrevWndProc)
				throw new DflException("Unable to initialize MdiClient.");
			mdiclientClassStyle = info.style;
		}
	}
	
	
	void _initRichtextbox()
	{
		if(!richtextboxPrevWndProc)
		{
			if(!hmodRichtextbox)
			{
				hmodRichtextbox = LoadLibraryA("riched20.dll");
				if(!hmodRichtextbox)
					throw new DflException("Unable to load 'riched20.dll'");
			}
			
			char[] classname;
			if(dfl.utf.useUnicode)
				classname = "RichEdit20W";
			else
				classname = "RichEdit20A";
			
			WNDCLASS info;
			richtextboxPrevWndProc = superClass(null, classname, RICHTEXTBOX_CLASSNAME, info);
			if(!richtextboxPrevWndProc)
				throw new DflException("Unable to initialize RichTextBox.");
			richtextboxClassStyle = info.style;
		}
	}
	
	
	void _initCombobox()
	{
		if(!comboboxPrevWndProc)
		{
			WNDCLASS info;
			comboboxPrevWndProc = superClass(null, "COMBOBOX", COMBOBOX_CLASSNAME, info);
			if(!comboboxPrevWndProc)
				throw new DflException("Unable to initialize ComboBox.");
			comboboxClassStyle = info.style;
		}
	}
	
	
	void _initTreeview()
	{
		if(!treeviewPrevWndProc)
		{
			_initCommonControls(ICC_TREEVIEW_CLASSES);
			
			WNDCLASS info;
			treeviewPrevWndProc = superClass(null, "SysTreeView32", TREEVIEW_CLASSNAME, info);
			if(!treeviewPrevWndProc)
				throw new DflException("Unable to initialize TreeView.");
			treeviewClassStyle = info.style;
		}
	}
	
	
	void _initTabcontrol()
	{
		if(!tabcontrolPrevWndProc)
		{
			_initCommonControls(ICC_TAB_CLASSES);
			
			WNDCLASS info;
			tabcontrolPrevWndProc = superClass(null, "SysTabControl32", TABCONTROL_CLASSNAME, info);
			if(!tabcontrolPrevWndProc)
				throw new DflException("Unable to initialize TabControl.");
			tabcontrolClassStyle = info.style;
		}
	}
	
	
	void _initListview()
	{
		if(!listviewPrevWndProc)
		{
			_initCommonControls(ICC_LISTVIEW_CLASSES);
			
			WNDCLASS info;
			listviewPrevWndProc = superClass(null, "SysListView32", LISTVIEW_CLASSNAME, info);
			if(!listviewPrevWndProc)
				throw new DflException("Unable to initialize ListView.");
			listviewClassStyle = info.style;
		}
	}
}


// Returns the old wndProc.
WNDPROC superClass(HINSTANCE hinst, char[] className, char[] newClassName, out WNDCLASS getInfo) // package
{
	WNDPROC wndProc;
	
	if(!GetClassInfoA(hinst, toStringz(className), &getInfo))
		throw new DflException("Unable to obtain information for window class '" ~ className ~ "'");
	
	wndProc = getInfo.lpfnWndProc;
	getInfo.lpfnWndProc = &dflWndProc;
	
	getInfo.style &= ~CS_GLOBALCLASS;
	getInfo.hCursor = null;
	getInfo.lpszClassName = toStringz(newClassName);
	getInfo.hInstance = GetModuleHandleA(null);
	
	if(!RegisterClassA(&getInfo))
		//throw new DflException("Unable to register window class '" ~ newClassName ~ "'.");
		return null;
	return wndProc;
}


WNDPROC superClass(HINSTANCE hinst, char[] className, char[] newClassName) // package
{
	WNDCLASS info;
	return superClass(hinst, className, newClassName, info);
}

