// Written by Christopher E. Miller
// See the included license.txt for copyright and license details.


// Not actually part of forms, but is handy.

module dfl.collections;


private extern(C) void* memmove(void*, void*, size_t len);


interface IEnumerator
{
	Object current(); // getter
	bool moveNext(); // Must be called to move to the first item as well.
	void reset();
	int opApply(int delegate(inout Object) dg); // mixin EnumeratorForeach;
	int opApply(int delegate(inout int, inout Object) dg); // mixin EnumeratorForeach;
}


template EnumeratorForeach(ItemType = Object)
{
	int opApply(int delegate(inout ItemType) dg)
	{
		int result = 0;
		ItemType tempo;
		
		reset();
		while(moveNext())
		{
			tempo = current;
			result = dg(tempo);
			if(result)
				break;
		}
		
		return result;
	}
	
	
	int opApply(int delegate(inout int i, inout ItemType) dg)
	{
		int result = 0;
		int i = 0, tempi;
		ItemType tempo;
		
		reset();
		while(moveNext())
		{
			tempo = current;
			tempi = i; // Use temp so caller can't modify.
			result = dg(tempi, tempo);
			if(result)
				break;
			i++;
		}
		
		return result;
	}
}


interface IEnumerable
{
	IEnumerator getEnumerator();
}


interface ICollection: IEnumerable
{
	int count(); // getter
	bool isSynchronized(); // getter
	Object syncRoot(); // getter
	void copyTo(Array array, int index);
}


interface IList: ICollection
{
	bool isFixedSize(); // getter
	bool isReadOnly(); // getter
	void opIndexAssign(Object value, int index); // setter
	Object opIndex(int index); // getter
	void add(Object value);
	void clear();
	bool contains(Object value); // opIn...
	int indexOf(Object value);
	void insert(int index, Object value);
	void remove(Object value);
	void removeAt(int index);
}


class Array: IList
{
	this(int numElements)
	{
		array = new Object[numElements];
	}
	
	
	bool isFixedSize() // getter
	{
		return true;
	}
	
	
	bool isReadOnly() // getter
	{
		return false;
	}
	
	
	bool isSynchronized() // getter
	{
		return false;
	}
	
	
	final int length() // getter
	{
		return array.length;
	}
	
	
	Object syncRoot() // getter
	{
		return this;
	}
	
	
	// Shallow copy.
	Array clone()
	{
		Array result;
		result = new Array;
		result.array = array.dup;
		return result;
	}
	
	
	// D-style.
	final Array dup() // getter
	{
		return clone();
	}
	
	
	static void copy(Object[] src, Array dest, int len)
	in
	{
		assert(len <= dest.length);
	}
	body
	{
		memmove(dest.array, src, len * Object.sizeof);
	}
	
	
	static void copy(Array src, Array dest, int len)
	{
		copy(src.array, dest, len);
	}
	
	
	static void copy(Object[] src, int srcIndex, Array dest, int destIndex, int len)
	in
	{
		assert(srcIndex + len <= src.length);
		assert(destIndex + len <= dest.length);
	}
	body
	{
		memmove(&dest.array[destIndex], &src[srcIndex], len * Object.sizeof);
	}
	
	
	static void copy(Array src, int srcIndex, Array dest, int destIndex, int len)
	{
		copy(src.array, srcIndex, dest, destIndex, len);
	}
	
	
	void copyTo(Array dest, int destIndex)
	{
		copy(array, 0, dest, destIndex, array.length);
	}
	
	
	static Array createInstance(int numElements)
	{
		return new Array(numElements);
	}
	
	
	IEnumerator getEnumerator()
	{
		return new AEnum(this);
	}
	
	
	private static int _indexOf(Object[] src, Object value)
	{
		size_t i;
		for(i = 0; i != src.length; i++)
		{
			if(value == src[i])
				return i;
		}
		return -1;
	}
	
	
	static int indexOf(Array src, Object value)
	{
		return _indexOf(src.array, value);
	}
	
	
	static int indexOf(Array src, Object value, int starti)
	{
		return _indexOf(src.array[starti .. src.array.length], value);
	}
	
	
	static int indexOf(Array src, Object value, int starti, int count)
	{
		return _indexOf(src.array[starti .. starti + count], value);
	}
	
	
	private static int _lastIndexOf(Object[] src, Object value)
	{
		size_t i;
		i = src.length;
		while(i)
		{
			i--;
			if(src[i] == value)
				return i;
		}
		return -1;
	}
	
	
	static int lastIndexOf(Array src, Object value)
	{
		return _lastIndexOf(src.array, value);
	}
	
	
	static int lastIndexOf(Array src, Object value, int starti)
	{
		return _lastIndexOf(src.array[starti .. src.array.length], value);
	}
	
	
	static int lastIndexOf(Array src, Object value, int starti, int count)
	{
		return _lastIndexOf(src.array[starti .. starti + count], value);
	}
	
	
	static void reverse(Array src)
	{
		src.array.reverse;
	}
	
	
	static void reverse(Array src, int starti, int count)
	{
		src.array[starti .. starti + count].reverse;
	}
	
	
	static void sort(Array src)
	{
		src.array.sort;
	}
	
	
	static void sort(Array src, int starti, int count)
	{
		src.array[starti .. starti + count].sort;
	}
	
	
	void add(Object value)
	{
		assert(0);
	}
	
	
	void clear()
	{
		foreach(inout Object obj; array)
		{
			obj = null; // ?
		}
	}
	
	
	bool contains(Object value)
	{
		return typeof(this).indexOf(this, value) != -1;
	}
	
	
	int /+ IList. +/ indexOf(Object value)
	{
		return typeof(this).indexOf(this, value);
	}
	
	
	void insert(int index, Object value)
	{
		assert(0);
	}
	
	
	void remove(Object value)
	{
		assert(0);
	}
	
	
	void removeAt(int index)
	{
		assert(0);
	}
	
	
	// This could be different from -length- if overridden.
	int count() // getter
	{
		return length;
	}
	
	
	void opIndexAssign(Object value, int index) // setter
	{
		array[index] = value;
	}
	
	
	Object opIndex(int index) // getter
	{
		return array[index];
	}
	
	
	private:
	Object[] array;
	
	
	this()
	{
	}
	
	
	static class AEnum: IEnumerator
	{
		mixin EnumeratorForeach;
		
		
		this(Array aobj)
		{
			this.aobj = aobj;
		}
		
		
		Object current() // getter
		{
			return aobj[idx];
		}
		
		
		bool moveNext()
		{
			idx++;
			if(idx >= aobj.length) // Note: signed-unsigned comparison.
			{
				idx = aobj.length; // Prevent overflow.
				return false;
			}
			return true;
		}
		
		
		void reset()
		{
			idx = -1;
		}
		
		
		private:
		Array aobj;
		int idx = -1;
	}
}


template removeIndex(T) // package
{
	T[] removeIndex(T[] array, size_t index)
	{
		if(!index)
			array = array[1 .. array.length];
		else if(index == array.length - 1)
			array = array[0 .. index];
		else
			array = array[0 .. index] ~ array[index + 1 .. array.length];
		return array;
	}
}


// Returns -1 if not found.
template findIsIndex(T) // package
{
	int findIsIndex(T[] array, T obj)
	{
		int idx;
		for(idx = 0; idx != array.length; idx++)
		{
			if(obj is array[idx])
				return idx;
		}
		return -1;
	}
}

