// Written by Christopher E. Miller
// See the included license.txt for copyright and license details.


module dfl.combobox;

private import std.string;

private import dfl.listbox, dfl.application, dfl.base, dfl.winapi;
private import dfl.event, dfl.drawing, dfl.collections, dfl.control;


private extern(Windows) void _initCombobox();


enum ComboBoxStyle: ubyte
{
	DROP_DOWN,
	DROP_DOWN_LIST,
	SIMPLE,
}


class ComboBox: ListControl // docmain
{
	this()
	{
		_initCombobox();
		
		wstyle |= WS_TABSTOP | WS_VSCROLL | CBS_DROPDOWN | CBS_AUTOHSCROLL | CBS_HASSTRINGS;
		wexstyle |= WS_EX_CLIENTEDGE;
		ctrlStyle |= ControlStyles.SELECTABLE;
		wclassStyle = comboboxClassStyle;
		
		icollection = createItemCollection();
	}
	
	
	final void dropDownStyle(ComboBoxStyle ddstyle) // setter
	{
		LONG st;
		st = _style() & ~(CBS_DROPDOWN | CBS_DROPDOWNLIST | CBS_SIMPLE);
		
		switch(ddstyle)
		{
			case ComboBoxStyle.DROP_DOWN:
				_style(st | CBS_DROPDOWN);
				break;
			
			case ComboBoxStyle.DROP_DOWN_LIST:
				_style(st | CBS_DROPDOWNLIST);
				break;
			
			case ComboBoxStyle.SIMPLE:
				_style(st | CBS_SIMPLE);
				break;
		}
	}
	
	
	final ComboBoxStyle dropDownStyle() // getter
	{
		LONG st;
		st = _style() & (CBS_DROPDOWN | CBS_DROPDOWNLIST | CBS_SIMPLE);
		
		switch(st)
		{
			case CBS_DROPDOWN:
				return ComboBoxStyle.DROP_DOWN;
			
			case CBS_DROPDOWNLIST:
				return ComboBoxStyle.DROP_DOWN_LIST;
			
			case CBS_SIMPLE:
				return ComboBoxStyle.SIMPLE;
		}
	}
	
	
	final void integralHeight(bool byes) //setter
	{
		if(byes)
			_style(_style() & ~CBS_NOINTEGRALHEIGHT);
		else
			_style(_style() | CBS_NOINTEGRALHEIGHT);
	}
	
	
	final bool integralHeight() // getter
	{
		return (_style() & CBS_NOINTEGRALHEIGHT) == 0;
	}
	
	
	// This function has no effect if the drawMode is OWNER_DRAW_VARIABLE.
	void itemHeight(int h) // setter
	{
		if(drawMode == DrawMode.OWNER_DRAW_VARIABLE)
			return;
		
		iheight = h;
		
		if(created)
			SendMessageA(hwnd, CB_SETITEMHEIGHT, 0, h);
	}
	
	
	// Return value is meaningless when drawMode is OWNER_DRAW_VARIABLE.
	int itemHeight() // getter
	{
		// Requesting it like this when owner draw variable doesn't work.
		/+
		if(!created)
			return iheight;
		
		int result = SendMessageA(handle, CB_GETITEMHEIGHT, 0, 0);
		if(result == CB_ERR)
			result = iheight; // ?
		else
			iheight = result;
		
		return result;
		+/
		
		return iheight;
	}
	
	
	void selectedIndex(int idx)
	{
		if(created)
		{
			SendMessageA(handle, CB_SETCURSEL, cast(WPARAM)idx, 0);
		}
	}
	
	
	int selectedIndex()
	{
		if(created)
		{
			LRESULT result;
			result = SendMessageA(handle, CB_GETCURSEL, 0, 0);
			if(CB_ERR != result) // Redundant.
				return cast(int)result;
		}
		return -1;
	}
	
	
	final void selectedItem(Object o) // setter
	{
		int i;
		i = items.indexOf(o);
		if(i != -1)
			selectedIndex = i;
	}
	
	
	final void selectedItem(char[] str) // setter
	{
		int i;
		i = items.indexOf(str);
		if(i != -1)
			selectedIndex = i;
	}
	
	
	final Object selectedItem() // getter
	{
		int idx;
		idx = selectedIndex;
		if(idx == -1)
			return null;
		return items[idx];
	}
	
	
	override void selectedValue(Object val) // setter
	{
		selectedItem = val;
	}
	
	
	override void selectedValue(char[] str) // setter
	{
		selectedItem = str;
	}
	
	
	override Object selectedValue() // getter
	{
		return selectedItem;
	}
	
	
	final void sorted(bool byes) // setter
	{
		if(byes)
			_style(_style() | CBS_SORT);
		else
			_style(_style() & ~CBS_SORT);
	}
	
	
	final bool sorted() // getter
	{
		return (_style() & CBS_SORT) != 0;
	}
	
	
	final void beginUpdate()
	{
		SendMessageA(handle, WM_SETREDRAW, false, 0);
	}
	
	
	final void endUpdate()
	{
		SendMessageA(handle, WM_SETREDRAW, true, 0);
		invalidate(true); // Show updates.
	}
	
	
	final int findString(char[] str, int startIndex)
	{
		// TODO: find string if control not created ?
		
		int result = NO_MATCHES;
		
		if(created)
		{
			result = SendMessageA(handle, CB_FINDSTRING, startIndex, cast(LPARAM)toStringz(str));
			if(result == CB_ERR) // Redundant.
				result = NO_MATCHES;
		}
		
		return result;
	}
	
	
	final int findString(char[] str)
	{
		return findString(str, -1); // Start at beginning.
	}
	
	
	final int findStringExact(char[] str, int startIndex)
	{
		// TODO: find string if control not created ?
		
		int result = NO_MATCHES;
		
		if(created)
		{
			result = SendMessageA(handle, CB_FINDSTRINGEXACT, startIndex, cast(LPARAM)toStringz(str));
			if(result == CB_ERR) // Redundant.
				result = NO_MATCHES;
		}
		
		return result;
	}
	
	
	final int findStringExact(char[] str)
	{
		return findStringExact(str, -1); // Start at beginning.
	}
	
	
	final int getItemHeight(int idx)
	{
		int result = SendMessageA(handle, CB_GETITEMHEIGHT, idx, 0);
		if(CB_ERR == result)
			throw new DflException("Unable to obtain item height.");
		return result;
	}
	
	
	final void drawMode(DrawMode dm) // setter
	{
		LONG wl = _style() & ~(CBS_OWNERDRAWVARIABLE | CBS_OWNERDRAWFIXED);
		
		switch(dm)
		{
			case DrawMode.OWNER_DRAW_VARIABLE:
				wl |= CBS_OWNERDRAWVARIABLE;
				break;
			
			case DrawMode.OWNER_DRAW_FIXED:
				wl |= CBS_OWNERDRAWFIXED;
				break;
			
			case DrawMode.NORMAL:
				break;
		}
		
		_style(wl);
	}
	
	
	final DrawMode drawMode() // getter
	{
		LONG wl = _style();
		
		if(wl & CBS_OWNERDRAWVARIABLE)
			return DrawMode.OWNER_DRAW_VARIABLE;
		if(wl & CBS_OWNERDRAWFIXED)
			return DrawMode.OWNER_DRAW_FIXED;
		return DrawMode.NORMAL;
	}
	
	
	final void selectAll()
	{
		if(created)
			SendMessageA(handle, CB_SETEDITSEL, 0, MAKELPARAM(0, cast(ushort)-1));
	}
	
	
	final void maxLength(uint len) // setter
	{
		if(!len)
			lim = 0x7FFFFFFE;
		else
			lim = len;
		
		if(created)
		{
			Message m;
			m = Message(handle, CB_LIMITTEXT, cast(WPARAM)lim, 0);
			prevWndProc(m);
		}
	}
	
	
	final uint maxLength() // getter
	{
		return lim;
	}
	
	
	final void selectionLength(uint len) // setter
	{
		if(created)
		{
			uint v1, v2;
			SendMessageA(handle, CB_GETEDITSEL, cast(WPARAM)&v1, cast(LPARAM)&v2);
			v2 = v1 + len;
			SendMessageA(handle, CB_SETEDITSEL, 0, MAKELPARAM(v1, v2));
		}
	}
	
	
	final uint selectionLength() // getter
	{
		if(created)
		{
			uint v1, v2;
			SendMessageA(handle, CB_GETEDITSEL, cast(WPARAM)&v1, cast(LPARAM)&v2);
			assert(v2 >= v1);
			return v2 - v1;
		}
		return 0;
	}
	
	
	final void selectionStart(uint pos) // setter
	{
		if(created)
		{
			uint v1, v2;
			SendMessageA(handle, CB_GETEDITSEL, cast(WPARAM)&v1, cast(LPARAM)&v2);
			assert(v2 >= v1);
			v2 = pos + (v2 - v1);
			SendMessageA(handle, CB_SETEDITSEL, 0, MAKELPARAM(pos, v2));
		}
	}
	
	
	final uint selectionStart() // getter
	{
		if(created)
		{
			uint v1, v2;
			SendMessageA(handle, CB_GETEDITSEL, cast(WPARAM)&v1, cast(LPARAM)&v2);
			return v1;
		}
		return 0;
	}
	
	
	final void droppedDown(bool byes) // setter
	{
		if(created)
			SendMessageA(handle, CB_SHOWDROPDOWN, cast(WPARAM)byes, 0);
	}
	
	
	final bool droppedDown() // getter
	{
		if(created)
			return SendMessageA(handle, CB_GETDROPPEDSTATE, 0, 0) != FALSE;
		return false;
	}
	
	
	final void dropDownWidth(int w) // setter
	{
		if(dropw == w)
			return;
		
		if(w < 0)
			w = 0;
		dropw = w;
		
		if(created)
		{
			if(dropw < width)
				SendMessageA(hwnd, CB_SETDROPPEDWIDTH, width, 0);
			else
				SendMessageA(hwnd, CB_SETDROPPEDWIDTH, dropw, 0);
		}
	}
	
	
	final int dropDownWidth() // getter
	{
		if(created)
		{
			int w;
			w = cast(int)SendMessageA(handle, CB_GETDROPPEDWIDTH, 0, 0);
			if(dropw != -1)
				dropw = w;
			return w;
		}
		else
		{
			if(dropw < width)
				return width;
			return dropw;
		}
	}
	
	
	final ObjectCollection items() // getter
	{
		return icollection;
	}
	
	
	const int DEFAULT_ITEM_HEIGHT = 13;
	const int NO_MATCHES = CB_ERR;
	
	
	static class ObjectCollection: IList
	{
		this(ComboBox lbox)
		{
			this.lbox = lbox;
		}
		
		
		this(ComboBox lbox, Object[] range)
		{
			this.lbox = lbox;
			addRange(range);
		}
		
		
		this(ComboBox lbox, char[][] range)
		{
			this.lbox = lbox;
			addRange(range);
		}
		
		
		this(ComboBox lbox, ObjectCollection range)
		{
			this.lbox = lbox;
			addRange(range);
		}
		
		
		bool isFixedSize() // getter
		{
			return false;
		}
		
		
		bool isReadOnly() // getter
		{
			return false;
		}
		
		
		bool isSynchronized() // getter
		{
			return false;
		}
		
		
		Object syncRoot() // getter
		{
			return this;
		}
		
		
		// Shallow copy.
		Array clone()
		{
			Array result;
			result = new Array(array.length);
			Array.copy(array, result, array.length);
			return result;
		}
		
		
		// D-style.
		final Array dup() // getter
		{
			return clone();
		}
		
		
		void copyTo(Array dest, int destIndex)
		{
			Array.copy(array, 0, dest, destIndex, array.length);
		}
		
		
		package final int add2(Object value)
		{
			int i;
			
			if(lbox.sorted)
			{
				// Insertion sort.
				
				for(i = 0; i != array.length; i++)
				{
					if(value < array[i])
						break;
				}
				
				if(i == array.length)
					//array ~= value;
					lbox._items ~= value;
				else
					array = array[0 .. i] ~ (&value)[0 .. 1] ~ array[i .. array.length];
			}
			else
			{
				i = array.length;
				//array ~= value;
				lbox._items ~= value;
			}
			
			if(lbox.created)
			{
				Message m;
				m = Message(lbox.handle, CB_INSERTSTRING, i, cast(LPARAM)toStringz(value.toString()));
				lbox.prevWndProc(m);
				
				i = m.result;
				//SendMessageA(lbox.handle, CB_SETITEMDATA, i, cast(LPARAM)cast(void*)value);
			}
			
			return i;
		}
		
		
		package final int add2(char[] value)
		{
			return add2(new ListString(value));
		}
		
		
		void add(Object value)
		{
			add2(value);
		}
		
		
		void add(char[] value)
		{
			add2(value);
		}
		
		
		void addRange(Object[] range)
		{
			if(lbox.sorted)
			{
				foreach(Object value; range)
				{
					add(value);
				}
			}
			else
			{
				//array ~= range;
				lbox._items ~= range;
				
				if(lbox.created)
				{
					Message m;
					m = Message(lbox.handle, CB_ADDSTRING, 0, 0);
					foreach(Object value; range)
					{
						m.lParam = cast(LPARAM)toStringz(value.toString());
						lbox.prevWndProc(m);
						
						//SendMessageA(lbox.handle, CB_SETITEMDATA, m.result, cast(LPARAM)cast(void*)value);
					}
				}
			}
		}
		
		
		void addRange(char[][] range)
		{
			foreach(char[] value; range)
			{
				add(value);
			}
		}
		
		
		void addRange(ObjectCollection range)
		{
			addRange(range.array);
		}
		
		
		void clear()
		{
			array = null; // ?
			
			if(lbox.created)
			{
				Message m;
				m = Message(lbox.handle, CB_RESETCONTENT, 0, 0);
				lbox.prevWndProc(m);
			}
		}
		
		
		bool contains(Object value)
		{
			return indexOf(value) != -1;
		}
		
		
		bool contains(char[] value)
		{
			return indexOf(value) != -1;
		}
		
		
		int indexOf(Object value)
		{
			size_t i;
			for(i = 0; i != array.length; i++)
			{
				if(value == array[i])
					return i;
			}
			return -1;
		}
		
		
		int indexOf(char[] value)
		{
			size_t i;
			ListString ls;
			for(i = 0; i != array.length; i++)
			{
				ls = cast(ListString)ls;
				if(ls)
				{
					if(value == ls.value)
						return i;
				}
			}
			return -1;
		}
		
		
		package final int insert2(int index, Object value)
		{
			if(index == array.length)
				//array ~= value;
				lbox._items ~= value;
			else
				array = array[0 .. index] ~ (&value)[0 .. 1] ~ array[index .. array.length];
			
			if(lbox.created)
			{
				Message m;
				m = Message(lbox.handle, CB_INSERTSTRING, index, cast(LPARAM)toStringz(value.toString()));
				lbox.prevWndProc(m);
				
				index = m.result;
				//SendMessageA(lbox.handle, CB_SETITEMDATA, index, cast(LPARAM)cast(void*)value);
			}
			
			return index;
		}
		
		
		package final int insert2(int index, char[] value)
		{
			return insert2(index, new ListString(value));
		}
		
		
		void insert(int index, Object value)
		{
			insert2(index, value);
		}
		
		
		void insert(int index, char[] value)
		{
			insert2(index, value);
		}
		
		
		void remove(Object value)
		{
			int i;
			i = indexOf(value);
			if(i != -1)
				removeAt(i);
		}
		
		
		void remove(char[] value)
		{
			int i;
			i = indexOf(value);
			if(i != -1)
				removeAt(i);
		}
		
		
		void removeAt(int index)
		{
			if(!index)
				array = array[1 .. array.length];
			else if(index == array.length - 1)
				array = array[0 .. index];
			else
				array = array[0 .. index] ~ array[index + 1 .. array.length];
			
			if(lbox.created)
			{
				Message m;
				m = Message(lbox.handle, CB_DELETESTRING, index, 0);
				lbox.prevWndProc(m);
			}
		}
		
		
		int count() // getter
		{
			return array.length;
		}
		
		
		void opIndexAssign(Object value, int index) // setter
		{
			//array[index] = value;
			
			removeAt(index);
			//add(value);
			insert(index, value);
		}
		
		
		Object opIndex(int index) // getter
		{
			return array[index];
		}
		
		
		IEnumerator getEnumerator()
		{
			return new ItemEnum(array());
		}
		
		
		private:
		ComboBox lbox;
		
		
		this()
		{
		}
		
		
		final Object[] array() // getter
		{
			return lbox._items;
		}
		
		
		final void array(Object[] setter) // setter
		{
			lbox._items = setter;
		}
		
		
		static class ItemEnum: IEnumerator
		{
			mixin EnumeratorForeach;
			
			
			this(Object[] items)
			{
				this.items = items;
			}
			
			
			Object current() // getter
			{
				return items[idx];
			}
			
			
			bool moveNext()
			{
				idx++;
				if(idx >= items.length) // Note: signed-unsigned comparison.
				{
					idx = items.length; // Prevent overflow.
					return false;
				}
				return true;
			}
			
			
			void reset()
			{
				idx = -1;
			}
			
			
			private:
			Object[] items;
			int idx = -1;
		}
	}
	
	
	protected ObjectCollection createItemCollection()
	{
		return new ObjectCollection(this);
	}
	
	
	protected override void onHandleCreated(EventArgs ea)
	{
		super.onHandleCreated(ea);
		
		// Set the Ctrl ID to the HWND so that it is unique
		// and WM_MEASUREITEM will work properly.
		SetWindowLongA(hwnd, GWL_ID, cast(LONG)hwnd);
		
		//SendMessageA(hwnd, EM_SETLIMITTEXT, cast(WPARAM)lim, 0);
		maxLength = lim; // Call virtual function.
		
		if(dropw < width)
			SendMessageA(hwnd, CB_SETDROPPEDWIDTH, width, 0);
		else
			SendMessageA(hwnd, CB_SETDROPPEDWIDTH, dropw, 0);
		
		if(iheight != DEFAULT_ITEM_HEIGHT)
			SendMessageA(hwnd, CB_SETITEMHEIGHT, 0, iheight);
		
		Message m;
		m.hWnd = hwnd;
		m.msg = CB_INSERTSTRING;
		foreach(int i, Object obj; _items)
		{
			m.wParam = i;
			m.lParam = cast(LPARAM)toStringz(obj.toString());
			
			prevWndProc(m);
			//if(CB_ERR == m.result || CB_ERRSPACE == m.result)
			if(m.result < 0)
				throw new DflException("Unable to add combo box item.");
			
			//SendMessageA(handle, CB_SETITEMDATA, m.result, cast(LPARAM)cast(void*)obj);
		}
		
		//redrawEntire();
	}
	
	
	package final bool hasDropList() // getter
	{
		return dropDownStyle != ComboBoxStyle.SIMPLE;
	}
	
	
	// This is needed for the SIMPLE style.
	protected override void onPaintBackground(PaintEventArgs pea)
	{
		RECT rect;
		pea.clipRectangle.getRect(&rect);
		FillRect(pea.graphics.handle, &rect, parent.hbrBg); // Hack.
	}
	
	
	/+
	override void createHandle()
	{
		if(isHandleCreated)
			return;
		
		// TODO: check if correct implementation.
		if(hasDropList)
			wrect.height = DEFAULT_ITEM_HEIGHT * 8;
		
		char[] ft;
		ft = wtext;
		
		createClassHandle(COMBOBOX_CLASSNAME);
		
		// Fix the cached window rect.
		RECT rect;
		GetWindowRect(hwnd, &rect);
		wrect = Rect(&rect);
		
		// Fix the combo box's text since the initial window
		// text isn't put in the edit box for some reason.
		Message m;
		m = Message(hwnd, WM_SETTEXT, 0, cast(LPARAM)toStringz(ft));
		prevWndProc(m);
		
		onHandleCreated(EventArgs.empty);
	}
	+/
	
	
	override void createHandle()
	{
		if(isHandleCreated)
			return;
		
		// TODO: check if correct implementation.
		if(hasDropList)
			wrect.height = DEFAULT_ITEM_HEIGHT * 8;
		
		char[] ft;
		ft = wtext;
		
		super.createHandle();
		
		// Fix the cached window rect.
		RECT rect;
		GetWindowRect(hwnd, &rect);
		wrect = Rect(&rect);
		
		// Fix the combo box's text since the initial window
		// text isn't put in the edit box for some reason.
		Message m;
		m = Message(hwnd, WM_SETTEXT, 0, cast(LPARAM)toStringz(ft));
		prevWndProc(m);
	}
	
	
	protected override void createParams(inout CreateParams cp)
	{
		super.createParams(cp);
		
		cp.className = COMBOBOX_CLASSNAME;
	}
	
	
	DrawItemEventHandler drawItem;
	MeasureItemEventHandler measureItem;
	
	
	protected:
	override Size defaultSize() // getter
	{
		return Size(120, 23); // ?
	}
	
	
	void onDrawItem(DrawItemEventArgs dieh)
	{
		drawItem(this, dieh);
	}
	
	
	void onMeasureItem(MeasureItemEventArgs miea)
	{
		measureItem(this, miea);
	}
	
	
	package final void _WmDrawItem(DRAWITEMSTRUCT* dis)
	in
	{
		assert(dis.hwndItem == handle);
		assert(dis.CtlType == ODT_COMBOBOX);
	}
	body
	{
		DrawItemState state;
		state = cast(DrawItemState)dis.itemState;
		
		if(dis.itemID == -1)
		{
			if(state & DrawItemState.FOCUS)
				DrawFocusRect(dis.hDC, &dis.rcItem);
		}
		else
		{
			DrawItemEventArgs diea;
			Color bc, fc;
			
			if(state & DrawItemState.SELECTED)
			{
				bc = Color.systemColor(COLOR_HIGHLIGHT);
				fc = Color.systemColor(COLOR_HIGHLIGHTTEXT);
			}
			else
			{
				bc = backColor;
				fc = foreColor;
			}
			
			prepareDc(dis.hDC);
			diea = new DrawItemEventArgs(new Graphics(dis.hDC, false), wfont,
				Rect(&dis.rcItem), dis.itemID, state, fc, bc);
			
			onDrawItem(diea);
		}
	}
	
	
	package final void _WmMeasureItem(MEASUREITEMSTRUCT* mis)
	in
	{
		assert(mis.CtlType == ODT_COMBOBOX);
	}
	body
	{
		MeasureItemEventArgs miea;
		auto Graphics gpx = new CommonGraphics(handle(), GetDC(handle));
		miea = new MeasureItemEventArgs(gpx, mis.itemID, /+ mis.itemHeight +/ iheight);
		miea.itemWidth = mis.itemWidth;
		
		onMeasureItem(miea);
		
		mis.itemHeight = miea.itemHeight;
		mis.itemWidth = miea.itemWidth;
	}
	
	
	override void prevWndProc(inout Message msg)
	{
		msg.result = CallWindowProcA(comboboxPrevWndProc, msg.hWnd, msg.msg, msg.wParam, msg.lParam);
	}
	
	
	/+package+/ override LRESULT customMsg(inout CustomMsg msg) // package
	{
		LRESULT result;
		result = super.customMsg(msg);
		
		switch(msg.message)
		{
			case WM_DRAWITEM:
				_WmDrawItem(cast(DRAWITEMSTRUCT*)msg.lParam);
				result = 1;
				break;
			
			case WM_MEASUREITEM:
				_WmMeasureItem(cast(MEASUREITEMSTRUCT*)msg.lParam);
				result = 1;
				break;
			
			case WM_COMMAND:
				//assert(cast(HWND)msg.lParam == handle); // Might be one of its children.
				switch(HIWORD(msg.wParam))
				{
					case CBN_SELCHANGE:
						/+
						if(drawMode != DrawMode.NORMAL)
						{
							// Hack.
							Object item = selectedItem;
							text = item ? item.toString() : cast(char[])null;
						}
						+/
						onSelectedIndexChanged(EventArgs.empty);
						break;
					
					default: ;
				}
				break;
			
			default: ;
		}
		
		return result;
	}
	
	
	override void wndProc(inout Message msg)
	{
		switch(msg.msg)
		{
			case CB_ADDSTRING:
				msg.result = icollection.add2(std.string.toString(cast(char*)msg.lParam).dup);
				return;
			
			case CB_INSERTSTRING:
				msg.result = icollection.insert2(msg.wParam, std.string.toString(cast(char*)msg.lParam).dup);
				return;
			
			case CB_DELETESTRING:
				icollection.removeAt(msg.wParam);
				msg.result = icollection.count;
				return;
			
			case CB_RESETCONTENT:
				icollection.clear();
				return;
			
			case CB_SETITEMDATA:
				// Cannot set item data from outside DFL.
				msg.result = CB_ERR;
				return;
			
			case CB_DIR:
				msg.result = CB_ERR;
				return;
			
			case CB_LIMITTEXT:
				maxLength = msg.wParam;
				return;
			
			default: ;
		}
		super.wndProc(msg);
	}
	
	
	private:
	int iheight = DEFAULT_ITEM_HEIGHT;
	int dropw = -1;
	ObjectCollection icollection;
	Object[] _items;
	package uint lim = 30_000; // Documented as default.
}

