// Written by Christopher E. Miller
// See the included license.txt for copyright and license details.


module dfl.control;

private import std.string;
	
private import dfl.base, dfl.menu, dfl.form, dfl.drawing;
private import dfl.winapi, dfl.application, dfl.event, dfl.label;
private import dfl.wincom, dfl.utf, dfl.collections;

version(NO_DRAG_DROP) {} else
	private import dfl.data;

version = RADIO_GROUP_LAYOUT;


enum AnchorStyles: ubyte
{
	NONE = 0,
	TOP = 1,
	BOTTOM = 2,
	LEFT = 4,
	RIGHT = 8,
}


enum BoundsSpecified: ubyte
{
	NONE = 0,
	X = 1,
	Y = 2,
	LOCATION = 1 | 2, //X | Y,
	WIDTH = 4,
	HEIGHT = 8,
	SIZE = 4 | 8, //WIDTH | HEIGHT,
	ALL = 1 | 2 | 4 | 8, //LOCATION | SIZE,
}


enum DockStyle: ubyte
{
	NONE,
	BOTTOM,
	FILL,
	LEFT,
	RIGHT,
	TOP,
}


private
{
	struct GetTabIndex
	{
		HWND hwnd;
		int index = -1;
		private int _tmp = 0;
	}
	
	
	extern(Windows) BOOL getTabIndexCallback(HWND hwnd, LPARAM lparam)
	{
		GetTabIndex* gti = cast(GetTabIndex*)lparam;
		if(hwnd == gti.hwnd)
		{
			gti.index = gti._tmp;
			return FALSE; // Stop, found it.
		}
		gti._tmp++;
		return TRUE; // Keep looking.
	}
}


enum DragDropEffects: DWORD
{
	NONE = 0,
	COPY = 1,
	MOVE = 2,
	LINK = 4,
	SCROLL = 0x80000000,
	ALL = COPY | MOVE | LINK | SCROLL,
}


enum DragAction: HRESULT
{
	CONTINUE = S_OK,
	CANCEL = DRAGDROP_S_CANCEL,
	DROP = DRAGDROP_S_DROP,
}


// Flags.
enum UICues: uint
{
	NONE = 0,
	SHOW_FOCUS = 1,
	SHOW_KEYBOARD = 2,
	SHOWN = SHOW_FOCUS | SHOW_KEYBOARD,
	CHANGE_FOCUS = 4,
	CHANGE_KEYBOARD = 8, // Key mnemonic underline cues are on.
	CHANGED = CHANGE_FOCUS | CHANGE_KEYBOARD,
}


// May be OR'ed together.
enum ControlStyles: uint
{
	NONE = 0,
	
	CONTAINER_CONTROL =                0x1, // Just container-like.
	USER_PAINT =                       0x2, // TODO: implement.
	OPAQUE =                           0x4,
	RESIZE_REDRAW =                    0x10,
	FIXED_WIDTH =                      0x20, // TODO: implement.
	FIXED_HEIGHT =                     0x40, // TODO: implement.
	STANDARD_CLICK =                   0x100,
	SELECTABLE =                       0x200, // Can be set focus.
	USER_MOUSE =                       0x400, // TODO: implement.
	SUPPORTS_TRANSPARENT_BACK_COLOR =  0x800, // Only if USER_PAINT and parent is derived from Control. TODO: implement.
	STANDARD_DOUBLE_CLICK =            0x1000,
	ALL_PAINTING_IN_WM_PAINT =         0x2000,
	CACHE_TEXT =                       0x4000,
	ENABLE_NOTIFY_MESSAGE =            0x8000, // Calls onNotifyMessage() for every message.
	DOUBLE_BUFFER =                    0x10000, // TODO: implement.
}


struct CreateParams
{
	char[] className;
	char[] caption;
	void* param;
	HWND parent;
	HMENU menu;
	HINSTANCE inst;
	int x;
	int y;
	int width;
	int height;
	DWORD classStyle;
	DWORD exStyle;
	DWORD style;
}


class UICuesEventArgs: EventArgs
{
	this(UICues uic)
	{
		chg = uic;
	}
	
	
	final UICues changed() // getter
	{
		return chg;
	}
	
	
	final bool changeFocus()
	{
		return (chg & UICues.CHANGE_FOCUS) != 0;
	}
	
	
	final bool changeKeyboard()
	{
		return (chg & UICues.CHANGE_KEYBOARD) != 0;
	}
	
	
	final bool showFocus()
	{
		return (chg & UICues.SHOW_FOCUS) != 0;
	}
	
	
	final bool showKeyboard()
	{
		return (chg & UICues.SHOW_KEYBOARD) != 0;
	}
	
	
	private:
	UICues chg;
}


// Usually fired on alt, tab or F10 keys.
alias Event!(UICuesEventArgs) UICuesEventHandler;


class ControlEventArgs: EventArgs
{
	this(Control ctrl)
	{
		this.ctrl = ctrl;
	}
	
	
	final Control control() // getter
	{
		return ctrl;
	}
	
	
	private:
	Control ctrl;
}


alias Event!(ControlEventArgs) ControlEventHandler;


class HelpEventArgs: EventArgs
{
	this(Point mousePos)
	{
		mpos = mousePos;
	}
	
	
	final void handled(bool byes) // setter
	{
		hand = byes;
	}
	
	
	final bool handled() // getter
	{
		return hand;
	}
	
	
	final Point mousePos() // getter
	{
		return mpos;
	}
	
	
	private:
	Point mpos;
	bool hand = false;
}


alias Event!(HelpEventArgs) HelpEventHandler;


class InvalidateEventArgs: EventArgs
{
	this(Rect invalidRect)
	{
		ir = invalidRect;
	}
	
	
	final Rect invalidRect() // getter
	{
		return ir;
	}
	
	
	private:
	Rect ir;
}


alias Event!(InvalidateEventArgs) InvalidateEventHandler;


class KeyEventArgs: EventArgs
{
	this(Keys keys)
	{
		ks = keys;
	}
	
	
	final bool alt() // getter
	{
		return (ks & Keys.ALT) != 0;
	}
	
	
	final bool control() // getter
	{
		return (ks & Keys.CONTROL) != 0;
	}
	
	
	final void handled(bool byes) // setter
	{
		hand = byes;
	}
	
	
	final bool handled() // getter
	{
		return hand;
	}
	
	
	final Keys keyCode() // getter
	{
		return ks & Keys.KEY_CODE;
	}
	
	
	final Keys keyData() // getter
	{
		return ks;
	}
	
	
	// -keyData- as an int.
	final int keyValue() // getter
	{
		return cast(int)ks;
	}
	
	
	final Keys modifiers() // getter
	{
		return ks & Keys.MODIFIERS;
	}
	
	
	final bool shift() // getter
	{
		return (ks & Keys.SHIFT) != 0;
	}
	
	
	private:
	Keys ks;
	bool hand = false;
}


alias Event!(KeyEventArgs) KeyEventHandler;


class LayoutEventArgs: EventArgs
{
	this(Control affectedControl)
	{
		ac = affectedControl;
	}
	
	
	final Control affectedControl()
	{
		return ac;
	}
	
	
	private:
	Control ac;
}


alias Event!(LayoutEventArgs) LayoutEventHandler;


class MouseEventArgs: EventArgs
{
	// -delta- is mouse wheel rotations.
	this(MouseButtons button, int clicks, int x, int y, int delta)
	{
		btn = button;
		clks = clicks;
		_x = x;
		_y = y;
		dlt = delta;
	}
	
	
	final MouseButtons button() // getter
	{
		return btn;
	}
	
	
	final int clicks() // getter
	{
		return clks;
	}
	
	
	final int delta() // getter
	{
		return dlt;
	}
	
	
	final int x() // getter
	{
		return _x;
	}
	
	
	final int y() // getter
	{
		return _y;
	}
	
	
	private:
	MouseButtons btn;
	int clks;
	int _x, _y;
	int dlt;
}


alias Event!(MouseEventArgs) MouseEventHandler;


class PaintEventArgs: EventArgs
{
	this(Graphics graphics, Rect clipRect)
	{
		g = graphics;
		cr = clipRect;
	}
	
	
	final Graphics graphics() // getter
	{
		return g;
	}
	
	
	final Rect clipRectangle() // getter
	{
		return cr;
	}
	
	
	private:
	Graphics g;
	Rect cr;
}


alias Event!(PaintEventArgs) PaintEventHandler;


class CancelEventArgs: EventArgs
{
	// Initialize cancel to false.
	this()
	{
		cncl = false;
	}
	
	
	this(bool cancel)
	{
		cncl = cancel;
	}
	
	
	final void cancel(bool byes) // setter
	{
		cncl = byes;
	}
	
	
	final bool cancel() // getter
	{
		return cncl;
	}
	
	
	private:
	bool cncl;
}


alias Event!(CancelEventArgs) CancelEventHandler;


version(NO_DRAG_DROP) {} else
{
	class DragEventArgs: EventArgs
	{
		this(dfl.data.IDataObject dataObj, int keyState, int x, int y,
			DragDropEffects allowedEffect, DragDropEffects effect)
		{
			_dobj = dataObj;
			_keyState = keyState;
			_x = x;
			_y = y;
			_allowedEffect = allowedEffect;
			_effect = effect;
		}
		
		
		final DragDropEffects allowedEffect() // getter
		{
			return _allowedEffect;
		}
		
		
		final void effect(DragDropEffects newEffect) // setter
		{
			_effect = newEffect;
		}
		
		
		final DragDropEffects effect() // getter
		{
			return _effect;
		}
		
		
		final dfl.data.IDataObject data() // getter
		{
			return _dobj;
		}
		
		
		// State of ctrl, alt, shift, and mouse buttons.
		final int keyState() // getter
		{
			return _keyState;
		}
		
		
		final int x() // getter
		{
			return _x;
		}
		
		
		final int y() // getter
		{
			return _y;
		}
		
		
		private:
		dfl.data.IDataObject _dobj;
		int _keyState;
		int _x, _y;
		DragDropEffects _allowedEffect, _effect;
	}
	
	
	alias Event!(DragEventArgs) DragEventHandler;
	
	
	class GiveFeedbackEventArgs: EventArgs
	{
		this(DragDropEffects effect, bool useDefaultCursors)
		{
			_effect = effect;
			udefcurs = useDefaultCursors;
		}
		
		
		final DragDropEffects effect() // getter
		{
			return _effect;
		}
		
		
		final void useDefaultCursors(bool byes) // setter
		{
			udefcurs = byes;
		}
		
		
		final bool useDefaultCursors() // getter
		{
			return udefcurs;
		}
		
		
		private:
		DragDropEffects _effect;
		bool udefcurs;
	}
	
	
	alias Event!(GiveFeedbackEventArgs) GiveFeedbackEventHandler;
	
	
	class QueryContinueDragEventArgs: EventArgs
	{
		this(int keyState, bool escapePressed, DragAction action)
		{
			_keyState = keyState;
			escp = escapePressed;
			_action = action;
		}
		
		
		final void action(DragAction newAction) // setter
		{
			_action = newAction;
		}
		
		
		final DragAction action() // getter
		{
			return _action;
		}
		
		
		final bool escapePressed() // getter
		{
			return escp;
		}
		
		
		// State of ctrl, alt and shift.
		final int keyState() // getter
		{
			return _keyState;
		}
		
		
		private:
		int _keyState;
		bool escp;
		DragAction _action;
	}
	
	
	alias Event!(QueryContinueDragEventArgs) QueryContinueDragEventHandler;
	;
}


package alias BOOL delegate(HWND) EnumWindowsCallback;


// Callback for EnumWindows() and EnumChildWindows().
private extern(Windows) BOOL enumingWindows(HWND hwnd, LPARAM lparam)
{
	EnumWindowsCallback dg = *(cast(EnumWindowsCallback*)lparam);
	return dg(hwnd);
}


private struct Efi
{
	HWND hwParent;
	EnumWindowsCallback dg;
}


// Callback for EnumChildWindows(). -lparam- = pointer to Efi;
private extern(Windows) BOOL enumingFirstWindows(HWND hwnd, LPARAM lparam)
{
	Efi* efi = cast(Efi*)lparam;
	if(efi.hwParent == GetParent(hwnd))
		return efi.dg(hwnd);
	return true;
}


package BOOL enumWindows(EnumWindowsCallback dg)
{
	static assert((&dg).sizeof <= LPARAM.sizeof);
	return EnumWindows(&enumingWindows, cast(LPARAM)&dg);
}


package BOOL enumChildWindows(HWND hwParent, EnumWindowsCallback dg)
{
	static assert((&dg).sizeof <= LPARAM.sizeof);
	return EnumChildWindows(hwParent, &enumingWindows, cast(LPARAM)&dg);
}


// Only the parent's children, not its children.
package BOOL enumFirstChildWindows(HWND hwParent, EnumWindowsCallback dg)
{
	Efi efi;
	efi.hwParent = hwParent;
	efi.dg = dg;
	return EnumChildWindows(hwParent, &enumingFirstWindows, cast(LPARAM)&efi);
}


class Control: IWin32Window // docmain
{
	static class ControlCollection
	{
		this(Control owner)
		{
			_owner = owner;
		}
		
		
		int count() // getter
		{
			if(_owner.created)
			{
				// Inefficient :(
				uint len = 0;
				foreach(Control ctrl; this)
				{
					len++;
				}
				return len;
			}
			else
			{
				return children.length;
			}
		}
		
		
		final bool isReadOnly() // getter
		{
			return false; // ?
		}
		
		
		bool isSynchronized() // getter
		{
			return false;
		}
		
		
		Control opIndex(int i) // getter
		{
			if(_owner.created)
			{
				int oni = 0;
				foreach(Control ctrl; this)
				{
					if(oni == i)
						return ctrl;
					oni++;
				}
				// Index out of bounds, bad things happen.
				assert(0);
				return null;
			}
			else
			{
				return children[i];
			}
		}
		
		
		void add(Control ctrl)
		{
			ctrl.parent = _owner;
		}
		
		
		// opIn ?
		bool contains(Control ctrl)
		{
			return indexOf(ctrl) != -1;
		}
		
		
		int indexOf(Control ctrl)
		{
			if(_owner.created)
			{
				int i = 0;
				int foundi = -1;
				
				
				BOOL enuming(HWND hwnd)
				{
					if(hwnd == ctrl.handle)
					{
						foundi = i;
						return false; // Stop.
					}
					
					i++;
					return true; // Continue.
				}
				
				
				enumFirstChildWindows(_owner.handle, &enuming);
				return foundi;
			}
			else
			{
				foreach(int i, Control onCtrl; children)
				{
					if(onCtrl == ctrl)
						return i;
				}
				return -1;
			}
		}
		
		
		void remove(Control ctrl)
		{
			if(_owner.created)
			{
				_removeCreated(ctrl.handle);
			}
			else
			{
				int i = indexOf(ctrl);
				if(i != -1)
					_removeNotCreated(i);
			}
		}
		
		
		private void _removeCreated(HWND hwnd)
		{
			DestroyWindow(hwnd); // ?
		}
		
		
		private void _removeNotCreated(int i)
		{
			if(!i)
				children = children[1 .. children.length];
			if(i == children.length - 1)
				children = children[0 .. i];
			else
				children = children[0 .. i] ~ children[i + 1 .. children.length];
		}
		
		
		void removeAt(int i)
		{
			if(_owner.created)
			{
				int ith = 0;
				HWND hwndith = null;
				
				
				BOOL enuming(HWND hwnd)
				{
					if(ith == i)
					{
						hwndith = hwnd;
						return false; // Stop.
					}
					
					ith++;
					return true; // Continue.
				}
				
				
				enumFirstChildWindows(_owner.handle, &enuming);
				if(hwndith)
					_removeCreated(hwndith);
			}
			else
			{
				_removeNotCreated(i);
			}
		}
		
		
		protected final Control owner() // getter
		{
			return _owner;
		}
		
		
		int opApply(int delegate(inout Control) dg)
		{
			int result = 0;
			
			if(_owner.created)
			{
				BOOL enuming(HWND hwnd)
				{
					Control ctrl = fromHandle(hwnd);
					if(ctrl)
					{
						result = dg(ctrl);
						if(result)
							return false; // Stop.
					}
					
					return true; // Continue.
				}
				
				
				enumFirstChildWindows(_owner.handle, &enuming);
			}
			else
			{
				foreach(Control ctrl; children)
				{
					result = dg(ctrl);
					if(result)
						break;
				}
			}
			
			return result;
		}
		
		
		package:
		Control _owner;
		Control[] children; // Only valid if -owner- isn't created yet.
	}
	
	
	private void onControlAdded(ControlEventArgs cea)
	{
		controlAdded(this, cea);
	}
	
	
	private void onControlRemoved(ControlEventArgs cea)
	{
		controlRemoved(this, cea);
	}
	
	
	final HWND handle() // IWin32Window getter
	{
		if(!created)
		{
			debug(APP_PRINT)
				printf("Control created due to handle request.\n");
			
			createHandle();
		}
		
		return hwnd;
	}
	
	
	version(NO_DRAG_DROP) {} else
	{
		void allowDrop(bool byes) // setter
		{
			/+
			if(dyes)
				_exStyle(_exStyle() | WS_EX_ACCEPTFILES);
			else
				_exStyle(_exStyle() & ~WS_EX_ACCEPTFILES);
			+/
			
			if(byes)
			{
				if(!droptarget)
				{
					droptarget = new DropTarget(this);
					if(created)
					{
						switch(RegisterDragDrop(hwnd, droptarget))
						{
							case S_OK:
							case DRAGDROP_E_ALREADYREGISTERED: // Hmm.
								break;
							
							default:
								droptarget = null;
								throw new DflException("Unable to register drag-drop.");
						}
					}
				}
			}
			else
			{
				delete droptarget;
				droptarget = null;
				RevokeDragDrop(hwnd);
			}
		}
		
		
		bool allowDrop() // getter
		{
			/+
			return (_exStyle() & WS_EX_ACCEPTFILES) != 0;
			+/
			
			return droptarget != null;
		}
	}
	
	
	deprecated void anchor(AnchorStyles a) // setter
	{
		/+
		anch = a;
		if(!(anch & (AnchorStyles.LEFT | AnchorStyles.RIGHT)))
			anch |= AnchorStyles.LEFT;
		if(!(anch & (AnchorStyles.TOP | AnchorStyles.BOTTOM)))
			anch |= AnchorStyles.TOP;
		+/
		
		sdock = cast(DockStyle)sdock.init; // Can't be set at the same time.
	}
	
	
	deprecated AnchorStyles anchor() // getter
	{
		/+
		return anch;
		+/
		return cast(AnchorStyles)(AnchorStyles.LEFT | AnchorStyles.TOP);
	}
	
	
	private void deleteThisBackgroundBrush()
	{
		DeleteObject(hbrBg);
	}
	
	
	protected void onBackColorChanged(EventArgs ea)
	{
		backColorChanged(this, ea);
	}
	
	
	void backColor(Color c) // setter
	{
		if(backc == c)
			return;
		
		if(created)
		{
			deleteThisBackgroundBrush();
			
			hbrBg = c.createBrush();
			backc = c;
			
			invalidate(true); // Redraw!
			
			onBackColorChanged(EventArgs.empty);
		}
		else
		{
			backc = c;
		}
	}
	
	
	Color backColor() // getter
	{
		return backc;
	}
	
	
	final int bottom() // getter
	{
		return wrect.bottom;
	}
	
	
	final void bounds(Rect r) // setter
	{
		setBoundsCore(r.x, r.y, r.width, r.height, BoundsSpecified.ALL);
	}
	
	
	final Rect bounds() // getter
	{
		return wrect;
	}
	
	
	protected void setBoundsCore(int x, int y, int width, int height, BoundsSpecified specified)
	{
		// Make sure at least one flag is set.
		//if(!(specified & BoundsSpecified.ALL))
		if(!specified)
			return;
		
		if(created)
		{
			UINT swpf = SWP_NOZORDER | SWP_NOACTIVATE | SWP_NOOWNERZORDER | SWP_NOMOVE | SWP_NOSIZE;
			
			if(specified & BoundsSpecified.X)
			{
				if(!(specified & BoundsSpecified.Y))
					y = this.top();
				swpf &= ~SWP_NOMOVE;
			}
			else if(specified & BoundsSpecified.Y)
			{
				x = this.left();
				swpf &= ~SWP_NOMOVE;
			}
			
			if(specified & BoundsSpecified.WIDTH)
			{
				if(!(specified & BoundsSpecified.HEIGHT))
					height = this.height();
				swpf &= ~SWP_NOSIZE;
			}
			else if(specified & BoundsSpecified.HEIGHT)
			{
				width = this.width();
				swpf &= ~SWP_NOSIZE;
			}
			
			SetWindowPos(hwnd, null, x, y, width, height, swpf);
			// Window events will update -wrect-.
		}
		else
		{
			if(specified & BoundsSpecified.X)
				wrect.x = x;
			if(specified & BoundsSpecified.Y)
				wrect.y = y;
			if(specified & BoundsSpecified.WIDTH)
			{
				if(width < 0)
					width = 0;
				
				wrect.width = width;
				wclientsz.width = width;
			}
			if(specified & BoundsSpecified.HEIGHT)
			{
				if(height < 0)
					height = 0;
				
				wrect.height = height;
				wclientsz.height = height;
			}
		}
	}
	
	
	final bool canFocus() // getter
	{
		LONG wl = _style();
		return /+ hwnd && +/ (wl & WS_VISIBLE) && !(wl & WS_DISABLED);
	}
	
	
	final bool canSelect() // getter
	{
		// All parent controls need to be visible and enabled, too.
		
		return (ctrlStyle & ControlStyles.SELECTABLE) &&
			IsWindowVisible(hwnd) && IsWindowEnabled(hwnd);
	}
	
	
	// Returns true if this control has the mouse capture.
	final bool capture() // getter
	{
		return created() && hwnd == GetCapture();
	}
	
	
	final void capture(bool cyes) // setter
	{
		if(cyes)
			SetCapture(hwnd);
		else
			ReleaseCapture();
	}
	
	
	// When true, validating and validated events are fired when the control
	// receives focus. Typically set to false for controls such as a Help button.
	// Default is true.
	final bool causesValidation() // getter
	{
		return cvalidation;
	}
	
	
	protected void onCausesValidationChanged(EventArgs ea)
	{
		causesValidationChanged(this, ea);
	}
	
	
	final void causesValidation(bool vyes) // setter
	{
		if(cvalidation == vyes)
			return;
		
		cvalidation = vyes;
		
		onCausesValidationChanged(EventArgs.empty);
	}
	
	
	final Rect clientRectangle() // getter
	{
		return Rect(Point(0, 0), wclientsz);
	}
	
	
	final bool contains(Control ctrl)
	{
		return ccollection.contains(ctrl);
	}
	
	
	final Size clientSize() // getter
	{
		return wclientsz;
	}
	
	
	final void clientSize(Size sz) // setter
	{
		setClientSizeCore(sz.width, sz.height);
	}
	
	
	protected void setClientSizeCore(int width, int height)
	{
		/+
		if(created)
			setBoundsCore(0, 0, width, height, BoundsSpecified.SIZE);
		
		//wclientsz = Size(width, height);
		+/
		
		RECT r;
		
		r.left = 0;
		r.top = 0;
		r.right = width;
		r.bottom = height;
		
		AdjustWindowRectEx(&r, _style(), FALSE, _exStyle());
		
		setBoundsCore(0, 0, r.right - r.left, r.bottom - r.top, BoundsSpecified.SIZE);
	}
	
	
	// This window or one of its children has focus.
	final bool containsFocus() // getter
	{
		if(!created)
			return false;
		
		return IsChild(hwnd, GetFocus()) != FALSE;
	}
	
	
	ContextMenu contextMenu() // getter
	{
		return cmenu;
	}
	
	
	protected void onContextMenuChanged(EventArgs ea)
	{
		contextMenuChanged(this, ea);
	}
	
	
	void contextMenu(ContextMenu menu) // setter
	{
		if(cmenu is menu)
			return;
		
		cmenu = menu;
		
		if(created)
		{
			onContextMenuChanged(EventArgs.empty);
		}
	}
	
	
	final ControlCollection controls() // getter
	{
		//return new ControlCollection(this);
		return ccollection;
	}
	
	
	final bool created() // getter
	{
		// return IsWindow(hwnd) != FALSE;
		return hwnd != null; // && !ccollection.children.length ?
	}
	
	
	protected void onCursorChanged(EventArgs ea)
	{
		cursorChanged(this, ea);
	}
	
	
	void cursor(Cursor cur) // setter
	{
		if(cur == wcurs)
			return;
		
		if(created)
		{
			wcurs = cur;
			
			onCursorChanged(EventArgs.empty);
		}
		else
		{
			wcurs = cur;
		}
	}
	
	
	Cursor cursor() // getter
	{
		return wcurs;
	}
	
	
	static Color defaultBackColor() // getter
	{
		return Color.systemColor(COLOR_BTNFACE);
	}
	
	
	static Color defaultForeColor() //getter
	{
		return Color.systemColor(COLOR_BTNTEXT);
	}
	
	
	static Font defaultFont() // getter
	{
		return new Font(cast(HFONT)GetStockObject(DEFAULT_GUI_FONT), false);
	}
	
	
	Rect displayRectangle() // getter
	{
		return clientRectangle;
	}
	
	
	protected void onDockChanged(EventArgs ea)
	{
		alayout(this);
		
		dockChanged(this, ea);
	}
	
	
	DockStyle dock() // getter
	{
		return sdock;
	}
	
	
	void dock(DockStyle ds) // setter
	{
		if(ds == sdock)
			return;
		
		sdock = ds;
		/+
		anch = cast(AnchorStyles)anch.init; // Can't be set at the same time.
		+/
		
		if(created)
		{
			onDockChanged(EventArgs.empty);
		}
	}
	
	
	final void enabled(bool byes) // setter
	{
		if(created)
		{
			if(byes == enabled)
				return;
			
			EnableWindow(hwnd, byes);
			// Window events will update -wstyle-.
		}
		else
		{
			if(byes)
				wstyle &= ~WS_DISABLED;
			else
				wstyle |= WS_DISABLED;
		}
	}
	
	
	final bool enabled() // getter
	{
		/*
		return IsWindowEnabled(hwnd) ? true : false;
		*/
		
		return (wstyle & WS_DISABLED) == 0;
	}
	
	
	bool focused() // getter
	{
		return created() && hwnd == GetFocus();
	}
	
	
	void font(Font f) // setter
	{
		if(wfont is f)
			return;
		
		wfont = f;
		
		if(created)
			SendMessageA(hwnd, WM_SETFONT, cast(WPARAM)f.handle, MAKELPARAM(true, 0));
		
		onFontChanged(EventArgs.empty);
	}
	
	
	Font font() // getter
	{
		return wfont;
	}
	
	
	protected void onForeColorChanged(EventArgs ea)
	{
		foreColorChanged(this, ea);
	}
	
	
	void foreColor(Color c) // setter
	{
		if(c == forec)
			return;
		
		forec = c;
		
		if(created)
		{
			invalidate(true);
			
			onForeColorChanged(EventArgs.empty);
		}
	}
	
	
	Color foreColor() // getter
	{
		return forec;
	}
	
	
	// Doesn't cause a ControlCollection to be constructed so
	// it could improve performance when walking through children.
	final bool hasChildren() // getter
	{
		//return created() && GetWindow(hwnd, GW_CHILD) != null;
		
		if(created)
		{
			return GetWindow(hwnd, GW_CHILD) != null;
		}
		else
		{
			return ccollection.children.length != 0;
		}
	}
	
	
	final void height(int h) // setter
	{
		/*
		RECT rect;
		GetWindowRect(hwnd, &rect);
		SetWindowPos(hwnd, null, 0, 0, rect.right - rect.left, h, SWP_NOACTIVATE | SWP_NOZORDER | SWP_NOMOVE);
		*/
		
		setBoundsCore(0, 0, 0, h, BoundsSpecified.HEIGHT);
	}
	
	
	final int height() // getter
	{
		return wrect.height;
	}
	
	
	final bool isHandleCreated()
	{
		return hwnd != null;
	}
	
	
	final void left(int l) // setter
	{
		/*
		RECT rect;
		GetWindowRect(hwnd, &rect);
		SetWindowPos(hwnd, null, l, rect.top, 0, 0, SWP_NOACTIVATE | SWP_NOZORDER | SWP_NOSIZE);
		*/
		
		setBoundsCore(l, 0, 0, 0, BoundsSpecified.X);
	}
	
	
	final int left() // getter
	{
		return wrect.x;
	}
	
	
	final void location(Point pt) // setter
	{
		/*
		SetWindowPos(hwnd, null, pt.x, pt.y, 0, 0, SWP_NOACTIVATE | SWP_NOZORDER | SWP_NOSIZE);
		*/
		
		setBoundsCore(pt.x, pt.y, 0, 0, BoundsSpecified.LOCATION);
	}
	
	
	final Point location() // getter
	{
		return wrect.location;
	}
	
	
	static Keys modifierKeys() // getter
	{
		// Is there a better way to do this?
		Keys ks = Keys.NONE;
		if(GetAsyncKeyState(VK_SHIFT) & 0x8000)
			ks |= Keys.SHIFT;
		if(GetAsyncKeyState(VK_MENU) & 0x8000)
			ks |= Keys.ALT;
		if(GetAsyncKeyState(VK_CONTROL) & 0x8000)
			ks|= Keys.CONTROL;
		return ks;
	}
	
	
	static MouseButtons mouseButtons() // getter
	{
		MouseButtons result;
		
		result = MouseButtons.NONE;
		if(GetSystemMetrics(SM_SWAPBUTTON))
		{
			if(GetAsyncKeyState(VK_LBUTTON) & 0x8000)
				result |= MouseButtons.RIGHT; // Swapped.
			if(GetAsyncKeyState(VK_RBUTTON) & 0x8000)
				result |= MouseButtons.LEFT; // Swapped.
		}
		else
		{
			if(GetAsyncKeyState(VK_LBUTTON) & 0x8000)
				result |= MouseButtons.LEFT;
			if(GetAsyncKeyState(VK_RBUTTON) & 0x8000)
				result |= MouseButtons.RIGHT;
		}
		if(GetAsyncKeyState(VK_MBUTTON) & 0x8000)
			result |= MouseButtons.MIDDLE;
		
		return result;
	}
	
	
	static Point mousePosition() // getter
	{
		Point pt;
		GetCursorPos(&pt.point);
		return pt;
	}
	
	
	// Name is usually the name used in code.
	// Probably used with debugging.
	
	final void name(char[] txt) // setter
	{
		formname = txt;
	}
	
	
	final char[] name() // getter
	{
		return formname;
	}
	
	
	protected void onParentChanged(EventArgs ea)
	{
		parentChanged(this, ea);
	}
	
	
	final void parent(Control c) // setter
	{
		if(c is wparent)
			return;
		
		if(!(_style() & WS_CHILD) || (_exStyle() & WS_EX_MDICHILD))
			throw new DflException("Cannot add a top level control to a control.");
		
		if(c)
		{
			wparent = c;
			
			// I want the destroy notification.
			c._exStyle(c._exStyle() & ~WS_EX_NOPARENTNOTIFY);
			
			if(c.created)
			{
				if(created)
				{
					SetParent(hwnd, c.hwnd);
				}
				else
				{
					// If the parent is created, create me!
					createControl();
				}
				
				onParentChanged(EventArgs.empty);
				c.onControlAdded(new ControlEventArgs(this));
				
				initLayout();
			}
			else
			{
				// If the parent exists and isn't created, need to add
				// -this- to its children array.
				c.ccollection.children ~= this;
			}
		}
		else
		{
			wparent = c;
			
			if(created)
				SetParent(hwnd, null);
		}
	}
	
	
	private final Control _fetchParent()
	{
		HWND hwParent = GetParent(hwnd);
		return fromHandle(hwParent);
	}
	
	
	final Control parent() // getter
	{
		return wparent;
	}
	
	
	// TODO: check implementation.
	private static HRGN dupHrgn(HRGN hrgn)
	{
		HRGN rdup = CreateRectRgn(0, 0, 1, 1);
		CombineRgn(rdup, hrgn, null, RGN_COPY);
		return rdup;
	}
	
	
	final void region(Region rgn) // setter
	{
		if(created)
		{
			// Need to make a copy of the region.
			SetWindowRgn(hwnd, dupHrgn(rgn.handle), true);
		}
		
		wregion = rgn;
	}
	
	
	private final Region _fetchRegion()
	{
		HRGN hrgn = CreateRectRgn(0, 0, 1, 1);
		GetWindowRgn(hwnd, hrgn);
		return new Region(hrgn); // Owned because GetWindowRgn() gives a copy.
	}
	
	
	final Region region() // getter
	{
		return wregion;
	}
	
	
	final int right() // getter
	{
		return wrect.right;
	}
	
	
	void rightToLeft(bool byes) // setter
	{
		LONG wl = _exStyle();
		if(byes)
			wl |= WS_EX_RTLREADING;
		else
			wl &= ~WS_EX_RTLREADING;
		_exStyle(wl);
	}
	
	
	bool rightToLeft() // getter
	{
		return (_exStyle() & WS_EX_RTLREADING) != 0;
	}
	
	
	final void size(Size sz) // setter
	{
		/*
		SetWindowPos(hwnd, null, 0, 0, sz.width, sz.height, SWP_NOACTIVATE | SWP_NOZORDER | SWP_NOMOVE);
		*/
		
		setBoundsCore(0, 0, sz.width, sz.height, BoundsSpecified.SIZE);
	}
	
	
	final Size size() // getter
	{
		return wrect.size; // struct Size, not sizeof.
	}
	
	
	/+
	final void tabIndex(int i) // setter
	{
		// TODO: ?
	}
	+/
	
	
	final int tabIndex() // getter
	{
		GetTabIndex gti;
		gti.hwnd = hwnd;
		EnumChildWindows(hwnd, &getTabIndexCallback, cast(LPARAM)&gti);
		return gti.index;
	}
	
	
	// True if control can be tabbed to.
	final void tabStop(bool byes) // setter
	{
		LONG wl = _style();
		if(byes)
			wl |= WS_TABSTOP;
		else
			wl &= ~WS_TABSTOP;
		_style(wl);
	}
	
	
	final bool tabStop() // getter
	{
		return (_style() & WS_TABSTOP) != 0;
	}
	
	
	final void tag(Object o) // setter
	{
		otag = o;
	}
	
	
	final Object tag() // getter
	{
		return otag;
	}
	
	
	void text(char[] txt) // setter
	{
		if(created)
		{
			dfl.utf.setWindowText(hwnd, txt);
		}
		else
		{
			wtext = txt;
		}
	}
	
	
	private final char[] _fetchText()
	{
		return dfl.utf.getWindowText(hwnd);
	}
	
	
	char[] text() // getter
	{
		if(created)
		{
			if(ctrlStyle & ControlStyles.CACHE_TEXT)
				return wtext;
			
			return _fetchText();
		}
		else
		{
			return wtext;
		}
	}
	
	
	final void top(int t) // setter
	{
		setBoundsCore(0, t, 0, 0, BoundsSpecified.Y);
	}
	
	
	final int top() // getter
	{
		return wrect.y;
	}
	
	
	// Returns the owner control that has no parent.
	// Returns this Control if no owner ?
	final Control topLevelControl() // getter
	{
		if(created)
		{
			HWND hwCurrent = hwnd;
			HWND hwParent;
			
			for(;;)
			{
				hwParent = (GetWindowLongA(hwCurrent, GWL_STYLE) & WS_CHILD) ? GetParent(hwCurrent) : GetWindow(hwCurrent, GW_OWNER);
				if(!hwParent)
					break;
				
				hwCurrent = hwParent;
			}
			
			return fromHandle(hwCurrent);
		}
		else
		{
			Control ctrl;
			ctrl = this;
			while(ctrl.parent)
			{
				ctrl = ctrl.parent;
			}
			return ctrl;
		}
	}
	
	
	final void visible(bool byes) // setter
	{
		setVisibleCore(byes);
	}
	
	
	private bool _fetchVisible()
	{
		return IsWindowVisible(hwnd) != FALSE;
	}
	
	
	final bool visible() // getter
	{
		return (wstyle & WS_VISIBLE) != 0;
	}
	
	
	final void width(int w) // setter
	{
		setBoundsCore(0, 0, w, 0, BoundsSpecified.WIDTH);
	}
	
	
	final int width() // getter
	{
		return wrect.width;
	}
	
	
	final void bringToFront()
	{
		//SetWindowPos(hwnd, HWND_TOP, 0, 0, 0, 0, SWP_NOMOVE | SWP_NOSIZE);
		BringWindowToTop(hwnd);
	}
	
	
	package final void createChildren()
	{
		assert(isHandleCreated);
		
		Control[] ctrls;
		ctrls = ccollection.children;
		ccollection.children = null;
		
		foreach(Control ctrl; ctrls)
		{
			assert(ctrl.parent is this);
			assert(!(ctrl is null));
			assert(ctrl);
			ctrl.createControl();
		}
	}
	
	
	// Force creation of the window and its child controls.
	final void createControl()
	{
		createHandle();
		
		// Called in WM_CREATE also.
		createChildren();
	}
	
	
	final Graphics createGraphics()
	{
		HDC hdc = GetDC(hwnd); // Or window DC?
		SetTextColor(hdc, forec.toRgb());
		return new CommonGraphics(hwnd, hdc);
	}
	
	
	version(NO_DRAG_DROP) {} else
	{
		private static class DropTarget: ComObject, IDropTarget
		{
			this(Control ctrl)
			{
				this.ctrl = ctrl;
			}
			
			
			extern(Windows):
			override HRESULT QueryInterface(IID* riid, void** ppv)
			{
				if(*riid == IID_IDropTarget)
				{
					*ppv = cast(IDropTarget)this;
					AddRef();
					return S_OK;
				}
				else if(*riid == IID_IUnknown)
				{
					*ppv = cast(IUnknown)this;
					AddRef();
					return S_OK;
				}
				else
				{
					*ppv = null;
					return E_NOINTERFACE;
				}
			}
			
			
			HRESULT DragEnter(dfl.wincom.IDataObject pDataObject, DWORD grfKeyState, POINTL pt, DWORD *pdwEffect)
			{
				HRESULT result;
				
				try
				{
					dataObj = new ComToDdataObject(pDataObject);
					
					DragEventArgs ea;
					ea = new DragEventArgs(dataObj, cast(int)grfKeyState, pt.x, pt.y, 
						cast(DragDropEffects)*pdwEffect, DragDropEffects.NONE); // ?
					ctrl.onDragEnter(ea);
					*pdwEffect = ea.effect;
					
					result = S_OK;
				}
				catch(Object e)
				{
					Application.onThreadException(e);
					
					result = E_UNEXPECTED;
				}
				
				return result;
			}
			
			
			HRESULT DragOver(DWORD grfKeyState, POINTL pt, DWORD *pdwEffect)
			{
				HRESULT result;
				
				try
				{
					DragEventArgs ea;
					ea = new DragEventArgs(dataObj, cast(int)grfKeyState, pt.x, pt.y, 
						cast(DragDropEffects)*pdwEffect, DragDropEffects.NONE); // ?
					ctrl.onDragOver(ea);
					*pdwEffect = ea.effect;
					
					result = S_OK;
				}
				catch(Object e)
				{
					Application.onThreadException(e);
					
					result = E_UNEXPECTED;
				}
				
				return result;
			}
			
			
			HRESULT DragLeave()
			{
				HRESULT result;
				
				try
				{
					ctrl.onDragLeave(EventArgs.empty);
					
					delete dataObj;
					dataObj = null;
					
					result = S_OK;
				}
				catch(Object e)
				{
					Application.onThreadException(e);
					
					result = E_UNEXPECTED;
				}
				
				return result;
			}
			
			
			HRESULT Drop(dfl.wincom.IDataObject pDataObject, DWORD grfKeyState, POINTL pt, DWORD *pdwEffect)
			{
				HRESULT result;
				
				try
				{
					DragEventArgs ea;
					ea = new DragEventArgs(dataObj, cast(int)grfKeyState, pt.x, pt.y, 
						cast(DragDropEffects)*pdwEffect, DragDropEffects.NONE); // ?
					ctrl.onDragDrop(ea);
					*pdwEffect = ea.effect;
					
					result = S_OK;
				}
				catch(Object e)
				{
					Application.onThreadException(e);
					
					result = E_UNEXPECTED;
				}
				
				return result;
			}
			
			
			private:
			Control ctrl;
			dfl.data.IDataObject dataObj;
		}
		
		
		protected void onDragLeave(EventArgs ea)
		{
			dragLeave(this, ea);
		}
		
		
		protected void onDragEnter(DragEventArgs ea)
		{
			dragEnter(this, ea);
		}
		
		
		protected void onDragOver(DragEventArgs ea)
		{
			dragOver(this, ea);
		}
		
		
		protected void onDragDrop(DragEventArgs ea)
		{
			dragDrop(this, ea);
		}
		
		
		private static class DropSource: ComObject, IDropSource
		{
			this(Control ctrl)
			{
				this.ctrl = ctrl;
				mbtns = Control.mouseButtons;
			}
			
			
			extern(Windows):
			override HRESULT QueryInterface(IID* riid, void** ppv)
			{
				if(*riid == IID_IDropSource)
				{
					*ppv = cast(IDropSource)this;
					AddRef();
					return S_OK;
				}
				else if(*riid == IID_IUnknown)
				{
					*ppv = cast(IUnknown)this;
					AddRef();
					return S_OK;
				}
				else
				{
					*ppv = null;
					return E_NOINTERFACE;
				}
			}
			
			
			HRESULT QueryContinueDrag(BOOL fEscapePressed, DWORD grfKeyState)
			{
				HRESULT result;
				
				try
				{
					QueryContinueDragEventArgs ea;
					DragAction act;
					
					if(fEscapePressed)
					{
						act = cast(DragAction)DragAction.CANCEL;
					}
					else
					{
						if(mbtns & MouseButtons.LEFT)
						{
							if(!(grfKeyState & MK_LBUTTON))
							{
								act = cast(DragAction)DragAction.DROP;
								goto qdoit;
							}
						}
						else
						{
							if(grfKeyState & MK_LBUTTON)
							{
								act = cast(DragAction)DragAction.CANCEL;
								goto qdoit;
							}
						}
						if(mbtns & MouseButtons.RIGHT)
						{
							if(!(grfKeyState & MK_RBUTTON))
							{
								act = cast(DragAction)DragAction.DROP;
								goto qdoit;
							}
						}
						else
						{
							if(grfKeyState & MK_RBUTTON)
							{
								act = cast(DragAction)DragAction.CANCEL;
								goto qdoit;
							}
						}
						if(mbtns & MouseButtons.MIDDLE)
						{
							if(!(grfKeyState & MK_MBUTTON))
							{
								act = cast(DragAction)DragAction.DROP;
								goto qdoit;
							}
						}
						else
						{
							if(grfKeyState & MK_MBUTTON)
							{
								act = cast(DragAction)DragAction.CANCEL;
								goto qdoit;
							}
						}
						
						act = cast(DragAction)DragAction.CONTINUE;
					}
					
					qdoit:
					ea = new QueryContinueDragEventArgs(cast(int)grfKeyState,
						fEscapePressed != FALSE, act); // ?
					ctrl.onQueryContinueDrag(ea);
					
					result = cast(HRESULT)ea.action;
				}
				catch(Object e)
				{
					Application.onThreadException(e);
					
					result = E_UNEXPECTED;
				}
				
				return result;
			}
			
			
			HRESULT GiveFeedback(DWORD dwEffect)
			{
				HRESULT result;
				
				try
				{
					GiveFeedbackEventArgs ea;
					
					ea = new GiveFeedbackEventArgs(cast(DragDropEffects)dwEffect, true);
					ctrl.onGiveFeedback(ea);
					
					result = ea.useDefaultCursors ? DRAGDROP_S_USEDEFAULTCURSORS : S_OK;
				}
				catch(Object e)
				{
					Application.onThreadException(e);
					
					result = E_UNEXPECTED;
				}
				
				return result;
			}
			
			
			private:
			Control ctrl;
			MouseButtons mbtns;
		}
		
		
		protected void onQueryContinueDrag(QueryContinueDragEventArgs ea)
		{
			queryContinueDrag(this, ea);
		}
		
		
		protected void onGiveFeedback(GiveFeedbackEventArgs ea)
		{
			giveFeedback(this, ea);
		}
		
		
		final DragDropEffects doDragDrop(dfl.data.IDataObject dataObj, DragDropEffects allowedEffects)
		{
			DWORD effect;
			DropSource dropsrc;
			dfl.wincom.IDataObject dropdata;
			
			dropsrc = new DropSource(this);
			dropdata = new DtoComDataObject(dataObj);
			switch(DoDragDrop(dropdata, dropsrc, cast(DWORD)allowedEffects, &effect))
			{
				case DRAGDROP_S_DROP: // All good.
					break;
				
				case DRAGDROP_S_CANCEL:
					return DragDropEffects.NONE; // ?
				
				default:
					throw new DflException("Unable to complete drag-drop operation.");
			}
			
			return cast(DragDropEffects)effect;
		}
		
		
		final DragDropEffects doDragDrop(Data obj, DragDropEffects allowedEffects)
		{
			dfl.data.IDataObject dd;
			dd = new DataObject;
			dd.setData(obj);
			return doDragDrop(dd, allowedEffects);
		}
	}
	
	
	override int opEquals(Object o)
	{
		Control ctrl = cast(Control)o;
		if(!ctrl)
			return 0; // Not equal.
		return opEquals(ctrl);
	}
	
	
	int opEquals(Control ctrl)
	{
		if(!created)
			return super.opEquals(ctrl);
		return hwnd == ctrl.hwnd;
	}
	
	
	final bool focus()
	{
		return SetFocus(hwnd) != null;
	}
	
	
	// Finds controls that own more than one handle.
	// A combo box has several HWNDs, this would return the
	// correct combo box control if any of those handles are
	// provided.
	static Control fromChildHandle(HWND hwChild)
	{
		// TODO: check if correct implementation.
		
		Control result;
		for(;;)
		{
			if(!hwChild)
				return null;
			
			result = fromHandle(hwChild);
			if(result)
				return result;
			
			hwChild = GetParent(hwChild);
		}
	}
	
	
	static Control fromHandle(HWND hw)
	{
		return Application.lookupHwnd(hw);
	}
	
	
	final Control getChildAtPoint(Point pt)
	{
		HWND hwChild = ChildWindowFromPoint(hwnd, pt.point);
		if(!hwChild)
			return null;
		return fromChildHandle(hwChild);
	}
	
	
	// Next control in the tab order relative to -ctrl-.
	final Control getNextControl(Control ctrl, bool forward)
	{
		// The tab order loops around.
		return fromHandle(GetNextDlgTabItem(hwnd, ctrl.hwnd, forward));
	}
	
	
	final void hide()
	{
		setVisibleCore(false);
	}
	
	
	package final void redrawEntire()
	{
		SetWindowPos(hwnd, null, 0, 0, 0, 0, SWP_FRAMECHANGED | SWP_DRAWFRAME | SWP_NOMOVE
			| SWP_NOSIZE | SWP_NOZORDER | SWP_NOACTIVATE);
	}
	
	
	final void invalidate()
	{
		if(!hwnd)
			return;
		
		RedrawWindow(hwnd, null, null, RDW_ERASE | RDW_INVALIDATE | RDW_NOCHILDREN);
	}
	
	
	final void invalidate(bool andChildren)
	{
		if(!hwnd)
			return;
		
		RedrawWindow(hwnd, null, null, RDW_ERASE | RDW_INVALIDATE | (andChildren ? RDW_ALLCHILDREN : RDW_NOCHILDREN));
	}
	
	
	final void invalidate(Rect r)
	{
		if(!hwnd)
			return;
		
		RECT rect;
		r.getRect(&rect);
		
		RedrawWindow(hwnd, &rect, null, RDW_ERASE | RDW_INVALIDATE | RDW_NOCHILDREN);
	}
	
	
	final void invalidate(Rect r, bool andChildren)
	{
		if(!hwnd)
			return;
		
		RECT rect;
		r.getRect(&rect);
		
		RedrawWindow(hwnd, &rect, null, RDW_ERASE | RDW_INVALIDATE | (andChildren ? RDW_ALLCHILDREN : RDW_NOCHILDREN));
	}
	
	
	final void invalidate(Region rgn)
	{
		if(!hwnd)
			return;
		
		RedrawWindow(hwnd, null, rgn.handle, RDW_ERASE | RDW_INVALIDATE | RDW_NOCHILDREN);
	}
	
	
	final void invalidate(Region rgn, bool andChildren)
	{
		if(!hwnd)
			return;
		
		RedrawWindow(hwnd, null, rgn.handle, RDW_ERASE | RDW_INVALIDATE | (andChildren ? RDW_ALLCHILDREN : RDW_NOCHILDREN));
	}
	
	
	// Returns true if the window does not belong to the current thread.
	bool invokeRequired() // getter
	{
		DWORD tid = GetWindowThreadProcessId(hwnd, null);
		return tid != GetCurrentThreadId();
	}
	
	
	private static void badInvokeHandle()
	{
		throw new DflException("Must invoke after creating handle.");
	}
	
	
	// Exceptions are propagated back to the caller of invoke().
	final Object invoke(Object delegate(Object[]) dg, Object[] args)
	{
		if(!hwnd)
			badInvokeHandle();
		
		InvokeData inv;
		inv.dg = dg;
		inv.args = args;
		
		if(LRESULT_DFL_INVOKE != SendMessageA(hwnd, wmDfl, WPARAM_DFL_INVOKE, cast(LRESULT)&inv))
			throw new DflException("Invoke failure.");
		if(inv.exception)
			throw inv.exception;
		
		return inv.result;
	}
	
	
	// Extra.
	// Exceptions will be passed to Application.onThreadException() and
	// trigger the threadException event or the default exception dialog.
	final void delayInvoke(void function() fn)
	{
		if(!hwnd)
			badInvokeHandle();
		
		static assert(fn.sizeof <= LPARAM.sizeof);
		PostMessageA(hwnd, wmDfl, WPARAM_DFL_DELAY_INVOKE, cast(LPARAM)fn);
	}
	
	
	static bool isMnemonic(char charCode, char[] text)
	{
		uint i;
		for(i = 0; i != text.length; i++)
		{
			if(text[i] == '&')
			{
				if(++i == text.length)
					break;
				return text[i] == charCode;
			}
		}
		return false;
	}
	
	
	/+
	// TODO: implement.
	// Force layout recalculation for all children.
	
	final void performLayout()
	{
		
	}
	
	
	final void performLayout(Control affectedControl)
	{
		
	}
	+/
	
	
	// Converts a screen Point to a client Point.
	final Point pointToClient(Point pt)
	{
		ScreenToClient(hwnd, &pt.point);
		return pt;
	}
	
	
	// Converts a client Point to a screen Point.
	final Point pointToScreen(Point pt)
	{
		ClientToScreen(hwnd, &pt.point);
		return pt;
	}
	
	
	// Return true if processed.
	bool preProcessMessage(inout Message msg)
	{
		return false;
	}
	
	
	// Converts a screen Rectangle to a client Rectangle.
	final Rect rectangleToClient(Rect r)
	{
		RECT rect;
		r.getRect(&rect);
		
		MapWindowPoints(null, hwnd, cast(POINT*)&rect, 2);
		return Rect(&rect);
	}
	
	
	// Converts a client Rectangle to a screen Rectangle.
	final Rect rectangleToScreen(Rect r)
	{
		RECT rect;
		r.getRect(&rect);
		
		MapWindowPoints(hwnd, null, cast(POINT*)&rect, 2);
		return Rect(&rect);
	}
	
	
	void refresh()
	{
		invalidate(true);
	}
	
	
	void resetBackColor()
	{
		backColor = defaultBackColor;
	}
	
	
	void resetCursor()
	{
		cursor = new Cursor(LoadCursorA(null, IDC_ARROW), false);
	}
	
	
	void resetFont()
	{
		font = defaultFont;
	}
	
	
	void resetForeColor()
	{
		foreColor = defaultForeColor;
	}
	
	
	void resetRightToLeft()
	{
		rightToLeft = false;
	}
	
	
	void resetText()
	{
		text = "";
	}
	
	// Just allow layout recalc, but don't do it right now.
	final void resumeLayout()
	{
		_allowLayout = true;
	}
	
	
	// Allow layout recalc, only do it now if -byes- is true.
	final void resumeLayout(bool byes)
	{
		_allowLayout = true;
		
		if(byes)
		//	onLayout(new LayoutEventArgs(null));
			alayout(null);
	}
	
	
	final void suspendLayout()
	{
		_allowLayout = false;
	}
	
	
	/+
	// TODO: implement.
	
	// Scale both height and width to -ratio-.
	final void scale(float ratio)
	{
		scaleCore(ratio, ratio);
	}
	
	
	// Scale -width- and -height- ratios.
	final void scale(float width, float height)
	{
		scaleCore(width, height);
	}
	
	
	// Also scales child controls recursively.
	protected void scaleCore(float width, float height)
	{
		suspendLayout();
		
		// ...
		
		resumeLayout();
	}
	+/
	
	
	final void select()
	{
		select(false, false);
	}
	
	
	// If -directed- is true, -forward- is used.
	// If -forward- is true, the next control in the tab order is selected,
	// otherwise the previous control in the tab order is selected.
	// Controls without style ControlStyles.SELECTABLE are skipped.
	protected void select(bool directed, bool forward)
	{
		if(!created)
			return;
		
		if(directed)
		{
			SendMessageA(hwnd, WM_NEXTDLGCTL, !forward, MAKELPARAM(false, 0));
		}
		else
		{
			SendMessageA(hwnd, WM_NEXTDLGCTL, cast(WPARAM)hwnd, MAKELPARAM(true, 0));
		}
	}
	
	
	/+
	// Controls without style ControlStyles.SELECTABLE are skipped.
	final bool selectNextControl(Control ctrlStart, bool forward, bool tabStopOnly, bool nested, bool wrap)
	{
		// TODO: implement. May need to manually loop through child windows.
	}
	+/
	
	
	final void sendToBack()
	{
		if(!created)
			return;
		
		SetWindowPos(hwnd, HWND_BOTTOM, 0, 0, 0, 0, SWP_NOMOVE | SWP_NOSIZE);
	}
	
	
	final void setBounds(int x, int y, int width, int height)
	{
		setBoundsCore(x, y, width, height, BoundsSpecified.ALL);
	}
	
	
	final void setBounds(int x, int y, int width, int height, BoundsSpecified specified)
	{
		setBoundsCore(x, y, width, height, specified);
	}
	
	
	final void show()
	{
		/*
		ShowWindow(hwnd, SW_SHOW);
		doShow();
		*/
		
		setVisibleCore(true);
	}
	
	
	override char[] toString()
	{
		return text;
	}
	
	
	final void update()
	{
		if(!created)
			return;
		
		UpdateWindow(hwnd);
	}
	
	
	void dispose()
	{
		dispose(true);
	}
	
	
	// If mouseEnter, mouseHover and mouseLeave events are supported.
	// Returns true on Windows 95 with IE 5.5, Windows 98+ or Windows NT 4.0+.
	static bool supportsMouseTracking() // getter
	{
		return trackMouseEvent != null;
	}
	
	
	package final Rect _fetchBounds()
	{
		RECT r;
		GetWindowRect(hwnd, &r);
		HWND hwParent = GetParent(hwnd);
		if(hwParent && (_style() & WS_CHILD))
			MapWindowPoints(null, hwParent, cast(POINT*)&r, 2);
		return Rect(&r);
	}
	
	
	package final Size _fetchClientSize()
	{
		RECT r;
		GetClientRect(hwnd, &r);
		return Size(r.right, r.bottom);
	}
	
	
	protected void onInvalidated(InvalidateEventArgs iea)
	{
		invalidated(this, iea);
	}
	
	
	protected void onPaint(PaintEventArgs pea)
	{
		paint(this, pea);
	}
	
	
	protected void onLocationChanged(EventArgs ea)
	{
		locationChanged(this, ea);
	}
	
	
	protected void onSizeChanged(EventArgs ea)
	{
		sizeChanged(this, ea);
	}
	
	
	protected void onResize(EventArgs ea)
	{
		resize(this, ea);
	}
	
	
	protected void onMouseEnter(MouseEventArgs mea)
	{
		mouseEnter(this, mea);
	}
	
	
	protected void onMouseMove(MouseEventArgs mea)
	{
		mouseMove(this, mea);
	}
	
	
	protected void onKeyDown(KeyEventArgs kea)
	{
		keyDown(this, kea);
	}
	
	
	protected void onKeyPress(KeyEventArgs kea)
	{
		keyPress(this, kea);
	}
	
	
	protected void onKeyUp(KeyEventArgs kea)
	{
		keyUp(this, kea);
	}
	
	
	protected void onMouseWheel(MouseEventArgs mea)
	{
		mouseWheel(this, mea);
	}
	
	
	protected void onMouseHover(MouseEventArgs mea)
	{
		mouseHover(this, mea);
	}
	
	
	protected void onMouseLeave(MouseEventArgs mea)
	{
		mouseLeave(this, mea);
	}
	
	
	protected void onMouseDown(MouseEventArgs mea)
	{
		mouseDown(this, mea);
	}
	
	
	protected void onMouseUp(MouseEventArgs mea)
	{
		mouseUp(this, mea);
	}
	
	
	protected void onClick(EventArgs ea)
	{
		click(this, ea);
	}
	
	
	protected void onDoubleClick(EventArgs ea)
	{
		doubleClick(this, ea);
	}
	
	
	protected void onEnter(EventArgs ea)
	{
		enter(this, ea);
	}
	
	
	protected void onGotFocus(EventArgs ea)
	{
		gotFocus(this, ea);
	}
	
	
	protected void onLeave(EventArgs ea)
	{
		leave(this, ea);
	}
	
	
	protected void onValidated(EventArgs ea)
	{
		validated(this, ea);
	}
	
	
	protected void onValidating(CancelEventArgs cea)
	{
		foreach(CancelEventHandler.Handler handler; validating.handlers())
		{
			handler(this, cea);
			
			if(cea.cancel)
				return; // Not validated.
		}
		
		onValidated(EventArgs.empty);
	}
	
	
	protected void onLostFocus(EventArgs ea)
	{
		lostFocus(this, ea);
	}
	
	
	protected void onEnabledChanged(EventArgs ea)
	{
		enabledChanged(this, ea);
	}
	
	
	protected void onTextChanged(EventArgs ea)
	{
		textChanged(this, ea);
	}
	
	
	protected void onFontChanged(EventArgs ea)
	{
		fontChanged(this, ea);
	}
	
	
	protected void onRightToLeftChanged(EventArgs ea)
	{
		rightToLeftChanged(this, ea);
	}
	
	
	protected void onVisibleChanged(EventArgs ea)
	{
		visibleChanged(this, ea);
	}
	
	
	protected void onHelpRequested(HelpEventArgs hea)
	{
		helpRequested(this, hea);
	}
	
	
	protected void onSystemColorsChanged(EventArgs ea)
	{
		systemColorsChanged(this, ea);
	}
	
	
	protected void onHandleCreated(EventArgs ea)
	{
		if(wfont)
			SendMessageA(hwnd, WM_SETFONT, cast(WPARAM)wfont.handle, 0);
		
		if(wregion)
		{
			// Need to make a copy of the region.
			SetWindowRgn(hwnd, dupHrgn(wregion.handle), true);
		}
		
		version(NO_DRAG_DROP) {} else
		{
			if(droptarget)
			{
				if(S_OK != RegisterDragDrop(hwnd, droptarget))
				{
					droptarget = null;
					throw new DflException("Unable to register drag-drop.");
				}
			}
		}
		
		handleCreated(this, ea);
	}
	
	
	protected void onHandleDestroyed(EventArgs ea)
	{
		handleDestroyed(this, ea);
	}
	
	
	protected void onPaintBackground(PaintEventArgs pea)
	{
		RECT rect;
		pea.clipRectangle.getRect(&rect);
		FillRect(pea.graphics.handle, &rect, hbrBg);
	}
	
	
	protected void onMove(EventArgs ea)
	{
		move(this, ea);
	}
	
	
	private static MouseButtons wparamMouseButtons(WPARAM wparam)
	{
		MouseButtons result;
		if(wparam & MK_LBUTTON)
			result |= MouseButtons.LEFT;
		if(wparam & MK_RBUTTON)
			result |= MouseButtons.RIGHT;
		if(wparam & MK_MBUTTON)
			result |= MouseButtons.MIDDLE;
		return result;
	}
	
	
	package final void prepareDc(HDC hdc)
	{
		// SetBkMode(hdc, );
		SetBkColor(hdc, backc.toRgb());
		SetTextColor(hdc, forec.toRgb());
	}
	
	
	// Message copy so it cannot be modified.
	protected void onNotifyMessage(Message msg)
	{
	}
	
	
	/+package+/ LRESULT customMsg(inout CustomMsg msg) // package
	{
		return 0;
	}
	
	
	protected void wndProc(inout Message msg)
	{
		if(ctrlStyle & ControlStyles.ENABLE_NOTIFY_MESSAGE)
			onNotifyMessage(msg);
		
		switch(msg.msg)
		{
			case WM_PAINT:
				{
					// This can't be done in BeginPaint() becuase part might get
					// validated during this event ?
					RECT uprect;
					GetUpdateRect(hwnd, &uprect, true);
					onInvalidated(new InvalidateEventArgs(Rect(&uprect)));
					
					PAINTSTRUCT ps;
					BeginPaint(msg.hWnd, &ps);
					try
					{
						//onInvalidated(new InvalidateEventArgs(Rect(&uprect)));
						
						PaintEventArgs pea = new PaintEventArgs(new Graphics(ps.hdc, false), Rect(&ps.rcPaint));
						
						// Probably because ControlStyles.ALL_PAINTING_IN_WM_PAINT.
						if(ps.fErase)
						{
							prepareDc(ps.hdc);
							onPaintBackground(pea);
						}
						
						prepareDc(ps.hdc);
						onPaint(pea);
					}
					finally
					{
						EndPaint(hwnd, &ps);
					}
				}
				return;
			
			case WM_ERASEBKGND:
				if(ctrlStyle & ControlStyles.OPAQUE)
				{
					msg.result = 1; // Erased.
				}
				else if(!(ctrlStyle & ControlStyles.ALL_PAINTING_IN_WM_PAINT))
				{
					RECT uprect;
					/+
					GetUpdateRect(hwnd, &uprect, false);
					+/
					uprect.left = 0;
					uprect.top = 0;
					uprect.right = clientSize.width;
					uprect.bottom = clientSize.height;
					
					prepareDc(cast(HDC)msg.wParam);
					onPaintBackground(new PaintEventArgs(new Graphics(cast(HDC)msg.wParam, false), Rect(&uprect)));
					msg.result = 1; // Erased.
				}
				return;
			
			case WM_PRINTCLIENT:
				prepareDc(cast(HDC)msg.wParam);
				onPaint(new PaintEventArgs(new Graphics(cast(HDC)msg.wParam, false), Rect(Point(0, 0), wclientsz)));
				return;
			
			// TODO: fix?
			// These should probably not be handled by the parent.
			// I can't find a better way yet, because hooks can't directly return values.
			case WM_CTLCOLORSTATIC:
			case WM_CTLCOLORLISTBOX:
			case WM_CTLCOLOREDIT:
			case WM_CTLCOLORSCROLLBAR:
			case WM_CTLCOLORBTN:
			//case WM_CTLCOLORDLG: // ?
				{
					Control ctrl = fromChildHandle(cast(HWND)msg.lParam);
					if(ctrl)
					{
						ctrl.prepareDc(cast(HDC)msg.wParam);
						msg.result = cast(LRESULT)ctrl.hbrBg;
						return;
					}
				}
				break;
			
			case WM_WINDOWPOSCHANGED:
				{
					WINDOWPOS* wp = cast(WINDOWPOS*)msg.lParam;
					bit needLayout = false;
					
					if(!(wp.flags & (SWP_NOZORDER | SWP_NOOWNERZORDER)))
					{
						if(wparent)
							wparent.vchanged();
					}
					
					if(!(wp.flags & SWP_NOMOVE))
					{
						onMove(EventArgs.empty);
						onLocationChanged(EventArgs.empty);
					}
					
					if(!(wp.flags & SWP_NOSIZE))
					{
						// I think...
						onResize(EventArgs.empty);
						onSizeChanged(EventArgs.empty);
						
						needLayout = true;
					}
					
					// Frame change results in a new client size.
					if(wp.flags & SWP_FRAMECHANGED)
					{
						needLayout = true;
					}
					
					if(needLayout)
						alayout(this);
					
					if(wparent)
						wparent.alayout(this);
					
					if(wp.flags & (SWP_HIDEWINDOW | SWP_SHOWWINDOW))
					{
						onVisibleChanged(EventArgs.empty);
					}
				}
				break;
			
			case WM_MOUSEMOVE:
				if(_clicking)
				{
					if(!(msg.wParam & MK_LBUTTON))
						_clicking = false;
				}
				
				if(trackMouseEvent) // Requires Windows 95 with IE 5.5, 98 or NT4.
				{
					if(!menter)
					{
						menter = true;
						
						POINT pt;
						GetCursorPos(&pt);
						MapWindowPoints(null, hwnd, &pt, 1);
						onMouseEnter(new MouseEventArgs(wparamMouseButtons(msg.wParam), 0, pt.x, pt.y, 0));
						
						TRACKMOUSEEVENT tme;
						tme.cbSize = TRACKMOUSEEVENT.sizeof;
						tme.dwFlags = TME_HOVER | TME_LEAVE;
						tme.hwndTrack = msg.hWnd;
						tme.dwHoverTime = HOVER_DEFAULT;
						trackMouseEvent(&tme);
					}
				}
				
				onMouseMove(new MouseEventArgs(wparamMouseButtons(msg.wParam), 0, cast(short)LOWORD(msg.lParam), cast(short)HIWORD(msg.lParam), 0));
				break;
			
			case WM_SETCURSOR:
				// Just update it so that Control.defWndProc() can set it correctly.
				if(cast(HWND)msg.wParam == hwnd)
				{
					if(wcurs)
					{
						if(cast(HCURSOR)GetClassLongA(hwnd, GCL_HCURSOR) != wcurs.handle)
							SetClassLongA(hwnd, GCL_HCURSOR, cast(LONG)wcurs.handle);
					}
					else
					{
						if(cast(HCURSOR)GetClassLongA(hwnd, GCL_HCURSOR) != null)
							SetClassLongA(hwnd, GCL_HCURSOR, cast(LONG)cast(HCURSOR)null);
					}
					Control.defWndProc(msg);
					return;
				}
				break;
			
			case WM_KEYDOWN:
				{
					KeyEventArgs kea = new KeyEventArgs(cast(Keys)(msg.wParam | modifierKeys));
					
					onKeyDown(kea);
					
					ushort repeat = msg.lParam & 0xFFFF; // First 16 bits.
					for(; repeat; repeat--)
					{
						onKeyPress(kea);
					}
				}
				break;
			
			case WM_KEYUP:
				// Repeat count is always 1 for key up.
				onKeyUp(new KeyEventArgs(cast(Keys)(msg.wParam | modifierKeys)));
				break;
			
			case WM_MOUSEWHEEL: // Requires Windows 98 or NT4.
				onMouseWheel(new MouseEventArgs(wparamMouseButtons(LOWORD(msg.wParam)), 0, cast(short)LOWORD(msg.lParam), cast(short)HIWORD(msg.lParam), cast(short)HIWORD(msg.wParam)));
				break;
			
			case WM_MOUSEHOVER: // Requires Windows 95 with IE 5.5, 98 or NT4.
				onMouseHover(new MouseEventArgs(wparamMouseButtons(msg.wParam), 0, cast(short)LOWORD(msg.lParam), cast(short)HIWORD(msg.lParam), 0));
				break;
			
			case WM_MOUSELEAVE: // Requires Windows 95 with IE 5.5, 98 or NT4.
				{
					menter = false;
					
					POINT pt;
					GetCursorPos(&pt);
					MapWindowPoints(null, hwnd, &pt, 1);
					onMouseLeave(new MouseEventArgs(wparamMouseButtons(msg.wParam), 0, pt.x, pt.y, 0));
				}
				break;
			
			case WM_LBUTTONDOWN:
				_clicking = true;
				
				onMouseDown(new MouseEventArgs(MouseButtons.LEFT, 0, cast(short)LOWORD(msg.lParam), cast(short)HIWORD(msg.lParam), 0));
				break;
			
			case WM_RBUTTONDOWN:
				onMouseDown(new MouseEventArgs(MouseButtons.RIGHT, 0, cast(short)LOWORD(msg.lParam), cast(short)HIWORD(msg.lParam), 0));
				break;
			
			case WM_MBUTTONDOWN:
				onMouseDown(new MouseEventArgs(MouseButtons.MIDDLE, 0, cast(short)LOWORD(msg.lParam), cast(short)HIWORD(msg.lParam), 0));
				break;
			
			case WM_LBUTTONUP:
				{
					if(msg.lParam == -1)
						break;
					
					// Use temp in case of exception.
					bit wasClicking = _clicking;
					_clicking = false;
					
					MouseEventArgs mea = new MouseEventArgs(MouseButtons.LEFT, 1, cast(short)LOWORD(msg.lParam), cast(short)HIWORD(msg.lParam), 0);
					onMouseUp(mea);
					
					if(wasClicking && (ctrlStyle & ControlStyles.STANDARD_CLICK))
					{
						// See if the mouse up was over the control.
						if(Rect(0, 0, wclientsz.width, wclientsz.height).contains(mea.x, mea.y))
						{
							// Now make sure there's no child in the way.
							if(ChildWindowFromPoint(hwnd, Point(mea.x, mea.y).point) == hwnd)
								onClick(EventArgs.empty);
						}
					}
				}
				break;
			
			version(CUSTOM_MSG_HOOK)
			{}
			else
			{
				case WM_DRAWITEM:
					{
						Control ctrl;
						
						DRAWITEMSTRUCT* dis = cast(DRAWITEMSTRUCT*)msg.lParam;
						if(dis.CtlType == ODT_MENU)
						{
							// dis.hwndItem is the HMENU.
						}
						else
						{
							ctrl = Control.fromChildHandle(dis.hwndItem);
							if(ctrl)
							{
								msg.result = ctrl.customMsg(*(cast(CustomMsg*)&msg));
								return;
							}
						}
					}
					break;
				
				case WM_MEASUREITEM:
					{
						Control ctrl;
						
						MEASUREITEMSTRUCT* mis = cast(MEASUREITEMSTRUCT*)msg.lParam;
						if(!(mis.CtlType == ODT_MENU))
						{
							ctrl = Control.fromChildHandle(cast(HWND)mis.CtlID);
							if(ctrl)
							{
								msg.result = ctrl.customMsg(*(cast(CustomMsg*)&msg));
								return;
							}
						}
					}
					break;
				
				case WM_COMMAND:
					{
						Control ctrl;
						
						ctrl = Control.fromChildHandle(cast(HWND)msg.lParam);
						if(ctrl)
						{
							msg.result = ctrl.customMsg(*(cast(CustomMsg*)&msg));
							return;
						}
						else
						{
							MenuItem m;
							
							m = cast(MenuItem)Application.lookupMenuID(LOWORD(msg.wParam));
							if(m)
							{
								msg.result = m.customMsg(*(cast(CustomMsg*)&msg));
								return;
							}
						}
					}
					break;
				
				case WM_NOTIFY:
					{
						Control ctrl;
						NMHDR* nmh;
						nmh = cast(NMHDR*)msg.lParam;
						
						ctrl = Control.fromChildHandle(nmh.hwndFrom);
						if(ctrl)
						{
							msg.result = ctrl.customMsg(*(cast(CustomMsg*)&msg));
							return;
						}
					}
					break;
				
				case WM_MENUSELECT:
					{
						UINT mflags;
						UINT uitem;
						int mid;
						MenuItem m;
						
						mflags = HIWORD(msg.wParam);
						uitem = LOWORD(msg.wParam); // Depends on the flags.
						
						if(mflags & MF_SYSMENU)
							break;
						
						if(mflags & MF_POPUP)
						{
							// -uitem- is an index.
							mid = GetMenuItemID(cast(HMENU)msg.lParam, uitem);
						}
						else
						{
							// -uitem- is the item identifier.
							mid = uitem;
						}
						
						m = cast(MenuItem)Application.lookupMenuID(mid);
						if(m)
						{
							msg.result = m.customMsg(*(cast(CustomMsg*)&msg));
							return;
						}
					}
					break;
				
				case WM_INITMENUPOPUP:
					if(HIWORD(msg.lParam))
					{
						// System menu.
					}
					else
					{
						MenuItem m;
						
						//m = cast(MenuItem)Application.lookupMenuID(GetMenuItemID(cast(HMENU)msg.wParam, LOWORD(msg.lParam)));
						m = cast(MenuItem)Application.lookupMenu(cast(HMENU)msg.wParam);
						if(m)
						{
							msg.result = m.customMsg(*(cast(CustomMsg*)&msg));
							return;
						}
					}
					break;
				
				case WM_INITMENU:
					{
						ContextMenu m;
						
						m = cast(ContextMenu)Application.lookupMenu(cast(HMENU)msg.wParam);
						if(m)
						{
							msg.result = m.customMsg(*(cast(CustomMsg*)&msg));
							return;
						}
					}
					break;
			}
			
			case WM_RBUTTONUP:
				onMouseUp(new MouseEventArgs(MouseButtons.RIGHT, 1, cast(short)LOWORD(msg.lParam), cast(short)HIWORD(msg.lParam), 0));
				break;
			
			case WM_MBUTTONUP:
				onMouseUp(new MouseEventArgs(MouseButtons.MIDDLE, 1, cast(short)LOWORD(msg.lParam), cast(short)HIWORD(msg.lParam), 0));
				break;
			
			case WM_LBUTTONDBLCLK:
				onMouseDown(new MouseEventArgs(MouseButtons.LEFT, 2, cast(short)LOWORD(msg.lParam), cast(short)HIWORD(msg.lParam), 0));
				
				if(ctrlStyle & (ControlStyles.STANDARD_CLICK | ControlStyles.STANDARD_DOUBLE_CLICK))
					onDoubleClick(EventArgs.empty);
				break;
			
			case WM_RBUTTONDBLCLK:
				onMouseDown(new MouseEventArgs(MouseButtons.RIGHT, 2, cast(short)LOWORD(msg.lParam), cast(short)HIWORD(msg.lParam), 0));
				break;
			
			case WM_MBUTTONDBLCLK:
				onMouseDown(new MouseEventArgs(MouseButtons.MIDDLE, 2, cast(short)LOWORD(msg.lParam), cast(short)HIWORD(msg.lParam), 0));
				break;
			
			case WM_SETFOCUS:
				onEnter(EventArgs.empty);
				
				onGotFocus(EventArgs.empty);
				
				// defWndProc* Form focuses a child.
				break;
			
			case WM_KILLFOCUS:
				_clicking = false;
				
				onLeave(EventArgs.empty);
				
				if(cvalidation)
					onValidating(new CancelEventArgs);
				
				onLostFocus(EventArgs.empty);
				break;
			
			case WM_ENABLE:
				onEnabledChanged(EventArgs.empty);
				
				// defWndProc*
				break;
			
			case WM_SETTEXT:
				defWndProc(msg);
				
				// Need to fetch it because cast(char*)lparam isn't always accessible ?
				// Should this go in _wndProc()? Need to defWndProc() first ?
				if(ctrlStyle & ControlStyles.CACHE_TEXT)
					wtext = _fetchText();
				
				onTextChanged(EventArgs.empty);
				return;
			
			case WM_SETFONT:
				// Don't replace -wfont- if it's the same one, beacuse the old Font
				// object will get garbage collected and probably delete the HFONT.
				
				//onFontChanged(EventArgs.empty);
				
				// defWndProc*
				return;
			
			case WM_STYLECHANGED:
				{
					//defWndProc(msg);
					
					STYLESTRUCT* ss = cast(STYLESTRUCT*)msg.lParam;
					
					if(msg.wParam == GWL_EXSTYLE)
					{
						DWORD exChanged = ss.styleOld ^ ss.styleNew;
						
						if(exChanged & WS_EX_RTLREADING)
							onRightToLeftChanged(EventArgs.empty);
					}
				}
				break;
			
			case WM_SHOWWINDOW:
				//defWndProc(msg);
				
				if(!msg.wParam) // Hiding.
					_clicking = false;
				
				//onVisibleChanged(EventArgs.empty); // Done in WM_WINDOWPOSCHANGED now.
				break;
			
			case WM_ACTIVATE:
				switch(LOWORD(msg.wParam))
				{
					case WA_INACTIVE:
						_clicking = false;
						break;
					
					default: ;
				}
				break;
			
			case WM_CONTEXTMENU:
				if(hwnd == cast(HWND)msg.wParam)
				{
					if(cmenu)
					{
						// Shift+F10 causes xPos and yPos to be -1.
						
						Point point;
						
						if(msg.lParam == -1)
							point = pointToScreen(Point(0, 0));
						else
							point = Point(cast(short)LOWORD(msg.lParam), cast(short)HIWORD(msg.lParam));
						
						SetFocus(handle); // ?
						cmenu.show(this, point);
						
						return;
					}
				}
				break;
			
			case WM_HELP:
				{
					HELPINFO* hi = cast(HELPINFO*)msg.lParam;
					
					onHelpRequested(new HelpEventArgs(Point(hi.MousePos.x, hi.MousePos.y)));
				}
				break;
			
			case WM_SYSCOLORCHANGE:
				onSystemColorsChanged(EventArgs.empty);
				
				// Need to send the message to children for some common controls to update properly.
				foreach(Control ctrl; ccollection)
				{
					SendMessageA(ctrl.handle, WM_SYSCOLORCHANGE, 0, 0);
				}
				break;
			
			case WM_PARENTNOTIFY:
				switch(LOWORD(msg.wParam))
				{
					case WM_DESTROY:
						Control ctrl = fromChildHandle(cast(HWND)msg.lParam);
						if(ctrl)
						{
							onControlRemoved(new ControlEventArgs(ctrl));
							
							// ?
							vchanged();
							//alayout(ctrl); // This is already being called from somewhere else..
						}
						break;
					
					/+
					case WM_CREATE:
						initLayout();
						break;
					+/
					
					default: ;
				}
				break;
			
			case WM_CREATE:
				if(wparent)
					initLayout(); // ?
				break;
			
			case WM_DESTROY:
				onHandleDestroyed(EventArgs.empty);
				break;
			
			default:
				//defWndProc(msg);
		}
		
		defWndProc(msg);
		
		if(msg.msg == WM_CREATE)
			onHandleCreated(EventArgs.empty);
	}
	
	
	void defWndProc(inout Message msg)
	{
		msg.result = DefWindowProcA(msg.hWnd, msg.msg, msg.wParam, msg.lParam);
	}
	
	
	// Always called right when destroyed before doing anything else.
	// hwnd is cleared after this step.
	void _destroying() // package
	{
		wparent = null; // ?
	}
	
	
	// This function must be called FIRST for EVERY message to this
	// window in order to keep the correct window state.
	// This function must not throw exceptions.
	package final void mustWndProc(inout Message msg)
	{
		if(needCalcSize)
		{
			needCalcSize = false;
			RECT crect;
			GetClientRect(msg.hWnd, &crect);
			wclientsz.width = crect.right;
			wclientsz.height = crect.bottom;
		}
		
		switch(msg.msg)
		{
			case WM_NCCALCSIZE:
				needCalcSize = true;
				break;
			
			case WM_WINDOWPOSCHANGED:
				{
					wstyle = GetWindowLongA(hwnd, GWL_STYLE);
					
					WINDOWPOS* wp = cast(WINDOWPOS*)msg.lParam;
					
					/+
					if(wp.flags & SWP_SHOWWINDOW)
						wstyle |= WS_VISIBLE;
					else if(wp.flags & SWP_HIDEWINDOW)
						wstyle &= ~WS_VISIBLE;
					+/
					
					//if(!(wp.flags & SWP_NOMOVE))
					//	wrect.location = Point(wp.x, wp.y);
					if(!(wp.flags & SWP_NOSIZE) || !(wp.flags & SWP_NOMOVE))
					{
						//wrect = _fetchBounds();
						wrect = Rect(wp.x, wp.y, wp.cx, wp.cy);
						wclientsz = _fetchClientSize();
					}
				}
				break;
			
			/+
			case WM_SETFONT:
				//wfont = _fetchFont();
				break;
			+/
			
			case WM_STYLECHANGED:
				{
					STYLESTRUCT* ss = cast(STYLESTRUCT*)msg.lParam;
					
					if(msg.wParam == GWL_STYLE)
						wstyle = ss.styleNew;
					else if(msg.wParam == GWL_EXSTYLE)
						wexstyle = ss.styleNew;
					
					/+
					wrect = _fetchBounds();
					wclientsz = _fetchClientSize();
					+/
				}
				break;
			
			case WM_SHOWWINDOW:
				if(cast(BOOL)msg.wParam)
				{
					wstyle |= WS_VISIBLE;
					
					try
					{
						createChildren(); // Might throw.
					}
					catch(Object e)
					{
						Application.onThreadException(e);
					}
				}
				else
				{
					wstyle &= ~WS_VISIBLE;
				}
				break;
			
			case WM_ENABLE:
				//if(IsWindowEnabled(hwnd))
				if(cast(BOOL)msg.wParam)
					wstyle &= ~WS_DISABLED;
				else
					wstyle |= WS_DISABLED;
				break;
			
			/+
			case WM_PARENTNOTIFY:
				switch(LOWORD(msg.wParam))
				{
					case WM_DESTROY:
						// ...
						break;
					
					default: ;
				}
				break;
			+/
			
			case WM_NCCREATE:
				{
					//hwnd = msg.hWnd;
					
					/+
					// Not using CREATESTRUCT for window bounds because it can contain
					// CW_USEDEFAULT and other magic values.
					
					CREATESTRUCTA* cs;
					cs = cast(CREATESTRUCTA*)msg.lParam;
					
					//wrect = Rect(cs.x, cs.y, cs.cx, cs.cy);
					+/
					
					wrect = _fetchBounds();
					wclientsz = _fetchClientSize();
				}
				break;
			
			case WM_CREATE:
				try
				{
					//hwnd = msg.hWnd;
					
					/+
					CREATESTRUCTA* cs;
					cs = cast(CREATESTRUCTA*)msg.lParam;
					
					// Done in WM_NCCREATE now.
					//wrect = _fetchBounds();
					wrect = Rect(cs.x, cs.y, cs.cx, cs.cy);
					wclientsz = _fetchClientSize();
					+/
					
					// If class style was changed, update.
					if(_fetchClassLong() != wclassStyle)
						SetClassLongA(hwnd, GCL_STYLE, wclassStyle);
					
					//finishCreating(msg.hWnd);
					
					if(!(ctrlStyle & ControlStyles.CACHE_TEXT))
						wtext = null;
					
					hbrBg = backc.createBrush();
					
					//createChildren(); // Might throw.
				}
				catch(Object e)
				{
					Application.onThreadException(e);
				}
				break;
			
			case WM_DESTROY:
				_destroying();
				if(!killing)
					fillRecreationData();
				Application.removeHwnd(hwnd);
				hwnd = null;
				break;
			
			default: ;
				/+
				if(msg.msg == wmDfl)
				{
					switch(msg.wParam)
					{
						case WPARAM_DFL_:
						
						default: ;
					}
				}
				+/
		}
	}
	
	
	package final void _wndProc(inout Message msg)
	{
		//mustWndProc(msg); // Done in dflWndProc() now.
		wndProc(msg);
	}
	
	
	package final void doShow()
	{
		//ShowWindow(hwnd, SW_SHOW);
		SetWindowPos(hwnd, HWND_TOP, 0, 0, 0, 0, SWP_NOSIZE | SWP_NOMOVE | SWP_SHOWWINDOW);
	}
	
	
	package final void doHide()
	{
		//ShowWindow(hwnd, SW_HIDE);
		SetWindowPos(hwnd, null, 0, 0, 0, 0, SWP_NOSIZE | SWP_NOMOVE | SWP_NOACTIVATE | SWP_HIDEWINDOW);
	}
	
	
	EventHandler backColorChanged;
	// EventHandler backgroundImageChanged;
	EventHandler causesValidationChanged;
	UICuesEventHandler changeUICues; // TODO: properly fire.
	EventHandler click;
	EventHandler contextMenuChanged;
	ControlEventHandler controlAdded;
	ControlEventHandler controlRemoved;
	EventHandler cursorChanged;
	EventHandler disposed;
	EventHandler dockChanged;
	EventHandler doubleClick;
	EventHandler enabledChanged;
	EventHandler enter; // Cascades up. TODO: fix implementation.
	EventHandler fontChanged;
	EventHandler foreColorChanged;
	EventHandler gotFocus; // After enter.
	EventHandler handleCreated;
	EventHandler handleDestroyed;
	HelpEventHandler helpRequested;
	InvalidateEventHandler invalidated;
	KeyEventHandler keyDown;
	KeyEventHandler keyPress;
	KeyEventHandler keyUp;
	LayoutEventHandler layout;
	EventHandler leave; // Cascades down. TODO: fix implementation.
	EventHandler locationChanged;
	EventHandler lostFocus;
	MouseEventHandler mouseDown;
	MouseEventHandler mouseEnter;
	MouseEventHandler mouseHover;
	MouseEventHandler mouseLeave;
	MouseEventHandler mouseMove;
	MouseEventHandler mouseUp;
	MouseEventHandler mouseWheel;
	EventHandler move;
	PaintEventHandler paint;
	EventHandler parentChanged;
	EventHandler resize;
	EventHandler rightToLeftChanged;
	EventHandler sizeChanged;
	// EventHandler styleChanged;
	EventHandler systemColorsChanged;
	// EventHandler tabIndexChanged;
	// EventHandler tabStopChanged;
	EventHandler textChanged;
	EventHandler validated;
	CancelEventHandler validating; // Once cancel is true, remaining events are suppressed (including validated).
	EventHandler visibleChanged;
	
	version(NO_DRAG_DROP) {} else
	{
		DragEventHandler dragDrop;
		DragEventHandler dragEnter;
		EventHandler dragLeave;
		DragEventHandler dragOver;
		GiveFeedbackEventHandler giveFeedback;
		QueryContinueDragEventHandler queryContinueDrag;
	}
	
	
	this()
	{
		wrect.size = defaultSize;
		backc = defaultBackColor;
		forec = defaultForeColor;
		wfont = defaultFont;
		
		wcurs = new Cursor(LoadCursorA(null, IDC_ARROW), false);
		
		ccollection = createControlsInstance();
	}
	
	
	this(char[] text)
	{
		this();
		wtext = text;
		
		ccollection = createControlsInstance();
	}
	
	
	this(Control cparent, char[] text)
	{
		this();
		wtext = text;
		parent = cparent;
		
		ccollection = createControlsInstance();
	}
	
	
	this(char[] text, int left, int top, int width, int height)
	{
		this();
		wtext = text;
		wrect = Rect(left, top, width, height);
		
		ccollection = createControlsInstance();
	}
	
	
	this(Control cparent, char[] text, int left, int top, int width, int height)
	{
		this();
		wtext = text;
		wrect = Rect(left, top, width, height);
		parent = cparent;
		
		ccollection = createControlsInstance();
	}
	
	
	/+
	// Used internally.
	this(HWND hwnd)
	in
	{
		assert(hwnd);
	}
	body
	{
		this.hwnd = hwnd;
		owned = false;
		
		ccollection = new ControlCollection(this);
	}
	+/
	
	
	~this()
	{
		debug(APP_PRINT)
			printf("~Control %p\n", cast(void*)this);
		
		//dispose(false);
		destroyHandle();
		deleteThisBackgroundBrush();
	}
	
	
	protected:
	Size defaultSize() // getter
	{
		return Size(0, 0);
	}
	
	
	/+
	// TODO: implement.
	EventHandlerList events() // getter
	{
	}
	+/
	
	
	/+
	// TODO: implement. Is this worth implementing?
	
	// Set to -1 to reset cache.
	final void fontHeight(int fh) // setter
	{
		
	}
	
	
	final int fontHeight() // getter
	{
		return fonth;
	}
	+/
	
	
	final void resizeRedraw(bool byes) // setter
	{
		LONG cl = _classStyle();
		if(byes)
			cl |= CS_HREDRAW | CS_VREDRAW;
		else
			cl &= ~(CS_HREDRAW | CS_VREDRAW);
		
		_classStyle(cl);
	}
	
	
	final bool resizeRedraw() // getter
	{
		return (_classStyle() & (CS_HREDRAW | CS_VREDRAW)) != 0;
	}
	
	
	ControlCollection createControlsInstance()
	{
		return new ControlCollection(this);
	}
	
	
	deprecated package final void createClassHandle(char[] className)
	{
		if(!wparent || !wparent.handle || killing)
		{
			create_err:
			throw new DflException("Control creation failure.");
		}
		
		// This is here because referencing wparent.handle might create me.
		//if(created)
		if(isHandleCreated)
			return;
		
		Application.creatingControl(this);
		hwnd = dfl.utf.createWindowEx(wexstyle, className, wtext, wstyle, wrect.x, wrect.y,
			wrect.width, wrect.height, wparent.handle, null, GetModuleHandleA(null), null);
		if(!hwnd)
			goto create_err;
	}
	
	
	// Override to change the creation parameters.
	// Be sure to call super.createParams() or all the create params will need to be filled.
	protected void createParams(inout CreateParams cp)
	{
		with(cp)
		{
			className = CONTROL_CLASSNAME;
			caption = wtext;
			param = null;
			//parent = wparent.handle;
			parent = wparent ? wparent.handle : null;
			menu = null;
			inst = GetModuleHandleA(null);
			x = wrect.x;
			y = wrect.y;
			width = wrect.width;
			height = wrect.height;
			classStyle = wclassStyle;
			exStyle = wexstyle;
			style = wstyle;
		}
	}
	
	
	protected void createHandle()
	{
		// Note: if modified, Form.createHandle() should be modified as well.
		
		if(isHandleCreated)
			return;
		
		//createClassHandle(CONTROL_CLASSNAME);
		
		/+
		if(!wparent || !wparent.handle || killing)
		{
			create_err:
			//throw new DflException("Control creation failure.");
			throw new DflException(Object.toString() ~ " creation failure."); // ?
		}
		+/
		
		if(killing)
		{
			create_err:
			throw new DflException("Control creation failure.");
			//throw new DflException(Object.toString() ~ " creation failure."); // ?
		}
		
		// Need the parent's handle to exist.
		if(wparent)
			wparent.createHandle();
		
		// This is here because wparent.createHandle() might create me.
		//if(created)
		if(isHandleCreated)
			return;
		
		CreateParams cp;
		DWORD prevClassStyle;
		prevClassStyle = wclassStyle;
		
		createParams(cp);
		assert(!isHandleCreated); // Make sure the handle wasn't created in createParams().
		
		with(cp)
		{
			wtext = caption;
			//wrect = Rect(x, y, width, height); // This gets updated in WM_CREATE.
			wclassStyle = classStyle;
			wexstyle = exStyle;
			wstyle = style;
			
			Application.creatingControl(this);
			hwnd = dfl.utf.createWindowEx(exStyle, className, caption, style, x, y,
				width, height, parent, menu, inst, param);
			if(!hwnd)
				goto create_err;
		}
		
		//onHandleCreated(EventArgs.empty); // Called in WM_CREATE now.
	}
	
	
	package final void _createHandle()
	{
		createHandle();
	}
	
	
	void destroyHandle()
	{
		if(!isHandleCreated)
			return;
		
		DestroyWindow(hwnd);
		
		// This stuff is done in WM_DESTROY because DestroyWindow() could be called elsewhere..
		//hwnd = null; // Done in WM_DESTROY.
		//onHandleDestroyed(EventArgs.empty); // Done in WM_DESTROY.
	}
	
	
	private final void fillRecreationData()
	{
		/+
		wtext = _fetchText();
		wclassStyle = _fetchClassLong();
		// Fetch children ?
		+/
	}
	
	
	protected void onDisposed(EventArgs ea)
	{
		disposed(this, ea);
	}
	
	
	// TODO: check if correct implementation.
	final void dispose(bool managedToo)
	{
		if(!isHandleCreated)
			return;
		
		if(managedToo)
		{
			cmenu = null;
			formname = null;
			otag = null;
			wcurs = null;
			wfont = null;
			wparent = null;
			wregion = null;
			wtext = null;
			
			killing = true;
		}
		
		destroyHandle();
		assert(hwnd == null);
		
		onDisposed(EventArgs.empty);
	}
	
	
	protected final bool getStyle(ControlStyles flag)
	{
		return (ctrlStyle & flag) != 0;
	}
	
	
	protected final void setStyle(ControlStyles flag, bool value)
	{
		if(flag & ControlStyles.CACHE_TEXT)
		{
			if(value)
				wtext = _fetchText();
			else
				wtext = null;
		}
		
		if(value)
			ctrlStyle |= flag;
		else
			ctrlStyle &= ~flag;
	}
	
	
	// Only for setStyle() styles that are part of hwnd and wndclass styles.
	protected final void updateStyles()
	{
		LONG newClassStyles = _classStyle();
		LONG newWndStyles = _style();
		
		if(ctrlStyle & ControlStyles.STANDARD_DOUBLE_CLICK)
			newClassStyles |= CS_DBLCLKS;
		else
			newClassStyles &= ~CS_DBLCLKS;
		
		if(ctrlStyle & ControlStyles.RESIZE_REDRAW)
			newClassStyles |= CS_HREDRAW | CS_VREDRAW;
		else
			newClassStyles &= ~(CS_HREDRAW | CS_VREDRAW);
		
		if(ctrlStyle & ControlStyles.SELECTABLE)
			newWndStyles |= WS_TABSTOP;
		else
			newWndStyles &= ~WS_TABSTOP;
		
		_classStyle(newClassStyles);
		_style(newWndStyles);
	}
	
	
	final bool getTopLevel()
	{
		// return GetParent(hwnd) == null;
		return wparent is null;
	}
	
	
	package final void alayout(Control ctrl)
	{
		if(_allowLayout)
		{
			//printf("alayout\n");
			auto LayoutEventArgs lea = new LayoutEventArgs(ctrl);
			onLayout(lea);
		}
	}
	
	
	// Z-order of controls has changed.
	package final void vchanged()
	{
		// Z-order can't change if it's not created or invisible.
		//if(!created() || !visible())
		//	return;
		
		version(RADIO_GROUP_LAYOUT)
		{
			//printf("vchanged\n");
			
			bit foundRadio = false;
			
			foreach(Control ctrl; ccollection)
			{
				if(!ctrl.visible)
					continue;
				
				if(ctrl._rtype() & 1)
				{
					LONG wlg;
					wlg = ctrl._style();
					if(foundRadio)
					{
						if(wlg & WS_GROUP)
							ctrl._style(wlg & ~WS_GROUP);
					}
					else
					{
						foundRadio = true;
						
						if(!(wlg & WS_GROUP))
							ctrl._style(wlg | WS_GROUP);
					}
				}
				else
				{
					// Found non-radio so reset group.
					// Update: only reset group if found ctrl with WS_EX_CONTROLPARENT.
					// TODO: check if correct implementation.
					if(ctrl._exStyle() & WS_EX_CONTROLPARENT)
						foundRadio = false;
				}
			}
		}
	}
	
	
	// Called after adding the control to a container.
	protected void initLayout()
	{
		/+
		if(_allowLayout)
			onLayout(new LayoutEventArgs(this));
		+/
		
		assert(wparent !is null);
		wparent.vchanged();
		wparent.alayout(this);
	}
	
	
	/+ package +/ protected int _rtype() // package
	{
		return 0;
	}
	
	
	protected void onLayout(LayoutEventArgs lea)
	{
		Rect area;
		area = displayRectangle;
		
		foreach(Control ctrl; ccollection)
		{
			if(!ctrl.visible)
				continue;
			if(ctrl._rtype() & (2 | 4)) // Mdichild | Tabpage
				continue;
			
			switch(ctrl.sdock)
			{
				case DockStyle.NONE:
					break;
				
				case DockStyle.LEFT:
					ctrl.setBoundsCore(area.x, area.y, 0, area.height, cast(BoundsSpecified)(BoundsSpecified.LOCATION | BoundsSpecified.HEIGHT));
					area.x = area.x + ctrl.width;
					area.width = area.width - ctrl.width;
					break;
				
				case DockStyle.TOP:
					ctrl.setBoundsCore(area.x, area.y, area.width, 0, cast(BoundsSpecified)(BoundsSpecified.LOCATION | BoundsSpecified.WIDTH));
					area.y = area.y + ctrl.height;
					area.height = area.height - ctrl.height;
					break;
				
				case DockStyle.FILL:
					//ctrl.bounds(Rect(area.x, area.y, area.width, area.height));
					ctrl.bounds = area;
					// area = ?
					break;
				
				case DockStyle.BOTTOM:
					ctrl.setBoundsCore(area.x, area.bottom - ctrl.height, area.width, 0, cast(BoundsSpecified)(BoundsSpecified.LOCATION | BoundsSpecified.WIDTH));
					area.height = area.height - ctrl.height;
					break;
				
				case DockStyle.RIGHT:
					ctrl.setBoundsCore(area.right - ctrl.width, area.y, 0, area.height, cast(BoundsSpecified)(BoundsSpecified.LOCATION | BoundsSpecified.HEIGHT));
					area.width = area.width - ctrl.width;
					break;
				
				default:
					assert(0);
			}
		}
		
		layout(this, lea);
	}
	
	
	// Not sure what to do here.
	deprecated bool isInputChar(char charCode)
	{
		return false;
	}
	
	
	void setVisibleCore(bool byes)
	{
		if(created)
		{
			if(visible == byes)
				return;
			
			//ShowWindow(hwnd, byes ? SW_SHOW : SW_HIDE);
			if(byes)
				doShow();
			else
				doHide();
			// Window events will update -wstyle-.
		}
		else
		{
			if(byes)
			{
				wstyle |= WS_VISIBLE;
				createControl();
			}
			else
			{
				wstyle &= ~WS_VISIBLE;
			}
		}
		
		if(wparent)
		{
			wparent.vchanged();
			wparent.alayout(this);
		}
	}
	
	
	deprecated bool processKeyPreview(inout Message m)
	{
		if(wparent)
			return wparent.processKeyPreview(m);
		return false;
	}
	
	
	deprecated protected bool processDialogChar(char charCode)
	{
		if(wparent)
			return wparent.processDialogChar(charCode);
		return false;
	}
	
	
	deprecated protected bool processMnemonic(char charCode)
	{
		return false;
	}
	
	
	package:
	final:
	HWND hwnd = null;
	/+
	AnchorStyles anch = cast(AnchorStyles)(AnchorStyles.TOP | AnchorStyles.LEFT);
	+/
	bit cvalidation = true;
	ContextMenu cmenu;
	DockStyle sdock = DockStyle.NONE;
	char[] formname;
	Object otag;
	bit menter = false; // Is mouse entered? Only valid if -trackMouseEvent- is non-null.
	bool killing = false;
	bit owned = true;
	bit _allowLayout = true;
	bit _clicking = false;
	bit needCalcSize = false;
	Color backc, forec;
	Rect wrect;
	Size wclientsz;
	Cursor wcurs;
	Font wfont;
	Control wparent;
	Region wregion;
	ControlCollection ccollection;
	char[] wtext; // After creation, this isn't used unless ControlStyles.CACHE_TEXT.
	ControlStyles ctrlStyle = ControlStyles.STANDARD_CLICK | ControlStyles.STANDARD_DOUBLE_CLICK /+ | ControlStyles.RESIZE_REDRAW +/ ;
	HBRUSH hbrBg;
	
	version(NO_DRAG_DROP) {} else
	{
		DropTarget droptarget = null;
	}
	
	LONG wstyle = WS_CHILD | WS_VISIBLE | WS_CLIPCHILDREN | WS_CLIPSIBLINGS; // Child, visible and enabled by default.
	LONG wexstyle;
	LONG wclassStyle = WNDCLASS_STYLE;
	
	
	// Helper functions.
	
	LONG _exStyle()
	{
		// return GetWindowLongA(hwnd, GWL_EXSTYLE);
		return wexstyle;
	}
	
	
	void _exStyle(LONG wl)
	{
		if(created)
		{
			SetWindowLongA(hwnd, GWL_EXSTYLE, wl);
		}
		
		wexstyle = wl;
	}
	
	
	LONG _style()
	{
		// return GetWindowLongA(hwnd, GWL_STYLE);
		return wstyle;
	}
	
	
	void _style(LONG wl)
	{
		if(created)
		{
			SetWindowLongA(hwnd, GWL_STYLE, wl);
		}
		
		wstyle = wl;
	}
	
	
	LRESULT defwproc(UINT msg, WPARAM wparam, LPARAM lparam)
	{
		return DefWindowProcA(hwnd, msg, wparam, lparam);
	}
	
	
	LONG _fetchClassLong()
	{
		return GetClassLongA(hwnd, GCL_STYLE);
	}
	
	
	LONG _classStyle()
	{
		// return GetClassLongA(hwnd, GCL_STYLE);
		// return wclassStyle;
		
		if(created)
		{
			// Always fetch because it's not guaranteed to be accurate.
			wclassStyle = _fetchClassLong();
		}
		
		return wclassStyle;
	}
	
	
	package void _classStyle(LONG cl)
	{
		if(created)
		{
			SetClassLongA(hwnd, GCL_STYLE, cl);
		}
		
		wclassStyle = cl;
	}
}


package abstract class ControlSuperClass: Control
{
	// Call previous wndProc().
	abstract protected void prevWndProc(inout Message msg);
	
	
	protected override void wndProc(inout Message msg)
	{
		switch(msg.msg)
		{
			case WM_PAINT:
				{
					RECT uprect;
					//if(msg.wParam)
					//	GetClientRect(hwnd, &uprect);
					//else
						GetUpdateRect(hwnd, &uprect, true);
					onInvalidated(new InvalidateEventArgs(Rect(&uprect)));
					
					//if(!msg.wParam)
						GetUpdateRect(hwnd, &uprect, false); // Preserve.
					
					prevWndProc(msg);
					
					// Now fake a normal paint event...
					
					auto Graphics gpx = new CommonGraphics(hwnd, GetDC(hwnd));
					//auto Graphics gpx = new CommonGraphics(hwnd, msg.wParam ? cast(HDC)msg.wParam : GetDC(hwnd), msg.wParam ? false : true);
					HRGN hrgn;
					
					hrgn = CreateRectRgnIndirect(&uprect);
					SelectClipRgn(gpx.handle, hrgn);
					DeleteObject(hrgn);
					
					PaintEventArgs pea = new PaintEventArgs(gpx, Rect(&uprect));
					
					// Can't erase the background now, Windows just painted..
					//if(ps.fErase)
					//{
					//	prepareDc(gpx.handle);
					//	onPaintBackground(pea);
					//}
					
					prepareDc(gpx.handle);
					onPaint(pea);
				}
				break;
			
			case WM_PRINTCLIENT:
				{
					prevWndProc(msg);
					
					auto Graphics gpx = new CommonGraphics(hwnd, GetDC(hwnd));
					PaintEventArgs pea = new PaintEventArgs(gpx,
						Rect(Point(0, 0), wclientsz));
					
					prepareDc(pea.graphics.handle);
					onPaint(pea);
				}
				break;
			
			case WM_PRINT:
				Control.defWndProc(msg);
				break;
			
			case WM_ERASEBKGND:
				Control.wndProc(msg);
				break;
			
			case WM_NCACTIVATE:
			case WM_NCCALCSIZE:
			case WM_NCCREATE:
			case WM_NCPAINT:
				prevWndProc(msg);
				break;
			
			default:
				prevWndProc(msg);
				super.wndProc(msg);
		}
	}
	
	
	override void defWndProc(inout Message msg)
	{
	}
	
	
	protected override void onPaintBackground(PaintEventArgs pea)
	{
		Message msg;
		
		msg.hWnd = handle;
		msg.msg = WM_ERASEBKGND;
		msg.wParam = cast(WPARAM)pea.graphics.handle;
		
		prevWndProc(msg);
		
		// Don't paint the background twice.
		//super.onPaintBackground(pea);
		
		// Event ?
		//paintBackground(this, pea);
	}
}


class ScrollableControl: Control // docmain
{
	void autoScroll(bool byes) // setter
	{
		autoscroll = byes;
	}
	
	
	bool autoScroll() // getter
	{
		return autoscroll;
	}
	
	
	final void autoScrollMargin(Size sz) // setter
	{
		scrollmargin = sz;
	}
	
	
	final Size autoScrollMargin() // getter
	{
		return scrollmargin;
	}
	
	
	// I don't understand what this scroll min size does.
	
	final void autoScrollMinSize(Size sz) // setter
	{
		scrollmin = sz;
	}
	
	
	final Size autoScrollMinSize() // getter
	{
		return scrollmin;
	}
	
	
	final void autoScrollPosition(Point pt) // setter
	{
		autoscrollpos = pt;
	}
	
	
	final Point autoScrollPosition() // getter
	{
		return autoscrollpos;
	}
	
	
	override final Rect displayRectangle() // getter
	{
		Rect result = clientRectangle;
		
		// Subtract dock padding.
		result.x = result.x + dpad.left;
		result.width = result.width - dpad.right - dpad.left;
		result.y = result.y + dpad.top;
		result.height = result.height - dpad.bottom - dpad.top;
		
		return result;
	}
	
	
	
	static class DockPaddingEdges
	{
		final:
		void all(int x) // setter
		{
			_bottom = _right = _top = _left = _all = x;
			
			changed();
		}
		
		
		final int all() // getter
		{
			return _all;
		}
		
		
		void left(int x) // setter
		{
			_left = x;
			
			changed();
		}
		
		
		int left() // getter
		{
			return _left;
		}
		
		
		void top(int x) // setter
		{
			_top = x;
			
			changed();
		}
		
		
		int top() // getter
		{
			return _top;
		}
		
		
		void right(int x) // setter
		{
			_right = x;
			
			changed();
		}
		
		
		int right() // getter
		{
			return _right;
		}
		
		
		void bottom(int x) // setter
		{
			_bottom = x;
			
			changed();
		}
		
		
		int bottom() // getter
		{
			return _bottom;
		}
		
		
		private:
		int _left, _top, _right, _bottom;
		int _all;
		package void delegate() changed;
	}
	
	
	final DockPaddingEdges dockPadding() // getter
	{
		return dpad;
	}
	
	
	final void setAutoScrollMargin(int x, int y)
	{
		// TODO:
		/*
		The margin sets the width and height of the border around each control. This margin is used
		to determine when scroll bars are needed on the container and where to scroll to when a
		control is selected.
		Note: If a negative number is passed in as the x or y values, the value will be reset to zero.
		*/
	}
	
	
	this()
	{
		super();
		_init();
	}
	
	
	/+
	// Used internally.
	this(HWND hwnd)
	{
		super(hwnd);
		_init();
	}
	+/
	
	
	protected:
	final void hScroll(bool byes) // setter
	{
		LONG wl = _style();
		if(byes)
			wl |= WS_HSCROLL;
		else
			wl &= ~WS_HSCROLL;
		_style(wl);
	}
	
	
	final bool hScroll() // getter
	{
		return (_style() & WS_HSCROLL) != 0;
	}
	
	
	final void vScroll(bool byes) // setter
	{
		LONG wl = _style();
		if(byes)
			wl |= WS_VSCROLL;
		else
			wl &= ~WS_VSCROLL;
		_style(wl);
	}
	
	
	final bool vScroll() // getter
	{
		return (_style() & WS_VSCROLL) != 0;
	}
	
	
	/+
	override void onLayout(LayoutEventArgs lea)
	{
		// ...
		super.onLayout(lea);
	}
	+/
	
	
	override void onMouseWheel(MouseEventArgs mea)
	{
		// ...
		super.onMouseWheel(mea);
	}
	
	
	override void onVisibleChanged(EventArgs ea)
	{
		// ...
		super.onVisibleChanged(ea);
	}
	
	
	/+
	override void scaleCore(float width, float height)
	{
		// Might not want to call super.scaleCore().
	}
	+/
	
	
	override void wndProc(inout Message msg)
	{
		super.wndProc(msg);
	}
	
	
	private:
	bool autoscroll = false;
	Size scrollmargin, scrollmin;
	Point autoscrollpos;
	DockPaddingEdges dpad;
	
	
	void _init()
	{
		dpad = new DockPaddingEdges;
		dpad.changed = &dpadChanged;
	}
	
	
	void dpadChanged()
	{
		if(_allowLayout)
		{
			//auto LayoutEventArgs lea = new LayoutEventArgs(this);
			//onLayout(lea);
			alayout(this);
		}
	}
}


interface IContainerControl // docmain
{
	Control activeControl(); // getter
	void activeControl(Control); // setter
	bool activateControl(Control);
}


// TODO: implement.
class ContainerControl: ScrollableControl, IContainerControl // docmain
{
	Control activeControl() // getter
	{
		// Not sure if this is correct.
		
		HWND hw = GetFocus();
		if(hw)
		{
			hw = GetParent(hw);
			while(hw)
			{
				if(hw == handle)
					return fromChildHandle(hw);
			}
		}
		
		return null;
	}
	
	
	void activeControl(Control ctrl) // setter
	{
		if(!activateControl(ctrl))
			throw new DflException("Unable to activate control.");
	}
	
	
	// Returns true if successfully activated.
	final bool activateControl(Control ctrl)
	{
		// Not sure if this is correct.
		
		if(!ctrl.canSelect)
			return false;
		//if(!SetActiveWindow(ctrl.handle))
		//	return false;
		ctrl.select();
		return true;
	}
	
	
	final Form parentForm() // getter
	{
		Control par = parent;
		Form f;
		
		while(par)
		{
			f = cast(Form)par;
			if(f)
				return f;
		}
		
		return null;
	}
	
	
	/+
	final bool validate()
	{
		// I don't know what this should do exactly.
	}
	+/
	
	
	this()
	{
		super();
		_init();
	}
	
	
	/+
	// Used internally.
	this(HWND hwnd)
	{
		super(hwnd);
		_init();
	}
	+/
	
	
	private void _init()
	{
		ctrlStyle |= ControlStyles.CONTAINER_CONTROL;
	}
	
	
	protected:
	override bool processDialogChar(char charCode)
	{
		// Not sure if this is correct.
		return false;
	}
	
	
	/+
	override bool processMnemonic(char charCode)
	{
		// ?
		return false;
	}
	+/
	
	
	bool processTabKey(bool forward)
	{
		// TODO: fix implementation.
		SendMessageA(hwnd, WM_NEXTDLGCTL, !forward, 0);
		return true;
	}
}

