// Written by Christopher E. Miller
// See the included license.txt for copyright and license details.


module dfl.data;

private import std.string, std.outofmemory;

private import dfl.base, dfl.winapi, dfl.wincom, dfl.application;
private import dfl.utf;


class DataFormats // docmain
{
	static class Format // docmain
	{
		final int id() // getter
		{
			return _id;
		}
		
		
		final char[] name() // getter
		{
			return _name;
		}
		
		
		package:
		int _id;
		char[] _name;
		
		
		this()
		{
		}
	}
	
	
	static:
	char[] bitmap() // getter
	{
		return getFormat(CF_BITMAP).name;
	}
	
	
	/+
	char[] commaSeparatedValue() // getter
	{
		return getFormat(?).name;
	}
	+/
	
	
	char[] dib() // getter
	{
		return getFormat(CF_DIB).name;
	}
	
	
	char[] dif() // getter
	{
		return getFormat(CF_DIF).name;
	}
	
	
	char[] enhandedMetaFile() // getter
	{
		return getFormat(CF_ENHMETAFILE).name;
	}
	
	
	char[] fileDrop() // getter
	{
		return getFormat(CF_HDROP).name;
	}
	
	
	char[] html() // getter
	{
		return getFormat("HTML Format").name;
	}
	
	
	char[] locale() // getter
	{
		return getFormat(CF_LOCALE).name;
	}
	
	
	char[] metafilePict() // getter
	{
		return getFormat(CF_METAFILEPICT).name;
	}
	
	
	char[] oemText() // getter
	{
		return getFormat(CF_OEMTEXT).name;
	}
	
	
	char[] palette() // getter
	{
		return getFormat(CF_PALETTE).name;
	}
	
	
	char[] penData() // getter
	{
		return getFormat(CF_PENDATA).name;
	}
	
	
	char[] riff() // getter
	{
		return getFormat(CF_RIFF).name;
	}
	
	
	char[] rtf() // getter
	{
		return getFormat("Rich Text Format").name;
	}
	
	
	/+
	char[] serializable() // getter
	{
		return getFormat(?).name;
	}
	+/
	
	
	char[] stringFormat() // getter
	{
		return getFormat("UTF-8").name; // ?
	}
	
	
	char[] symbolicLink() // getter
	{
		return getFormat(CF_SYLK).name;
	}
	
	
	char[] text() // getter
	{
		return getFormat(CF_TEXT).name;
	}
	
	
	char[] tiff() // getter
	{
		return getFormat(CF_TIFF).name;
	}
	
	
	char[] unicodeText() // getter
	{
		return getFormat(CF_UNICODETEXT).name;
	}
	
	
	char[] waveAudio() // getter
	{
		return getFormat(CF_WAVE).name;
	}
	
	
	// Assumes _init() was already called and
	// -id- is not in -fmts-.
	private Format _didntFindId(int id)
	{
		Format result;
		result = new Format;
		result._id = id;
		result._name = getName(id);
		//synchronized // _init() would need to be synchronized with it.
		{
			fmts[id] = result;
		}
		return result;
	}
	
	
	Format getFormat(int id)
	{
		_init();
		
		if(id in fmts)
			return fmts[id];
		
		return _didntFindId(id);
	}
	
	
	// Creates the format name if it doesn't exist.
	Format getFormat(char[] name)
	{
		_init();
		foreach(Format onfmt; fmts)
		{
			if(!icmp(name, onfmt.name))
				return onfmt;
		}
		// Didn't find it.
		return _didntFindId(dfl.utf.registerClipboardFormat(name));
	}
	
	
	// Extra.
	Format getFormat(TypeInfo type)
	{
		return getFormatFromType(type);
	}
	
	
	private:
	Format[int] fmts; // Indexed by identifier. Must _init() before accessing!
	
	
	void _init()
	{
		if(fmts.length)
			return;
		
		
		void initfmt(int id, char[] name)
		in
		{
			assert(!(id in fmts));
		}
		body
		{
			Format fmt;
			fmt = new Format;
			fmt._id = id;
			fmt._name = name;
			fmts[id] = fmt;
		}
		
		
		initfmt(CF_BITMAP, "Bitmap");
		initfmt(CF_DIB, "DeviceIndependentBitmap");
		initfmt(CF_DIF, "DataInterchangeFormat");
		initfmt(CF_ENHMETAFILE, "EnhancedMetafile");
		initfmt(CF_HDROP, "FileDrop");
		initfmt(CF_LOCALE, "Locale");
		initfmt(CF_METAFILEPICT, "MetaFilePict");
		initfmt(CF_OEMTEXT, "OEMText");
		initfmt(CF_PALETTE, "Palette");
		initfmt(CF_PENDATA, "PenData");
		initfmt(CF_RIFF, "RiffAudio");
		initfmt(CF_SYLK, "SymbolicLink");
		initfmt(CF_TEXT, "Text");
		initfmt(CF_TIFF, "TaggedImageFileFormat");
		initfmt(CF_UNICODETEXT, "UnicodeText");
		initfmt(CF_WAVE, "WaveAudio");
		
		fmts.rehash;
	}
	
	
	// Does not get the name of one of the predefined constant ones.
	char[] getName(int id)
	{
		char[] result;
		result = dfl.utf.getClipboardFormatName(id);
		if(!result.length)
			throw new DflException("Unable to get format.");
		return result;
	}
	
	
	package Format getFormatFromType(TypeInfo type)
	{
		if(type == typeid(ubyte[]))
			return getFormat(text);
		if(type == typeid(char[]))
			return getFormat(stringFormat);
		if(type == typeid(wchar[]))
			return getFormat(unicodeText);
		//if(type == typeid(Bitmap))
		//	return getFormat(bitmap);
		
		throw new DflException("Unknown data format.");
	}
	
	
	// Convert clipboard -value- to Data.
	Data getDataFromFormat(int id, void[] value)
	{
		switch(id)
		{
			case CF_TEXT:
				return Data(stopAtNull!(ubyte)(cast(ubyte[])value));
			
			case CF_UNICODETEXT:
				return Data(stopAtNull!(wchar)(cast(wchar[])value));
			
			default:
				if(id == getFormat(stringFormat).id)
					return Data(stopAtNull!(char)(cast(char[])value));
		}
		
		//throw new DflException("Unknown data format.");
		return Data(value); // ?
	}
	
	
	// Value the clipboard wants.
	void[] getClipboardValueFromData(Data data)
	{
		if(data.info == typeid(ubyte[]))
		{
			// ANSI text.
			const ubyte[] UBYTE_ZERO = [0];
			return data.getText() ~ UBYTE_ZERO;
		}
		else if(data.info == typeid(char[]))
		{
			// UTF-8 string.
			char[] str;
			str = data.getString();
			return toStringz(str)[0 .. str.length + 1];
		}
		else if(data.info == typeid(wchar[]))
		{
			// Unicode string.
			return data.getUnicodeText() ~ cast(wchar[])\0;
		}
		else
		{
			return data.value; // ?
		}
	}
	
	
	this()
	{
	}
}


private template stopAtNull(T)
{
	T[] stopAtNull(T[] array)
	{
		int i;
		for(i = 0; i != array.length; i++)
		{
			if(!array[i])
				return array[0 .. i];
		}
		//return null;
		throw new DflException("Invalid data."); // ?
	}
}


struct Data // docmain
{
	TypeInfo info() // getter
	{
		return _info;
	}
	
	
	void[] value() // getter
	{
		return _value[0 .. _info.tsize()];
	}
	
	
	static Data opCall(...)
	in
	{
		assert(_arguments.length == 1);
	}
	body
	{
		Data result;
		result._info = _arguments[0];
		result._value = _argptr[0 .. result._info.tsize()].dup;
		return result;
	}
	
	
	// UTF-8.
	char[] getString()
	{
		assert(_info == typeid(char[]));
		return *cast(char[]*)_value;
	}
	
	
	// ANSI text.
	ubyte[] getText()
	{
		assert(_info == typeid(ubyte[]));
		return *cast(ubyte[]*)_value;
	}
	
	
	wchar[] getUnicodeText()
	{
		assert(_info == typeid(wchar[]));
		return *cast(wchar[]*)_value;
	}
	
	
	int getInt()
	{
		assert(_info == typeid(int));
		return *cast(int*)_value;
	}
	
	
	int getUint()
	{
		assert(_info == typeid(uint));
		return *cast(uint*)_value;
	}
	
	
	Object getObject()
	{
		assert(cast(TypeInfo_Class)_info !is null);
		return cast(Object)*cast(Object**)_value;
	}
	
	
	private:
	TypeInfo _info;
	void* _value;
}


/+
interface IDataFormat
{
	
}
+/


interface IDataObject // docmain
{
	Data getData(char[] fmt);
	Data getData(TypeInfo type);
	Data getData(char[] fmt, bool doConvert);
	bool getDataPresent(char[] fmt); // Check.
	bool getDataPresent(TypeInfo type); // Check.
	bool getDataPresent(char[] fmt, bool canConvert); // Check.
	char[][] getFormats();
	char[][] getFormats(bool onlyNative);
	void setData(Data obj);
	void setData(char[] fmt, Data obj);
	void setData(TypeInfo type, Data obj);
	void setData(char[] fmt, bool canConvert, Data obj);
}


class DataObject: dfl.data.IDataObject // docmain
{
	Data getData(char[] fmt)
	{
		return getData(fmt, false);
	}
	
	
	Data getData(TypeInfo type)
	{
		return getData(DataFormats.getFormat(type).name);
	}
	
	
	Data getData(char[] fmt, bool doConvert)
	{
		// doConvert ...
		
		//printf("Looking for format '%.*s'.\n", fmt);
		int i;
		i = find(fmt);
		if(i == -1)
			throw new DflException("Data format not present.");
		return all[i].obj;
	}
	
	
	bool getDataPresent(char[] fmt)
	{
		return getDataPresent(fmt, false);
	}
	
	
	bool getDataPresent(TypeInfo type)
	{
		return getDataPresent(DataFormats.getFormat(type).name);
	}
	
	
	bool getDataPresent(char[] fmt, bool canConvert)
	{
		// canConvert ...
		return find(fmt) != -1;
	}
	
	
	char[][] getFormats()
	{
		char[][] result;
		result = new char[][all.length];
		foreach(int i, inout char[] fmt; result)
		{
			fmt = all[i].fmt;
		}
		return result;
	}
	
	
	char[][] getFormats(bool onlyNative)
	{
		return getFormats(); // ?
	}
	
	
	void setData(Data obj)
	{
		setData(DataFormats.getFormat(obj.info).name, obj);
	}
	
	
	void setData(char[] fmt, Data obj)
	{
		int i;
		i = find(fmt);
		if(i != -1)
		{
			all[i].obj = obj;
		}
		else
		{
			Pair pair;
			pair.fmt = fmt;
			pair.obj = obj;
			all ~= pair;
		}
	}
	
	
	void setData(TypeInfo type, Data obj)
	{
		setData(DataFormats.getFormat(type).name, obj);
	}
	
	
	void setData(char[] fmt, bool canConvert, Data obj)
	{
		setData(fmt, obj); // ?
	}
	
	
	private:
	struct Pair
	{
		char[] fmt;
		Data obj;
	}
	
	
	Pair[] all;
	
	
	int find(char[] fmt)
	{
		int i;
		for(i = 0; i != all.length; i++)
		{
			if(!icmp(all[i].fmt, fmt))
				return i;
		}
		return -1;
	}
}


package class DtoComDataObject: ComObject, dfl.wincom.IDataObject
{
	this(dfl.data.IDataObject dataObj)
	{
		this.dataObj = dataObj;
	}
	
	
	extern(Windows):
	override HRESULT QueryInterface(IID* riid, void** ppv)
	{
		if(*riid == IID_IDataObject)
		{
			*ppv = cast(dfl.wincom.IDataObject)this;
			AddRef();
			return S_OK;
		}
		else if(*riid == IID_IUnknown)
		{
			*ppv = cast(IUnknown)this;
			AddRef();
			return S_OK;
		}
		else
		{
			*ppv = null;
			return E_NOINTERFACE;
		}
	}
	
	
	HRESULT GetData(FORMATETC *pFormatetc, STGMEDIUM *pmedium)
	{
		char[] fmt;
		HRESULT result = S_OK;
		Data data;
		
		try
		{
			if(pFormatetc.lindex != -1)
			{
				result = DV_E_LINDEX;
			}
			else if(!(pFormatetc.tymed & TYMED_HGLOBAL))
			{
				// Unsupported medium type.
				result = DV_E_TYMED;
			}
			else if(!(pFormatetc.dwAspect & DVASPECT_CONTENT))
			{
				// What about the other aspects?
				result = DV_E_DVASPECT;
			}
			else
			{
				fmt = DataFormats.getFormat(pFormatetc.cfFormat).name;
				data = dataObj.getData(fmt, true); // Should this be convertable?
				
				HGLOBAL hg;
				void* pmem;
				void[] src;
				
				//src = data.value;
				src = DataFormats.getClipboardValueFromData(data);
				hg = GlobalAlloc(GMEM_SHARE, src.length);
				if(!hg)
				{
					result = STG_E_MEDIUMFULL;
				}
				else
				{
					pmem = GlobalLock(hg);
					if(!hg)
					{
						result = E_UNEXPECTED;
						GlobalFree(hg);
					}
					else
					{
						pmem[0 .. src.length] = src;
						GlobalUnlock(hg);
						
						pmedium.tymed = TYMED_HGLOBAL;
						pmedium.hGlobal = hg;
						pmedium.pUnkForRelease = null; // ?
					}
				}
			}
		}
		catch(DflException e)
		{
			//Application.onThreadException(e);
			
			result = DV_E_FORMATETC;
		}
		catch(OutOfMemoryException e)
		{
			Application.onThreadException(e);
			
			result = E_OUTOFMEMORY;
		}
		catch(Object e)
		{
			Application.onThreadException(e);
			
			result = E_UNEXPECTED;
		}
		
		return result;
	}
	
	
	HRESULT GetDataHere(FORMATETC *pFormatetc, STGMEDIUM *pmedium)
	{
		return E_UNEXPECTED; // TODO: finish.
	}
	
	
	HRESULT QueryGetData(FORMATETC *pFormatetc)
	{
		char[] fmt;
		HRESULT result = S_OK;
		
		try
		{
			if(pFormatetc.lindex != -1)
			{
				result = DV_E_LINDEX;
			}
			else if(!(pFormatetc.tymed & TYMED_HGLOBAL))
			{
				// Unsupported medium type.
				result = DV_E_TYMED;
			}
			else if(!(pFormatetc.dwAspect & DVASPECT_CONTENT))
			{
				// What about the other aspects?
				result = DV_E_DVASPECT;
			}
			else
			{
				fmt = DataFormats.getFormat(pFormatetc.cfFormat).name;
				
				if(!dataObj.getDataPresent(fmt))
					result = S_FALSE; // ?
			}
		}
		catch(DflException e)
		{
			//Application.onThreadException(e);
			
			result = DV_E_FORMATETC;
		}
		catch(OutOfMemoryException e)
		{
			Application.onThreadException(e);
			
			result = E_OUTOFMEMORY;
		}
		catch(Object e)
		{
			Application.onThreadException(e);
			
			result = E_UNEXPECTED;
		}
		
		return result;
	}
	
	
	HRESULT GetCanonicalFormatEtc(FORMATETC *pFormatetcIn, FORMATETC *pFormatetcOut)
	{
		// TODO: finish.
		
		pFormatetcOut.ptd = null;
		return E_NOTIMPL;
	}
	
	
	HRESULT SetData(FORMATETC *pFormatetc, STGMEDIUM *pmedium, BOOL fRelease)
	{
		return E_UNEXPECTED; // TODO: finish.
	}
	
	
	HRESULT EnumFormatEtc(DWORD dwDirection, IEnumFORMATETC *ppenumFormatetc)
	{
		// SHCreateStdEnumFmtEtc() requires Windows 2000 +
		
		HRESULT result;
		
		try
		{
			if(dwDirection == DATADIR_GET)
			{
				*ppenumFormatetc = new EnumDataObjectFORMATETC(dataObj);
				result = S_OK;
			}
			else
			{
				result = E_NOTIMPL;
			}
		}
		catch(Object e)
		{
			Application.onThreadException(e);
			
			result = E_UNEXPECTED;
		}
		
		return result;
	}
	
	
	HRESULT DAdvise(FORMATETC *pFormatetc, DWORD advf, IAdviseSink pAdvSink, DWORD *pdwConnection)
	{
		return E_UNEXPECTED; // TODO: finish.
	}
	
	
	HRESULT DUnadvise(DWORD dwConnection)
	{
		return E_UNEXPECTED; // TODO: finish.
	}
	
	
	HRESULT EnumDAdvise(IEnumSTATDATA *ppenumAdvise)
	{
		return E_UNEXPECTED; // TODO: finish.
	}
	
	
	private:
	dfl.data.IDataObject dataObj;
}


private class EnumDataObjectFORMATETC: ComObject, IEnumFORMATETC
{
	this(dfl.data.IDataObject dataObj, char[][] fmts, ULONG start)
	{
		this.dataObj = dataObj;
		this.fmts = fmts;
		idx = start;
	}
	
	
	this(dfl.data.IDataObject dataObj)
	{
		this(dataObj, dataObj.getFormats(), 0);
	}
	
	
	extern(Windows):
	override HRESULT QueryInterface(IID* riid, void** ppv)
	{
		if(*riid == IID_IEnumFORMATETC)
		{
			*ppv = cast(IEnumFORMATETC)this;
			AddRef();
			return S_OK;
		}
		else if(*riid == IID_IUnknown)
		{
			*ppv = cast(IUnknown)this;
			AddRef();
			return S_OK;
		}
		else
		{
			*ppv = null;
			return E_NOINTERFACE;
		}
	}
	
	
	HRESULT Next(ULONG celt, FORMATETC* rgelt, ULONG* pceltFetched)
	{
		HRESULT result;
		
		try
		{
			if(idx < fmts.length)
			{
				ULONG end;
				end = idx + celt;
				if(end > fmts.length)
				{
					result = S_FALSE; // ?
					end = fmts.length;
					
					if(pceltFetched)
						*pceltFetched = end - idx;
				}
				else
				{
					result = S_OK;
					
					if(pceltFetched)
						*pceltFetched = celt;
				}
				
				for(; idx != end; idx++)
				{
					rgelt.cfFormat = DataFormats.getFormat(fmts[idx]).id;
					rgelt.ptd = null;
					rgelt.dwAspect = DVASPECT_CONTENT; // ?
					rgelt.lindex = -1;
					//rgelt.tymed = TYMED_NULL;
					rgelt.tymed = TYMED_HGLOBAL;
					
					rgelt++;
				}
			}
			else
			{
				if(pceltFetched)
					*pceltFetched = 0;
				result = S_FALSE;
			}
		}
		catch(Object e)
		{
			Application.onThreadException(e);
			
			result = E_UNEXPECTED;
		}
		
		return result;
	}
	
	
	HRESULT Skip(ULONG celt)
	{
		idx += celt;
		return (idx > fmts.length) ? S_FALSE : S_OK;
	}
	
	
	HRESULT Reset()
	{
		HRESULT result;
		
		try
		{
			idx = 0;
			fmts = dataObj.getFormats();
			
			result = S_OK;
		}
		catch(Object e)
		{
			Application.onThreadException(e);
			
			result = E_UNEXPECTED;
		}
		
		return result;
	}
	
	
	HRESULT Clone(IEnumFORMATETC* ppenum)
	{
		HRESULT result;
		
		try
		{
			*ppenum = new EnumDataObjectFORMATETC(dataObj, fmts, idx);
			result = S_OK;
		}
		catch(Object e)
		{
			Application.onThreadException(e);
			
			result = E_UNEXPECTED;
		}
		
		return result;
	}
	
	
	private:
	dfl.data.IDataObject dataObj;
	char[][] fmts;
	ULONG idx;
}


package class ComToDdataObject: dfl.data.IDataObject
{
	this(dfl.wincom.IDataObject dataObj)
	{
		this.dataObj = dataObj;
		dataObj.AddRef();
	}
	
	
	~this()
	{
		dataObj.Release(); // Must get called...
	}
	
	
	private Data _getData(int id)
	{
		FORMATETC fmte;
		STGMEDIUM stgm;
		void[] mem;
		void* plock;
		
		fmte.cfFormat = id;
		fmte.ptd = null;
		fmte.dwAspect = DVASPECT_CONTENT; // ?
		fmte.lindex = -1;
		fmte.tymed = TYMED_HGLOBAL; // ?
		
		if(S_OK != dataObj.GetData(&fmte, &stgm))
			throw new DflException("Unable to get data.");
		
		
		void release()
		{
			//ReleaseStgMedium(&stgm);
			if(stgm.pUnkForRelease)
				stgm.pUnkForRelease.Release();
			else
				GlobalFree(stgm.hGlobal);
		}
		
		
		plock = GlobalLock(stgm.hGlobal);
		if(!plock)
		{
			release();
			throw new DflException("Error obtaining data.");
		}
		
		mem = new ubyte[GlobalSize(stgm.hGlobal)];
		mem[] = plock[0 .. mem.length];
		GlobalUnlock(stgm.hGlobal);
		release();
		
		return DataFormats.getDataFromFormat(id, mem);
	}
	
	
	Data getData(char[] fmt)
	{
		return _getData(DataFormats.getFormat(fmt).id);
	}
	
	
	Data getData(TypeInfo type)
	{
		return _getData(DataFormats.getFormatFromType(type).id);
	}
	
	
	Data getData(char[] fmt, bool doConvert)
	{
		return getData(fmt); // ?
	}
	
	
	private bool _getDataPresent(int id)
	{
		FORMATETC fmte;
		
		fmte.cfFormat = id;
		fmte.ptd = null;
		fmte.dwAspect = DVASPECT_CONTENT; // ?
		fmte.lindex = -1;
		fmte.tymed = TYMED_HGLOBAL; // ?
		
		return S_OK == dataObj.QueryGetData(&fmte);
	}
	
	
	bool getDataPresent(char[] fmt)
	{
		return _getDataPresent(DataFormats.getFormat(fmt).id);
	}
	
	
	bool getDataPresent(TypeInfo type)
	{
		return _getDataPresent(DataFormats.getFormatFromType(type).id);
	}
	
	
	bool getDataPresent(char[] fmt, bool canConvert)
	{
		return getDataPresent(fmt); // ?
	}
	
	
	char[][] getFormats()
	{
		IEnumFORMATETC fenum;
		FORMATETC fmte;
		char[][] result;
		ULONG nfetched = 1; // ?
		
		if(S_OK != dataObj.EnumFormatEtc(1, &fenum))
			throw new DflException("Unable to get formats.");
		
		fenum.AddRef(); // ?
		for(;;)
		{
			if(S_OK != fenum.Next(1, &fmte, &nfetched))
				break;
			if(!nfetched)
				break;
			//printf("\t\t{getFormats:%d}\n", fmte.cfFormat);
			result ~= DataFormats.getFormat(fmte.cfFormat).name;
		}
		fenum.Release(); // ?
		
		return result;
	}
	
	
	char[][] getFormats(bool onlyNative)
	{
		return getFormats(); // ?
	}
	
	
	private void _setData(int id, Data obj)
	{
		/+
		FORMATETC fmte;
		STGMEDIUM stgm;
		HANDLE hmem;
		void[] mem;
		void* pmem;
		
		mem = DataFormats.getClipboardValueFromData(obj);
		
		hmem = GlobalAlloc(GMEM_SHARE, mem.length);
		if(!hmem)
		{
			//printf("Unable to GlobalAlloc().\n");
			err_set:
			throw new DflException("Unable to set data.");
		}
		pmem = GlobalLock(hmem);
		if(!pmem)
		{
			//printf("Unable to GlobalLock().\n");
			GlobalFree(hmem);
			goto err_set;
		}
		pmem[0 .. mem.length] = mem;
		GlobalUnlock(hmem);
		
		fmte.cfFormat = id;
		fmte.ptd = null;
		fmte.dwAspect = DVASPECT_CONTENT; // ?
		fmte.lindex = -1;
		fmte.tymed = TYMED_HGLOBAL;
		
		stgm.tymed = TYMED_HGLOBAL;
		stgm.hGlobal = hmem;
		stgm.pUnkForRelease = null;
		
		// -dataObj- now owns the handle.
		HRESULT hr = dataObj.SetData(&fmte, &stgm, true);
		if(S_OK != hr)
		{
			//printf("Unable to IDataObject::SetData() = %d (0x%X).\n", hr, hr);
			// Failed, need to free it..
			GlobalFree(hmem);
			goto err_set;
		}
		+/
	}
	
	
	void setData(Data obj)
	{
		_setData(DataFormats.getFormatFromType(obj.info).id, obj);
	}
	
	
	void setData(char[] fmt, Data obj)
	{
		_setData(DataFormats.getFormat(fmt).id, obj);
	}
	
	
	void setData(TypeInfo type, Data obj)
	{
		_setData(DataFormats.getFormatFromType(type).id, obj);
	}
	
	
	void setData(char[] fmt, bool canConvert, Data obj)
	{
		setData(fmt, obj); // ?
	}
	
	
	private:
	dfl.wincom.IDataObject dataObj;
}


/+
// Shut up linker: Error 42: Symbol Undefined __init_11TypeInfo_Av
class TypeInfo_Av: TypeInfo
{
}
+/

