// Written by Christopher E. Miller
// See the included license.txt for copyright and license details.


// Not actually part of forms, but is handy.

module dfl.environment;

private import std.string, std.path, std.conv;
private import std.c.stdlib;

private import dfl.winapi, dfl.base, dfl.utf;


class Environment // docmain
{
	private this() {}
	
	
	static:
	char[] commandLine() // getter
	{
		return dfl.utf.getCommandLine();
	}
	
	
	void currentDirectory(char[] cd) // setter
	{
		if(!dfl.utf.setCurrentDirectory(cd))
			throw new DflException("Unable to set current directory.");
	}
	
	
	char[] currentDirectory() // getter
	{
		return dfl.utf.getCurrentDirectory();
	}
	
	
	char[] machineName() // getter
	{
		char[] result;
		result = dfl.utf.getComputerName();
		if(!result.length)
			throw new DflException("Unable to obtain machine name.");
		return result;
	}
	
	
	char[] newLine() // getter
	{
		return std.path.linesep; // Why is this in std.path?
	}
	
	
	OperatingSystem osVersion() // getter
	{
		OSVERSIONINFOA osi;
		Version ver;
		
		osi.dwOSVersionInfoSize = osi.sizeof;
		if(!GetVersionExA(&osi))
			throw new DflException("Unable to obtain operating system version information.");
		
		int build;
		
		switch(osi.dwPlatformId)
		{
			case VER_PLATFORM_WIN32_NT:
				ver = new Version(osi.dwMajorVersion, osi.dwMinorVersion, osi.dwBuildNumber);
				break;
			
			case VER_PLATFORM_WIN32_WINDOWS:
				ver = new Version(osi.dwMajorVersion, osi.dwMinorVersion, LOWORD(osi.dwBuildNumber));
				break;
			
			default:
				ver = new Version(osi.dwMajorVersion, osi.dwMinorVersion);
		}
		
		return new OperatingSystem(cast(PlatformId)osi.dwPlatformId, ver);
	}
	
	
	char[] systemDirectory() // getter
	{
		char[] result;
		result = dfl.utf.getSystemDirectory();
		if(!result.length)
			throw new DflException("Unable to obtain system directory.");
		return result;
	}
	
	
	// Should return int ?
	DWORD tickCount() // getter
	{
		return GetTickCount();
	}
	
	
	char[] userName() // getter
	{
		char[] result;
		result = dfl.utf.getUserName();
		if(!result.length)
			throw new DflException("Unable to obtain user name.");
		return result;
	}
	
	
	void exit(int code)
	{
		// This is probably better than ExitProcess(code).
		std.c.stdlib.exit(code);
	}
	
	
	char[] expandEnvironmentVariables(char[] str)
	{
		char[] result;
		if(!dfl.utf.expandEnvironmentStrings(str, result))
			throw new DflException("Unable to expand environment variables.");
		return result;
	}
	
	
	char[][] getCommandLineArgs()
	{
		return parseArgs(commandLine);
	}
	
	
	char[] getEnvironmentVariable(char[] name)
	{
		char[] result;
		result = dfl.utf.getEnvironmentVariable(name);
		if(!result.length)
			throw new DflException("Unable to obtain environment variable.");
		return result;
	}
	
	
	//char[][char[]] getEnvironmentVariables()
	//char[][] getEnvironmentVariables()
	
	
	char[][] getLogicalDrives()
	{
		DWORD dr = GetLogicalDrives();
		char[][] result;
		int i;
		char[4] tmp = " :\\\0";
		
		for(i = 0; dr; i++)
		{
			if(dr & 1)
			{
				char[] s = tmp.dup[0 .. 3];
				s[0] = 'A' + i;
				result ~= s;
			}
			dr >>= 1;
		}
		
		return result;
	}
}


package char[][] parseArgs(char[] args)
{
	char[][] result;
	uint i;
	bit inQuote = false;
	bit findStart = true;
	uint startIndex = 0;
	
	for(i = 0;; i++)
	{
		if(i == args.length)
		{
			if(findStart)
				startIndex = i;
			break;
		}
		
		if(findStart)
		{
			if(args[i] == ' ' || args[i] == '\t')
				continue;
			findStart = false;
			startIndex = i;
		}
		
		if(args[i] == '"')
		{
			inQuote = !inQuote;
			if(!inQuote) //matched quotes
			{
				result.length = result.length + 1;
				result[result.length - 1] = args[startIndex .. i];
				findStart = true;
			}
			else //starting quote
			{
				if(startIndex != i) //must be a quote stuck to another word, separate them
				{
					result.length = result.length + 1;
					result[result.length - 1] = args[startIndex .. i];
					startIndex = i + 1;
				}
				else
				{
					startIndex++; //exclude the quote
				}
			}
		}
		else if(!inQuote)
		{
			if(args[i] == ' ' || args[i] == '\t')
			{
				result.length = result.length + 1;
				result[result.length - 1] = args[startIndex .. i];
				findStart = true;
			}
		}
	}
	
	if(startIndex != i)
	{
		result.length = result.length + 1;
		result[result.length - 1] = args[startIndex .. i];
	}
	
	return result;
}


unittest
{
	char[][] args;
	
	args = parseArgs(`"foo" bar`);
	assert(args.length == 2);
	assert(args[0] == "foo");
	assert(args[1] == "bar");
	
	args = parseArgs(`"environment"`);
	assert(args.length == 1);
	assert(args[0] == "environment");
	
	/+
	writefln("commandLine = '%s'", Environment.commandLine);
	foreach(char[] arg; Environment.getCommandLineArgs())
	{
		writefln("\t'%s'", arg);
	}
	+/
}


// Any version, not just the operating system.
class Version // docmain ?
{
	final:
	this()
	{
	}
	
	
	// A string containing "major.minor.build.revision".
	// 2 to 4 parts expected.
	this(char[] str)
	{
		char[][] stuff = std.string.split(str, ".");
		
		// Note: fallthrough.
		switch(stuff.length)
		{
			case 4:
				_revision = std.conv.toInt(stuff[3]);
			case 3:
				_build = std.conv.toInt(stuff[2]);
			case 2:
				_minor = std.conv.toInt(stuff[1]);
				_major = std.conv.toInt(stuff[0]);
			default:
				throw new DflException("Invalid version parameter.");
		}
	}
	
	
	this(int major, int minor)
	{
		_major = major;
		_minor = minor;
	}
	
	
	this(int major, int minor, int build)
	{
		_major = major;
		_minor = minor;
		_build = build;
	}
	
	
	this(int major, int minor, int build, int revision)
	{
		_major = major;
		_minor = minor;
		_build = build;
		_revision = revision;
	}
	
	
	invariant
	{
		assert(_major >= 0);
		assert(_minor >= 0);
		assert(_build >= -1);
		assert(_revision >= -1);
	}
	
	
	override char[] toString()
	{
		char[] result;
		
		result = std.string.toString(_major) ~ "." ~ std.string.toString(_minor);
		if(_build != -1)
			result ~= "." ~ std.string.toString(_build);
		if(_revision != -1)
			result ~= "." ~ std.string.toString(_revision);
		
		return result;
	}
	
	
	int major() // getter
	{
		return _major;
	}
	
	
	int minor() // getter
	{
		return _minor;
	}
	
	
	// -1 if no build.
	int build() // getter
	{
		return _build;
	}
	
	
	// -1 if no revision.
	int revision() // getter
	{
		return _revision;
	}
	
	
	private:
	int _major = 0, _minor = 0;
	int _build = -1, _revision = -1;
}


enum PlatformId: DWORD
{
	WIN_CE = cast(DWORD)-1,
	WIN32s = VER_PLATFORM_WIN32s,
	WIN32_WINDOWS = VER_PLATFORM_WIN32_WINDOWS,
	WIN32_NT = VER_PLATFORM_WIN32_NT,
}


final class OperatingSystem // docmain
{
	final:
	this(PlatformId platId, Version ver)
	{
		this.platId = platId;
		this.vers = ver;
	}
	
	
	override char[] toString()
	{
		char[] result;
		
		// DMD 0.92 says error: cannot implicitly convert uint to PlatformId
		switch(cast(DWORD)platId)
		{
			case PlatformId.WIN32_NT:
				result = "Microsoft Windows NT ";
				break;
			
			case PlatformId.WIN32_WINDOWS:
				result = "Microsoft Windows 95 ";
				break;
			
			case PlatformId.WIN32s:
				result = "Microsoft Win32s ";
				break;
			
			case PlatformId.WIN_CE:
				result = "Microsoft Windows CE ";
				break;
			
			default:
				throw new DflException("Unknown platform ID.");
		}
		
		result ~= vers.toString();
		return result;
	}
	
	
	PlatformId platform() // getter
	{
		return platId;
	}
	
	
	// Should be version() :p
	Version ver() // getter
	{
		return ver;
	}
	
	
	private:
	PlatformId platId;
	Version vers;
}

