// Written by Christopher E. Miller
// See the included license.txt for copyright and license details.


module dfl.folderdialog;

private import std.string, std.c.stdlib;

private import dfl.commondialog, dfl.base, dfl.winapi, dfl.wincom;
private import dfl.utf, dfl.application;


private extern(Windows)
{
	alias LPITEMIDLIST function(LPBROWSEINFOW lpbi) SHBrowseForFolderWProc;
	alias BOOL function(LPCITEMIDLIST pidl, LPWSTR pszPath) SHGetPathFromIDListWProc;
}


class FolderBrowserDialog: CommonDialog // docmain
{
	this()
	{
		// Flag BIF_NEWDIALOGSTYLE requires OleInitialize().
		//OleInitialize(null);
		
		Application.ppin(cast(void*)this);
		
		bi.ulFlags = INIT_FLAGS;
		bi.lParam = cast(typeof(bi.lParam))cast(void*)this;
		///////bi.lpfn = ;
	}
	
	
	~this()
	{
		//OleUninitialize();
	}
	
	
	override DialogResult showDialog()
	{
		if(!runDialog(GetActiveWindow()))
			return DialogResult.CANCEL;
		return DialogResult.OK;
	}
	
	
	override DialogResult showDialog(IWin32Window owner)
	{
		if(!runDialog(owner ? owner.handle : GetActiveWindow()))
			return DialogResult.CANCEL;
		return DialogResult.OK;
	}
	
	
	override void reset()
	{
		bi.ulFlags = INIT_FLAGS;
		_desc = null;
		_selpath = null;
	}
	
	
	final void description(char[] desc) // setter
	{
		// lpszTitle
		
		_desc = desc;
	}
	
	
	final char[] description() // getter
	{
		return _desc;
	}
	
	
	final void selectedPath(char[] selpath) // setter
	{
		// pszDisplayName
		
		_selpath = selpath;
	}
	
	
	final char[] selectedPath() // getter
	{
		return _selpath;
	}
	
	
	/+
	// Currently only works for shell32.dll version 6.0+.
	final void showNewFolderButton(bool byes) // setter
	{
		// BIF_NONEWFOLDERBUTTON exists with shell 6.0+.
		// Might need to enum child windows looking for window title
		// "&New Folder" and hide it, then shift "OK" and "Cancel" over.
		
		if(byes)
			bi.ulFlags |= BIF_NONEWFOLDERBUTTON;
		else
			bi.ulFlags &= ~BIF_NONEWFOLDERBUTTON;
	}
	
	
	final bool showNewFolderButton() // getter
	{
		return (bi.ulFlags & BIF_NONEWFOLDERBUTTON) == 0;
	}
	+/
	
	
	private void _errPathTooLong()
	{
		throw new DflException("Path name is too long.");
	}
	
	
	private void _errNoGetPath()
	{
		throw new DflException("Unable to obtain path.");
	}
	
	
	private void _errNoShMalloc()
	{
		throw new DflException("Unable to get shell memory allocator.");
	}
	
	
	protected override bool runDialog(HWND owner)
	{
		IMalloc shmalloc;
		
		bi.hwndOwner = owner;
		
		// Using size of wchar so that the buffer works for ansi and unicode.
		void* pdescz = std.c.stdlib.alloca(wchar.sizeof * MAX_PATH);
		if(!pdescz)
			throw new DflException("Out of memory."); // Stack overflow ?
		
		if(dfl.utf.useUnicode)
		{
			const char[] BROWSE_NAME = "SHBrowseForFolderW";
			const char[] PATH_NAME = "SHGetPathFromIDListW";
			static SHBrowseForFolderWProc browseproc = null;
			static SHGetPathFromIDListWProc pathproc = null;
			
			if(!browseproc)
			{
				HMODULE hmod;
				hmod = GetModuleHandleA("shell32.dll");
				
				browseproc = cast(SHBrowseForFolderWProc)GetProcAddress(hmod, BROWSE_NAME);
				if(!browseproc)
					throw new Exception("Unable to load procedure " ~ BROWSE_NAME ~ ".");
				
				pathproc = cast(SHGetPathFromIDListWProc)GetProcAddress(hmod, PATH_NAME);
				if(!pathproc)
					throw new Exception("Unable to load procedure " ~ PATH_NAME ~ ".");
			}
			
			biw.lpszTitle = dfl.utf.toUnicodez(_desc);
			
			biw.pszDisplayName = cast(wchar*)pdescz;
			if(_desc.length)
			{
				wchar[] tmp;
				tmp = dfl.utf.toUnicode(_desc);
				if(tmp.length >= MAX_PATH)
					_errPathTooLong();
				biw.pszDisplayName[0 .. tmp.length] = tmp;
				biw.pszDisplayName[tmp.length] = 0;
			}
			else
			{
				biw.pszDisplayName[0] = 0;
			}
			
			// Show the dialog!
			LPITEMIDLIST result;
			result = browseproc(&biw);
			
			if(!result)
			{
				biw.lpszTitle = null;
				return false;
			}
			
			if(NOERROR != SHGetMalloc(&shmalloc))
				_errNoShMalloc();
			
			wchar* wbuf = cast(wchar*)std.c.stdlib.alloca(wchar.sizeof * MAX_PATH);
			if(!pathproc(result, wbuf))
			{
				shmalloc.Free(result);
				shmalloc.Release();
				_errNoGetPath();
				assert(0);
			}
			
			_selpath = dfl.utf.fromUnicodez(wbuf); // Assumes fromUnicodez() copies.
			
			shmalloc.Free(result);
			shmalloc.Release();
			
			biw.lpszTitle = null;
		}
		else
		{
			bia.lpszTitle = dfl.utf.toAnsiz(_desc);
			
			bia.pszDisplayName = cast(char*)pdescz;
			if(_desc.length)
			{
				char[] tmp; // ansi.
				tmp = dfl.utf.toAnsi(_desc);
				if(tmp.length >= MAX_PATH)
					_errPathTooLong();
				bia.pszDisplayName[0 .. tmp.length] = tmp;
				bia.pszDisplayName[tmp.length] = 0;
			}
			else
			{
				bia.pszDisplayName[0] = 0;
			}
			
			// Show the dialog!
			LPITEMIDLIST result;
			result = SHBrowseForFolderA(&bia);
			
			if(!result)
			{
				bia.lpszTitle = null;
				return false;
			}
			
			if(NOERROR != SHGetMalloc(&shmalloc))
				_errNoShMalloc();
			
			char* abuf = cast(char*)std.c.stdlib.alloca(char.sizeof * MAX_PATH);
			if(!SHGetPathFromIDListA(result, abuf))
			{
				shmalloc.Free(result);
				shmalloc.Release();
				_errNoGetPath();
				assert(0);
			}
			
			_selpath = dfl.utf.fromAnsiz(abuf); // Assumes fromAnsiz() copies.
			
			shmalloc.Free(result);
			shmalloc.Release();
			
			bia.lpszTitle = null;
		}
		
		return true;
	}
	
	
	private:
	union
	{
		BROWSEINFOW biw;
		BROWSEINFOA bia;
		alias biw bi;
		
		static assert(BROWSEINFOW.sizeof == BROWSEINFOA.sizeof);
		static assert(BROWSEINFOW.ulFlags.offsetof == BROWSEINFOA.ulFlags.offsetof);
	}
	
	char[] _desc;
	char[] _selpath;
	
	
	const UINT INIT_FLAGS = BIF_RETURNONLYFSDIRS | BIF_NEWDIALOGSTYLE;
}

