// Written by Christopher E. Miller
// See the included license.txt for copyright and license details.


module dfl.form;

private import std.string;

private import dfl.control, dfl.winapi, dfl.event, dfl.drawing;
private import dfl.application, dfl.base, dfl.menu, dfl.utf;
private import dfl.collections;


private extern(Windows) void _initMdiclient();


enum FormBorderStyle: ubyte //: BorderStyle
{
	NONE = BorderStyle.NONE,
	
	FIXED_3D = BorderStyle.FIXED_3D,
	FIXED_SINGLE = BorderStyle.FIXED_SINGLE,
	FIXED_DIALOG,
	SIZABLE,
	FIXED_TOOLWINDOW,
	SIZABLE_TOOLWINDOW,
}


deprecated enum SizeGripStyle: ubyte
{
	AUTO,
	HIDE,
	SHOW,
}


enum FormStartPosition: ubyte
{
	CENTER_PARENT,
	CENTER_SCREEN,
	MANUAL,
	WINDOWS_DEFAULT_BOUNDS,
	WINDOWS_DEFAULT_LOCATION,
}


enum FormWindowState: ubyte
{
	MAXIMIZED,
	MINIMIZED,
	NORMAL,
}


enum MdiLayout: ubyte
{
	ARRANGE_ICONS,
	CASCADE,
	TILE_HORIZONTAL,
	TILE_VERTICAL,
}


// The Form's shortcut was pressed.
class FormShortcutEventArgs: EventArgs
{
	this(Keys shortcut)
	{
		this._shortcut = shortcut;
	}
	
	
	final Keys shortcut() // getter
	{
		return _shortcut;
	}
	
	
	private:
	Keys _shortcut;
}


// DMD 0.93 crashes if this is placed in Form.
//private import dfl.button;


class Form: ContainerControl // docmain
{
	final void acceptButton(IButtonControl btn) // setter
	{
		if(acceptBtn)
			acceptBtn.notifyDefault(false);
		
		acceptBtn = btn;
		
		if(btn)
			btn.notifyDefault(true);
	}
	
	
	final IButtonControl acceptButton() // getter
	{
		return acceptBtn;
	}
	
	
	final void cancelButton(IButtonControl btn) // setter
	{
		cancelBtn = btn;
	}
	
	
	final IButtonControl cancelButton() // getter
	{
		return cancelBtn;
	}
	
	
	// An exception is thrown if the shortcut was already added.
	final void addShortcut(Keys shortcut, void delegate(Object sender, FormShortcutEventArgs ea) pressed)
	in
	{
		assert(shortcut & Keys.KEY_CODE); // At least one key code.
		assert(pressed !is null);
	}
	body
	{
		if(shortcut in _shortcuts)
			throw new DflException("Shortcut key conflict.");
		
		_shortcuts[shortcut] = pressed;
	}
	
	
	final void removeShortcut(Keys shortcut)
	{
		//delete _shortcuts[shortcut];
		_shortcuts.remove(shortcut);
	}
	
	
	static Form activeForm() // getter
	{
		return cast(Form)fromHandle(GetActiveWindow());
	}
	
	
	final Form getActiveMdiChild() // getter
	{
		return cast(Form)fromHandle(cast(HWND)SendMessageA(handle, WM_MDIGETACTIVE, 0, 0));
	}
	
	
	protected override Size defaultSize() // getter
	{
		return Size(300, 300);
	}
	
	
	protected override void createParams(inout CreateParams cp)
	{
		super.createParams(cp);
		
		Control cwparent;
		if(cp.style & WS_CHILD)
			cwparent = wparent;
		else
			cwparent = wowner;
		
		cp.className = FORM_CLASSNAME;
		cp.menu = wmenu ? wmenu.handle : null;
		
		//cp.parent = wparent ? wparent.handle : null;
		//if(!(cp.style & WS_CHILD))
		//	cp.parent = wowner ? wowner.handle : null;
		cp.parent = cwparent ? cwparent.handle : null;
		version(DFL_PARK_WINDOW)
		{
			if(!cp.parent && !showInTaskbar())
				cp.parent = getParkHwnd();
		}
		
		switch(startpos)
		{
			case FormStartPosition.CENTER_PARENT:
				if(cwparent && cwparent.visible())
				{
					cp.x = cwparent.left + ((cwparent.width - cp.width) / 2);
					cp.y = cwparent.top + ((cwparent.height - cp.height) / 2);
					
					// Make sure part of the form isn't off the screen.
					RECT area;
					SystemParametersInfoA(SPI_GETWORKAREA, 0, &area, FALSE);
					if(cp.x < area.left)
						cp.x = area.left;
					else if(cp.x + cp.width > area.right)
						cp.x = area.right - cp.width;
					if(cp.y < area.top)
						cp.y = area.top;
					else if(cp.y + cp.height > area.bottom)
						cp.y = area.bottom - cp.height;
					break;
				}
				// No parent so use the screen.
			case FormStartPosition.CENTER_SCREEN:
				{
					// TODO: map to client coords if MDI child.
					
					RECT area;
					SystemParametersInfoA(SPI_GETWORKAREA, 0, &area, FALSE);
					
					cp.x = area.left + (((area.right - area.left) - cp.width) / 2);
					cp.y = area.top + (((area.bottom - area.top) - cp.height) / 2);
				}
				break;
			
			case FormStartPosition.WINDOWS_DEFAULT_BOUNDS:
				// WM_CREATE fixes these.
				cp.width = CW_USEDEFAULT;
				cp.height = CW_USEDEFAULT;
				//break; // WINDOWS_DEFAULT_BOUNDS assumes default location.
			case FormStartPosition.WINDOWS_DEFAULT_LOCATION:
				// WM_CREATE fixes these.
				cp.x = CW_USEDEFAULT;
				//cp.y = CW_USEDEFAULT;
				cp.y = visible ? SW_SHOW : SW_HIDE;
				break;
			
			default: ;
		}
	}
	
	
	protected override void createHandle()
	{
		// This code is reimplemented to allow some tricks.
		
		if(isHandleCreated)
			return;
		
		if(killing)
		{
			create_err:
			throw new DflException("Form creation failure.");
			//throw new DflException(Object.toString() ~ " creation failure."); // ?
		}
		
		// Need the owner's handle to exist.
		if(wowner)
		//	wowner.createHandle(); // DMD 0.111: class dfl.control.Control member createHandle is not accessible
			wowner._createHandle();
		
		// This is here because wowner.createHandle() might create me.
		//if(created)
		if(isHandleCreated)
			return;
		
		DWORD vis;
		CreateParams cp;
		
		createParams(cp);
		assert(!isHandleCreated); // Make sure the handle wasn't created in createParams().
		
		with(cp)
		{
			wtext = caption;
			//wrect = Rect(x, y, width, height); // Avoid CW_USEDEFAULT problems. This gets updated in WM_CREATE.
			wclassStyle = classStyle;
			wexstyle = exStyle;
			wstyle = style;
			
			// Use local var to avoid changing -cp- at this point.
			int ly;
			ly = y;
			
			// Delay setting visible.
			vis = wstyle;
			if(x == CW_USEDEFAULT)
				ly = SW_HIDE;
			
			Application.creatingControl(this);
			hwnd = dfl.utf.createWindowEx(exStyle, className, caption, wstyle & ~WS_VISIBLE,
				x, ly, width, height, parent, menu, inst, param);
			if(!hwnd)
				goto create_err;
		}
		
		if(setLayeredWindowAttributes)
		{
			BYTE alpha = opacityToAlpha(opa);
			DWORD flags = 0;
			
			if(alpha != BYTE.max)
				flags |= LWA_ALPHA;
			
			if(transKey != Color.empty)
				flags |= LWA_COLORKEY;
			
			if(flags)
			{
				//_exStyle(_exStyle() | WS_EX_LAYERED); // Should already be set.
				setLayeredWindowAttributes(hwnd, transKey.toRgb(), alpha, flags);
			}
		}
		
		//createChildren();
		try
		{
			createChildren(); // Might throw.
		}
		catch(Object e)
		{
			Application.onThreadException(e);
		}
		alayout(this);
		
		if(vis & WS_VISIBLE)
		{
			wstyle |= WS_VISIBLE;
			doShow();
		}
		
		Application.addMessageFilter(mfilter = new FormMessageFilter(this)); // To process IsDialogMessage().
		
		//onHandleCreated(EventArgs.empty); // Called in Control's WM_CREATE now.
		
		onLoad(EventArgs.empty);
	}
	
	
	/+
	final void autoScale(bool byes) // setter
	{
		autoscale = byes;
	}
	
	
	final bool autoScale() // getter
	{
		return autoscale;
	}
	
	
	void autoScaleBaseSize(Size sz) // setter
	{
		// Can't be changed after creation.
		if(!created)
			autoscaleBase = sz;
	}
	
	
	Size autoScaleBaseSize() // getter
	{
		return autoscaleBase;
	}
	+/
	
	
	/+
	// Focused children are scrolled into view.
	override void autoScroll(bool byes) // setter
	{
		super.autoScroll(byes);
	}
	
	
	override bool autoScroll() // getter
	{
		return super.autoScroll(byes);
	}
	+/
	
	
	// This only works if the windows version is
	// set to 4.0 or higher.
	
	final void controlBox(bool byes) // setter
	{
		if(byes)
			_style(_style() | WS_SYSMENU);
		else
			_style(_style() & ~WS_SYSMENU);
		
		// Update taskbar button.
		if(created)
		{
			if(visible)
			{
				//hide();
				//show();
				// Do it directly so that DFL code can't prevent it.
				doHide();
				doShow();
			}
		}
	}
	
	
	final bool controlBox() // getter
	{
		return (_style() & WS_SYSMENU) != 0;
	}
	
	
	final void desktopBounds(Rect r) // setter
	{
		RECT rect;
		if(r.width < 0)
			r.width = 0;
		if(r.height < 0)
			r.height = 0;
		r.getRect(&rect);
		
		//Control par = parent;
		//if(par) // Convert from screen coords to parent coords.
		//	MapWindowPoints(null, par.handle, cast(POINT*)&rect, 2);
		
		setBoundsCore(rect.left, rect.top, rect.right - rect.left, rect.bottom - rect.top, BoundsSpecified.ALL);
	}
	
	
	final Rect desktopBounds() // getter
	{
		RECT r;
		GetWindowRect(handle, &r);
		return Rect(&r);
	}
	
	
	final void desktopLocation(Point dp) // setter
	{
		//Control par = parent;
		//if(par) // Convert from screen coords to parent coords.
		//	MapWindowPoints(null, par.handle, &dp.point, 1);
		
		setBoundsCore(dp.x, dp.y, 0, 0, BoundsSpecified.LOCATION);
	}
	
	
	final Point desktopLocation() // getter
	{
		RECT r;
		GetWindowRect(handle, &r);
		return Point(r.left, r.top);
	}
	
	
	final void dialogResult(DialogResult dr) // setter
	{
		fresult = dr;
	}
	
	
	final DialogResult dialogResult() // getter
	{
		return fresult;
	}
	
	
	final void formBorderStyle(FormBorderStyle bstyle) // setter
	{
		bool vis = false;
		
		if(created() && visible())
		{
			vis = true;
			// Do it directly so that DFL code can't prevent it.
			doHide();
		}
		
		LONG st;
		LONG exst;
		//Size csz;
		st = _style();
		exst = _exStyle();
		//csz = clientSize;
		
		switch(bstyle)
		{
			case FormBorderStyle.FIXED_3D:
				st &= ~(WS_BORDER | WS_THICKFRAME | WS_DLGFRAME);
				exst &= ~(WS_EX_TOOLWINDOW | WS_EX_STATICEDGE);
				
				st |= WS_CAPTION;
				exst |= WS_EX_CLIENTEDGE | WS_EX_DLGMODALFRAME | WS_EX_WINDOWEDGE;
				break;
			
			case FormBorderStyle.FIXED_DIALOG:
				st &= ~(WS_BORDER | WS_THICKFRAME);
				exst &= ~(WS_EX_TOOLWINDOW | WS_EX_CLIENTEDGE | WS_EX_STATICEDGE);
				
				st |= WS_CAPTION | WS_DLGFRAME;
				exst |= WS_EX_DLGMODALFRAME | WS_EX_WINDOWEDGE;
				break;
			
			case FormBorderStyle.FIXED_SINGLE:
				st &= ~(WS_THICKFRAME | WS_DLGFRAME);
				exst &= ~(WS_EX_TOOLWINDOW | WS_EX_CLIENTEDGE | WS_EX_WINDOWEDGE | WS_EX_STATICEDGE);
				
				st |= WS_BORDER | WS_CAPTION;
				exst |= WS_EX_DLGMODALFRAME;
				break;
			
			case FormBorderStyle.FIXED_TOOLWINDOW:
				st &= ~(WS_BORDER | WS_THICKFRAME | WS_DLGFRAME);
				exst &= ~(WS_EX_CLIENTEDGE | WS_EX_STATICEDGE);
				
				st |= WS_CAPTION;
				exst |= WS_EX_TOOLWINDOW | WS_EX_WINDOWEDGE | WS_EX_DLGMODALFRAME;
				break;
			
			case FormBorderStyle.SIZABLE:
				st &= ~(WS_BORDER | WS_DLGFRAME);
				exst &= ~(WS_EX_TOOLWINDOW | WS_EX_CLIENTEDGE | WS_EX_DLGMODALFRAME | WS_EX_STATICEDGE);
				
				st |= WS_THICKFRAME | WS_CAPTION;
				exst |= WS_EX_WINDOWEDGE;
				break;
			
			case FormBorderStyle.SIZABLE_TOOLWINDOW:
				st &= ~(WS_BORDER | WS_DLGFRAME);
				exst &= ~(WS_EX_CLIENTEDGE | WS_EX_DLGMODALFRAME | WS_EX_STATICEDGE);
				
				st |= WS_THICKFRAME | WS_CAPTION;
				exst |= WS_EX_TOOLWINDOW | WS_EX_WINDOWEDGE;
				break;
			
			case FormBorderStyle.NONE:
				st &= ~(WS_BORDER | WS_THICKFRAME | WS_CAPTION | WS_DLGFRAME);
				exst &= ~(WS_EX_TOOLWINDOW | WS_EX_CLIENTEDGE | WS_EX_DLGMODALFRAME | WS_EX_STATICEDGE | WS_EX_WINDOWEDGE);
				break;
		}
		
		_style(st);
		_exStyle(exst);
		//clientSize = csz;
		
		// Update taskbar button.
		if(created)
		{
			if(vis)
			{
				//hide();
				//show();
				SetWindowPos(hwnd, null, 0, 0, 0, 0, SWP_FRAMECHANGED | SWP_NOMOVE
					| SWP_NOSIZE | SWP_NOZORDER | SWP_NOACTIVATE); // Recalculate the frame while hidden.
				// Do it directly so that DFL code can't prevent it.
				doShow();
			}
			else
			{
				SetWindowPos(hwnd, null, 0, 0, 0, 0, SWP_FRAMECHANGED | SWP_NOMOVE
					| SWP_NOSIZE | SWP_NOZORDER | SWP_NOACTIVATE); // Recalculate the frame.
			}
		}
	}
	
	
	final FormBorderStyle formBorderStyle() // getter
	{
		LONG st = _style();
		LONG exst = _exStyle();
		
		if(exst & WS_EX_TOOLWINDOW)
		{
			if(st & WS_THICKFRAME)
				return FormBorderStyle.SIZABLE_TOOLWINDOW;
			else
				return FormBorderStyle.FIXED_TOOLWINDOW;
		}
		else
		{
			if(st & WS_THICKFRAME)
			{
				return FormBorderStyle.SIZABLE;
			}
			else
			{
				if(exst & WS_EX_CLIENTEDGE)
					return FormBorderStyle.FIXED_3D;
				
				if(exst & WS_EX_WINDOWEDGE)
					return FormBorderStyle.FIXED_DIALOG;
				
				if(st & WS_BORDER)
					return FormBorderStyle.FIXED_SINGLE;
			}
		}
		
		return FormBorderStyle.NONE;
	}
	
	
	// Ignored if min and max buttons are enabled.
	final void helpButton(bool byes) // setter
	{
		if(byes)
			_exStyle(_exStyle() | WS_EX_CONTEXTHELP);
		else
			_exStyle(_exStyle() & ~WS_EX_CONTEXTHELP);
		
		redrawEntire();
	}
	
	
	final bool helpButton() // getter
	{
		return (_exStyle() & WS_EX_CONTEXTHELP) != 0;
	}
	
	
	private void _setIcon()
	{
		HICON hico, hicoSm;
		
		if(wicon)
		{
			hico = wicon.handle;
			
			int smx, smy;
			smx = GetSystemMetrics(SM_CXSMICON);
			smy = GetSystemMetrics(SM_CYSMICON);
			hicoSm = CopyImage(hico, IMAGE_ICON, smx, smy, LR_COPYFROMRESOURCE);
			if(!hicoSm)
				hicoSm = CopyImage(hico, IMAGE_ICON, smx, smy, 0);
			if(hicoSm)
			{
				wiconSm = new Icon(hicoSm);
			}
			else
			{
				wiconSm = null;
				hicoSm = hico;
			}
		}
		else
		{
			hico = null;
			hicoSm = null;
			
			wiconSm = null;
		}
		
		SendMessageA(hwnd, WM_SETICON, ICON_BIG, cast(LPARAM)hico);
		SendMessageA(hwnd, WM_SETICON, ICON_SMALL, cast(LPARAM)hicoSm);
		
		if(visible)
			redrawEntire();
	}
	
	
	final void icon(Icon ico) // setter
	{
		wicon = ico;
		
		if(created)
			_setIcon();
	}
	
	
	final Icon icon() // getter
	{
		return wicon;
	}
	
	
	// TODO: implement.
	// keyPreview
	
	
	final bool isMdiChild() // getter
	{
		return (_exStyle() & WS_EX_MDICHILD) != 0;
	}
	
	
	version(NO_MDI) {} else
	{
		final void isMdiContainer(bool byes) // setter
		{
			if(mdiClient)
			{
				if(!byes)
				{
					// Remove MDI client.
					mdiClient.dispose();
					mdiClient = null;
				}
			}
			else
			{
				if(byes)
				{
					// Create MDI client.
					mdiClient = new MdiClient;
					mdiClient.parent = this;
				}
			}
		}
		
		
		final bool isMdiContainer() // getter
		{
			return !(mdiClient is null);
		}
		
		
		final Form[] mdiChildren() // getter
		{
			/+
			if(!mdiClient)
				return null;
			+/
			
			return _mdiChildren;
		}
		
		
		// parent is the MDI client and mdiParent is the MDI frame.
		
		
		final void mdiParent(Form frm) // setter
		in
		{
			if(frm)
			{
				assert(frm.isMdiContainer);
				assert(!(frm.mdiClient is null));
			}
		}
		/+out
		{
			if(frm)
			{
				bit found = false;
				foreach(Form elem; frm._mdiChildren)
				{
					if(elem is this)
					{
						found = true;
						break;
					}
				}
				assert(found);
			}
		}+/
		body
		{
			if(wmdiparent is frm)
				return;
			
			_removeFromOldOwner();
			wowner = null;
			wmdiparent = null; // Safety in case of exception.
			
			if(frm)
			{
				if(created)
				{
					frm.createControl();
					frm.mdiClient.createControl(); // Should already be done from frm.createControl().
				}
				
				// Copy so that old mdiChildren arrays won't get overwritten.
				Form[] _thisa = new Form[1]; // DMD 0.123: this can't be a static array or the append screws up.
				_thisa[0] = this;
				frm._mdiChildren = frm._mdiChildren ~ _thisa;
				
				_style((_style() | WS_CHILD) & ~WS_POPUP);
				_exStyle(_exStyle() | WS_EX_MDICHILD);
				
				wparent = frm.mdiClient;
				wmdiparent = frm;
				if(created)
					SetParent(hwnd, frm.mdiClient.hwnd);
			}
			else
			{
				_exStyle(_exStyle() & ~WS_EX_MDICHILD);
				_style((_style() | WS_POPUP) & ~WS_CHILD);
				
				if(created)
					SetParent(hwnd, null);
				wparent = null;
				
				//wmdiparent = null;
			}
		}
		
		
		final Form mdiParent() // getter
		{
			//if(isMdiChild)
				return wmdiparent;
			return null;
		}
	}
	
	
	final void maximizeBox(bool byes) // setter
	{
		if(byes)
			_style(_style() | WS_MAXIMIZEBOX);
		else
			_style(_style() & ~WS_MAXIMIZEBOX);
		
		redrawEntire();
	}
	
	
	final bool maximizeBox() // getter
	{
		return (_style() & WS_MAXIMIZEBOX) != 0;
	}
	
	
	final void minimizeBox(bool byes) // setter
	{
		if(byes)
			_style(_style() | WS_MINIMIZEBOX);
		else
			_style(_style() & ~WS_MINIMIZEBOX);
		
		redrawEntire();
	}
	
	
	final bool minimizeBox() // getter
	{
		return (_style() & WS_MINIMIZEBOX) != 0;
	}
	
	
	protected override void onHandleCreated(EventArgs ea)
	{
		super.onHandleCreated(ea);
		
		if(wmenu)
			wmenu._setHwnd(handle);
		
		_setIcon();
		
		//SendMessageA(handle, DM_SETDEFID, IDOK, 0);
	}
	
	
	protected override void onResize(EventArgs ea)
	{
		super.onResize(ea);
		
		if(_isPaintingSizeGrip)
		{
			RECT rect;
			_getSizeGripArea(&rect);
			InvalidateRect(hwnd, &rect, TRUE);
		}
	}
	
	
	private void _getSizeGripArea(RECT* rect)
	{
		rect.right = clientSize.width;
		rect.bottom = clientSize.height;
		rect.left = rect.right - GetSystemMetrics(SM_CXVSCROLL);
		rect.top = rect.bottom - GetSystemMetrics(SM_CYHSCROLL);
	}
	
	
	private bool _isPaintingSizeGrip()
	{
		if(grip)
		{
			if(wstyle & WS_THICKFRAME)
			{
				return !(wstyle & (WS_MINIMIZE | WS_MAXIMIZE |
					WS_VSCROLL | WS_HSCROLL));
			}
		}
		return false;
	}
	
	
	protected override void onPaint(PaintEventArgs ea)
	{
		super.onPaint(ea);
		
		if(_isPaintingSizeGrip)
		{
			/+
			RECT rect;
			_getSizeGripArea(&rect);
			DrawFrameControl(ea.graphics.handle, &rect, DFC_SCROLL, DFCS_SCROLLSIZEGRIP);
			+/
			
			ea.graphics.drawSizeGrip(clientSize.width, clientSize.height);
		}
	}
	
	
	final void menu(MainMenu menu) // setter
	{
		if(created)
		{
			HWND hwnd;
			hwnd = handle;
			
			if(menu)
			{
				SetMenu(hwnd, menu.handle);
				menu._setHwnd(hwnd);
			}
			else
			{
				SetMenu(hwnd, null);
			}
			
			if(wmenu)
				wmenu._setHwnd(HWND.init);
			wmenu = menu;
			
			DrawMenuBar(hwnd);
		}
		else
		{
			wmenu = menu;
			_recalcClientSize();
		}
	}
	
	
	final MainMenu menu() // getter
	{
		return wmenu;
	}
	
	
	/+
	final MainMenu mergedMenu() // getter
	{
		// Return menu belonging to active MDI child if maximized ?
	}
	+/
	
	
	final void minimumSize(Size min) // setter
	{
		minsz = min;
		// Restrict now ?
	}
	
	
	final Size minimumSize() // getter
	{
		return minsz;
	}
	
	
	final void maximumSize(Size max) // setter
	{
		maxsz = max;
		// Restrict now ?
	}
	
	
	final Size maximumSize() // getter
	{
		return maxsz;
	}
	
	
	final bool modal() // getter
	{
		return wmodal;
	}
	
	
	// If opacity and transparency are supported.
	static bool supportsOpacity() // getter
	{
		return setLayeredWindowAttributes != null;
	}
	
	
	private static BYTE opacityToAlpha(double opa)
	{
		return cast(BYTE)(opa * BYTE.max);
	}
	
	
	// 1.0 is 100%, 0.0 is 0%, 0.75 is 75%.
	// Does nothing if not supported.
	final void opacity(double opa) // setter
	{
		if(setLayeredWindowAttributes)
		{
			BYTE alpha;
			
			if(opa >= 1.0)
			{
				this.opa = 1.0;
				alpha = BYTE.max;
			}
			else if(opa <= 0.0)
			{
				this.opa = 0.0;
				alpha = BYTE.min;
			}
			else
			{
				this.opa = opa;
				alpha = opacityToAlpha(opa);
			}
			
			if(alpha == BYTE.max) // Disable
			{
				if(transKey == Color.empty)
					_exStyle(_exStyle() & ~WS_EX_LAYERED);
				else
					setLayeredWindowAttributes(handle, transKey.toRgb(), 0, LWA_COLORKEY);
			}
			else
			{
				_exStyle(_exStyle() | WS_EX_LAYERED);
				if(created)
				{
					//_exStyle(_exStyle() | WS_EX_LAYERED);
					if(transKey == Color.empty)
						setLayeredWindowAttributes(handle, 0, alpha, LWA_ALPHA);
					else
						setLayeredWindowAttributes(handle, transKey.toRgb(), alpha, LWA_ALPHA | LWA_COLORKEY);
				}
			}
		}
	}
	
	
	final double opacity() // getter
	{
		return opa;
	}
	
	
	/+
	final Form[] ownedForms() // getter
	{
		// TODO: implement.
	}
	+/
	
	
	// the "old owner" is the current -wowner- or -wmdiparent-.
	// If neither are set, nothing happens.
	private void _removeFromOldOwner()
	{
		int idx;
		
		if(wmdiparent)
		{
			idx = findIsIndex!(Form)(wmdiparent._mdiChildren, this);
			if(-1 != idx)
				wmdiparent._mdiChildren = removeIndex!(Form)(wmdiparent._mdiChildren, idx);
			//else
			//	assert(0);
		}
		else if(wowner)
		{
			idx = findIsIndex!(Form)(wowner._owned, this);
			if(-1 != idx)
				wowner._owned = removeIndex!(Form)(wowner._owned, idx);
			//else
			//	assert(0);
		}
	}
	
	
	final void owner(Form frm) // setter
	/+out
	{
		if(frm)
		{
			bit found = false;
			foreach(Form elem; frm._owned)
			{
				if(elem is this)
				{
					found = true;
					break;
				}
			}
			assert(found);
		}
	}+/
	body
	{
		if(wowner is frm)
			return;
		
		// Remove from old owner.
		_removeFromOldOwner();
		wmdiparent = null;
		wowner = null; // Safety in case of exception.
		_exStyle(_exStyle() & ~WS_EX_MDICHILD);
		_style((_style() | WS_POPUP) & ~WS_CHILD);
		
		// Add to new owner.
		if(frm)
		{
			if(created)
			{
				frm.createControl();
			}
			
			// Copy so that old ownedForms arrays won't get overwritten.
			Form[] _thisa = new Form[1]; // DMD 0.123: this can't be a static array or the append screws up.
			_thisa[0] = this;
			frm._owned = frm._owned ~ _thisa;
			
			wowner = frm;
			if(created)
				SetParent(hwnd, frm.hwnd);
		}
		else
		{
			if(created)
				SetParent(hwnd, null);
		}
		
		//wowner = frm;
	}
	
	
	final Form owner() // getter
	{
		return wowner;
	}
	
	
	// This function does not work in all cases.
	final void showInTaskbar(bool byes) // setter
	{
		if(created)
		{
			bool vis;
			vis = visible;
			
			if(vis)
			{
				//hide();
				// Do it directly so that DFL code can't prevent it.
				doHide();
			}
			
			if(byes)
			{
				_exStyle(_exStyle() | WS_EX_APPWINDOW);
				
				version(DFL_PARK_WINDOW)
				{
					if(_hwPark && GetParent(handle) == _hwPark)
						SetParent(handle, null);
				}
			}
			else
			{
				_exStyle(_exStyle() & ~WS_EX_APPWINDOW);
				
				version(DFL_PARK_WINDOW)
				{
					if(!GetParent(handle))
						SetParent(handle, getParkHwnd());
				}
			}
			
			if(vis)
			{
				//show();
				// Do it directly so that DFL code can't prevent it.
				doShow();
			}
		}
		else
		{
			if(byes)
				wexstyle |= WS_EX_APPWINDOW;
			else
				wexstyle &= ~WS_EX_APPWINDOW;
		}
	}
	
	
	final bool showInTaskbar() // getter
	{
		return (_exStyle() & WS_EX_APPWINDOW) != 0;
	}
	
	
	final void sizingGrip(bool byes) // setter
	{
		if(grip == byes)
			return;
		
		this.grip = byes;
		
		if(created)
		{
			RECT rect;
			_getSizeGripArea(&rect);
			
			InvalidateRect(hwnd, &rect, TRUE);
		}
	}
	
	
	final bool sizingGrip() // getter
	{
		return grip;
	}
	
	
	final void startPosition(FormStartPosition startpos) // setter
	{
		this.startpos = startpos;
	}
	
	
	final FormStartPosition startPosition() // getter
	{
		return startpos;
	}
	
	
	final void topMost(bool byes) // setter
	{
		/+
		if(byes)
			_exStyle(_exStyle() | WS_EX_TOPMOST);
		else
			_exStyle(_exStyle() & ~WS_EX_TOPMOST);
		+/
		
		if(created)
		{
			SetWindowPos(handle, byes ? HWND_TOPMOST : HWND_NOTOPMOST, 0, 0, 0, 0, SWP_NOMOVE | SWP_NOSIZE);
		}
		else
		{
			if(byes)
				wexstyle |= WS_EX_TOPMOST;
			else
				wexstyle &= ~WS_EX_TOPMOST;
		}
	}
	
	
	final bool topMost() // getter
	{
		return (_exStyle() & WS_EX_TOPMOST) != 0;
	}
	
	
	final void transparencyKey(Color c) // setter
	{
		if(setLayeredWindowAttributes)
		{
			transKey = c;
			BYTE alpha = opacityToAlpha(opa);
			
			if(c == Color.empty) // Disable
			{
				if(alpha == BYTE.max)
					_exStyle(_exStyle() & ~WS_EX_LAYERED);
				else
					setLayeredWindowAttributes(handle, 0, alpha, LWA_ALPHA);
			}
			else
			{
				_exStyle(_exStyle() | WS_EX_LAYERED);
				if(created)
				{
					//_exStyle(_exStyle() | WS_EX_LAYERED);
					if(alpha == BYTE.max)
						setLayeredWindowAttributes(handle, c.toRgb(), 0, LWA_COLORKEY);
					else
						setLayeredWindowAttributes(handle, c.toRgb(), alpha, LWA_COLORKEY | LWA_ALPHA);
				}
			}
		}
	}
	
	
	final Color transparencyKey() // getter
	{
		return transKey;
	}
	
	
	final void windowState(FormWindowState state) // setter
	{
		// Not sure if visible should be checked here..
		if(created() && visible())
		{
			switch(state)
			{
				case FormWindowState.MAXIMIZED:
					ShowWindow(handle, SW_MAXIMIZE);
					break;
				
				case FormWindowState.MINIMIZED:
					ShowWindow(handle, SW_MINIMIZE);
					break;
				
				case FormWindowState.NORMAL:
					ShowWindow(handle, SW_RESTORE);
					break;
			}
		}
		else
		{
			switch(state)
			{
				case FormWindowState.MAXIMIZED:
					_style(_style() & ~WS_MINIMIZE | WS_MAXIMIZE);
					break;
				
				case FormWindowState.MINIMIZED:
					_style(_style() | WS_MINIMIZE & ~WS_MAXIMIZE);
					break;
				
				case FormWindowState.NORMAL:
					_style(_style() & ~(WS_MINIMIZE | WS_MAXIMIZE));
					break;
			}
		}
	}
	
	
	final FormWindowState windowState() // getter
	{
		LONG wl = _style();
		
		if(wl & WS_MAXIMIZE)
			return FormWindowState.MAXIMIZED;
		else if(wl & WS_MINIMIZE)
			return FormWindowState.MINIMIZED;
		else
			return FormWindowState.NORMAL;
	}
	
	
	final void activate()
	{
		// ?
		if(!visible)
			show();
		
		SetActiveWindow(handle);
	}
	
	
	override void destroyHandle()
	{
		if(!isHandleCreated)
			return;
		
		if(isMdiChild)
			DefMDIChildProcA(hwnd, WM_CLOSE, 0, 0);
		DestroyWindow(hwnd);
	}
	
	
	final void close()
	{
		CancelEventArgs cea = new CancelEventArgs;
		onClosing(cea);
		if(!cea.cancel)
		{
			destroyHandle();
		}
	}
	
	
	final void layoutMdi(MdiLayout lay)
	{
		switch(lay)
		{
			case MdiLayout.ARRANGE_ICONS:
				SendMessageA(handle, WM_MDIICONARRANGE, 0, 0);
				break;
			
			case MdiLayout.CASCADE:
				SendMessageA(handle, WM_MDICASCADE, 0, 0);
				break;
			
			case MdiLayout.TILE_HORIZONTAL:
				SendMessageA(handle, WM_MDITILE, MDITILE_HORIZONTAL, 0);
				break;
			
			case MdiLayout.TILE_VERTICAL:
				SendMessageA(handle, WM_MDITILE, MDITILE_VERTICAL, 0);
				break;
		}
	}
	
	
	final void setDesktopBounds(int x, int y, int width, int height)
	{
		desktopBounds = Rect(x, y, width, height);
	}
	
	
	final void setDesktopLocation(int x, int y)
	{
		desktopLocation = Point(x, y);
	}
	
	
	final void showDialog()
	{
		// Use active window as the owner.
		this.sowner = GetActiveWindow();
		showDialog2();
	}
	
	
	final void showDialog(IWin32Window iwsowner)
	{
		this.sowner = iwsowner ? iwsowner.handle : GetActiveWindow();
		showDialog2();
	}
	
	
	// Used internally.
	package final void showDialog2()
	{
		LONG wl = _style();
		sownerEnabled = false;
		
		if(wl & WS_DISABLED)
			throw new DflException("Unable to show dialog because it is disabled.");
		
		if(created)
		{
			if(wl & WS_VISIBLE)
				throw new DflException("Unable to show dialog because it is already visible.");
			
			if(sowner == hwnd)
			{
				bad_owner:
				throw new DflException("Invalid dialog owner.");
			}
			
			//owner = null;
			//_exStyle(_exStyle() & ~WS_EX_MDICHILD);
			//_style((_style() | WS_POPUP) & ~WS_CHILD);
			//SetParent(hwnd, sowner);
		}
		
		if(sowner)
		{
			LONG owl = GetWindowLongA(sowner, GWL_STYLE);
			if(owl & WS_CHILD)
				goto bad_owner;
			
			wowner = cast(Form)fromHandle(sowner);
			
			if(!(owl & WS_DISABLED))
			{
				sownerEnabled = true;
				EnableWindow(sowner, false);
			}
		}
		
		try
		{
			show();
			
			wmodal = true;
			for(;;)
			{
				Application.doEvents();
				if(!wmodal)
					break;
				Application.waitMsg();
			}
		}
		finally
		{
			/+
			if(sownerEnabled)
				EnableWindow(sowner, true);
			SetFocus(sowner);
			+/
			
			//if(!wmodal)
			//	DestroyWindow(hwnd);
			
			wmodal = false;
			
			//hide();
			// Do it directly so that DFL code can't prevent it.
			doHide();
		}
	}
	
	
	EventHandler activated;
	EventHandler deactivate;
	EventHandler closed;
	CancelEventHandler closing;
	EventHandler load;
	
	
	protected void onActivated(EventArgs ea)
	{
		activated(this, ea);
	}
	
	
	protected void onDeactivate(EventArgs ea)
	{
		deactivate(this, ea);
	}
	
	
	/+
	protected void onInputLanguageChanged(InputLanguageChangedEventArgs ilcea)
	{
		inputLanguageChanged(this, ilcea);
	}
	
	
	protected void onInputLanguageChanging(InputLanguageChangingEventArgs ilcea)
	{
		inputLanguageChanging(this, ilcea);
	}
	+/
	
	
	protected void onLoad(EventArgs ea)
	{
		load(this, ea);
	}
	
	
	private void _init()
	{
		_recalcClientSize();
		
		wicon = new Icon(LoadIconA(null, IDI_APPLICATION), false);
		transKey = Color.empty;
	}
	
	
	this()
	{
		super();
		
		// Default border: FormBorderStyle.SIZABLE.
		wstyle = WS_VISIBLE | WS_OVERLAPPEDWINDOW | WS_CLIPCHILDREN | WS_CLIPSIBLINGS;
		wexstyle = WS_EX_CONTROLPARENT | WS_EX_WINDOWEDGE | WS_EX_APPWINDOW;
		
		_init();
	}
	
	
	/+
	// Used internally.
	this(HWND hwnd)
	{
		super(hwnd);
		_init();
	}
	+/
	
	
	protected override void wndProc(inout Message msg)
	{
		switch(msg.msg)
		{
			case WM_COMMAND:
				// Don't let Control handle the WM_COMMAND if it's a default or cancel button;
				// otherwise, the events will be fired twice.
				switch(LOWORD(msg.wParam))
				{
					case IDOK:
						if(acceptBtn)
						{
							acceptBtn.performClick();
							//return;
						}
						//break;
						return;
					
					case IDCANCEL:
						if(cancelBtn)
						{
							cancelBtn.performClick();
							//return;
						}
						//break;
						return;
					
					default: ;
				}
				break;
			
			case WM_CREATE:
				// Make sure Windows doesn't magically change the styles.
				SetWindowLongA(hwnd, GWL_EXSTYLE, wexstyle);
				SetWindowLongA(hwnd, GWL_STYLE, wstyle & ~WS_VISIBLE);
				
				SetWindowPos(hwnd, null, 0, 0, 0, 0, SWP_FRAMECHANGED | SWP_NOMOVE
					| SWP_NOSIZE | SWP_NOZORDER | SWP_NOACTIVATE); // Recalculate the frame.
				break;
			
			case WM_CLOSE:
				// Check for this first because defWndProc() will destroy the window.
				fresult = DialogResult.CANCEL;
				if(wmodal)
				{
					doHide();
				}
				else
				{
					close();
				}
				return;
			
			default: ;
		}
		
		super.wndProc(msg);
		
		switch(msg.msg)
		{
			case WM_NCHITTEST:
				//if(msg.result == HTCLIENT || msg.result == HTBORDER)
				if(msg.result != HTNOWHERE && msg.result != HTERROR)
				{
					if(_isPaintingSizeGrip)
					{
						RECT rect;
						_getSizeGripArea(&rect);
						
						Point pt;
						pt.x = LOWORD(msg.lParam);
						pt.y = HIWORD(msg.lParam);
						pt = pointToClient(pt);
						
						if(pt.x >= rect.left && pt.y >= rect.top)
							msg.result = HTBOTTOMRIGHT;
					}
				}
				break;
			
			case WM_ACTIVATE:
				switch(LOWORD(msg.wParam))
				{
					case WA_ACTIVE:
					case WA_CLICKACTIVE:
						onActivated(EventArgs.empty);
						break;
					
					case WA_INACTIVE:
						onDeactivate(EventArgs.empty);
						break;
					
					default: ;
				}
				break;
			
			case WM_WINDOWPOSCHANGING:
				{
					WINDOWPOS* wp = cast(WINDOWPOS*)msg.lParam;
					
					if(!(wp.flags & SWP_NOSIZE))
					{
						if(_isPaintingSizeGrip)
						{
							// This comparison is needed to prevent some painting glitches
							// when moving the window...
							if(width != wp.cx || height != wp.cy)
							{
								RECT rect;
								_getSizeGripArea(&rect);
								InvalidateRect(hwnd, &rect, TRUE);
							}
						}
					}
					
					if(wp.flags & SWP_HIDEWINDOW)
					{
						if(sownerEnabled)
						{
							EnableWindow(sowner, true);
							SetActiveWindow(sowner);
							SetFocus(sowner);
						}
						
						wmodal = false;
					}
				}
				break;
			
			/+
			case WM_CREATE:
				/+
				// mustWndProc() does this now.
				if(startpos != FormStartPosition.MANUAL)
				{
					wrect = _fetchBounds();
					wclientsz = _fetchClientSize();
					
					if(Control._allowLayout)
						onLayout(new LayoutEventArgs(null));
				}
				+/
				
				// Called in createHandle() now.
				//onLoad(EventArgs.empty);
				break;
			+/
			
			case WM_DESTROY:
				onClosed(EventArgs.empty);
				break;
			
			default: ;
		}
	}
	
	
	/+ package +/ void _destroying() // package
	{
		_removeFromOldOwner();
		wowner = null;
		wmdiparent = null;
		
		Application.removeMessageFilter(mfilter);
		mfilter = null;
		
		if(wmenu)
			wmenu._setHwnd(HWND.init);
		
		super._destroying();
	}
	
	
	/+ package +/ protected int _rtype() // package
	{
		return isMdiChild ? 2 : 0;
	}
	
	
	protected override void defWndProc(inout Message msg)
	{
		switch(msg.msg)
		{
			/+
			// Not handled by defWndProc() anymore..
			
			case WM_PAINT:
			case WM_PRINT:
			case WM_PRINTCLIENT:
			case WM_ERASEBKGND:
				// DefDlgProc() doesn't let you use a custom background
				// color, so call the default window proc instead.
				super.defWndProc(msg);
				break;
			+/
			
			default:
				version(NO_MDI)
				{
					msg.result = DefDlgProcA(msg.hWnd, msg.msg, msg.wParam, msg.lParam);
				}
				else
				{
					if(mdiClient && mdiClient.created() && msg.msg != WM_SIZE)
						msg.result = DefFrameProcA(msg.hWnd, mdiClient.handle, msg.msg, msg.wParam, msg.lParam);
					else if(isMdiChild)
						msg.result = DefMDIChildProcA(msg.hWnd, msg.msg, msg.wParam, msg.lParam);
					else
						msg.result = DefDlgProcA(msg.hWnd, msg.msg, msg.wParam, msg.lParam);
				}
		}
	}
	
	
	protected:
	void onClosing(CancelEventArgs cea)
	{
		closing(this, cea);
	}
	
	
	void onClosed(EventArgs ea)
	{
		closed(this, ea);
	}
	
	
	override void setClientSizeCore(int width, int height)
	{
		RECT r;
		
		r.left = 0;
		r.top = 0;
		r.right = width;
		r.bottom = height;
		
		LONG wl = _style();
		AdjustWindowRectEx(&r, wl, !(wl & WS_CHILD) && wmenu, _exStyle());
		
		setBoundsCore(0, 0, r.right - r.left, r.bottom - r.top, BoundsSpecified.SIZE);
	}
	
	
	protected override void setBoundsCore(int x, int y, int width, int height, BoundsSpecified specified)
	{
		if(created)
		{
			super.setBoundsCore(x, y, width, height, specified);
		}
		else
		{
			if(specified & BoundsSpecified.X)
				wrect.x = x;
			if(specified & BoundsSpecified.Y)
				wrect.y = y;
			if(specified & BoundsSpecified.WIDTH)
			{
				if(width < 0)
					width = 0;
				
				wrect.width = width;
			}
			if(specified & BoundsSpecified.HEIGHT)
			{
				if(height < 0)
					height = 0;
				
				wrect.height = height;
			}
			
			_recalcClientSize();
		}
	}
	
	
	private:
	final:
	IButtonControl acceptBtn, cancelBtn;
	bool autoscale = true;
	Size autoscaleBase;
	DialogResult fresult = DialogResult.NONE;
	Icon wicon, wiconSm;
	MainMenu wmenu;
	Size minsz, maxsz; // {0, 0} means none.
	bit wmodal = false;
	bit sownerEnabled;
	HWND sowner = null;
	double opa = 1.0; // Opacity.
	Color transKey;
	bool grip = false;
	FormStartPosition startpos = FormStartPosition.WINDOWS_DEFAULT_LOCATION;
	FormMessageFilter mfilter;
	bit _loaded = false;
	void delegate(Object sender, FormShortcutEventArgs ea)[Keys] _shortcuts;
	Form[] _owned, _mdiChildren; // Always set because they can be created and destroyed at any time.
	Form wowner = null, wmdiparent = null;
	
	version(NO_MDI) {} else
	{
		MdiClient mdiClient = null; // null == not MDI container.
	}
	
	
	package static bool wantsAllKeys(HWND hwnd)
	{
		return (SendMessageA(hwnd, WM_GETDLGCODE, 0, 0) &
			DLGC_WANTALLKEYS) != 0;
	}
	
	
	private class FormMessageFilter: IMessageFilter
	{
		protected bool preFilterMessage(inout Message m)
		{
			switch(m.msg)
			{
				// Process shortcut keys.
				// This should be better than TranslateAccelerator().
				case WM_SYSKEYDOWN:
				case WM_KEYDOWN:
					if(m.hWnd == form.handle || IsChild(form.handle, m.hWnd))
					{
						void delegate(Object sender, FormShortcutEventArgs ea)* ppressed;
						Keys k;
						
						k = cast(Keys)m.wParam | Control.modifierKeys;
						ppressed = k in form._shortcuts;
						
						if(ppressed)
						{
							auto FormShortcutEventArgs ea = new FormShortcutEventArgs(k);
							(*ppressed)(form, ea);
							return true; // Prevent.
						}
					}
					break;
				
				default: ;
			}
			
			//if(!form.isMdiChild && !form.isMdiContainer)
			{
				if(IsDialogMessageA(form.handle, &m._winMsg))
					return true; // Prevent.
			}
			
			return false; // Continue.
		}
		
		
		this(Form form)
		{
			this.form = form;
		}
		
		
		private:
		Form form;
	}
	
	
	void _recalcClientSize()
	{
		RECT r;
		r.left = 0;
		r.right = wrect.width;
		r.top = 0;
		r.bottom = wrect.height;
		
		LONG wl = _style();
		AdjustWindowRectEx(&r, wl, wmenu && !(wl & WS_CHILD), _exStyle());
		
		// Subtract the difference.
		wclientsz = Size(wrect.width - ((r.right - r.left) - wrect.width), wrect.height - ((r.bottom - r.top) - wrect.height));
	}
}


version(NO_MDI) {} else
{
	class MdiClient: ControlSuperClass
	{
		private this()
		{
			_initMdiclient();
			
			wclassStyle = mdiclientClassStyle;
			wexstyle |= WS_EX_CLIENTEDGE;
			
			backc = defaultBackColor;
			forec = defaultForeColor;
			
			dock = DockStyle.FILL;
		}
		
		
		/+
		override void createHandle()
		{
			//if(created)
			if(isHandleCreated)
				return;
			
			if(!wowner || killing)
			{
				create_err:
				throw new DflException("MDI client creation failure.");
			}
			
			CLIENTCREATESTRUCT ccs;
			ccs.hWindowMenu = null; //wowner.menu ? wowner.menu.handle : null;
			ccs.idFirstChild = 10000;
			
			Application.creatingControl(this);
			hwnd = dfl.utf.createWindowEx(wexstyle, MDICLIENT_CLASSNAME, wtext, wstyle, wrect.x, wrect.y,
				wrect.width, wrect.height, wparent.handle, null, GetModuleHandleA(null), &ccs);
			if(!hwnd)
				goto create_err;
			
			onHandleCreated(EventArgs.empty);
		}
		+/
		
		
		protected override void createParams(inout CreateParams cp)
		{
			if(!wparent)
				throw new DflException("Invalid MDI child parent.");
			
			super.createParams(cp);
			
			cp.className = MDICLIENT_CLASSNAME;
			
			ccs.hWindowMenu = null; //wowner.menu ? wowner.menu.handle : null;
			ccs.idFirstChild = 10000;
			cp.param = &ccs;
		}
		
		
		static Color defaultBackColor() // getter
		{
			return Color.systemColor(COLOR_APPWORKSPACE);
		}
		
		
		static Color defaultForeColor() //getter
		{
			return Color.systemColor(COLOR_WINDOWTEXT);
		}
		
		
		protected override void prevWndProc(inout Message msg)
		{
			msg.result = CallWindowProcA(mdiclientPrevWndProc, msg.hWnd, msg.msg, msg.wParam, msg.lParam);
		}
		
		
		private:
		CLIENTCREATESTRUCT ccs;
	}
}


private:

version(DFL_PARK_WINDOW)
{
	static assert(0);
	
	
	HWND getParkHwnd()
	{
		if(!_hwPark)
			_makePark();
		return _hwPark;
	}
	
	
	void _makePark()
	{
		WNDCLASSEXA wce;
		(cast(ubyte*)&wce)[0 .. wce.sizeof] = 0;
		wce.cbSize = wce.sizeof;
		wce.style = CS_DBLCLKS;
		wce.lpszClassName = PARK_CLASSNAME;
		wce.lpfnWndProc = &DefWindowProcA;
		wce.hInstance = GetModuleHandleA(null);
		
		if(!RegisterClassExA(&wce))
		{
			debug(APP_PRINT)
				printf("RegisterClassEx() failed for park class.\n");
			
			init_err:
			//throw new DflException("Unable to initialize forms library.");
			throw new DflException("Unable to create park window.");
		}
		
		_hwPark = CreateWindowExA(WS_EX_WINDOWEDGE | WS_EX_CONTROLPARENT, PARK_CLASSNAME, "",
			WS_OVERLAPPEDWINDOW | WS_CLIPSIBLINGS | WS_CLIPCHILDREN, 0, 0, 0, 0,
			null, null, wce.hInstance, null);
		if(!_hwPark)
		{
			debug(APP_PRINT)
				printf("CreateWindowEx() failed for park window.\n");
			
			goto init_err;
		}
	}
	
	
	const char[] PARK_CLASSNAME = "DFL_Parking";
	
	HWND _hwPark; // Don't use directly; use getParkHwnd().
}

