// Written by Christopher E. Miller
// See the included license.txt for copyright and license details.


module dfl.listbox;

private import std.string;

private import dfl.winapi, dfl.control, dfl.base, dfl.application;
private import dfl.drawing, dfl.event, dfl.collections;


private extern(C) void* memmove(void*, void*, size_t len);

private extern(Windows) void _initListbox();


alias StringObject ListString;


abstract class ListControl: ControlSuperClass // docmain
{
	final char[] getItemText(Object item)
	{
		return item.toString();
	}
	
	
	EventHandler selectedValueChanged;
	
	
	abstract void selectedIndex(int idx); // setter
	abstract int selectedIndex(); // getter
	abstract void selectedValue(Object val); // setter
	abstract void selectedValue(char[] str); // setter
	abstract Object selectedValue(); // getter
	
	
	static Color defaultBackColor() // getter
	{
		return SystemColors.window;
	}
	
	
	static Color defaultForeColor() //getter
	{
		return SystemColors.windowText;
	}
	
	
	this()
	{
		backc = defaultBackColor;
		forec = defaultForeColor;
	}
	
	
	protected:
	void onSelectedValueChanged(EventArgs ea)
	{
		selectedValueChanged(this, ea);
	}
	
	
	// Index change causes the value to be changed.
	void onSelectedIndexChanged(EventArgs ea)
	{
		onSelectedValueChanged(ea); // This appears to be correct.
	}
}


enum SelectionMode: ubyte
{
	ONE,
	NONE,
	MULTI_SIMPLE,
	MULTI_EXTENDED,
}


class ListBox: ListControl // docmain
{
	static class SelectedIndexCollection
	{
		int count() // getter
		{
			return SendMessageA(lbox.handle, LB_GETSELCOUNT, 0, 0);
		}
		
		
		bool isReadOnly() // getter
		{
			return true;
		}
		
		
		int opIndex(int idx)
		{
			foreach(int onidx; this)
			{
				if(!idx)
					return onidx;
				idx--;
			}
			
			// If it's not found it's out of bounds and bad things happen.
			assert(0);
			return -1;
		}
		
		
		bool contains(int idx)
		{
			return indexOf(idx) != -1;
		}
		
		
		int indexOf(int idx)
		{
			int i = 0;
			foreach(int onidx; this)
			{
				if(onidx == idx)
					return i;
				i++;
			}
			return -1;
		}
		
		
		int opApply(int delegate(inout int) dg)
		{
			int result = 0;
			int[] items;
			
			items = new int[count];
			if(items.length != SendMessageA(lbox.handle, LB_GETSELITEMS, items.length, cast(LPARAM)cast(int*)items))
				throw new DflException("Unable to enumerate selected list items.");
			foreach(int _idx; items)
			{
				int idx = _idx; // Prevent inout.
				result = dg(idx);
				if(result)
					break;
			}
			return result;
		}
		
		
		this(ListBox lb)
		{
			lbox = lb;
		}
		
		
		package:
		ListBox lbox;
	}
	
	
	static class SelectedObjectCollection
	{
		int count() // getter
		{
			return SendMessageA(lbox.handle, LB_GETSELCOUNT, 0, 0);
		}
		
		
		bool isReadOnly() // getter
		{
			return true;
		}
		
		
		Object opIndex(int idx)
		{
			foreach(Object obj; this)
			{
				if(!idx)
					return obj;
				idx--;
			}
			
			// If it's not found it's out of bounds and bad things happen.
			assert(0);
			return null;
		}
		
		
		bool contains(Object obj)
		{
			return indexOf(obj) != -1;
		}
		
		
		bool contains(char[] str)
		{
			return indexOf(str) != -1;
		}
		
		
		int indexOf(Object obj)
		{
			int idx = 0;
			foreach(Object onobj; this)
			{
				if(onobj == obj) // Not using is.
					return idx;
				idx++;
			}
			return -1;
		}
		
		
		int indexOf(char[] str)
		{
			int idx = 0;
			foreach(Object onobj; this)
			{
				//if(onobj.toString() is str && onobj.toString().length == str.length)
				if(onobj.toString() == str)
					return idx;
				idx++;
			}
			return -1;
		}
		
		
		int opApply(int delegate(inout Object) dg)
		{
			int result = 0;
			int[] items;
			
			items = new int[count];
			if(items.length != SendMessageA(lbox.handle, LB_GETSELITEMS, items.length, cast(LPARAM)cast(int*)items))
				throw new DflException("Unable to enumerate selected list items.");
			foreach(int idx; items)
			{
				Object obj;
				obj = lbox.items[idx];
				result = dg(obj);
				if(result)
					break;
			}
			return result;
		}
		
		
		int opApply(int delegate(inout char[]) dg)
		{
			int result = 0;
			int[] items;
			
			items = new int[count];
			if(items.length != SendMessageA(lbox.handle, LB_GETSELITEMS, items.length, cast(LPARAM)cast(int*)items))
				throw new DflException("Unable to enumerate selected list items.");
			foreach(int idx; items)
			{
				char[] str;
				str = lbox.items[idx].toString();
				result = dg(str);
				if(result)
					break;
			}
			return result;
		}
		
		
		this(ListBox lb)
		{
			lbox = lb;
		}
		
		
		package:
		ListBox lbox;
	}
	
	
	const int DEFAULT_ITEM_HEIGHT = 13;
	const int NO_MATCHES = LB_ERR;
	
	
	protected override Size defaultSize() // getter
	{
		return Size(120, 95);
	}
	
	
	void borderStyle(BorderStyle bs) // setter
	{
		switch(bs)
		{
			case BorderStyle.FIXED_3D:
				_style(_style() & ~WS_BORDER);
				_exStyle(_exStyle() | WS_EX_CLIENTEDGE);
				break;
				
			case BorderStyle.FIXED_SINGLE:
				_exStyle(_exStyle() & ~WS_EX_CLIENTEDGE);
				_style(_style() | WS_BORDER);
				break;
				
			case BorderStyle.NONE:
				_style(_style() & ~WS_BORDER);
				_exStyle(_exStyle() & ~WS_EX_CLIENTEDGE);
				break;
		}
		
		if(created)
		{
			redrawEntire();
		}
	}
	
	
	BorderStyle borderStyle() // getter
	{
		if(_exStyle() & WS_EX_CLIENTEDGE)
			return BorderStyle.FIXED_3D;
		else if(_style() & WS_BORDER)
			return BorderStyle.FIXED_SINGLE;
		return BorderStyle.NONE;
	}
	
	
	void drawMode(DrawMode dm) // setter
	{
		LONG wl = _style() & ~(LBS_OWNERDRAWVARIABLE | LBS_OWNERDRAWFIXED);
		
		switch(dm)
		{
			case DrawMode.OWNER_DRAW_VARIABLE:
				wl |= LBS_OWNERDRAWVARIABLE;
				break;
			
			case DrawMode.OWNER_DRAW_FIXED:
				wl |= LBS_OWNERDRAWFIXED;
				break;
			
			case DrawMode.NORMAL:
				break;
		}
		
		_style(wl);
	}
	
	
	DrawMode drawMode() // getter
	{
		LONG wl = _style();
		
		if(wl & LBS_OWNERDRAWVARIABLE)
			return DrawMode.OWNER_DRAW_VARIABLE;
		if(wl & LBS_OWNERDRAWFIXED)
			return DrawMode.OWNER_DRAW_FIXED;
		return DrawMode.NORMAL;
	}
	
	
	final void horizontalExtent(int he) // setter
	{
		if(created)
			SendMessageA(handle, LB_SETHORIZONTALEXTENT, he, 0);
		
		hextent = he;
	}
	
	
	final int horizontalExtent() // getter
	{
		if(created)
			hextent = cast(int)SendMessageA(handle, LB_GETHORIZONTALEXTENT, 0, 0);
		return hextent;
	}
	
	
	final void horizontalScrollbar(bool byes) // setter
	{
		if(byes)
			_style(_style() | WS_HSCROLL);
		else
			_style(_style() & ~WS_HSCROLL);
	}
	
	
	final bool horizontalScrollbar() // getter
	{
		return (_style() & WS_HSCROLL) != 0;
	}
	
	
	final void integralHeight(bool byes) //setter
	{
		if(byes)
			_style(_style() & ~LBS_NOINTEGRALHEIGHT);
		else
			_style(_style() | LBS_NOINTEGRALHEIGHT);
	}
	
	
	final bool integralHeight() // getter
	{
		return (_style() & LBS_NOINTEGRALHEIGHT) == 0;
	}
	
	
	// This function has no effect if the drawMode is OWNER_DRAW_VARIABLE.
	final void itemHeight(int h) // setter
	{
		if(drawMode == DrawMode.OWNER_DRAW_VARIABLE)
			return;
		
		iheight = h;
		
		if(created)
			SendMessageA(hwnd, LB_SETITEMHEIGHT, 0, MAKELPARAM(h, 0));
	}
	
	
	// Return value is meaningless when drawMode is OWNER_DRAW_VARIABLE.
	final int itemHeight() // getter
	{
		// Requesting it like this when owner draw variable doesn't work.
		/+
		if(!created)
			return iheight;
		
		int result = SendMessageA(handle, LB_GETITEMHEIGHT, 0, 0);
		if(result == LB_ERR)
			result = iheight; // ?
		else
			iheight = result;
		
		return result;
		+/
		
		return iheight;
	}
	
	
	final ObjectCollection items() // getter
	{
		return icollection;
	}
	
	
	final void multiColumn(bool byes) // setter
	{
		// TODO: is this the correct implementation?
		
		if(byes)
			_style(_style() | LBS_MULTICOLUMN | WS_HSCROLL);
		else
			_style(_style() & ~(LBS_MULTICOLUMN | WS_HSCROLL));
	}
	
	
	final bool multiColumn() // getter
	{
		return (_style() & LBS_MULTICOLUMN) != 0;
	}
	
	
	final void scrollAlwaysVisible(bool byes) // setter
	{
		if(byes)
			_style(_style() | LBS_DISABLENOSCROLL);
		else
			_style(_style() & ~LBS_DISABLENOSCROLL);
	}
	
	
	final bool scrollAlwaysVisible() // getter
	{
		return (_style() & LBS_DISABLENOSCROLL) != 0;
	}
	
	
	override void selectedIndex(int idx) // setter
	{
		if(created)
		{
			if(isMultSel())
			{
				if(idx == -1)
				{
					// Remove all selection.
					
					// Not working right.
					//SendMessageA(handle, LB_SELITEMRANGE, false, MAKELPARAM(0, ushort.max));
					
					// Get the indices directly because deselecting them during
					// selidxcollection.foreach could screw it up.
					
					int[] items;
					
					items = new int[selidxcollection.count];
					if(items.length != SendMessageA(handle, LB_GETSELITEMS, items.length, cast(LPARAM)cast(int*)items))
						throw new DflException("Unable to clear selected list items.");
					
					foreach(int idx; items)
					{
						SendMessageA(handle, LB_SETSEL, false, idx);
					}
				}
				else
				{
					// ?
					SendMessageA(handle, LB_SETSEL, true, idx);
				}
			}
			else
			{
				SendMessageA(handle, LB_SETCURSEL, idx, 0);
			}
		}
	}
	
	
	override int selectedIndex() // getter
	{
		if(created)
		{
			if(isMultSel())
			{
				if(selidxcollection.count)
					return selidxcollection[0];
			}
			else
			{
				LRESULT result;
				result = SendMessageA(handle, LB_GETCURSEL, 0, 0);
				if(LB_ERR != result) // Redundant.
					return cast(int)result;
			}
		}
		return -1;
	}
	
	
	final void selectedItem(Object o) // setter
	{
		int i;
		i = items.indexOf(o);
		if(i != -1)
			selectedIndex = i;
	}
	
	
	final void selectedItem(char[] str) // setter
	{
		int i;
		i = items.indexOf(str);
		if(i != -1)
			selectedIndex = i;
	}
	
	
	final Object selectedItem() // getter
	{
		int idx;
		idx = selectedIndex;
		if(idx == -1)
			return null;
		return items[idx];
	}
	
	
	override void selectedValue(Object val) // setter
	{
		selectedItem = val;
	}
	
	
	override void selectedValue(char[] str) // setter
	{
		selectedItem = str;
	}
	
	
	override Object selectedValue() // getter
	{
		return selectedItem;
	}
	
	
	final SelectedIndexCollection selectedIndices() // getter
	{
		return selidxcollection;
	}
	
	
	final SelectedObjectCollection selectedItems() // getter
	{
		return selobjcollection;
	}
	
	
	void selectionMode(SelectionMode selmode) // setter
	{
		LONG wl = _style() & ~(LBS_NOSEL | LBS_EXTENDEDSEL | LBS_MULTIPLESEL);
		
		switch(selmode)
		{
			case SelectionMode.ONE:
				break;
			
			case SelectionMode.MULTI_SIMPLE:
				wl |= LBS_MULTIPLESEL;
				break;
			
			case SelectionMode.MULTI_EXTENDED:
				wl |= LBS_EXTENDEDSEL;
				break;
			
			case SelectionMode.NONE:
				wl |= LBS_NOSEL;
				break;
		}
		
		_style(wl);
	}
	
	
	SelectionMode selectionMode() // getter
	{
		LONG wl = _style();
		
		if(wl & LBS_NOSEL)
			return SelectionMode.NONE;
		if(wl & LBS_EXTENDEDSEL)
			return SelectionMode.MULTI_EXTENDED;
		if(wl & LBS_MULTIPLESEL)
			return SelectionMode.MULTI_SIMPLE;
		return SelectionMode.ONE;
	}
	
	
	final void sorted(bool byes) // setter
	{
		if(byes)
			_style(_style() | LBS_SORT);
		else
			_style(_style() & ~LBS_SORT);
	}
	
	
	final bool sorted() // getter
	{
		return (_style() & LBS_SORT) != 0;
	}
	
	
	final void topIndex(int idx) // setter
	{
		if(created)
			SendMessageA(handle, LB_SETTOPINDEX, idx, 0);
	}
	
	
	final int topIndex() // getter
	{
		if(created)
			return SendMessageA(handle, LB_GETTOPINDEX, 0, 0);
		return 0;
	}
	
	
	final void useTabStops(bool byes) // setter
	{
		if(byes)
			_style(_style() | LBS_USETABSTOPS);
		else
			_style(_style() & ~LBS_USETABSTOPS);
	}
	
	
	final bool useTabStops() // getter
	{
		return (_style() & LBS_USETABSTOPS) != 0;
	}
	
	
	final void beginUpdate()
	{
		SendMessageA(handle, WM_SETREDRAW, false, 0);
	}
	
	
	final void endUpdate()
	{
		SendMessageA(handle, WM_SETREDRAW, true, 0);
		invalidate(true); // Show updates.
	}
	
	
	package final bool isMultSel()
	{
		return (_style() & (LBS_EXTENDEDSEL | LBS_MULTIPLESEL)) != 0;
	}
	
	
	final void clearSelected()
	{
		if(created)
			selectedIndex = -1;
	}
	
	
	final int findString(char[] str, int startIndex)
	{
		// TODO: find string if control not created ?
		
		int result = NO_MATCHES;
		
		if(created)
		{
			result = SendMessageA(handle, LB_FINDSTRING, startIndex, cast(LPARAM)toStringz(str));
			if(result == LB_ERR) // Redundant.
				result = NO_MATCHES;
		}
		
		return result;
	}
	
	
	final int findString(char[] str)
	{
		return findString(str, -1); // Start at beginning.
	}
	
	
	final int findStringExact(char[] str, int startIndex)
	{
		// TODO: find string if control not created ?
		
		int result = NO_MATCHES;
		
		if(created)
		{
			result = SendMessageA(handle, LB_FINDSTRINGEXACT, startIndex, cast(LPARAM)toStringz(str));
			if(result == LB_ERR) // Redundant.
				result = NO_MATCHES;
		}
		
		return result;
	}
	
	
	final int findStringExact(char[] str)
	{
		return findStringExact(str, -1); // Start at beginning.
	}
	
	
	final int getItemHeight(int idx)
	{
		int result = SendMessageA(handle, LB_GETITEMHEIGHT, idx, 0);
		if(LB_ERR == result)
			throw new DflException("Unable to obtain item height.");
		return result;
	}
	
	
	final Rect getItemRectangle(int idx)
	{
		RECT rect;
		if(LB_ERR == SendMessageA(handle, LB_GETITEMRECT, idx, cast(LPARAM)&rect))
			throw new DflException("Unable to obtain item rectangle.");
		return Rect(&rect);
	}
	
	
	final bool getSelected(int idx)
	{
		return SendMessageA(handle, LB_GETSEL, idx, 0) > 0;
	}
	
	
	final int indexFromPoint(int x, int y)
	{
		// LB_ITEMFROMPOINT is "nearest", so also check with the item rectangle to
		// see if the point is directly in the item.
		
		// Maybe use LBItemFromPt() from common controls.
		
		int result = NO_MATCHES;
		
		if(created)
		{
			result = SendMessageA(handle, LB_ITEMFROMPOINT, 0, MAKELPARAM(x, y));
			if(!HIWORD(result)) // In client area
			{
				//result = LOWORD(result); // High word already 0.
				if(result < 0 || !getItemRectangle(result).contains(x, y))
					result = NO_MATCHES;
			}
			else // Outside client area.
			{
				result = NO_MATCHES;
			}
		}
		
		return result;
	}
	
	
	final int indexFromPoint(Point pt)
	{
		return indexFromPoint(pt.x, pt.y);
	}
	
	
	final void setSelected(int idx, bool byes)
	{
		if(created)
			SendMessageA(handle, LB_SETSEL, byes, idx);
	}
	
	
	protected ObjectCollection createItemCollection()
	{
		return new ObjectCollection(this);
	}
	
	
	/+
	protected void sort()
	{
		
	}
	+/
	
	
	static class ObjectCollection: IList
	{
		this(ListBox lbox)
		{
			this.lbox = lbox;
		}
		
		
		this(ListBox lbox, Object[] range)
		{
			this.lbox = lbox;
			addRange(range);
		}
		
		
		this(ListBox lbox, char[][] range)
		{
			this.lbox = lbox;
			addRange(range);
		}
		
		
		this(ListBox lbox, ObjectCollection range)
		{
			this.lbox = lbox;
			addRange(range);
		}
		
		
		bool isFixedSize() // getter
		{
			return false;
		}
		
		
		bool isReadOnly() // getter
		{
			return false;
		}
		
		
		bool isSynchronized() // getter
		{
			return false;
		}
		
		
		Object syncRoot() // getter
		{
			return this;
		}
		
		
		// Shallow copy.
		Array clone()
		{
			Array result;
			result = new Array(array.length);
			Array.copy(array, result, array.length);
			return result;
		}
		
		
		// D-style.
		final Array dup() // getter
		{
			return clone();
		}
		
		
		void copyTo(Array dest, int destIndex)
		{
			Array.copy(array, 0, dest, destIndex, array.length);
		}
		
		
		package final int add2(Object value)
		{
			int i;
			
			if(lbox.sorted)
			{
				// Insertion sort.
				
				for(i = 0; i != array.length; i++)
				{
					if(value < array[i])
						break;
				}
				
				if(i == array.length)
					//array ~= value;
					lbox._items ~= value;
				else
					array = array[0 .. i] ~ (&value)[0 .. 1] ~ array[i .. array.length];
			}
			else
			{
				i = array.length;
				//array ~= value;
				lbox._items ~= value;
			}
			
			if(lbox.created)
			{
				Message m;
				m = Message(lbox.handle, LB_INSERTSTRING, i, cast(LPARAM)toStringz(value.toString()));
				lbox.prevWndProc(m);
				
				i = m.result;
				//SendMessageA(lbox.handle, LB_SETITEMDATA, i, cast(LPARAM)cast(void*)value);
			}
			
			return i;
		}
		
		
		package final int add2(char[] value)
		{
			return add2(new ListString(value));
		}
		
		
		void add(Object value)
		{
			add2(value);
		}
		
		
		void add(char[] value)
		{
			add2(value);
		}
		
		
		void addRange(Object[] range)
		{
			if(lbox.sorted)
			{
				foreach(Object value; range)
				{
					add(value);
				}
			}
			else
			{
				//array ~= range;
				lbox._items ~= range;
				
				if(lbox.created)
				{
					Message m;
					m = Message(lbox.handle, LB_ADDSTRING, 0, 0);
					foreach(Object value; range)
					{
						m.lParam = cast(LPARAM)toStringz(value.toString());
						lbox.prevWndProc(m);
						
						//SendMessageA(lbox.handle, LB_SETITEMDATA, m.result, cast(LPARAM)cast(void*)value);
					}
				}
			}
		}
		
		
		void addRange(char[][] range)
		{
			foreach(char[] value; range)
			{
				add(value);
			}
		}
		
		
		void addRange(ObjectCollection range)
		{
			addRange(range.array);
		}
		
		
		void clear()
		{
			array = null; // ?
			
			if(lbox.created)
			{
				Message m;
				m = Message(lbox.handle, LB_RESETCONTENT, 0, 0);
				lbox.prevWndProc(m);
			}
		}
		
		
		bool contains(Object value)
		{
			return indexOf(value) != -1;
		}
		
		
		bool contains(char[] value)
		{
			return indexOf(value) != -1;
		}
		
		
		int indexOf(Object value)
		{
			size_t i;
			for(i = 0; i != array.length; i++)
			{
				if(value == array[i])
					return i;
			}
			return -1;
		}
		
		
		int indexOf(char[] value)
		{
			size_t i;
			ListString ls;
			for(i = 0; i != array.length; i++)
			{
				ls = cast(ListString)ls;
				if(ls)
				{
					if(value == ls.value)
						return i;
				}
			}
			return -1;
		}
		
		
		package final int insert2(int index, Object value)
		{
			if(index < 0)
				index = 0;
			if(index >= cast(int)array.length)
				//array ~= value;
				lbox._items ~= value;
			else
				array = array[0 .. index] ~ (&value)[0 .. 1] ~ array[index .. array.length];
			
			if(lbox.created)
			{
				Message m;
				m = Message(lbox.handle, LB_INSERTSTRING, index, cast(LPARAM)toStringz(value.toString()));
				lbox.prevWndProc(m);
				
				index = m.result;
				//SendMessageA(lbox.handle, LB_SETITEMDATA, index, cast(LPARAM)cast(void*)value);
			}
			
			return index;
		}
		
		
		package final int insert2(int index, char[] value)
		{
			return insert2(index, new ListString(value));
		}
		
		
		void insert(int index, Object value)
		{
			insert2(index, value);
		}
		
		
		void insert(int index, char[] value)
		{
			insert2(index, value);
		}
		
		
		void remove(Object value)
		{
			int i;
			i = indexOf(value);
			if(i != -1)
				removeAt(i);
		}
		
		
		void remove(char[] value)
		{
			int i;
			i = indexOf(value);
			if(i != -1)
				removeAt(i);
		}
		
		
		void removeAt(int index)
		{
			if(!index)
				array = array[1 .. array.length];
			else if(index == array.length - 1)
				array = array[0 .. index];
			else
				array = array[0 .. index] ~ array[index + 1 .. array.length];
			
			if(lbox.created)
			{
				Message m;
				m = Message(lbox.handle, LB_DELETESTRING, index, 0);
				lbox.prevWndProc(m);
			}
		}
		
		
		int count() // getter
		{
			return array.length;
		}
		
		
		void opIndexAssign(Object value, int index) // setter
		{
			//array[index] = value;
			
			removeAt(index);
			//add(value);
			insert(index, value);
		}
		
		
		Object opIndex(int index) // getter
		{
			return array[index];
		}
		
		
		IEnumerator getEnumerator()
		{
			return new ItemEnum(array());
		}
		
		
		private:
		ListBox lbox;
		
		
		this()
		{
		}
		
		
		final Object[] array() // getter
		{
			return lbox._items;
		}
		
		
		final void array(Object[] setter) // setter
		{
			lbox._items = setter;
		}
		
		
		class ItemEnum: IEnumerator
		{
			mixin EnumeratorForeach;
			
			
			this(Object[] items)
			{
				this.items = items;
			}
			
			
			Object current() // getter
			{
				return items[idx];
			}
			
			
			bool moveNext()
			{
				idx++;
				if(idx >= items.length) // Note: signed-unsigned comparison.
				{
					idx = items.length; // Prevent overflow.
					return false;
				}
				return true;
			}
			
			
			void reset()
			{
				idx = -1;
			}
			
			
			private:
			Object[] items;
			int idx = -1;
		}
	}
	
	
	this()
	{
		_initListbox();
		
		// Default useTabStops and vertical scrolling.
		wstyle |= WS_TABSTOP | LBS_USETABSTOPS | LBS_HASSTRINGS | WS_VSCROLL | LBS_NOTIFY;
		wexstyle |= WS_EX_CLIENTEDGE;
		ctrlStyle |= ControlStyles.SELECTABLE;
		wclassStyle = listboxClassStyle;
		
		icollection = createItemCollection();
		selidxcollection = new SelectedIndexCollection(this);
		selobjcollection = new SelectedObjectCollection(this);
	}
	
	
	protected override void onHandleCreated(EventArgs ea)
	{
		super.onHandleCreated(ea);
		
		// Set the Ctrl ID to the HWND so that it is unique
		// and WM_MEASUREITEM will work properly.
		SetWindowLongA(hwnd, GWL_ID, cast(LONG)hwnd);
		
		if(hextent != 0)
			SendMessageA(hwnd, LB_SETHORIZONTALEXTENT, hextent, 0);
		
		if(iheight != DEFAULT_ITEM_HEIGHT)
			SendMessageA(hwnd, LB_SETITEMHEIGHT, 0, MAKELPARAM(iheight, 0));
		
		Message m;
		m.hWnd = handle;
		m.msg = LB_INSERTSTRING;
		foreach(int i, Object obj; _items)
		{
			m.wParam = i;
			m.lParam = cast(LPARAM)toStringz(obj.toString());
			
			prevWndProc(m);
			//if(LB_ERR == m.result || LB_ERRSPACE == m.result)
			if(m.result < 0)
				throw new DflException("Unable to add list item.");
			
			//SendMessageA(handle, LB_SETITEMDATA, m.result, cast(LPARAM)cast(void*)obj);
		}
		
		//redrawEntire();
	}
	
	
	/+
	override void createHandle()
	{
		if(isHandleCreated)
			return;
		
		createClassHandle(LISTBOX_CLASSNAME);
		
		onHandleCreated(EventArgs.empty);
	}
	+/
	
	
	protected override void createParams(inout CreateParams cp)
	{
		super.createParams(cp);
		
		cp.className = LISTBOX_CLASSNAME;
	}
	
	
	DrawItemEventHandler drawItem;
	MeasureItemEventHandler measureItem;
	
	
	protected:
	void onDrawItem(DrawItemEventArgs dieh)
	{
		drawItem(this, dieh);
	}
	
	
	void onMeasureItem(MeasureItemEventArgs miea)
	{
		measureItem(this, miea);
	}
	
	
	package final void _WmDrawItem(DRAWITEMSTRUCT* dis)
	in
	{
		assert(dis.hwndItem == handle);
		assert(dis.CtlType == ODT_LISTBOX);
	}
	body
	{
		DrawItemState state;
		state = cast(DrawItemState)dis.itemState;
		
		if(dis.itemID == -1)
		{
			FillRect(dis.hDC, &dis.rcItem, hbrBg);
			if(state & DrawItemState.FOCUS)
				DrawFocusRect(dis.hDC, &dis.rcItem);
		}
		else
		{
			DrawItemEventArgs diea;
			Color bc, fc;
			
			if(state & DrawItemState.SELECTED)
			{
				bc = Color.systemColor(COLOR_HIGHLIGHT);
				fc = Color.systemColor(COLOR_HIGHLIGHTTEXT);
			}
			else
			{
				bc = backColor;
				fc = foreColor;
			}
			
			prepareDc(dis.hDC);
			diea = new DrawItemEventArgs(new Graphics(dis.hDC, false), wfont,
				Rect(&dis.rcItem), dis.itemID, state, fc, bc);
			
			onDrawItem(diea);
		}
	}
	
	
	package final void _WmMeasureItem(MEASUREITEMSTRUCT* mis)
	in
	{
		assert(mis.CtlType == ODT_LISTBOX);
	}
	body
	{
		MeasureItemEventArgs miea;
		auto Graphics gpx = new CommonGraphics(handle(), GetDC(handle));
		miea = new MeasureItemEventArgs(gpx, mis.itemID, /+ mis.itemHeight +/ iheight);
		miea.itemWidth = mis.itemWidth;
		
		onMeasureItem(miea);
		
		mis.itemHeight = miea.itemHeight;
		mis.itemWidth = miea.itemWidth;
	}
	
	
	override void prevWndProc(inout Message msg)
	{
		msg.result = CallWindowProcA(listboxPrevWndProc, msg.hWnd, msg.msg, msg.wParam, msg.lParam);
	}
	
	
	/+package+/ override LRESULT customMsg(inout CustomMsg msg) // package
	{
		LRESULT result;
		result = super.customMsg(msg);
		
		switch(msg.message)
		{
			case WM_DRAWITEM:
				_WmDrawItem(cast(DRAWITEMSTRUCT*)msg.lParam);
				result = 1;
				break;
			
			case WM_MEASUREITEM:
				_WmMeasureItem(cast(MEASUREITEMSTRUCT*)msg.lParam);
				result = 1;
				break;
			
			case WM_COMMAND:
				assert(cast(HWND)msg.lParam == handle);
				switch(HIWORD(msg.wParam))
				{
					case LBN_SELCHANGE:
						onSelectedIndexChanged(EventArgs.empty);
						break;
					
					case LBN_SELCANCEL:
						onSelectedIndexChanged(EventArgs.empty);
						break;
					
					default: ;
				}
				break;
			
			default: ;
		}
		
		return result;
	}
	
	
	override void wndProc(inout Message msg)
	{
		switch(msg.msg)
		{
			case LB_ADDSTRING:
				msg.result = icollection.add2(std.string.toString(cast(char*)msg.lParam).dup);
				return;
			
			case LB_INSERTSTRING:
				msg.result = icollection.insert2(msg.wParam, std.string.toString(cast(char*)msg.lParam).dup);
				return;
			
			case LB_DELETESTRING:
				icollection.removeAt(msg.wParam);
				msg.result = icollection.count;
				return;
			
			case LB_RESETCONTENT:
				icollection.clear();
				return;
			
			case LB_SETITEMDATA:
				// Cannot set item data from outside DFL.
				msg.result = LB_ERR;
				return;
			
			case LB_ADDFILE:
				msg.result = LB_ERR;
				return;
			
			case LB_DIR:
				msg.result = LB_ERR;
				return;
			
			default:
				super.wndProc(msg);
				return;
		}
		prevWndProc(msg);
	}
	
	
	private:
	int hextent = 0;
	int iheight = DEFAULT_ITEM_HEIGHT;
	ObjectCollection icollection;
	SelectedIndexCollection selidxcollection;
	SelectedObjectCollection selobjcollection;
	Object[] _items;
}

