// Written by Christopher E. Miller
// See the included license.txt for copyright and license details.


module dfl.listview;

private import std.string;

private import dfl.base, dfl.control, dfl.winapi, dfl.application;
private import dfl.event, dfl.drawing, dfl.collections, dfl.utf;


private extern(Windows) void _initListview();


enum ListViewAlignment: ubyte
{
	TOP,
	DEFAULT, // Stays where dropped.
	LEFT,
	SNAP_TO_GRID,
}


private union CallText
{
	char* ansi;
	wchar* unicode;
}


private CallText getCallText(char[] text)
{
	CallText result;
	if(text is null)
	{
		if(useUnicode)
			result.unicode = null;
		else
			result.ansi = null;
	}
	else
	{
		if(useUnicode)
			result.unicode = toUnicodez(text);
		else
			result.ansi = toAnsiz(text);
	}
	return result;
}


class ListViewSubItem
{
	this()
	{
		Application.ppin(cast(void*)this);
	}
	
	
	// Extra.
	this(char[] thisSubItemText)
	{
		this();
		
		settextin(thisSubItemText);
	}
	
	
	this(ListViewItem owner, char[] thisSubItemText)
	{
		this();
		
		settextin(thisSubItemText);
		if(owner)
		{
			this._item = owner;
			owner.subItems.add(this);
		}
	}
	
	
	package final void settextin(char[] newText)
	{
		calltxt = getCallText(newText);
		_txt = newText;
	}
	
	
	override char[] toString()
	{
		return text;
	}
	
	
	override int opEquals(Object o)
	{
		return text == o.toString();
	}
	
	
	override int opCmp(Object o)
	{
		return std.string.icmp(text, o.toString());
	}
	
	
	final void text(char[] newText) // setter
	{
		settextin(newText);
		
		if(_item && _item.lview && _item.lview.created)
		{
			int ii, subi;
			ii = _item.lview.items.indexOf(_item);
			assert(-1 != ii);
			subi = _item.subItems.indexOf(this);
			assert(-1 != subi);
			_item.lview.updateItemText(ii, newText, subi + 1); // Sub items really start at 1 in the list view.
		}
	}
	
	
	final char[] text() // getter
	{
		return _txt;
	}
	
	
	private:
	package ListViewItem _item;
	char[] _txt;
	package CallText calltxt;
}


class ListViewItem
{
	static class ListViewSubItemCollection//: IList
	{
		this(ListViewItem owner)
		in
		{
			assert(!owner.isubs);
		}
		body
		{
			_item = owner;
		}
		
		
		bool isFixedSize() // getter
		{
			return false;
		}
		
		
		bool isReadOnly() // getter
		{
			return false;
		}
		
		
		bool isSynchronized() // getter
		{
			return false;
		}
		
		
		Object syncRoot() // getter
		{
			return this;
		}
		
		
		// Shallow copy.
		Array clone()
		{
			Array result;
			result = new Array(_subs.length);
			Array.copy(_subs, result, _subs.length);
			return result;
		}
		
		
		// D-style.
		final Array dup() // getter
		{
			return clone();
		}
		
		
		void copyTo(Array dest, int destIndex)
		{
			Array.copy(_subs, 0, dest, destIndex, _subs.length);
		}
		
		
		void add(ListViewSubItem subItem)
		{
			insert(cast(int)_subs.length, subItem);
		}
		
		
		void add(char[] text)
		{
			return add(new ListViewSubItem(text));
		}
		
		
		void add(Object o)
		{
			ListViewSubItem subItem;
			subItem = cast(ListViewSubItem)o;
			if(subItem)
			{
				return add(subItem);
			}
			else
			{
				return add(o.toString()); // ?
			}
		}
		
		
		void addRange(Object[] range)
		{
			foreach(Object o; range)
			{
				add(o);
			}
		}
		
		
		void addRange(ListViewSubItem[] range)
		{
			foreach(ListViewSubItem subItem; range)
			{
				add(subItem);
			}
		}
		
		
		void addRange(char[][] range)
		{
			foreach(char[] s; range)
			{
				add(s);
			}
		}
		
		
		/+
		void clear()
		{
			// TODO: ?
			assert(0);
		}
		+/
		
		
		bool contains(Object value)
		{
			return indexOf(value) != -1;
		}
		
		
		bool contains(ListViewSubItem value)
		{
			return indexOf(value) != -1;
		}
		
		
		int indexOf(Object value)
		{
			ListViewSubItem subItem;
			subItem = cast(ListViewSubItem)value;
			if(subItem)
			{
				return indexOf(subItem);
			}
			else
			{
				foreach(int i, ListViewSubItem subItem; _subs)
				{
					if(value == subItem) // ?
						return i;
				}
				return -1;
			}
		}
		
		
		int indexOf(ListViewSubItem value)
		{
			int result = -1;
			foreach(int i, ListViewSubItem subItem; _subs)
			{
				//if(value == subItem)
				{
					if(value is subItem)
					{
						result = i;
						break;
					}
					
					//if(result == -1)
					//	result = i;
				}
			}
			return result;
		}
		
		
		void insert(int index, Object value)
		{
			ListViewSubItem subItem;
			subItem = cast(ListViewSubItem)value;
			if(subItem)
			{
				insert(index, subItem);
			}
			else
			{
				insert(index, value.toString()); // ?
			}
		}
		
		
		void insert(int index, char[] value)
		{
			insert(index, new ListViewSubItem(value));
		}
		
		
		private void verifyNoParent(ListViewSubItem subItem)
		{
			if(subItem._item)
				throw new DflException("ListViewSubItem already belongs to a ListViewItem.");
		}
		
		
		void insert(int i, ListViewSubItem subItem)
		{
			verifyNoParent(subItem);
			
			subItem._item = _item;
			
			if(i < 0)
				i = 0;
			/+if(i <= 0)
				_subs = (&_subs)[0 .. 1].dup ~ _subs; // DMD 0.124: .dup needed for some reason.
			else+/ if(i >= cast(int)_subs.length)
				_subs ~= subItem;
			else
				_subs = _subs[0 .. i] ~ (&subItem)[0 .. 1] ~ _subs[i .. _subs.length];
			
			if(_item && _item.lview && _item.lview.created)
			{
				int owneri;
				owneri = _item.index;
				assert(-1 != owneri);
				//i = _item.lview._ins(owneri, subItem, i);
				i = _item.lview._ins(owneri, subItem, i + 1); // Sub items really start at 1 in the list view.
				assert(-1 != i);
			}
		}
		
		
		/+
		void remove(Object value)
		{
			// TODO: ?
			assert(0);
		}
		
		
		void remove(ListViewSubItem subItem)
		{
			// TODO: ?
			assert(0);
			
			// subItem._item = null;
		}
		
		
		void removeAt(int index)
		{
			// TODO: ?
			assert(0);
		}
		+/
		
		
		int count() // getter
		{
			return _subs.length;
		}
		
		
		/+
		void opIndexAssign(Object value, int index) // setter
		{
			// TODO: ?
			assert(0);
		}
		+/
		
		
		// Covariant return with interface IList.
		ListViewSubItem opIndex(int index) // getter
		{
			return _subs[index];
		}
		
		
		IEnumerator getEnumerator()
		{
			return new SubItemEnum(_subs);
		}
		
		
		private:
		ListViewItem _item;
		package ListViewSubItem[] _subs;
		
		
		class SubItemEnum: IEnumerator
		{
			mixin EnumeratorForeach;
			//mixin EnumeratorForeach!(ListViewSubItem);
			
			
			this(ListViewSubItem[] subItems)
			{
				this.subItems = subItems;
			}
			
			
			// Covariant return with IEnumerator.
			//ListViewSubItem current() // getter
			Object current() // getter
			{
				return subItems[idx];
			}
			
			
			bool moveNext()
			{
				idx++;
				if(idx >= subItems.length) // Note: signed-unsigned comparison.
				{
					idx = subItems.length; // Prevent overflow.
					return false;
				}
				return true;
			}
			
			
			void reset()
			{
				idx = -1;
			}
			
			
			private:
			ListViewSubItem[] subItems;
			int idx = -1;
		}
	}
	
	
	this()
	{
		Application.ppin(cast(void*)this);
		
		isubs = new ListViewSubItemCollection(this);
	}
	
	
	this(char[] text)
	{
		this();
		
		settextin(text);
	}
	
	
	package final void settextin(char[] newText)
	{
		calltxt = getCallText(newText);
		_txt = newText;
	}
	
	
	override char[] toString()
	{
		return text;
	}
	
	
	override int opEquals(Object o)
	{
		return text == o.toString();
	}
	
	
	override int opCmp(Object o)
	{
		return std.string.icmp(text, o.toString());
	}
	
	
	final Rect bounds() // getter
	{
		if(lview)
		{
			int i = index;
			assert(-1 != i);
			return lview.getItemRect(i);
		}
		return Rect(0, 0, 0, 0);
	}
	
	
	final int index() // getter
	{
		if(lview)
			return lview.litems.indexOf(this);
		return -1;
	}
	
	
	final void text(char[] newText) // setter
	{
		settextin(newText);
		
		if(lview && lview.created)
			lview.updateItemText(this, newText);
	}
	
	
	final char[] text() // getter
	{
		return _txt;
	}
	
	
	final ListView listView() // getter
	{
		return lview;
	}
	
	
	final void tag(Object obj) // setter
	{
		_tag = obj;
	}
	
	
	final Object tag() // getter
	{
		return _tag;
	}
	
	
	final ListViewSubItemCollection subItems() // getter
	{
		return isubs;
	}
	
	
	private:
	package ListView lview = null;
	char[] _txt;
	package CallText calltxt;
	Object _tag = null;
	package ListViewSubItemCollection isubs = null;
}


class ColumnHeader
{
	/+
	this(char[] text)
	{
		this();
		
		this._txt = text;
	}
	+/
	
	
	this()
	{
		Application.ppin(cast(void*)this);
	}
	
	
	final ListView listView() // getter
	{
		return lview;
	}
	
	
	final void text(char[] newText) // setter
	{
		_txt = newText;
		
		if(lview && lview.created)
		{
			lview.updateColumnText(this, newText);
		}
	}
	
	
	final char[] text() // getter
	{
		return _txt;
	}
	
	
	final int index() // getter
	{
		if(lview)
			lview.cols.indexOf(this);
		return -1;
	}
	
	
	final void textAlign(HorizontalAlignment halign) // setter
	{
		_align = halign;
		
		if(lview && lview.created)
		{
			lview.updateColumnAlign(this, halign);
		}
	}
	
	
	final HorizontalAlignment textAlign() // getter
	{
		return _align;
	}
	
	
	final void width(int w) // setter
	{
		_width = w;
		
		if(lview && lview.created)
		{
			lview.updateColumnWidth(this, w);
		}
	}
	
	
	final int width() // getter
	{
		return _width;
	}
	
	
	private:
	package ListView lview;
	char[] _txt;
	int _width;
	HorizontalAlignment _align;
}


class ListView: ControlSuperClass // docmain
{
	static class ListViewItemCollection//: IList
	{
		this(ListView lv)
		in
		{
			assert(lv.litems is null);
		}
		body
		{
			this.lv = lv;
		}
		
		
		bool isFixedSize() // getter
		{
			return false;
		}
		
		
		bool isReadOnly() // getter
		{
			return false;
		}
		
		
		bool isSynchronized() // getter
		{
			return false;
		}
		
		
		Object syncRoot() // getter
		{
			return this;
		}
		
		
		// Shallow copy.
		Array clone()
		{
			Array result;
			result = new Array(_items.length);
			Array.copy(_items, result, _items.length);
			return result;
		}
		
		
		// D-style.
		final Array dup() // getter
		{
			return clone();
		}
		
		
		void copyTo(Array dest, int destIndex)
		{
			Array.copy(_items, 0, dest, destIndex, _items.length);
		}
		
		
		void add(Object o)
		{
			ListViewItem item;
			item = cast(ListViewItem)o;
			if(item)
			{
				return add(item);
			}
			else
			{
				return add(o.toString()); // ?
			}
		}
		
		
		package final bool created() // getter
		{
			return lv && lv.created();
		}
		
		
		void add(ListViewItem item)
		{
			int ii = -1; // Insert index.
			
			switch(lv.sorting)
			{
				case SortOrder.NONE: // Add to end.
					ii = _items.length;
					break;
				
				case SortOrder.ASCENDING: // Insertion sort.
					for(ii = 0; ii != _items.length; ii++)
					{
						if(item < _items[ii])
							break;
					}
					break;
				
				case SortOrder.DESCENDING: // Insertion sort.
					for(ii = 0; ii != _items.length; ii++)
					{
						if(item >= _items[ii])
							break;
					}
					break;
				
				default:
					assert(0);
			}
			
			assert(-1 != ii);
			insert(ii, item);
		}
		
		
		void add(char[] text)
		{
			return add(new ListViewItem(text));
		}
		
		
		void addRange(Object[] range)
		{
			foreach(Object o; range)
			{
				add(o);
			}
		}
		
		
		void addRange(ListViewItem[] range)
		{
			foreach(ListViewItem item; range)
			{
				add(item);
			}
		}
		
		
		void addRange(char[][] range)
		{
			foreach(char[] s; range)
			{
				add(s);
			}
		}
		
		
		void clear()
		{
			_items = null;
			
			if(created)
			{
				lv.prevwproc(LVM_DELETEALLITEMS, 0, 0);
			}
		}
		
		
		bool contains(Object value)
		{
			return indexOf(value) != -1;
		}
		
		
		bool contains(ListViewItem value)
		{
			return indexOf(value) != -1;
		}
		
		
		int indexOf(Object value)
		{
			ListViewItem item;
			item = cast(ListViewItem)item;
			if(item)
			{
				return indexOf(item);
			}
			else
			{
				foreach(int i, ListViewItem item; _items)
				{
					if(value == item)
						return i;
				}
				return -1;
			}
		}
		
		
		int indexOf(ListViewItem value)
		{
			int result = -1;
			foreach(int i, ListViewItem item; _items)
			{
				//if(value == item)
				{
					if(value is item)
					{
						result = i;
						break;
					}
					
					//if(result == -1)
					//	result = i;
				}
			}
			return result;
		}
		
		
		void insert(int index, Object value)
		{
			ListViewItem item;
			item = cast(ListViewItem)value;
			if(item)
			{
				insert(index, item);
			}
			else
			{
				insert(index, value.toString()); // ?
			}
		}
		
		
		private void verifyNoParent(ListViewItem item)
		{
			if(item.lview)
				throw new DflException("ListViewItem already belongs to a ListView.");
		}
		
		
		void insert(int index, char[] value)
		{
			insert(index, new ListViewItem(value));
		}
		
		
		void insert(int i, ListViewItem item)
		{
			verifyNoParent(item);
			
			item.lview = lv;
			
			if(i < 0)
				i = 0;
			/+if(i <= 0)
				_items = (&_items)[0 .. 1].dup ~ _items; // DMD 0.124: .dup needed for some reason.
			else+/ if(i >= cast(int)_items.length)
				_items ~= item;
			else
				_items = _items[0 .. i] ~ (&item)[0 .. 1] ~ _items[i .. _items.length];
			
			if(created)
			{
				i = lv._ins(i, item);
				assert(-1 != i);
				
				/+
				// Add sub items.
				foreach(int subi, ListViewSubItem subItem; item.isubs._subs)
				{
					lv._ins(i, subItem, subi + 1); // Sub items really start at 1 in the list view.
				}
				+/
			}
		}
		
		
		void remove(Object value)
		{
			int i;
			i = indexOf(value);
			if(i != -1)
				removeAt(i);
		}
		
		
		void remove(ListViewItem item)
		{
			int i;
			i = indexOf(item);
			if(i != -1)
				removeAt(i);
		}
		
		
		void removeAt(int index)
		{
			ListViewItem item;
			item = _items[index];
			
			item.lview = null;
			
			if(!index)
				_items = _items[1 .. _items.length];
			else if(index == _items.length - 1)
				_items = _items[0 .. index];
			else if(index > 0 && index < cast(int)_items.length)
				_items = _items[0 .. index] ~ _items[index + 1 .. _items.length];
			
			if(created)
			{
				lv.prevwproc(LVM_DELETEITEM, cast(WPARAM)index, 0);
			}
		}
		
		
		int count() // getter
		{
			return _items.length;
		}
		
		
		void opIndexAssign(Object value, int index) // setter
		{
			//array[index] = value;
			
			removeAt(index);
			//add(value);
			insert(index, value);
		}
		
		
		// Covariant return with interface IList.
		ListViewItem opIndex(int index) // getter
		{
			return _items[index];
		}
		
		
		IEnumerator getEnumerator()
		{
			return new ListItemEnum(_items);
		}
		
		
		private:
		ListView lv;
		package ListViewItem[] _items;
		
		
		package final void doListItems() // DMD 0.125: this member is not accessible when private.
		in
		{
			assert(created);
		}
		body
		{
			int ii;
			foreach(int i, ListViewItem item; _items)
			{
				ii = lv._ins(i, item);
				//assert(-1 != ii);
				assert(i == ii);
				
				/+
				// Add sub items.
				foreach(int subi, ListViewSubItem subItem; item.isubs._subs)
				{
					lv._ins(i, subItem, subi + 1); // Sub items really start at 1 in the list view.
				}
				+/
			}
		}
		
		
		static class ListItemEnum: IEnumerator
		{
			mixin EnumeratorForeach;
			//mixin EnumeratorForeach!(ListViewItem);
			
			
			this(ListViewItem[] items)
			{
				this.items = items;
			}
			
			
			// Covariant return with IEnumerator.
			//ListViewItem current() // getter
			Object current() // getter
			{
				return items[idx];
			}
			
			
			bool moveNext()
			{
				idx++;
				if(idx >= items.length) // Note: signed-unsigned comparison.
				{
					idx = items.length; // Prevent overflow.
					return false;
				}
				return true;
			}
			
			
			void reset()
			{
				idx = -1;
			}
			
			
			private:
			ListViewItem[] items;
			int idx = -1;
		}
	}
	
	
	static class ColumnHeaderCollection//: IList
	{
		this(ListView owner)
		in
		{
			assert(!owner.cols);
		}
		body
		{
			lv = owner;
		}
		
		
		bool isFixedSize() // getter
		{
			return false;
		}
		
		
		bool isReadOnly() // getter
		{
			return false;
		}
		
		
		bool isSynchronized() // getter
		{
			return false;
		}
		
		
		Object syncRoot() // getter
		{
			return this;
		}
		
		
		// Shallow copy.
		Array clone()
		{
			Array result;
			result = new Array(_headers.length);
			Array.copy(_headers, result, _headers.length);
			return result;
		}
		
		
		// D-style.
		final Array dup() // getter
		{
			return clone();
		}
		
		
		void copyTo(Array dest, int destIndex)
		{
			Array.copy(_headers, 0, dest, destIndex, _headers.length);
		}
		
		
		void add(Object o)
		{
			ColumnHeader header;
			header = cast(ColumnHeader)o;
			if(header)
			{
				return add(header);
			}
			else
			{
				return add(o.toString()); // ?
			}
		}
		
		
		package final bool created() // getter
		{
			return lv && lv.created();
		}
		
		
		void add(ColumnHeader header)
		{
			insert(cast(int)_headers.length, header);
		}
		
		
		void add(char[] text)
		{
			ColumnHeader col;
			col = new ColumnHeader;
			col.text = text;
			return add(col);
		}
		
		
		void addRange(Object[] range)
		{
			foreach(Object o; range)
			{
				add(o);
			}
		}
		
		
		void addRange(ColumnHeader[] range)
		{
			foreach(ColumnHeader header; range)
			{
				add(header);
			}
		}
		
		
		void addRange(char[][] range)
		{
			foreach(char[] s; range)
			{
				add(s);
			}
		}
		
		
		void clear()
		{
			size_t len;
			len = _headers.length;
			_headers = null;
			
			if(created)
			{
				while(len)
				{
					len--;
					lv.prevwproc(LVM_DELETECOLUMN, cast(WPARAM)len, 0);
				}
			}
		}
		
		
		bool contains(Object value)
		{
			return indexOf(value) != -1;
		}
		
		
		bool contains(ColumnHeader value)
		{
			return indexOf(value) != -1;
		}
		
		
		int indexOf(Object value)
		{
			ColumnHeader header;
			header = cast(ColumnHeader)header;
			if(header)
			{
				return indexOf(header);
			}
			else
			{
				foreach(int i, ColumnHeader header; _headers)
				{
					if(value == header)
						return i;
				}
				return -1;
			}
		}
		
		
		int indexOf(ColumnHeader value)
		{
			int result = -1;
			foreach(int i, ColumnHeader header; _headers)
			{
				//if(value == header)
				{
					if(value is header)
					{
						result = i;
						break;
					}
					
					//if(result == -1)
					//	result = i;
				}
			}
			return result;
		}
		
		
		void insert(int index, Object value)
		{
			ColumnHeader header;
			header = cast(ColumnHeader)value;
			if(header)
			{
				insert(index, header);
			}
			else
			{
				insert(index, value.toString()); // ?
			}
		}
		
		
		private void verifyNoParent(ColumnHeader header)
		{
			if(header.lview)
				throw new DflException("ColumnHeader already belongs to a ListView.");
		}
		
		
		void insert(int index, char[] value)
		{
			ColumnHeader col;
			col = new ColumnHeader;
			col.text = value;
			insert(index, col);
		}
		
		
		void insert(int i, ColumnHeader header)
		{
			verifyNoParent(header);
			
			header.lview = lv;
			
			if(i < 0)
				i = 0;
			/+if(i <= 0)
				_headers = (&_headers)[0 .. 1].dup ~ _headers; // DMD 0.124: .dup needed for some reason.
			else+/ if(i >= cast(int)_headers.length)
				_headers ~= header;
			else
				_headers = _headers[0 .. i] ~ (&header)[0 .. 1] ~ _headers[i .. _headers.length];
			
			if(created)
			{
				i = lv._ins(i, header);
				assert(-1 != i);
			}
		}
		
		
		void remove(Object value)
		{
			int i;
			i = indexOf(value);
			if(i != -1)
				removeAt(i);
		}
		
		
		void remove(ColumnHeader header)
		{
			int i;
			i = indexOf(header);
			if(i != -1)
				removeAt(i);
		}
		
		
		void removeAt(int index)
		{
			ColumnHeader header;
			header = _headers[index];
			
			header.lview = null;
			
			if(!index)
				_headers = _headers[1 .. _headers.length];
			else if(index == _headers.length - 1)
				_headers = _headers[0 .. index];
			else if(index > 0 && index < cast(int)_headers.length)
				_headers = _headers[0 .. index] ~ _headers[index + 1 .. _headers.length];
			
			if(created)
			{
				lv.prevwproc(LVM_DELETECOLUMN, cast(WPARAM)index, 0);
			}
		}
		
		
		int count() // getter
		{
			return _headers.length;
		}
		
		
		void opIndexAssign(Object value, int index) // setter
		{
			//array[index] = value;
			
			removeAt(index);
			//add(value);
			insert(index, value);
		}
		
		
		// Covariant return with interface IList.
		ColumnHeader opIndex(int index) // getter
		{
			return _headers[index];
		}
		
		
		IEnumerator getEnumerator()
		{
			return new ColEnum(_headers);
		}
		
		
		private:
		ListView lv;
		ColumnHeader[] _headers;
		
		
		package final void doListHeaders() // DMD 0.125: this member is not accessible when private.
		in
		{
			assert(created);
		}
		body
		{
			int ii;
			foreach(int i, ColumnHeader header; _headers)
			{
				ii = lv._ins(i, header);
				assert(-1 != ii);
				//assert(i == ii);
			}
		}
		
		
		static class ColEnum: IEnumerator
		{
			mixin EnumeratorForeach;
			//mixin EnumeratorForeach!(ColumnHeader);
			
			
			this(ColumnHeader[] headers)
			{
				this.headers = headers;
			}
			
			
			// Covariant return with IEnumerator.
			//ColumnHeader current() // getter
			Object current() // getter
			{
				return headers[idx];
			}
			
			
			bool moveNext()
			{
				idx++;
				if(idx >= headers.length) // Note: signed-unsigned comparison.
				{
					idx = headers.length; // Prevent overflow.
					return false;
				}
				return true;
			}
			
			
			void reset()
			{
				idx = -1;
			}
			
			
			private:
			ColumnHeader[] headers;
			int idx = -1;
		}
	}
	
	
	static class SelectedIndexCollection
	{
		int count() // getter
		{
			if(!lview.created)
				return 0;
			
			int result = 0;
			foreach(int onidx; this)
			{
				result++;
			}
			return result;
		}
		
		
		bool isReadOnly() // getter
		{
			return true;
		}
		
		
		int opIndex(int idx)
		{
			foreach(int onidx; this)
			{
				if(!idx)
					return onidx;
				idx--;
			}
			
			// If it's not found it's out of bounds and bad things happen.
			assert(0);
			return -1;
		}
		
		
		bool contains(int idx)
		{
			return indexOf(idx) != -1;
		}
		
		
		int indexOf(int idx)
		{
			int i = 0;
			foreach(int onidx; this)
			{
				if(onidx == idx)
					return i;
				i++;
			}
			return -1;
		}
		
		
		int opApply(int delegate(inout int) dg)
		{
			if(!lview.created)
				return 0;
			
			/+
			int result = 0;
			int idx;
			UINT istate;
			// Make sure nothing screws up if the user modifies the list while iterating.
			for(idx = 0; idx < lview.litems._items.length; idx++)
			{
				assert(idx >= 0);
				istate = cast(UINT)lview.prevwproc(LVM_GETITEMSTATE, cast(WPARAM)idx, cast(LPARAM)LVIS_SELECTED);
				if(istate & LVIS_SELECTED)
				{
					int dgidx = idx; // Prevent inout.
					result = dg(dgidx);
					if(result)
						break;
				}
			}
			return result;
			+/
			int result = 0;
			int idx = -1;
			for(;;)
			{
				idx = cast(int)lview.prevwproc(LVM_GETNEXTITEM, cast(WPARAM)idx, MAKELPARAM(cast(UINT)LVNI_SELECTED, 0));
				if(-1 == idx) // Done.
					break;
				int dgidx = idx; // Prevent inout.
				result = dg(dgidx);
				if(result)
					break;
			}
			return result;
		}
		
		
		this(ListView lv)
		{
			lview = lv;
		}
		
		
		package:
		ListView lview;
	}
	
	
	deprecated alias SelectedItemCollection SelectedListViewItemCollection;
	
	static class SelectedItemCollection
	{
		int count() // getter
		{
			if(!lview.created)
				return 0;
			
			int result = 0;
			foreach(ListViewItem onitem; this)
			{
				result++;
			}
			return result;
		}
		
		
		bool isReadOnly() // getter
		{
			return true;
		}
		
		
		ListViewItem opIndex(int idx)
		{
			foreach(ListViewItem onitem; this)
			{
				if(!idx)
					return onitem;
				idx--;
			}
			
			// If it's not found it's out of bounds and bad things happen.
			assert(0);
			return null;
		}
		
		
		bool contains(ListViewItem item)
		{
			return indexOf(item) != -1;
		}
		
		
		int indexOf(ListViewItem item)
		{
			int i = 0;
			foreach(ListViewItem onitem; this)
			{
				if(onitem == item) // Not using is.
					return i;
				i++;
			}
			return -1;
		}
		
		
		int opApply(int delegate(inout ListViewItem) dg)
		{
			if(!lview.created)
				return 0;
			
			/+
			int result = 0;
			int idx;
			UINT istate;
			// Make sure nothing screws up if the user modifies the list while iterating.
			for(idx = 0; idx < lview.litems._items.length; idx++)
			{
				assert(idx >= 0);
				istate = cast(UINT)lview.prevwproc(LVM_GETITEMSTATE, cast(WPARAM)idx, cast(LPARAM)LVIS_SELECTED);
				if(istate & LVIS_SELECTED)
				{
					ListViewItem litem = lview.litems._items[idx]; // Prevent inout.
					result = dg(litem);
					if(result)
						break;
				}
			}
			return result;
			+/
			int result = 0;
			int idx = -1;
			for(;;)
			{
				idx = cast(int)lview.prevwproc(LVM_GETNEXTITEM, cast(WPARAM)idx, MAKELPARAM(cast(UINT)LVNI_SELECTED, 0));
				if(-1 == idx) // Done.
					break;
				ListViewItem litem = lview.litems._items[idx]; // Prevent inout.
				result = dg(litem);
				if(result)
					break;
			}
			return result;
		}
		
		
		this(ListView lv)
		{
			lview = lv;
		}
		
		
		package:
		ListView lview;
	}
	
	
	this()
	{
		_initListview();
		
		litems = new ListViewItemCollection(this);
		cols = new ColumnHeaderCollection(this);
		selidxcollection = new SelectedIndexCollection(this);
		selobjcollection = new SelectedItemCollection(this);
		
		backc = defaultBackColor;
		forec = defaultForeColor;
		
		wstyle |= WS_TABSTOP | LVS_ALIGNTOP | LVS_AUTOARRANGE; // LVS_SHAREIMAGELISTS
		wexstyle |= WS_EX_CLIENTEDGE;
		ctrlStyle |= ControlStyles.SELECTABLE;
		wclassStyle = listviewClassStyle;
	}
	
	
	final void activation(ItemActivation ia) // setter
	{
		switch(ia)
		{
			case ItemActivation.STANDARD:
				_lvexstyle(LVS_EX_ONECLICKACTIVATE | LVS_EX_TWOCLICKACTIVATE, 0);
				break;
			
			case ItemActivation.ONE_CLICK:
				_lvexstyle(LVS_EX_ONECLICKACTIVATE | LVS_EX_TWOCLICKACTIVATE, LVS_EX_ONECLICKACTIVATE);
				break;
			
			case ItemActivation.TWO_CLICK:
				_lvexstyle(LVS_EX_ONECLICKACTIVATE | LVS_EX_TWOCLICKACTIVATE, LVS_EX_TWOCLICKACTIVATE);
				break;
			
			default:
				assert(0);
		}
	}
	
	
	final ItemActivation activation() // getter
	{
		DWORD lvex;
		lvex = _lvexstyle();
		if(lvex & LVS_EX_ONECLICKACTIVATE)
			return ItemActivation.ONE_CLICK;
		if(lvex & LVS_EX_TWOCLICKACTIVATE)
			return ItemActivation.TWO_CLICK;
		return ItemActivation.STANDARD;
	}
	
	
	/+
	final void alignment(ListViewAlignment lva)
	{
		// TODO
		
		switch(lva)
		{
			case ListViewAlignment.TOP:
				_style((_style() & ~(LVS_ALIGNLEFT | foo)) | LVS_ALIGNTOP);
				break;
			
			default:
				assert(0);
		}
	}
	
	
	final ListViewAlignment alignment() // getter
	{
		// TODO
	}
	+/
	
	
	final void allowColumnReorder(bool byes) // setter
	{
		_lvexstyle(LVS_EX_HEADERDRAGDROP, byes ? LVS_EX_HEADERDRAGDROP : 0);
	}
	
	
	final bool allowColumnReorder() // getter
	{
		return (_lvexstyle() & LVS_EX_HEADERDRAGDROP) == LVS_EX_HEADERDRAGDROP;
	}
	
	
	final void autoArrange(bool byes) // setter
	{
		if(byes)
			_style(_style() | LVS_AUTOARRANGE);
		else
			_style(_style() & ~LVS_AUTOARRANGE);
	}
	
	
	final bool autoArrange() // getter
	{
		return (_style() & LVS_AUTOARRANGE) == LVS_AUTOARRANGE;
	}
	
	
	alias Control.backColor backColor; // Override.
	
	override void backColor(Color c) // setter
	{
		if(created)
		{
			COLORREF cref;
			if(Color.empty == c)
				cref = CLR_NONE;
			else
				cref = c.toRgb();
			prevwproc(LVM_SETBKCOLOR, 0, cast(LPARAM)cref);
			prevwproc(LVM_SETTEXTBKCOLOR, 0, cast(LPARAM)cref);
		}
		
		super.backColor = c;
	}
	
	
	final void borderStyle(BorderStyle bs) // setter
	{
		switch(bs)
		{
			case BorderStyle.FIXED_3D:
				_style(_style() & ~WS_BORDER);
				_exStyle(_exStyle() | WS_EX_CLIENTEDGE);
				break;
				
			case BorderStyle.FIXED_SINGLE:
				_exStyle(_exStyle() & ~WS_EX_CLIENTEDGE);
				_style(_style() | WS_BORDER);
				break;
				
			case BorderStyle.NONE:
				_style(_style() & ~WS_BORDER);
				_exStyle(_exStyle() & ~WS_EX_CLIENTEDGE);
				break;
		}
		
		if(created)
		{
			redrawEntire();
		}
	}
	
	
	final BorderStyle borderStyle() // getter
	{
		if(_exStyle() & WS_EX_CLIENTEDGE)
			return BorderStyle.FIXED_3D;
		else if(_style() & WS_BORDER)
			return BorderStyle.FIXED_SINGLE;
		return BorderStyle.NONE;
	}
	
	
	final void checkBoxes(bool byes) // setter
	{
		_lvexstyle(LVS_EX_CHECKBOXES, byes ? LVS_EX_CHECKBOXES : 0);
	}
	
	
	final bool checkBoxes() // getter
	{
		return (_lvexstyle() & LVS_EX_CHECKBOXES) == LVS_EX_CHECKBOXES;
	}
	
	
	/+
	// ListView.CheckedIndexCollection
	final CheckedIndexCollection checkedIndices() // getter
	{
		// TODO
	}
	+/
	
	
	/+
	// ListView.CheckedListViewItemCollection
	final CheckedListViewItemCollection checkedItems() // getter
	{
		// TODO
	}
	+/
	
	
	final ColumnHeaderCollection columns() // getter
	{
		return cols;
	}
	
	
	// Extra.
	final int focusedIndex() // getter
	{
		if(!created)
			return -1;
		return cast(int)prevwproc(LVM_GETNEXTITEM, cast(WPARAM)-1, MAKELPARAM(cast(UINT)LVNI_FOCUSED, 0));
	}
	
	
	final ListViewItem focusedItem() // getter
	{
		int i;
		i = focusedIndex;
		if(-1 == i)
			return null;
		return litems._items[i];
	}
	
	
	alias Control.foreColor foreColor; // Override.
	
	override void foreColor(Color c) // setter
	{
		if(created)
			prevwproc(LVM_SETTEXTCOLOR, 0, cast(LPARAM)c.toRgb());
		
		super.foreColor = c;
	}
	
	
	final void fullRowSelect(bool byes) // setter
	{
		_lvexstyle(LVS_EX_FULLROWSELECT, byes ? LVS_EX_FULLROWSELECT : 0);
	}
	
	
	final bool fullRowSelect() // getter
	{
		return (_lvexstyle() & LVS_EX_FULLROWSELECT) == LVS_EX_FULLROWSELECT;
	}
	
	
	final void gridLines(bool byes) // setter
	{
		_lvexstyle(LVS_EX_GRIDLINES, byes ? LVS_EX_GRIDLINES : 0);
	}
	
	
	final bool gridLines() // getter
	{
		return (_lvexstyle() & LVS_EX_GRIDLINES) == LVS_EX_GRIDLINES;
	}
	
	
	/+
	final void headerStyle(ColumnHeaderStyle chs) // setter
	{
		// TODO: LVS_NOCOLUMNHEADER ... default is clickable.
	}
	
	
	final ColumnHeaderStyle headerStyle() // getter
	{
		// TODO
	}
	+/
	
	
	final void hideSelection(bool byes) // setter
	{
		if(byes)
			_style(_style() & ~LVS_SHOWSELALWAYS);
		else
			_style(_style() | LVS_SHOWSELALWAYS);
	}
	
	
	final bool hideSelection() // getter
	{
		return (_style() & LVS_SHOWSELALWAYS) != LVS_SHOWSELALWAYS;
	}
	
	
	final void hoverSelection(bool byes) // setter
	{
		_lvexstyle(LVS_EX_TRACKSELECT, byes ? LVS_EX_TRACKSELECT : 0);
	}
	
	
	final bool hoverSelection() // getter
	{
		return (_lvexstyle() & LVS_EX_TRACKSELECT) == LVS_EX_TRACKSELECT;
	}
	
	
	final ListViewItemCollection items() // getter
	{
		return litems;
	}
	
	
	final void labelEdit(bool byes) // setter
	{
		if(byes)
			_style(_style() | LVS_EDITLABELS);
		else
			_style(_style() & ~LVS_EDITLABELS);
	}
	
	
	final bool labelEdit() // getter
	{
		return (_style() & LVS_EDITLABELS) == LVS_EDITLABELS;
	}
	
	
	final void labelWrap(bool byes) // setter
	{
		if(byes)
			_style(_style() & ~LVS_NOLABELWRAP);
		else
			_style(_style() | LVS_NOLABELWRAP);
	}
	
	
	final bool labelWrap() // getter
	{
		return (_style() & LVS_NOLABELWRAP) != LVS_NOLABELWRAP;
	}
	
	
	// TODO: final get/set listViewItemSorter: IComparer
	
	
	final void multiSelect(bool byes) // setter
	{
		if(byes)
			_style(_style() & ~LVS_SINGLESEL);
		else
			_style(_style() | LVS_SINGLESEL);
	}
	
	
	final bool multiSelect() // getter
	{
		return (_style() & LVS_SINGLESEL) != LVS_SINGLESEL;
	}
	
	
	// Note: scrollable=false is not compatible with the list or details(report) styles(views).
	// See Knowledge Base Article Q137520.
	final void scrollable(bool byes) // setter
	{
		if(byes)
			_style(_style() & ~LVS_NOSCROLL);
		else
			_style(_style() | LVS_NOSCROLL);
	}
	
	
	final bool scrollable() // getter
	{
		return (_style() & LVS_NOSCROLL) != LVS_NOSCROLL;
	}
	
	
	final SelectedIndexCollection selectedIndices() // getter
	{
		return selidxcollection;
	}
	
	
	final SelectedItemCollection selectedItems() // getter
	{
		return selobjcollection;
	}
	
	
	final void view(View v) // setter
	{
		switch(v)
		{
			case View.LARGE_ICON:
				_style(_style() & ~(LVS_SMALLICON | LVS_LIST | LVS_REPORT));
				break;
			
			case View.SMALL_ICON:
				_style((_style() & ~(LVS_LIST | LVS_REPORT)) | LVS_SMALLICON);
				break;
			
			case View.LIST:
				_style((_style() & ~(LVS_SMALLICON | LVS_REPORT)) | LVS_LIST);
				break;
			
			case View.DETAILS:
				_style((_style() & ~(LVS_SMALLICON | LVS_LIST)) | LVS_REPORT);
				break;
			
			default:
				assert(0);
		}
		
		/+
		if(created)
			invalidate(false);
		+/
	}
	
	
	final View view() // getter
	{
		LONG st;
		st = _style();
		if(st & LVS_SMALLICON)
			return View.SMALL_ICON;
		if(st & LVS_LIST)
			return View.LIST;
		if(st & LVS_REPORT)
			return View.DETAILS;
		return View.LARGE_ICON;
	}
	
	
	final void sorting(SortOrder so) // setter
	{
		/+
		switch(so)
		{
			case SortOrder.NONE:
				_style(_style() & ~(LVS_SORTASCENDING | LVS_SORTDESCENDING));
				break;
			
			case SortOrder.ASCENDING:
				_style((_style() & ~LVS_SORTDESCENDING) | LVS_SORTASCENDING);
				break;
			
			case SortOrder.DESCENDING:
				_style((_style() & ~LVS_SORTASCENDING) | LVS_SORTDESCENDING);
				break;
			
			default:
				assert(0);
		}
		+/
		_sortorder = so;
	}
	
	
	final SortOrder sorting() // getter
	{
		/+
		LONG st;
		st = _style();
		if(st & LVS_SORTASCENDING)
			return SortOrder.ASCENDING;
		if(st & LVS_SORTDESCENDING)
			return SortOrder.DESCENDING;
		return SortOrder.NONE;
		+/
		return _sortorder;
	}
	
	
	/+
	// Gets the first visible item.
	final ListViewItem topItem() // getter
	{
		if(!created)
			return null;
		// TODO: LVM_GETTOPINDEX
	}
	+/
	
	
	final void arrangeIcons()
	{
		if(created)
		//	SendMessageA(hwnd, LVM_ARRANGE, LVA_DEFAULT, 0);
			prevwproc(LVM_ARRANGE, LVA_DEFAULT, 0);
	}
	
	
	final void arrangeIcons(ListViewAlignment a)
	{
		if(created)
		{
			switch(a)
			{
				case ListViewAlignment.TOP:
					//SendMessageA(hwnd, LVM_ARRANGE, LVA_ALIGNTOP, 0);
					prevwproc(LVM_ARRANGE, LVA_ALIGNTOP, 0);
					break;
				
				case ListViewAlignment.DEFAULT:
					//SendMessageA(hwnd, LVM_ARRANGE, LVA_DEFAULT, 0);
					prevwproc(LVM_ARRANGE, LVA_DEFAULT, 0);
					break;
				
				case ListViewAlignment.LEFT:
					//SendMessageA(hwnd, LVM_ARRANGE, LVA_ALIGNLEFT, 0);
					prevwproc(LVM_ARRANGE, LVA_ALIGNLEFT, 0);
					break;
				
				case ListViewAlignment.SNAP_TO_GRID:
					//SendMessageA(hwnd, LVM_ARRANGE, LVA_SNAPTOGRID, 0);
					prevwproc(LVM_ARRANGE, LVA_SNAPTOGRID, 0);
					break;
				
				default:
					assert(0);
			}
		}
	}
	
	
	final void beginUpdate()
	{
		SendMessageA(handle, WM_SETREDRAW, false, 0);
	}
	
	
	final void endUpdate()
	{
		SendMessageA(handle, WM_SETREDRAW, true, 0);
		invalidate(true); // Show updates.
	}
	
	
	final void clear()
	{
		litems.clear();
	}
	
	
	final void ensureVisible(int index)
	{
		// Can only be visible if it's created. Check if correct implementation.
		createControl();
		
		//if(created)
		//	SendMessageA(hwnd, LVM_ENSUREVISIBLE, cast(WPARAM)index, FALSE);
			prevwproc(LVM_ENSUREVISIBLE, cast(WPARAM)index, FALSE);
	}
	
	
	/+
	// Returns null if no item is at this location.
	final ListViewItem getItemAt(int x, int y)
	{
		// LVM_FINDITEM LVFI_NEARESTXY ? since it's nearest, need to see if it's really at that location.
		// TODO
	}
	+/
	
	
	final Rect getItemRect(int index)
	{
		if(created)
		{
			RECT rect;
			rect.left = LVIR_BOUNDS;
			if(prevwproc(LVM_GETITEMRECT, cast(WPARAM)index, cast(LPARAM)&rect))
				return Rect(&rect);
		}
		return Rect(0, 0, 0, 0);
	}
	
	
	final Rect getItemRect(int index, ItemBoundsPortion ibp)
	{
		if(created)
		{
			RECT rect;
			switch(ibp)
			{
				case ItemBoundsPortion.ENTIRE:
					rect.left = LVIR_BOUNDS;
					break;
				
				case ItemBoundsPortion.ICON:
					rect.left = LVIR_ICON;
					break;
				
				case ItemBoundsPortion.ITEM_ONLY:
					rect.left = LVIR_SELECTBOUNDS; // ?
					break;
				
				case ItemBoundsPortion.LABEL:
					rect.left = LVIR_LABEL;
					break;
				
				default:
					assert(0);
			}
			if(prevwproc(LVM_GETITEMRECT, cast(WPARAM)index, cast(LPARAM)&rect))
				return Rect(&rect);
		}
		return Rect(0, 0, 0, 0);
	}
	
	
	/+
	final void sort()
	{
		// TODO: LVM_SORTITEMS, lparam, sortproc using comparer interface
	}
	+/
	
	
	// TODO:
	//LabelEditEventHandler afterLabelEdit;
	//LabelEditEventHandler beforeLabelEdit;
	//ColumnClickEventHandler columnClick;
	//  itemActivate, itemCheck, itemDrag
	//EventHandler selectedIndexChanged;
	//CancelEventHandler selectedIndexChanging; // ?
	
	
	protected override Size defaultSize() // getter
	{
		return Size(120, 95);
	}
	
	
	static Color defaultBackColor() // getter
	{
		return SystemColors.window;
	}
	
	
	static Color defaultForeColor() //getter
	{
		return SystemColors.windowText;
	}
	
	
	protected override void createParams(inout CreateParams cp)
	{
		super.createParams(cp);
		
		cp.className = LISTVIEW_CLASSNAME;
	}
	
	
	protected override void prevWndProc(inout Message msg)
	{
		msg.result = CallWindowProcA(listviewPrevWndProc, msg.hWnd, msg.msg, msg.wParam, msg.lParam);
	}
	
	
	protected override void wndProc(inout Message m)
	{
		// TODO: support the listview messages.
		
		switch(m.msg)
		{
			case WM_PAINT:
				// This seems to be the only way to display columns correctly.
				prevWndProc(m);
				return;
			
			case LVM_ARRANGE:
				m.result = FALSE;
				return;
			
			case LVM_DELETEALLITEMS:
				litems.clear();
				m.result = TRUE;
				return;
			
			case LVM_DELETECOLUMN:
				cols.removeAt(cast(int)m.wParam);
				m.result = TRUE;
				return;
			
			case LVM_DELETEITEM:
				litems.removeAt(cast(int)m.wParam);
				m.result = TRUE;
				return;
			
			case LVM_INSERTCOLUMNA:
			case LVM_INSERTCOLUMNW:
				m.result = -1;
				return;
			
			case LVM_INSERTITEMA:
			case LVM_INSERTITEMW:
				m.result = -1;
				return;
			
			case LVM_SETBKCOLOR:
				backColor = Color(cast(COLORREF)m.lParam);
				m.result = TRUE;
				return;
			
			case LVM_SETCALLBACKMASK:
				m.result = FALSE;
				return;
			
			case LVM_SETCOLUMNA:
			case LVM_SETCOLUMNW:
				m.result = FALSE;
				return;
			
			case LVM_SETCOLUMNWIDTH:
				return;
			
			case LVM_SETIMAGELIST:
				m.result = cast(LRESULT)null;
				return;
			
			case LVM_SETITEMA:
				m.result = FALSE;
				return;
			
			case LVM_SETITEMSTATE:
				m.result = FALSE;
				return;
			
			case LVM_SETITEMTEXTA:
			case LVM_SETITEMTEXTW:
				m.result = FALSE;
				return;
			
			//case LVM_SETTEXTBKCOLOR:
			
			case LVM_SETTEXTCOLOR:
				foreColor = Color(cast(COLORREF)m.lParam);
				m.result = TRUE;
				return;
			
			case LVM_SORTITEMS:
				m.result = FALSE;
				return;
			
			default: ;
		}
		super.wndProc(m);
	}
	
	
	protected override void onHandleCreated(EventArgs ea)
	{
		super.onHandleCreated(ea);
		
		//SendMessageA(hwnd, LVM_SETEXTENDEDLISTVIEWSTYLE, wlvexstyle, wlvexstyle);
		prevwproc(LVM_SETEXTENDEDLISTVIEWSTYLE, 0, wlvexstyle); // wparam=0 sets all.
		//redrawEntire(); // Need to recalc the frame ?
		
		Color color;
		COLORREF cref;
		
		color = backColor;
		if(Color.empty == color)
			cref = CLR_NONE;
		else
			cref = color.toRgb();
		prevwproc(LVM_SETBKCOLOR, 0, cast(LPARAM)cref);
		prevwproc(LVM_SETTEXTBKCOLOR, 0, cast(LPARAM)cref);
		
		//prevwproc(LVM_SETTEXTCOLOR, 0, foreColor.toRgb()); // DMD 0.125: cast(Control )(this).foreColor() is not an lvalue
		color = foreColor;
		prevwproc(LVM_SETTEXTCOLOR, 0, cast(LPARAM)color.toRgb());
		
		cols.doListHeaders();
		litems.doListItems();
	}
	
	
	/+package+/ override LRESULT customMsg(inout CustomMsg m) // package
	{
		switch(m.message)
		{
			case WM_NOTIFY:
				{
					NMHDR* nmh;
					nmh = cast(NMHDR*)m.lParam;
					switch(nmh.code)
					{
						case LVN_GETDISPINFOA:
							{
								LV_DISPINFOA* lvdi;
								lvdi = cast(LV_DISPINFOA*)nmh;
								
								// Note: might want to verify it's a valid ListViewItem.
								
								ListViewItem item;
								item = cast(ListViewItem)cast(void*)lvdi.item.lParam;
								
								if(!lvdi.item.iSubItem) // Item.
								{
									lvdi.item.pszText = item.calltxt.ansi;
								}
								else // Sub item.
								{
									if(lvdi.item.iSubItem < item.subItems.count)
										lvdi.item.pszText = item.subItems[lvdi.item.iSubItem - 1].calltxt.ansi;
								}
							}
							break;
						
						case LVN_GETDISPINFOW:
							{
								char[] text;
								LV_DISPINFOW* lvdi;
								lvdi = cast(LV_DISPINFOW*)nmh;
								
								// Note: might want to verify it's a valid ListViewItem.
								
								ListViewItem item;
								item = cast(ListViewItem)cast(void*)lvdi.item.lParam;
								
								if(!lvdi.item.iSubItem) // Item.
								{
									lvdi.item.pszText = item.calltxt.unicode;
								}
								else // Sub item.
								{
									if(lvdi.item.iSubItem < item.subItems.count)
										lvdi.item.pszText = item.subItems[lvdi.item.iSubItem - 1].calltxt.unicode;
								}
							}
							break;
						
						default: ;
					}
				}
				break;
			
			default: ;
		}
		return 0;
	}
	
	
	private:
	DWORD wlvexstyle = 0;
	ListViewItemCollection litems;
	ColumnHeaderCollection cols;
	SelectedIndexCollection selidxcollection;
	SelectedItemCollection selobjcollection;
	SortOrder _sortorder = SortOrder.NONE;
	
	
	DWORD _lvexstyle()
	{
		//if(created)
		//	wlvexstyle = cast(DWORD)SendMessageA(hwnd, LVM_GETEXTENDEDLISTVIEWSTYLE, 0, 0);
		//	wlvexstyle = cast(DWORD)prevwproc(LVM_GETEXTENDEDLISTVIEWSTYLE, 0, 0);
		return wlvexstyle;
	}
	
	
	void _lvexstyle(DWORD flags)
	{
		DWORD _b4;
		_b4 = wlvexstyle;
		
		wlvexstyle = flags;
		if(created)
		{
			// hwnd, msg, mask, flags
			//SendMessageA(hwnd, LVM_SETEXTENDEDLISTVIEWSTYLE, flags ^ _b4, wlvexstyle);
			prevwproc(LVM_SETEXTENDEDLISTVIEWSTYLE, flags ^ _b4, wlvexstyle);
			//redrawEntire(); // Need to recalc the frame ?
		}
	}
	
	
	void _lvexstyle(DWORD mask, DWORD flags)
	in
	{
		assert(mask);
	}
	body
	{
		wlvexstyle = (wlvexstyle & ~mask) | (flags & mask);
		if(created)
		{
			// hwnd, msg, mask, flags
			//SendMessageA(hwnd, LVM_SETEXTENDEDLISTVIEWSTYLE, mask, flags);
			prevwproc(LVM_SETEXTENDEDLISTVIEWSTYLE, mask, flags);
			//redrawEntire(); // Need to recalc the frame ?
		}
	}
	
	
	// If -subItemIndex- is 0 it's an item not a sub item.
	// Returns the insertion index or -1 on failure.
	package final LRESULT _ins(int index, LPARAM lparam, char[] itemText, int subItemIndex)
	in
	{
		assert(created);
	}
	body
	{
		/+
		printf("^ Insert item:  index=%d, lparam=0x%X, text='%.*s', subItemIndex=%d\n",
			index, lparam, itemText.length > 20 ? 20 : itemText.length, cast(char*)itemText, subItemIndex);
		+/
		
		LV_ITEMA lvi;
		lvi.mask = LVIF_TEXT | LVIF_PARAM;
		lvi.iItem = index;
		lvi.iSubItem = subItemIndex;
		//lvi.pszText = toStringz(itemText);
		lvi.pszText = LPSTR_TEXTCALLBACKA;
		lvi.lParam = lparam;
		return prevwproc(LVM_INSERTITEMA, 0, cast(LPARAM)&lvi);
	}
	
	
	package final LRESULT _ins(int index, ListViewItem item)
	{
		return _ins(index, cast(LPARAM)cast(void*)item, item.text, 0);
	}
	
	
	package final LRESULT _ins(int index, ListViewSubItem subItem, int subItemIndex)
	in
	{
		assert(subItemIndex > 0);
	}
	body
	{
		return _ins(index, cast(LPARAM)cast(void*)subItem, subItem.text, subItemIndex);
	}
	
	
	package final LRESULT _ins(int index, ColumnHeader header)
	{
		// TODO: column inserted at index 0 can only be left aligned, so will need to
		// insert a dummy column to change the alignment, then delete the dummy column.
		
		LV_COLUMNA lvc;
		lvc.mask = LVCF_FMT | LVCF_SUBITEM | LVCF_TEXT | LVCF_WIDTH;
		switch(header.textAlign)
		{
			case HorizontalAlignment.RIGHT:
				lvc.fmt = LVCFMT_RIGHT;
				break;
			
			case HorizontalAlignment.CENTER:
				lvc.fmt = LVCFMT_CENTER;
				break;
			
			default:
				lvc.fmt = LVCFMT_LEFT;
		}
		lvc.cx = header.width;
		lvc.pszText = toStringz(header.text);
		lvc.iSubItem = index; // iSubItem is probably only used when retrieving column info.
		return prevwproc(LVM_INSERTCOLUMNA, cast(WPARAM)index, cast(LPARAM)&lvc);
	}
	
	
	// If -subItemIndex- is 0 it's an item not a sub item.
	// Returns FALSE on failure.
	LRESULT updateItemText(int index, char[] newText, int subItemIndex = 0)
	in
	{
		assert(created);
	}
	body
	{
		return prevwproc(LVM_REDRAWITEMS, cast(WPARAM)index, cast(LPARAM)index);
	}
	
	
	LRESULT updateItemText(ListViewItem item, char[] newText, int subItemIndex = 0)
	{
		int index;
		index = item.index;
		assert(-1 != index);
		return updateItemText(index, newText, subItemIndex);
	}
	
	
	LRESULT updateColumnText(int colIndex, char[] newText)
	{
		LV_COLUMNA lvc;
		lvc.mask = LVCF_TEXT;
		lvc.pszText = toStringz(newText);
		return prevwproc(LVM_SETCOLUMNA, cast(WPARAM)colIndex, cast(LPARAM)&lvc);
	}
	
	
	LRESULT updateColumnText(ColumnHeader col, char[] newText)
	{
		int colIndex;
		colIndex = columns.indexOf(col);
		assert(-1 != colIndex);
		return updateColumnText(colIndex, newText);
	}
	
	
	LRESULT updateColumnAlign(int colIndex, HorizontalAlignment halign)
	{
		LV_COLUMNA lvc;
		lvc.mask = LVCF_FMT;
		switch(halign)
		{
			case HorizontalAlignment.RIGHT:
				lvc.fmt = LVCFMT_RIGHT;
				break;
			
			case HorizontalAlignment.CENTER:
				lvc.fmt = LVCFMT_CENTER;
				break;
			
			default:
				lvc.fmt = LVCFMT_LEFT;
		}
		return prevwproc(LVM_SETCOLUMNA, cast(WPARAM)colIndex, cast(LPARAM)&lvc);
	}
	
	
	LRESULT updateColumnAlign(ColumnHeader col, HorizontalAlignment halign)
	{
		int colIndex;
		colIndex = columns.indexOf(col);
		assert(-1 != colIndex);
		return updateColumnAlign(colIndex, halign);
	}
	
	
	LRESULT updateColumnWidth(int colIndex, int w)
	{
		LV_COLUMNA lvc;
		lvc.mask = LVCF_WIDTH;
		lvc.cx = w;
		return prevwproc(LVM_SETCOLUMNA, cast(WPARAM)colIndex, cast(LPARAM)&lvc);
	}
	
	
	LRESULT updateColumnWidth(ColumnHeader col, int w)
	{
		int colIndex;
		colIndex = columns.indexOf(col);
		assert(-1 != colIndex);
		return updateColumnWidth(colIndex, w);
	}
	
	
	package:
	final:
	LRESULT prevwproc(UINT msg, WPARAM wparam, LPARAM lparam)
	{
		return CallWindowProcA(listviewPrevWndProc, hwnd, msg, wparam, lparam);
	}
}

