// Written by Christopher E. Miller
// See the included license.txt for copyright and license details.


module dfl.menu;

private import std.string;

private import dfl.winapi, dfl.control, dfl.base, dfl.event;
private import dfl.drawing, dfl.application;


class ContextMenu: Menu // docmain
{
	final void show(Control control, Point pos)
	{
		SetForegroundWindow(control.handle);
		TrackPopupMenu(hmenu, TPM_LEFTALIGN | TPM_LEFTBUTTON | TPM_RIGHTBUTTON,
			pos.x, pos.y, 0, control.handle, null);
	}
	
	
	EventHandler popup;
	
	
	// Used internally.
	this(HMENU hmenu, bool owned = true)
	{
		super(hmenu, owned);
		
		_init();
	}
	
	
	this()
	{
		super(CreatePopupMenu());
		
		_init();
	}
	
	
	~this()
	{
		Application.removeMenu(this);
		
		debug(APP_PRINT)
			printf("~ContextMenu\n");
	}
	
	
	package LRESULT customMsg(inout CustomMsg msg)
	{
		LRESULT result = 0;
		
		switch(msg.message)
		{
			case WM_INITMENU:
				assert(cast(HMENU)msg.wParam == handle);
				
				//onPopup(EventArgs.empty);
				popup(this, EventArgs.empty);
				break;
			
			default: ;
		}
		
		return result;
	}
	
	
	private:
	void _init()
	{
		Application.addContextMenu(this);
	}
}


class MenuItem: Menu // docmain
{
	final void text(char[] txt) // setter
	{
		if(!menuItems.count && txt == SEPARATOR_TEXT)
		{
			_type(_type() | MFT_SEPARATOR);
		}
		else
		{
			if(mparent)
			{
				MENUITEMINFOA mii;
				
				if(fType & MFT_SEPARATOR)
					fType = ~MFT_SEPARATOR;
				mii.cbSize = mii.sizeof;
				mii.fMask = MIIM_TYPE;
				mii.fType = fType;
				mii.dwTypeData = toStringz(txt);
				
				mparent._setInfo(mid, false, &mii);
			}
		}
		
		mtext = txt;
	}
	
	
	final char[] text() // getter
	{
		// if(mparent) fetch text ?
		return mtext;
	}
	
	
	final Menu parent()
	{
		return mparent;
	}
	
	
	package final void _setParent(Menu newParent)
	{
		assert(!mparent);
		mparent = newParent;
		
		if(cast(uint)mindex > mparent.menuItems.count)
			mindex = mparent.menuItems.count;
		
		_setParent();
	}
	
	
	private void _setParent()
	{
		MENUITEMINFOA mii;
		MenuItem miparent;
		
		mii.cbSize = mii.sizeof;
		mii.fMask = MIIM_TYPE | MIIM_STATE | MIIM_ID | MIIM_SUBMENU;
		mii.fType = fType;
		mii.fState = fState;
		mii.wID = mid;
		mii.hSubMenu = handle;
		if(!(fType & MFT_SEPARATOR))
			mii.dwTypeData = toStringz(mtext);
		miparent = cast(MenuItem)mparent;
		if(miparent && !miparent.hmenu)
		{
			miparent.hmenu = CreatePopupMenu();
			
			if(miparent.parent() && miparent.parent.hmenu)
			{
				MENUITEMINFOA miiPopup;
				
				miiPopup.cbSize = miiPopup.sizeof;
				miiPopup.fMask = MIIM_SUBMENU;
				miiPopup.hSubMenu = miparent.hmenu;
				miparent.parent._setInfo(miparent._menuID, false, &miiPopup);
			}
		}
		mparent._insert(mindex, true, &mii);
	}
	
	
	package final void _unsetParent()
	{
		assert(mparent);
		assert(mparent.menuItems.count > 0);
		assert(mparent.hmenu);
		
		// Last child menu item, make the parent non-popup now.
		if(mparent.menuItems.count == 1)
		{
			MenuItem miparent;
			
			miparent = cast(MenuItem)mparent;
			if(miparent.parent() && miparent.parent.hmenu)
			{
				MENUITEMINFOA miiPopup;
				
				miiPopup.cbSize = miiPopup.sizeof;
				miiPopup.fMask = MIIM_SUBMENU;
				miiPopup.hSubMenu = null;
				miparent.parent._setInfo(miparent._menuID, false, &miiPopup);
				
				miparent.hmenu = null;
			}
		}
		
		mparent = null;
	}
	
	
	final void barBreak(bool byes) // setter
	{
		if(byes)
			_type(_type() | MFT_MENUBARBREAK);
		else
			_type(_type() & ~MFT_MENUBARBREAK);
	}
	
	
	final bool barBreak() // getter
	{
		return (_type() & MFT_MENUBARBREAK) != 0;
	}
	
	
	// Can't be break().
	
	final void breakItem(bool byes) // setter
	{
		if(byes)
			_type(_type() | MFT_MENUBREAK);
		else
			_type(_type() & ~MFT_MENUBREAK);
	}
	
	
	final bool breakItem() // getter
	{
		return (_type() & MFT_MENUBREAK) != 0;
	}
	
	
	final void checked(bool byes) // setter
	{
		if(byes)
			_state(_state() | MFS_CHECKED);
		else
			_state(_state() & ~MFS_CHECKED);
	}
	
	
	final bool checked() // getter
	{
		return (_state() & MFS_CHECKED) != 0;
	}
	
	
	final void defaultItem(bool byes) // setter
	{
		if(byes)
			_state(_state() | MFS_DEFAULT);
		else
			_state(_state() & ~MFS_DEFAULT);
	}
	
	
	final bool defaultItem() // getter
	{
		return (_state() & MFS_DEFAULT) != 0;
	}
	
	
	final void enabled(bool byes) // setter
	{
		if(byes)
			_state(_state() & ~MFS_GRAYED);
		else
			_state(_state() | MFS_GRAYED);
	}
	
	
	final bool enabled() // getter
	{
		return (_state() & MFS_GRAYED) == 0;
	}
	
	
	final void index(int idx) // setter
	{
		if(mparent)
		{
			if(cast(uint)idx > mparent.menuItems.count)
				throw new DflException("Invalid menu index.");
			
			//RemoveMenu(mparent.handle, mid, MF_BYCOMMAND);
			mparent._remove(mid, MF_BYCOMMAND);
			
			mindex = idx;
			_setParent();
		}
		
		mindex = idx;
	}
	
	
	final int index() // getter
	{
		return mindex;
	}
	
	
	override bool isParent() // getter
	{
		return handle != null; // ?
	}
	
	
	final void mergeOrder(int ord) // setter
	{
		mergeord = ord;
	}
	
	
	final int mergeOrder() // getter
	{
		return mergeord;
	}
	
	
	// TODO: mergeType().
	
	
	// Returns a NUL char if none.
	final char mnemonic() // getter
	{
		bool singleAmp = false;
		
		foreach(char ch; mtext)
		{
			if(singleAmp)
			{
				if(ch == '&')
					singleAmp = false;
				else
					return ch;
			}
			else
			{
				if(ch == '&')
					singleAmp = true;
			}
		}
		
		return 0;
	}
	
	
	/+
	// TODO: implement owner drawn menus.
	
	final void ownerDraw(bool byes)
	{
		
	}
	
	
	final bool ownerDraw() // getter
	{
		
	}
	+/
	
	
	final void radioCheck(bool byes) // setter
	{
		if(byes)
			_type(_type() | MFT_RADIOCHECK);
		else
			_type(_type() & ~MFT_RADIOCHECK);
	}
	
	
	final bool radioCheck() // getter
	{
		return (_type() & MFT_RADIOCHECK) != 0;
	}
	
	
	// TODO: shortcut(), showShortcut().
	
	
	/+
	// TODO: need to fake this ?
	
	final void visible(bool byes) // setter
	{
		// ?
		mvisible = byes;
	}
	
	
	final bool visible() // getter
	{
		return mvisible;
	}
	+/
	
	
	final void performClick()
	{
		onClick(EventArgs.empty);
	}
	
	
	final void performSelect()
	{
		onSelect(EventArgs.empty);
	}
	
	
	// Used internally.
	this(HMENU hmenu, bool owned = true)
	{
		_init();
		
		super(hmenu, owned);
	}
	
	
	this(MenuItem[] items)
	{
		_init();
		
		HMENU hm = CreatePopupMenu();
		menuItems.addRange(items);
		
		super(hm);
	}
	
	
	this()
	{
		_init();
	}
	
	
	~this()
	{
		Application.removeMenu(this);
		
		debug(APP_PRINT)
			printf("~MenuItem\n");
	}
	
	
	package LRESULT customMsg(inout CustomMsg msg)
	{
		LRESULT result = 0;
		
		switch(msg.message)
		{
			case WM_COMMAND:
				assert(LOWORD(msg.wParam) == mid);
				
				onClick(EventArgs.empty);
				break;
			
			case WM_MENUSELECT:
				onSelect(EventArgs.empty);
				break;
			
			case WM_INITMENUPOPUP:
				assert(!HIWORD(msg.lParam));
				//assert(cast(HMENU)msg.wParam == mparent.handle);
				assert(cast(HMENU)msg.wParam == handle);
				//assert(GetMenuItemID(mparent.handle, LOWORD(msg.lParam)) == mid);
				
				onPopup(EventArgs.empty);
				break;
			
			default: ;
		}
		
		return result;
	}
	
	
	EventHandler click;
	EventHandler popup;
	EventHandler select;
	
	
	protected:
	final int menuID() // getter
	{
		return mid;
	}
	
	
	package final int _menuID()
	{
		return mid;
	}
	
	
	void onClick(EventArgs ea)
	{
		click(this, ea);
	}
	
	
	void onPopup(EventArgs ea)
	{
		popup(this, ea);
	}
	
	
	void onSelect(EventArgs ea)
	{
		popup(this, ea);
	}
	
	
	private:
	int mid; // Menu ID.
	char[] mtext;
	Menu mparent;
	UINT fType = 0; // MFT_*
	UINT fState = 0;
	int mindex = 0;
	int mergeord = 0;
	
	const char[] SEPARATOR_TEXT = "-";
	
	static assert(!MFS_UNCHECKED);
	static assert(!MFT_STRING);
	
	
	void _init()
	{
		mid = Application.addMenuItem(this);
	}
	
	
	void _type(UINT newType) // setter
	{
		if(mparent)
		{
			MENUITEMINFOA mii;
			
			mii.cbSize = mii.sizeof;
			mii.fMask = MIIM_TYPE;
			mii.fType = newType;
			
			mparent._setInfo(mid, false, &mii);
		}
		
		fType = newType;
	}
	
	
	UINT _type() // getter
	{
		// if(mparent) fetch value ?
		return fType;
	}
	
	
	void _state(UINT newState) // setter
	{
		if(mparent)
		{
			MENUITEMINFOA mii;
			
			mii.cbSize = mii.sizeof;
			mii.fMask = MIIM_STATE;
			mii.fState = newState;
			
			mparent._setInfo(mid, false, &mii);
		}
		
		fState = newState;
	}
	
	
	UINT _state() // getter
	{
		// if(mparent) fetch value ?
		return fState;
	}
}


abstract class Menu // docmain
{
	static class MenuItemCollection
	{
		this(Menu owner)
		{
			_owner = owner;
		}
		
		
		int count()
		{
			// return GetMenuItemCount();
			return items.length;
		}
		
		
		MenuItem opIndex(int idx)
		{
			return items[idx];
		}
		
		
		void add(MenuItem mi)
		{
			mi._setParent(_owner);
			items ~= mi;
		}
		
		
		void addRange(MenuItem[] mis)
		{
			int i;
			
			try
			{
				for(i = 0; i != mis.length; i++)
				{
					mis[i]._setParent(_owner);
				}
			}
			finally
			{
				items ~= mis[0 .. i];
			}
		}
		
		
		void clear()
		{
			/+
			foreach(MenuItem mi; items)
			{
				//RemoveMenu(_owner.handle, mi.menuID, MF_BYCOMMAND);
				_owner._remove(mi.menuID, MF_BYCOMMAND);
			}
			+/
			
			int i;
			
			try
			{
				for(i = 0; i != items.length; i++)
				{
					items[i]._unsetParent();
					//RemoveMenu(_owner.handle, 0, MF_BYPOSITION);
					_owner._remove(0, MF_BYPOSITION);
				}
			}
			finally
			{
				items = items[i .. items.length];
			}
		}
		
		
		void remove(MenuItem mi)
		{
			foreach(int i, MenuItem onmi; items)
			{
				if(mi == onmi)
				{
					removeAt(i);
					break;
				}
			}
		}
		
		
		void removeAt(int idx)
		{
			items[idx]._unsetParent();
			//RemoveMenu(_owner.handle, items[idx]._menuID, MF_BYCOMMAND);
			_owner._remove(items[idx]._menuID, MF_BYCOMMAND);
			
			if(!idx)
				items = items[1 .. items.length];
			else if(idx == items.length - 1)
				items = items[0 .. idx];
			else
				items = items[0 .. idx] ~ items[idx + 1 .. items.length];
		}
		
		
		// TODO: finish.
		
		
		private:
		Menu _owner;
		MenuItem[] items; // Kept populated so the menu can be moved around.
	}
	
	
	private void _init()
	{
		items = new MenuItemCollection(this);
	}
	
	
	// Menu item that isn't popup (yet).
	private this()
	{
		_init();
	}
	
	
	// Used internally.
	this(HMENU hmenu, bool owned = true)
	{
		this.hmenu = hmenu;
		this.owned = owned;
		
		_init();
	}
	
	
	// Used internally.
	this(HMENU hmenu, MenuItem[] items)
	{
		this.owned = true;
		this.hmenu = hmenu;
		
		_init();
		
		menuItems.addRange(items);
	}
	
	
	// Don't call directly.
	this(MenuItem[] items)
	{
		/+
		this.owned = true;
		
		_init();
		
		menuItems.addRange(items);
		+/
		
		assert(0);
	}
	
	
	~this()
	{
		if(owned)
			DestroyMenu(hmenu);
	}
	
	
	final HMENU handle() // getter
	{
		return hmenu;
	}
	
	
	final MenuItemCollection menuItems() // getter
	{
		return items;
	}
	
	
	bool isParent() // getter
	{
		return false;
	}
	
	
	/+ package +/ protected void _setInfo(UINT uItem, BOOL fByPosition, LPMENUITEMINFOA lpmii) // package
	{
		SetMenuItemInfoA(hmenu, uItem, fByPosition, lpmii);
	}
	
	
	/+ package +/ protected void _insert(UINT uItem, BOOL fByPosition, LPMENUITEMINFOA lpmii) // package
	{
		InsertMenuItemA(hmenu, uItem, fByPosition, lpmii);
	}
	
	
	/+ package +/ protected void _remove(UINT uPosition, UINT uFlags) // package
	{
		RemoveMenu(hmenu, uPosition, uFlags);
	}
	
	
	package HMENU hmenu;
	
	private:
	bool owned = true;
	MenuItemCollection items;
}


class MainMenu: Menu // docmain
{
	// Used internally.
	this(HMENU hmenu, bool owned = true)
	{
		super(hmenu, owned);
	}
	
	
	this()
	{
		super(CreateMenu());
	}
	
	
	this(MenuItem[] items)
	{
		super(CreateMenu(), items);
	}
	
	
	/+ package +/ protected override void _setInfo(UINT uItem, BOOL fByPosition, LPMENUITEMINFOA lpmii) // package
	{
		Menu._setInfo(uItem, fByPosition, lpmii);
		
		if(hwnd)
			DrawMenuBar(hwnd);
	}
	
	
	/+ package +/ protected override void _insert(UINT uItem, BOOL fByPosition, LPMENUITEMINFOA lpmii) // package
	{
		Menu._insert(uItem, fByPosition, lpmii);
		
		if(hwnd)
			DrawMenuBar(hwnd);
	}
	
	
	/+ package +/ protected override void _remove(UINT uPosition, UINT uFlags) // package
	{
		Menu._remove(uPosition, uFlags);
		
		if(hwnd)
			DrawMenuBar(hwnd);
	}
	
	
	private:
	HWND hwnd = null;
	
	
	package final void _setHwnd(HWND hwnd)
	{
		this.hwnd = hwnd;
	}
}

