// Written by Christopher E. Miller
// See the included license.txt for copyright and license details.


module dfl.richtextbox;

private import dfl.textbox, dfl.winapi, dfl.event, dfl.application;
private import dfl.base, dfl.drawing, dfl.menu, dfl.data;
private import dfl.control;


private extern(C) char* strcpy(char*, char*);


private extern(Windows) void _initRichtextbox();


class LinkClickedEventArgs: EventArgs
{
	this(char[] linkText)
	{
		_linktxt = linkText;
	}
	
	
	final char[] linkText() // getter
	{
		return _linktxt;
	}
	
	
	private:
	char[] _linktxt;
}


alias Event!(LinkClickedEventArgs) LinkClickedEventHandler;


enum RichTextBoxScrollBars: ubyte
{
	NONE,
	HORIZONTAL,
	VERTICAL,
	BOTH,
	FORCED_HORIZONTAL,
	FORCED_VERTICAL,
	FORCED_BOTH,
}


class RichTextBox: TextBoxBase // docmain
{
	this()
	{
		super();
		
		_initRichtextbox();
		
		wstyle |= ES_MULTILINE | ES_WANTRETURN | ES_AUTOHSCROLL | ES_AUTOVSCROLL | WS_HSCROLL | WS_VSCROLL;
		wcurs = null; // So that the control can change it accordingly.
		wclassStyle = richtextboxClassStyle;
		
		with(miredo = new MenuItem)
		{
			text = "&Redo";
			click ~= &menuRedo;
			index = 1;
			contextMenu.menuItems.add(miredo);
		}
		
		contextMenu.popup ~= &menuPopup2;
	}
	
	
	private
	{
		void menuRedo(Object sender, EventArgs ea)
		{
			redo();
		}
		
		
		void menuPopup2(Object sender, EventArgs ea)
		{
			miredo.enabled = canRedo;
		}
		
		
		MenuItem miredo;
	}
	
	
	alias TextBoxBase.selectedText selectedText; // Overload.
	
	override char[] selectedText() // getter
	{
		if(created)
		{
			/+
			uint len = selectionLength + 1;
			char[] result = new char[len];
			len = SendMessageA(handle, EM_GETSELTEXT, 0, cast(LPARAM)cast(char*)result);
			assert(!result[len]);
			return result[0 .. len];
			+/
			
			return dfl.utf.emGetSelText(hwnd, selectionLength + 1);
		}
		return null;
	}
	
	
	override void selectionLength(uint len) // setter
	{
		if(created)
		{
			CHARRANGE chrg;
			SendMessageA(handle, EM_EXGETSEL, 0, cast(LPARAM)&chrg);
			chrg.cpMax = chrg.cpMin + len;
			SendMessageA(handle, EM_EXSETSEL, 0, cast(LPARAM)&chrg);
		}
	}
	
	
	// Current selection length, in characters.
	// This does not necessarily correspond to the length of chars; some characters use multiple chars.
	// An end of line (\r\n) takes up 2 characters.
	override uint selectionLength() // getter
	{
		if(created)
		{
			CHARRANGE chrg;
			SendMessageA(handle, EM_EXGETSEL, 0, cast(LPARAM)&chrg);
			assert(chrg.cpMax >= chrg.cpMin);
			return chrg.cpMax - chrg.cpMin;
		}
		return 0;
	}
	
	
	override void selectionStart(uint pos) // setter
	{
		if(created)
		{
			CHARRANGE chrg;
			SendMessageA(handle, EM_EXGETSEL, 0, cast(LPARAM)&chrg);
			assert(chrg.cpMax >= chrg.cpMin);
			chrg.cpMax = pos + (chrg.cpMax - chrg.cpMin);
			chrg.cpMin = pos;
			SendMessageA(handle, EM_EXSETSEL, 0, cast(LPARAM)&chrg);
		}
	}
	
	
	// Current selection starting index, in characters.
	// This does not necessarily correspond to the index of chars; some characters use multiple chars.
	// An end of line (\r\n) takes up 2 characters.
	override uint selectionStart() // getter
	{
		if(created)
		{
			CHARRANGE chrg;
			SendMessageA(handle, EM_EXGETSEL, 0, cast(LPARAM)&chrg);
			return chrg.cpMin;
		}
		return 0;
	}
	
	
	alias TextBoxBase.maxLength maxLength; // Overload.
	
	override void maxLength(uint len) // setter
	{
		lim = len;
		
		if(created)
			SendMessageA(handle, EM_EXLIMITTEXT, 0, cast(LPARAM)len);
	}
	
	
	override Size defaultSize() // getter
	{
		return Size(120, 120); // ?
	}
	
	
	alias TextBoxBase.backColor backColor; // Overload.
	
	override void backColor(Color c) // setter
	{
		if(created)
		{
			if(c._systemColorIndex == COLOR_WINDOW)
				SendMessageA(handle, EM_SETBKGNDCOLOR, 1, 0);
			else
				SendMessageA(handle, EM_SETBKGNDCOLOR, 0, cast(LPARAM)c.toRgb());
		}
		
		super.backColor(c);
	}
	
	
	alias TextBoxBase.foreColor foreColor; // Overload.
	
	override void foreColor(Color c) // setter
	{
		if(created)
		{
			CHARFORMAT2A cf;
			
			cf.cbSize = cf.sizeof;
			cf.dwMask = CFM_COLOR;
			if(c._systemColorIndex == COLOR_WINDOWTEXT)
				cf.dwEffects = CFE_AUTOCOLOR;
			else
				cf.crTextColor = c.toRgb();
			
			_setFormat(&cf, SCF_ALL);
		}
		
		super.foreColor(c);
	}
	
	
	final bool canRedo() // getter
	{
		if(!created)
			false;
		return SendMessageA(handle, EM_CANREDO, 0, 0) != 0;
	}
	
	
	final bool canPaste(DataFormats.Format df)
	{
		if(created)
		{
			if(SendMessageA(handle, EM_CANPASTE, df.id, 0))
				return true;
		}
		
		return false;
	}
	
	
	final void redo()
	{
		if(created)
			SendMessageA(handle, EM_REDO, 0, 0);
	}
	
	
	alias TextBoxBase.paste paste; // Overload.
	
	// "Paste special."
	final void paste(DataFormats.Format df)
	{
		if(created)
		{
			SendMessageA(handle, EM_PASTESPECIAL, df.id, cast(LPARAM)null);
		}
	}
	
	
	final void selectionCharOffset(int yoffset) // setter
	{
		if(!created)
			return;
		
		CHARFORMAT2A cf;
		
		cf.cbSize = cf.sizeof;
		cf.dwMask = CFM_OFFSET;
		cf.yOffset = yoffset;
		
		_setFormat(&cf);
	}
	
	
	final int selectionCharOffset() // getter
	{
		if(created)
		{
			CHARFORMAT2A cf;
			cf.cbSize = cf.sizeof;
			cf.dwMask = CFM_OFFSET;
			_getFormat(&cf);
			return cf.yOffset;
		}
		return 0;
	}
	
	
	final void selectionColor(Color c) // setter
	{
		if(!created)
			return;
		
		CHARFORMAT2A cf;
		
		cf.cbSize = cf.sizeof;
		cf.dwMask = CFM_COLOR;
		if(c._systemColorIndex == COLOR_WINDOWTEXT)
			cf.dwEffects = CFE_AUTOCOLOR;
		else
			cf.crTextColor = c.toRgb();
		
		_setFormat(&cf);
	}
	
	
	final Color selectionColor() // getter
	{
		if(created)
		{
			CHARFORMAT2A cf;
			
			cf.cbSize = cf.sizeof;
			cf.dwMask = CFM_COLOR;
			_getFormat(&cf);
			
			if(cf.dwMask & CFM_COLOR)
			{
				if(cf.dwEffects & CFE_AUTOCOLOR)
					return Color.systemColor(COLOR_WINDOWTEXT);
				return Color(cf.crTextColor);
			}
		}
		return Color.empty;
	}
	
	
	final void selectionBackColor(Color c) // setter
	{
		if(!created)
			return;
		
		CHARFORMAT2A cf;
		
		cf.cbSize = cf.sizeof;
		cf.dwMask = CFM_BACKCOLOR;
		if(c._systemColorIndex == COLOR_WINDOW)
			cf.dwEffects = CFE_AUTOBACKCOLOR;
		else
			cf.crBackColor = c.toRgb();
		
		_setFormat(&cf);
	}
	
	
	final Color selectionBackColor() // getter
	{
		if(created)
		{
			CHARFORMAT2A cf;
			
			cf.cbSize = cf.sizeof;
			cf.dwMask = CFM_BACKCOLOR;
			_getFormat(&cf);
			
			if(cf.dwMask & CFM_BACKCOLOR)
			{
				if(cf.dwEffects & CFE_AUTOBACKCOLOR)
					return Color.systemColor(COLOR_WINDOW);
				return Color(cf.crBackColor);
			}
		}
		return Color.empty;
	}
	
	
	final void selectionSubscript(bool byes) // setter
	{
		if(!created)
			return;
		
		CHARFORMAT2A cf;
		
		cf.cbSize = cf.sizeof;
		cf.dwMask = CFM_SUPERSCRIPT | CFM_SUBSCRIPT;
		if(byes)
		{
			cf.dwEffects = CFE_SUBSCRIPT;
		}
		else
		{
			// Make sure it doesn't accidentally unset superscript.
			CHARFORMAT2A cf2get;
			cf2get.cbSize = cf2get.sizeof;
			cf2get.dwMask = CFM_SUPERSCRIPT | CFM_SUBSCRIPT;
			_getFormat(&cf2get);
			if(cf2get.dwEffects & CFE_SUPERSCRIPT)
				return; // Superscript is set, so don't bother.
			if(!(cf2get.dwEffects & CFE_SUBSCRIPT))
				return; // Don't need to unset twice.
		}
		
		_setFormat(&cf);
	}
	
	
	final bool selectionSubscript() // getter
	{
		if(created)
		{
			CHARFORMAT2A cf;
			
			cf.cbSize = cf.sizeof;
			cf.dwMask = CFM_SUPERSCRIPT | CFM_SUBSCRIPT;
			_getFormat(&cf);
			
			return (cf.dwEffects & CFE_SUBSCRIPT) == CFE_SUBSCRIPT;
		}
		return false;
	}
	
	
	final void selectionSuperscript(bool byes) // setter
	{
		if(!created)
			return;
		
		CHARFORMAT2A cf;
		
		cf.cbSize = cf.sizeof;
		cf.dwMask = CFM_SUPERSCRIPT | CFM_SUBSCRIPT;
		if(byes)
		{
			cf.dwEffects = CFE_SUPERSCRIPT;
		}
		else
		{
			// Make sure it doesn't accidentally unset subscript.
			CHARFORMAT2A cf2get;
			cf2get.cbSize = cf2get.sizeof;
			cf2get.dwMask = CFM_SUPERSCRIPT | CFM_SUBSCRIPT;
			_getFormat(&cf2get);
			if(cf2get.dwEffects & CFE_SUBSCRIPT)
				return; // Subscript is set, so don't bother.
			if(!(cf2get.dwEffects & CFE_SUPERSCRIPT))
				return; // Don't need to unset twice.
		}
		
		_setFormat(&cf);
	}
	
	
	final bool selectionSuperscript() // getter
	{
		if(created)
		{
			CHARFORMAT2A cf;
			
			cf.cbSize = cf.sizeof;
			cf.dwMask = CFM_SUPERSCRIPT | CFM_SUBSCRIPT;
			_getFormat(&cf);
			
			return (cf.dwEffects & CFE_SUPERSCRIPT) == CFE_SUPERSCRIPT;
		}
		return false;
	}
	
	
	private const DWORD FONT_MASK = CFM_BOLD | CFM_ITALIC | CFM_STRIKEOUT |
		CFM_UNDERLINE | CFM_CHARSET | CFM_FACE | CFM_SIZE | CFM_UNDERLINETYPE | CFM_WEIGHT;
	
	final void selectionFont(Font f) // setter
	{
		if(created)
		{
			CHARFORMAT2A cf;
			LOGFONTA lf;
			
			f._info(&lf);
			
			cf.cbSize = cf.sizeof;
			cf.dwMask = FONT_MASK;
			
			//cf.dwEffects = 0;
			if(lf.lfWeight >= FW_BOLD)
				cf.dwEffects |= CFE_BOLD;
			if(lf.lfItalic)
				cf.dwEffects |= CFE_ITALIC;
			if(lf.lfStrikeOut)
				cf.dwEffects |= CFE_STRIKEOUT;
			if(lf.lfUnderline)
				cf.dwEffects |= CFE_UNDERLINE;
			cf.yHeight = cast(typeof(cf.yHeight))Font.getEmSize(lf.lfHeight, GraphicsUnit.TWIP);
			cf.bCharSet = lf.lfCharSet;
			strcpy(cf.szFaceName, lf.lfFaceName);
			cf.bUnderlineType = CFU_UNDERLINE;
			cf.wWeight = lf.lfWeight;
			
			_setFormat(&cf);
		}
	}
	
	
	// Returns null if the selection has different fonts.
	final Font selectionFont() // getter
	{
		if(created)
		{
			CHARFORMAT2A cf;
			
			cf.cbSize = cf.sizeof;
			cf.dwMask = FONT_MASK;
			_getFormat(&cf);
			
			if((cf.dwMask & FONT_MASK) == FONT_MASK)
			{
				LOGFONTA lf;
				with(lf)
				{
					lfHeight = -Font.getLfHeight(cast(float)cf.yHeight, GraphicsUnit.TWIP);
					lfWidth = 0; // ?
					lfEscapement = 0; // ?
					lfOrientation = 0; // ?
					lfWeight = cf.wWeight;
					if(cf.dwEffects & CFE_BOLD)
					{
						if(lfWeight < FW_BOLD)
							lfWeight = FW_BOLD;
					}
					lfItalic = (cf.dwEffects & CFE_ITALIC) != 0;
					lfUnderline = (cf.dwEffects & CFE_UNDERLINE) != 0;
					lfStrikeOut = (cf.dwEffects & CFE_STRIKEOUT) != 0;
					lfCharSet = cf.bCharSet;
					strcpy(lfFaceName, cf.szFaceName);
					lfOutPrecision = OUT_DEFAULT_PRECIS;
					lf.lfClipPrecision = CLIP_DEFAULT_PRECIS;
					lf.lfQuality = DEFAULT_QUALITY;
					lf.lfPitchAndFamily = DEFAULT_PITCH | FF_DONTCARE;
				}
				return new Font(Font._create(&lf));
			}
		}
		
		return null;
	}
	
	
	final void scrollBars(RichTextBoxScrollBars sb) // setter
	{
		LONG st;
		st = _style() & ~(ES_DISABLENOSCROLL | WS_HSCROLL | WS_VSCROLL |
			ES_AUTOHSCROLL | ES_AUTOVSCROLL);
		
		switch(sb)
		{
			case RichTextBoxScrollBars.FORCED_BOTH:
				st |= ES_DISABLENOSCROLL;
			case RichTextBoxScrollBars.BOTH:
				st |= WS_HSCROLL | WS_VSCROLL | ES_AUTOHSCROLL | ES_AUTOVSCROLL;
				break;
			
			case RichTextBoxScrollBars.FORCED_HORIZONTAL:
				st |= ES_DISABLENOSCROLL;
			case RichTextBoxScrollBars.HORIZONTAL:
				st |= WS_HSCROLL | ES_AUTOHSCROLL;
				break;
			
			case RichTextBoxScrollBars.FORCED_VERTICAL:
				st |= ES_DISABLENOSCROLL;
			case RichTextBoxScrollBars.VERTICAL:
				st |= WS_VSCROLL | ES_AUTOVSCROLL;
				break;
			
			case RichTextBoxScrollBars.NONE:
				break;
		}
		
		_style(st);
	}
	
	
	final RichTextBoxScrollBars scrollBars() // getter
	{
		LONG wl = _style();
		
		if(wl & WS_HSCROLL)
		{
			if(wl & WS_VSCROLL)
			{
				if(wl & ES_DISABLENOSCROLL)
					return RichTextBoxScrollBars.FORCED_BOTH;
				return RichTextBoxScrollBars.BOTH;
			}
			
			if(wl & ES_DISABLENOSCROLL)
				return RichTextBoxScrollBars.FORCED_HORIZONTAL;
			return RichTextBoxScrollBars.HORIZONTAL;
		}
		
		if(wl & WS_VSCROLL)
		{
			if(wl & ES_DISABLENOSCROLL)
				return RichTextBoxScrollBars.FORCED_VERTICAL;
			return RichTextBoxScrollBars.VERTICAL;
		}
		
		return RichTextBoxScrollBars.NONE;
	}
	
	
	private void _getFormat(CHARFORMAT2A* cf, BOOL selection = TRUE)
	in
	{
		assert(created);
	}
	body
	{
		//SendMessageA(handle, EM_GETCHARFORMAT, selection, cast(LPARAM)cf);
		CallWindowProcA(richtextboxPrevWndProc, hwnd, EM_GETCHARFORMAT, selection, cast(LPARAM)cf);
	}
	
	
	private void _setFormat(CHARFORMAT2A* cf, WPARAM scf = SCF_SELECTION)
	in
	{
		assert(created);
	}
	body
	{
		//if(!SendMessageA(handle, EM_SETCHARFORMAT, scf, cast(LPARAM)cf))
		if(!CallWindowProcA(richtextboxPrevWndProc, hwnd, EM_SETCHARFORMAT, scf, cast(LPARAM)cf))
			throw new DflException("Unable to set text formatting.");
	}
	
	
	private struct _StreamStr
	{
		char[] str;
	}
	
	
	// Note: RTF should only be ASCII so no conversions are necessary.
	// TODO: verify this; I'm not certain.
	
	private void _streamIn(UINT fmt, char[] str)
	in
	{
		assert(created);
	}
	body
	{
		_StreamStr si;
		EDITSTREAM es;
		
		si.str = str;
		es.dwCookie = cast(DWORD)&si;
		es.pfnCallback = &_streamingInStr;
		
		//if(SendMessageA(handle, EM_STREAMIN, cast(WPARAM)fmt, cast(LPARAM)&es) != str.length)
		//	throw new DflException("Unable to set RTF.");
		
		SendMessageA(handle, EM_STREAMIN, cast(WPARAM)fmt, cast(LPARAM)&es);
	}
	
	
	private char[] _streamOut(UINT fmt)
	in
	{
		assert(created);
	}
	body
	{
		_StreamStr so;
		EDITSTREAM es;
		
		so.str = null;
		es.dwCookie = cast(DWORD)&so;
		es.pfnCallback = &_streamingOutStr;
		
		SendMessageA(handle, EM_STREAMOUT, cast(WPARAM)fmt, cast(LPARAM)&es);
		return so.str;
	}
	
	
	final void selectedRtf(char[] rtf) // setter
	{
		_streamIn(SF_RTF | SFF_SELECTION, rtf);
	}
	
	
	final char[] selectedRtf() // getter
	{
		return _streamOut(SF_RTF | SFF_SELECTION);
	}
	
	
	final void rtf(char[] newRtf) // setter
	{
		_streamIn(SF_RTF, rtf);
	}
	
	
	final void detectUrls(bool byes) // setter
	{
		autoUrl = byes;
		
		if(created)
		{
			SendMessageA(handle, EM_AUTOURLDETECT, byes, 0);
		}
	}
	
	
	final bool detectUrls() // getter
	{
		return autoUrl;
	}
	
	
	final char[] rtf() // getter
	{
		return _streamOut(SF_RTF);
	}
	
	
	/+
	override void createHandle()
	{
		if(isHandleCreated)
			return;
		
		createClassHandle(RICHTEXTBOX_CLASSNAME);
		
		onHandleCreated(EventArgs.empty);
	}
	+/
	
	
	override void createHandle()
	{
		char[] txt;
		txt = wtext;
		
		super.createHandle();
		
		dfl.utf.setWindowText(hwnd, txt);
	}
	
	
	protected override void createParams(inout CreateParams cp)
	{
		super.createParams(cp);
		
		cp.className = RICHTEXTBOX_CLASSNAME;
		cp.caption = null; // Set in createHandle() to allow larger buffers.
	}
	
	
	LinkClickedEventHandler linkClicked;
	
	
	protected:
	void onLinkClicked(LinkClickedEventArgs ea)
	{
		linkClicked(this, ea);
	}
	
	
	private char[] _getRange(LONG min, LONG max)
	/+ in
	{
		assert(created);
		assert(max >= 0);
		assert(min >= max);
	}
	body +/
	{
		TEXTRANGEA tr;
		char[] s;
		
		tr.chrg.cpMin = min;
		tr.chrg.cpMax = max;
		s = new char[max - min + 1];
		tr.lpstrText = s;
		
		return s[0 .. SendMessageA(handle, EM_GETTEXTRANGE, 0, cast(LPARAM)&tr)];
	}
	
	
	/+package+/ override LRESULT customMsg(inout CustomMsg msg) // package
	{
		LRESULT result;
		result = super.customMsg(msg);
		
		switch(msg.message)
		{
			case WM_NOTIFY:
				{
					NMHDR* nmh;
					nmh = cast(NMHDR*)msg.lParam;
					
					assert(nmh.hwndFrom == handle);
					
					switch(nmh.code)
					{
						case EN_LINK:
							{
								ENLINK* enl;
								enl = cast(ENLINK*)nmh;
								
								if(enl.msg == WM_LBUTTONUP)
								{
									if(!selectionLength)
										onLinkClicked(new LinkClickedEventArgs(_getRange(enl.chrg.cpMin, enl.chrg.cpMax)));
								}
							}
							break;
							
						default: ;
					}
				}
				break;
			
			default: ;
		}
		
		return result;
	}
	
	
	override void onHandleCreated(EventArgs ea)
	{
		super.onHandleCreated(ea);
		
		SendMessageA(handle, EM_AUTOURLDETECT, autoUrl, 0);
		
		// DMD 0.102: need 'this' to access member X
		this.backColor = this.backColor;
		
		//Application.doEvents(); // foreColor won't work otherwise.. seems to work now
		this.foreColor = this.foreColor;
		
		SendMessageA(handle, EM_SETEVENTMASK, 0, ENM_CHANGE | ENM_CHANGE | ENM_LINK | ENM_PROTECTED);
	}
	
	
	override void prevWndProc(inout Message m)
	{
		m.result = CallWindowProcA(richtextboxPrevWndProc, m.hWnd, m.msg, m.wParam, m.lParam);
	}
	
	
	private:
	bit autoUrl = true;
}


private extern(Windows) DWORD _streamingInStr(DWORD dwCookie, LPBYTE pbBuff, LONG cb, LONG* pcb)
{
	RichTextBox._StreamStr* si;
	si = cast(typeof(si))dwCookie;
	
	if(!si.str.length)
	{
		*pcb = 0;
		return 1; // ?
	}
	else if(cb >= si.str.length)
	{
		pbBuff[0 .. si.str.length] = cast(BYTE[])si.str;
		*pcb = si.str.length;
		si.str = null;
	}
	else
	{
		pbBuff[0 .. cb] = cast(BYTE[])si.str[0 .. cb];
		*pcb = cb;
		si.str = si.str[cb .. si.str.length];
	}
	
	return 0;
}


private extern(Windows) DWORD _streamingOutStr(DWORD dwCookie, LPBYTE pbBuff, LONG cb, LONG* pcb)
{
	RichTextBox._StreamStr* so;
	so = cast(typeof(so))dwCookie;
	
	so.str ~= cast(char[])pbBuff[0 .. cb];
	*pcb = cb;
	
	return 0;
}

