// Written by Christopher E. Miller
// See the included license.txt for copyright and license details.


module dfl.tooltip;


private import std.string, std.outofmemory;
private import std.c.stdlib;

private import dfl.control, dfl.base, dfl.application, dfl.winapi;


class ToolTip // docmain
{
	package this(DWORD style)
	{
		_initCommonControls(ICC_TREEVIEW_CLASSES); // Includes tooltip.
		
		hwtt = CreateWindowExA(WS_EX_TOPMOST | WS_EX_TOOLWINDOW, _TOOLTIPS_CLASSA,
			"", style, 0, 0, 50, 50, null, null, null, null);
		if(!hwtt)
			throw new DflException("Unable to create tooltip.");
	}
	
	
	this()
	{
		this(cast(DWORD)WS_POPUP);
	}
	
	
	~this()
	{
		removeAll(); // Fixes ref count.
		DestroyWindow(hwtt);
	}
	
	
	final HWND handle() // getter
	{
		return hwtt;
	}
	
	
	final void active(bool byes) // setter
	{
		SendMessageA(hwtt, TTM_ACTIVATE, byes, 0); // ?
		_active = byes;
	}
	
	
	final bool active() // getter
	{
		return _active;
	}
	
	
	// Sets autoPopDelay, initialDelay and reshowDelay.
	final void automaticDelay(DWORD ms) // setter
	{
		SendMessageA(hwtt, TTM_SETDELAYTIME, TTDT_AUTOMATIC, ms);
	}
	
	
	/+
	final DWORD automaticDelay() // getter
	{
	}
	+/
	
	
	final void autoPopDelay(DWORD ms) // setter
	{
		SendMessageA(hwtt, TTM_SETDELAYTIME, TTDT_AUTOPOP, ms);
	}
	
	
	/+
	final DWORD autoPopDelay() // getter
	{
	}
	+/
	
	
	final void initialDelay(DWORD ms) // setter
	{
		SendMessageA(hwtt, TTM_SETDELAYTIME, TTDT_INITIAL, ms);
	}
	
	
	/+
	final DWORD initialDelay() // getter
	{
	}
	+/
	
	
	final void reshowDelay(DWORD ms) // setter
	{
		SendMessageA(hwtt, TTM_SETDELAYTIME, TTDT_RESHOW, ms);
	}
	
	
	/+
	final DWORD reshowDelay() // getter
	{
	}
	+/
	
	
	final void showAlways(bool byes) // setter
	{
		LONG wl;
		wl = GetWindowLongA(hwtt, GWL_STYLE);
		if(byes)
		{
			if(wl & TTS_ALWAYSTIP)
				return;
			wl |= TTS_ALWAYSTIP;
		}
		else
		{
			if(!(wl & TTS_ALWAYSTIP))
				return;
			wl &= ~TTS_ALWAYSTIP;
		}
		SetWindowLongA(hwtt, GWL_STYLE, wl);
	}
	
	
	final bool showAlways() // getter
	{
		return (GetWindowLongA(hwtt, GWL_STYLE) & TTS_ALWAYSTIP) != 0;
	}
	
	
	// WARNING: possible buffer overflow.
	final char[] getToolTip(Control ctrl)
	{
		char[] result;
		TOOLINFOA tool;
		tool.cbSize = TOOLINFOA.sizeof;
		tool.uFlags = TTF_IDISHWND;
		tool.hwnd = ctrl.handle;
		tool.uId = cast(UINT)ctrl.handle;
		//tool.lpszText = cast(char*)alloca(MAX_TIP_TEXT_LENGTH + 1); // Overwriting the stack is the worst.
		tool.lpszText = cast(char*)malloc(MAX_TIP_TEXT_LENGTH + 1);
		if(!tool.lpszText)
			throw new OutOfMemoryException;
		try
		{
			tool.lpszText[0] = 0;
			SendMessageA(hwtt, TTM_GETTEXTA, 0, cast(LPARAM)&tool);
			if(!tool.lpszText[0])
				result = null;
			else
				result = std.string.toString(tool.lpszText).dup;
		}
		finally
		{
			free(tool.lpszText);
		}
		return result;
	}
	
	
	// Remove all tooltip text associated with this instance.
	final void removeAll()
	{
		TOOLINFOA tool;
		tool.cbSize = TOOLINFOA.sizeof;
		while(SendMessageA(hwtt, TTM_ENUMTOOLSA, 0, cast(LPARAM)&tool))
		{
			SendMessageA(hwtt, TTM_DELTOOLA, 0, cast(LPARAM)&tool);
			Application.refCountDec(cast(void*)this);
		}
	}
	
	
	final void setToolTip(Control ctrl, char[] text)
	in
	{
		try
		{
			ctrl.createControl();
		}
		catch(Object o)
		{
			assert(0); // If -ctrl- is a child, make sure the parent is set before setting tool tip text.
			//throw o;
		}
	}
	body
	{
		TOOLINFOA tool;
		tool.cbSize = TOOLINFOA.sizeof;
		tool.uFlags = TTF_IDISHWND;
		tool.hwnd = ctrl.handle;
		tool.uId = cast(UINT)ctrl.handle;
		
		if(!text.length)
		{
			if(SendMessageA(hwtt, TTM_GETTOOLINFOA, 0, cast(LPARAM)&tool))
			{
				// Remove.
				
				SendMessageA(hwtt, TTM_DELTOOLA, 0, cast(LPARAM)&tool);
				
				Application.refCountDec(cast(void*)this);
			}
			return;
		}
		
		// Hack to help prevent getToolTip() overflow.
		if(text.length > MAX_TIP_TEXT_LENGTH)
			text = text[0 .. MAX_TIP_TEXT_LENGTH];
		
		if(SendMessageA(hwtt, TTM_GETTOOLINFOA, 0, cast(LPARAM)&tool))
		{
			// Update.
			
			tool.lpszText = toStringz(text);
			SendMessageA(hwtt, TTM_UPDATETIPTEXTA, 0, cast(LPARAM)&tool);
		}
		else
		{
			// Add.
			
			/+
			// TOOLINFOA.rect is ignored if TTF_IDISHWND.
			tool.rect.left = 0;
			tool.rect.top = 0;
			tool.rect.right = ctrl.clientSize.width;
			tool.rect.bottom = ctrl.clientSize.height;
			+/
			tool.lpszText = toStringz(text);
			tool.uFlags |= TTF_SUBCLASS; // Not a good idea ?
			//if(!SendMessageA(hwtt, TTM_ADDTOOLA, 0, cast(LPARAM)&tool))
			//	throw new DflException("Unable to set tooltip text.");
			SendMessageA(hwtt, TTM_ADDTOOLA, 0, cast(LPARAM)&tool);
			
			Application.refCountInc(cast(void*)this);
		}
	}
	
	
	private:
	const char[] _TOOLTIPS_CLASSA = "tooltips_class32";
	const size_t MAX_TIP_TEXT_LENGTH = 2045;
	
	HWND hwtt; // Tooltip control handle.
	bool _active = true;
}

