// Written by Christopher E. Miller
// See the included license.txt for copyright and license details.


module dfl.treeview;

private import std.string;

private import dfl.control, dfl.application, dfl.base, dfl.winapi;
private import dfl.event, dfl.drawing, dfl.collections;


private extern(Windows) void _initTreeview();


enum TreeViewAction: ubyte
{
	UNKNOWN,
	COLLAPSE,
	EXPAND,
	BY_KEYBOARD,
	BY_MOUSE,
}


class TreeViewCancelEventArgs: CancelEventArgs
{
	this(TreeNode node, bool cancel, TreeViewAction action)
	{
		super(cancel);
		
		_node = node;
		_action = action;
	}
	
	
	final TreeViewAction action() // getter
	{
		return _action;
	}
	
	
	final TreeNode node() // getter
	{
		return _node;
	}
	
	
	private:
	TreeNode _node;
	TreeViewAction _action;
}


alias Event!(TreeViewCancelEventArgs) TreeViewCancelEventHandler;


class TreeViewEventArgs: EventArgs
{
	this(TreeNode node, TreeViewAction action)
	{
		_node = node;
		_action = action;
	}
	
	
	this(TreeNode node)
	{
		_node = node;
		//_action = TreeViewAction.UNKNOWN;
	}
	
	
	final TreeViewAction action() // getter
	{
		return _action;
	}
	
	
	final TreeNode node() // getter
	{
		return _node;
	}
	
	
	private:
	TreeNode _node;
	TreeViewAction _action = TreeViewAction.UNKNOWN;
}


alias Event!(TreeViewEventArgs) TreeViewEventHandler;


class NodeLabelEditEventArgs: EventArgs
{
	this(TreeNode node, char[] label)
	{
		_node = node;
		_label = label;
	}
	
	
	this(TreeNode node)
	{
		_node = node;
	}
	
	
	final TreeNode node() // getter
	{
		return _node;
	}
	
	
	final char[] label() // getter
	{
		return _label;
	}
	
	
	final void cancelEdit(bool byes) // setter
	{
		_cancel = byes;
	}
	
	
	final bool cancelEdit() // getter
	{
		return _cancel;
	}
	
	
	private:
	TreeNode _node;
	char[] _label;
	bool _cancel = false;
}


alias Event!(NodeLabelEditEventArgs) NodeLabelEditEventHandler;


class TreeNode
{
	this(char[] labelText)
	{
		this();
		
		ttext = labelText;
	}
	
	
	this(char[] labelText, TreeNode[] children)
	{
		this();
		
		ttext = labelText;
		tchildren.addRange(children);
	}
	
	
	this()
	{
		Application.ppin(cast(void*)this);
		
		/+
		bcolor = Color.empty;
		fcolor = Color.empty;
		+/
		
		tchildren = new TreeNodeCollection(tview, this);
	}
	
	
	/+
	final void backColor(Color c) // setter
	{
		bcolor = c;
	}
	
	
	final Color backColor() // getter
	{
		return bcolor;
	}
	+/
	
	
	final Rect bounds() // getter
	{
		Rect result;
		
		if(created)
		{
			RECT rect;
			*(cast(HTREEITEM*)&rect) = hnode;
			if(SendMessageA(tview.handle, TVM_GETITEMRECT, FALSE, cast(LPARAM)&rect))
			{
				result = Rect(&rect);
			}
		}
		
		return result;
	}
	
	
	final TreeNode firstNode() // getter
	{
		if(tchildren.count)
			return tchildren._nodes[0];
		return null;
	}
	
	
	/+
	final void foreColor(Color c) // setter
	{
		fcolor = c;
	}
	
	
	final Color foreColor() // getter
	{
		return fcolor;
	}
	+/
	
	
	// Path from the root to this node.
	final char[] fullPath() // getter
	{
		if(!tparent)
			return ttext;
		
		// Might want to manually loop through parents and preallocate the whole buffer.
		assert(tview !is null);
		char[1] pathsep;
		pathsep[0] = tview.pathSeparator;
		return tparent.fullPath ~ pathsep ~ ttext;
	}
	
	
	final HTREEITEM handle() // getter
	{
		return hnode;
	}
	
	
	// Index of this node in the parent node.
	final int index() // getter
	{
		int result = -1;
		if(tparent)
		{
			result = tparent.tchildren.indexOf(this);
			assert(result != -1);
		}
		return result;
	}
	
	
	/+
	final bool isEditing() // getter
	{
	}
	+/
	
	
	final bool isExpanded() // getter
	{
		return isState(TVIS_EXPANDED);
	}
	
	
	final bool isSelected() // getter
	{
		return isState(TVIS_SELECTED);
	}
	
	
	/+
	final bool isVisible() // getter
	{
	}
	+/
	
	
	final TreeNode lastNode() // getter
	{
		if(tchildren.count)
			return tchildren._nodes[tchildren.count - 1];
		return null;
	}
	
	
	// Next sibling node.
	final TreeNode nextNode() // getter
	{
		if(tparent)
		{
			int i;
			i = tparent.tchildren.indexOf(this);
			assert(i != -1);
			
			i++;
			if(i != tparent.tchildren.count)
				return tparent.tchildren._nodes[i];
		}
		return null;
	}
	
	
	/+
	final void nodeFont(Font f) // setter
	{
		tfont = f;
	}
	
	
	final Font nodeFont() // getter
	{
		return tfont;
	}
	+/
	
	
	final TreeNodeCollection nodes() // getter
	{
		return tchildren;
	}
	
	
	final TreeNode parent() // getter
	{
		return tparent;
	}
	
	
	// Previous sibling node.
	final TreeNode prevNode() // getter
	{
		if(tparent)
		{
			int i;
			i = tparent.tchildren.indexOf(this);
			assert(i != -1);
			
			if(i)
			{
				i--;
				return tparent.tchildren._nodes[i];
			}
		}
		return null;
	}
	
	
	final void tag(Object o) // setter
	{
		ttag = o;
	}
	
	
	final Object tag() // getter
	{
		return ttag;
	}
	
	
	final void text(char[] newText) // setter
	{
		ttext = newText;
		
		if(created)
		{
			TV_ITEMA item;
			Message m;
			
			item.mask = TVIF_HANDLE | TVIF_TEXT;
			item.hItem = hnode;
			item.pszText = toStringz(ttext);
			//item.cchTextMax = ttext.length; // ?
			
			m = Message(tview.handle, TVM_SETITEMA, 0, cast(LPARAM)&item);
			tview.prevWndProc(m);
		}
	}
	
	
	final char[] text() // getter
	{
		return ttext;
	}
	
	
	// Get the TreeView control this node belongs to.
	final TreeView treeView() // getter
	{
		return tview;
	}
	
	
	final void beginEdit()
	{
		if(created)
		{
			SetFocus(tview.hwnd); // Needs to have focus.
			HWND hwEdit;
			hwEdit = cast(HWND)SendMessageA(tview.hwnd, TVM_EDITLABELA, 0, cast(LPARAM)hnode);
			if(!hwEdit)
				goto err_edit;
		}
		else
		{
			err_edit:
			throw new DflException("Unable to edit TreeNode.");
		}
	}
	
	
	final void collapse()
	{
		if(created)
		{
			SendMessageA(tview.hwnd, TVM_EXPAND, TVE_COLLAPSE, cast(LPARAM)hnode);
		}
	}
	
	
	/+
	final void endEdit(bool cancel)
	{
		// ?
	}
	+/
	
	
	final void ensureVisible()
	{
		if(created)
		{
			SendMessageA(tview.hwnd, TVM_ENSUREVISIBLE, 0, cast(LPARAM)hnode);
		}
	}
	
	
	final void expand()
	{
		if(created)
		{
			SendMessageA(tview.hwnd, TVM_EXPAND, TVE_EXPAND, cast(LPARAM)hnode);
		}
	}
	
	
	final void expandAll()
	{
		if(created)
		{
			SendMessageA(tview.hwnd, TVM_EXPAND, TVE_EXPAND, cast(LPARAM)hnode);
			
			foreach(TreeNode node; tchildren._nodes)
			{
				node.expandAll();
			}
		}
	}
	
	
	static TreeNode fromHandle(TreeView tree, HTREEITEM handle)
	{
		return tree.treeNodeFromHandle(handle);
	}
	
	
	final void remove()
	{
		if(tparent)
			tparent.tchildren.remove(this);
		else if(tview) // It's a top level node.
			tview.tchildren.remove(this);
	}
	
	
	final void toggle()
	{
		if(created)
		{
			SendMessageA(tview.hwnd, TVM_EXPAND, TVE_TOGGLE, cast(LPARAM)hnode);
		}
	}
	
	
	override char[] toString()
	{
		return ttext;
	}
	
	
	override int opEquals(Object o)
	{
		return ttext == o.toString(); // ?
	}
	
	
	int opEquals(TreeNode node)
	{
		return ttext == node.ttext;
	}
	
	
	override int opCmp(Object o)
	{
		return std.string.icmp(ttext, o.toString()); // ?
	}
	
	
	int opCmp(TreeNode node)
	{
		return std.string.icmp(ttext, node.ttext);
	}
	
	
	private:
	char[] ttext;
	TreeNode tparent;
	TreeNodeCollection tchildren;
	Object ttag;
	HTREEITEM hnode;
	TreeView tview;
	/+
	Color bcolor, fcolor;
	Font tfont;
	+/
	
	
	package final bool created() // getter
	{
		if(tview && tview.created())
		{
			assert(hnode);
			return true;
		}
		return false;
	}
	
	
	bool isState(UINT state)
	{
		if(created)
		{
			TV_ITEMA ti;
			ti.mask = TVIF_HANDLE | TVIF_STATE;
			ti.hItem = hnode;
			ti.stateMask = state;
			if(SendMessageA(tview.handle, TVM_GETITEMA, 0, cast(LPARAM)&ti))
			{
				if(ti.state & state)
					return true;
			}
		}
		return false;
	}
	
	
	void _reset()
	{
		hnode = null;
		tview = null;
		tparent = null;
	}
}


// Will implement IList when interfaces allow covariant returns.
class TreeNodeCollection //: IList
{
	bool isFixedSize() // getter
	{
		return false;
	}
	
	
	bool isReadOnly() // getter
	{
		return false;
	}
	
	
	bool isSynchronized() // getter
	{
		return false;
	}
	
	
	Object syncRoot() // getter
	{
		return this;
	}
	
	
	// Shallow copy.
	Array clone()
	{
		Array result;
		result = new Array(_nodes.length);
		Array.copy(_nodes, result, _nodes.length);
		return result;
	}
	
	
	// D-style.
	final Array dup() // getter
	{
		return clone();
	}
	
	
	void copyTo(Array dest, int destIndex)
	{
		Array.copy(_nodes, 0, dest, destIndex, _nodes.length);
	}
	
	
	void add(Object o)
	{
		TreeNode node;
		node = cast(TreeNode)o;
		if(node)
		{
			return add(node);
		}
		else
		{
			return add(o.toString()); // ?
		}
	}
	
	
	private void verifyNoParent(TreeNode node)
	{
		if(node.tparent)
			throw new DflException("TreeNode already belongs to a TreeView.");
	}
	
	
	void add(TreeNode node)
	{
		//printf("Adding node %p '%.*s'\n", cast(void*)node, node.toString());
		
		verifyNoParent(node);
		
		int i;
		bool atEnd = false;
		
		if(tview && tview.sorted())
		{
			// Insertion sort.
			
			for(i = 0; i != _nodes.length; i++)
			{
				if(node < _nodes[i])
					break;
			}
			
			if(i == _nodes.length)
			{
				atEnd = true;
				_nodes ~= node;
			}
			else
			{
				_nodes = _nodes[0 .. i] ~ (&node)[0 .. 1] ~ _nodes[i .. _nodes.length];
			}
		}
		else
		{
			atEnd = true;
			i = _nodes.length;
			_nodes ~= node;
		}
		
		node.tparent = tparent;
		node.tview = tview;
		node.tchildren.setTreeView(tview);
		
		if(created)
		{
			TV_INSERTSTRUCTA tis;
			tis.hParent = tparent ? tparent.handle : TVI_ROOT;
			if(!i)
				tis.hInsertAfter = TVI_FIRST;
			//else if(i == _nodes.length - 1)
			else if(atEnd)
				tis.hInsertAfter = TVI_LAST;
			else
				tis.hInsertAfter = _nodes[i - 1].handle;
			assert(tis.hInsertAfter);
			populateChildNode(tis.item, node);
			
			Message m;
			m = Message(tview.handle, TVM_INSERTITEMA, 0, cast(LPARAM)&tis);
			tview.prevWndProc(m);
			assert(m.result);
			node.hnode = cast(HTREEITEM)m.result;
			
			node.tchildren.doNodes();
			
			if(tparent)
				tview.invalidate(tparent.bounds);
		}
		
		//return i;
	}
	
	
	void add(char[] text)
	{
		return add(new TreeNode(text));
	}
	
	
	void addRange(Object[] range)
	{
		foreach(Object o; range)
		{
			add(o);
		}
	}
	
	
	void addRange(TreeNode[] range)
	{
		foreach(TreeNode node; range)
		{
			add(node);
		}
	}
	
	
	void addRange(char[][] range)
	{
		foreach(char[] s; range)
		{
			add(s);
		}
	}
	
	
	void clear()
	{
		TreeNode[] nodes = _nodes;
		_nodes = null;
		if(created)
		{
			Message m;
			m.hWnd = tview.handle;
			m.msg = TVM_DELETEITEM;
			m.wParam = 0;
			foreach(TreeNode node; nodes)
			{
				assert(node.handle !is null);
				m.lParam = cast(LPARAM)node.handle;
				tview.prevWndProc(m);
				
				node._reset();
			}
		}
	}
	
	
	// Like clear but doesn't bother removing stuff from the lists.
	// Used when a parent is being removed and the children only
	// need to be reset.
	private void _reset()
	{
		foreach(TreeNode node; _nodes)
		{
			node._reset();
		}
	}
	
	
	// Clear node handles when the TreeView window is destroyed so
	// that it can be reconstructed.
	private void _resetHandles()
	{
		foreach(TreeNode node; _nodes)
		{
			node.tchildren._resetHandles();
			node.hnode = null;
		}
	}
	
	
	bool contains(Object value)
	{
		return indexOf(value) != -1;
	}
	
	
	bool contains(TreeNode value)
	{
		return indexOf(value) != -1;
	}
	
	
	int indexOf(Object value)
	{
		TreeNode node;
		node = cast(TreeNode)value;
		if(node)
		{
			return indexOf(node);
		}
		else
		{
			foreach(int i, TreeNode node; _nodes)
			{
				if(value == node) // ?
					return i;
			}
			return -1;
		}
	}
	
	
	int indexOf(TreeNode value)
	{
		int result = -1;
		foreach(int i, TreeNode node; _nodes)
		{
			//if(value == node)
			{
				if(value is node)
				{
					result = i;
					break;
				}
				
				//if(result == -1)
				//	result = i;
			}
		}
		return result;
	}
	
	
	void insert(int index, Object value)
	{
		TreeNode node;
		node = cast(TreeNode)value;
		if(node)
		{
			insert(index, node);
		}
		else
		{
			insert(index, value.toString()); // ?
		}
	}
	
	
	void insert(int index, char[] value)
	{
		insert(index, new TreeNode(value));
	}
	
	
	void insert(int i, TreeNode node)
	{
		verifyNoParent(node);
		
		node.tparent = tparent;
		node.tview = tview;
		node.tchildren.setTreeView(tview);
		
		TV_INSERTSTRUCTA tis;
		
		if(i <= 0)
		{
			tis.hInsertAfter = TVI_FIRST;
			_nodes = (&node)[0 .. 1].dup ~ _nodes; // DMD 0.124: .dup needed for some reason.
		}
		else if(i >= cast(int)_nodes.length)
		{
			tis.hInsertAfter = TVI_LAST;
			_nodes ~= node;
		}
		else
		{
			tis.hInsertAfter = _nodes[i - 1].handle;
			_nodes = _nodes[0 .. i] ~ (&node)[0 .. 1] ~ _nodes[i .. _nodes.length];
		}
		
		if(created)
		{
			tis.hParent = tparent ? tparent.handle : TVI_ROOT;
			assert(tis.hInsertAfter);
			populateChildNode(tis.item, node);
			
			Message m;
			m = Message(tview.handle, TVM_INSERTITEMA, 0, cast(LPARAM)&tis);
			tview.prevWndProc(m);
			assert(m.result);
			node.hnode = cast(HTREEITEM)m.result;
			
			node.tchildren.doNodes();
			
			if(tparent)
				tview.invalidate(tparent.bounds);
		}
		
		//return i;
	}
	
	
	void remove(Object value)
	{
		int i;
		i = indexOf(value);
		if(i != -1)
			removeAt(i);
	}
	
	
	void remove(TreeNode node)
	{
		int i;
		i = indexOf(node);
		if(i != -1)
			removeAt(i);
	}
	
	
	void removeAt(int index)
	{
		TreeNode node;
		node = _nodes[index];
		
		if(!index)
			_nodes = _nodes[1 .. _nodes.length];
		else if(index == _nodes.length - 1)
			_nodes = _nodes[0 .. index];
		else if(index > 0 && index < cast(int)_nodes.length)
			_nodes = _nodes[0 .. index] ~ _nodes[index + 1 .. _nodes.length];
		
		if(created)
		{
			assert(node.hnode);
			Message m;
			m = Message(tview.handle, TVM_DELETEITEM, 0, cast(LPARAM)node.hnode);
			tview.prevWndProc(m);
		}
		
		// Clear children.
		node._reset();
	}
	
	
	int count() // getter
	{
		return _nodes.length;
	}
	
	
	void opIndexAssign(Object value, int index) // setter
	{
		//array[index] = value;
		
		removeAt(index);
		//add(value);
		insert(index, value);
	}
	
	
	// Covariant return with interface IList.
	TreeNode opIndex(int index) // getter
	{
		return _nodes[index];
	}
	
	
	IEnumerator getEnumerator()
	{
		return new NodeEnum(_nodes);
	}
	
	
	private:
	TreeView tview; // null if not assigned to a TreeView yet.
	TreeNode tparent; // null if root. The parent of -_nodes-.
	TreeNode[] _nodes;
	
	
	static class NodeEnum: IEnumerator
	{
		mixin EnumeratorForeach;
		//mixin EnumeratorForeach!(TreeNode);
		
		
		this(TreeNode[] nodes)
		{
			this.nodes = nodes;
		}
		
		
		// Covariant return with IEnumerator.
		//TreeNode current() // getter
		Object current() // getter
		{
			return nodes[idx];
		}
		
		
		bool moveNext()
		{
			idx++;
			if(idx >= nodes.length) // Note: signed-unsigned comparison.
			{
				idx = nodes.length; // Prevent overflow.
				return false;
			}
			return true;
		}
		
		
		void reset()
		{
			idx = -1;
		}
		
		
		private:
		TreeNode[] nodes;
		int idx = -1;
	}
	
	
	package this(TreeView treeView, TreeNode parentNode)
	{
		tview = treeView;
		tparent = parentNode;
	}
	
	
	package final void setTreeView(TreeView treeView)
	{
		tview = treeView;
		foreach(TreeNode node; _nodes)
		{
			node.tchildren.setTreeView(treeView);
		}
	}
	
	
	package final bool created() // getter
	{
		return tview && tview.created();
	}
	
	
	void doNodes()
	in
	{
		assert(created);
	}
	body
	{
		TV_INSERTSTRUCTA tis;
		Message m;
		
		tis.hInsertAfter = TVI_LAST;
		
		m.hWnd = tview.handle;
		m.msg = TVM_INSERTITEMA;
		m.wParam = 0;
		
		foreach(TreeNode node; _nodes)
		{
			assert(!node.handle);
			
			tis.hParent = tparent ? tparent.handle : TVI_ROOT;
			populateChildNode(tis.item, node);
			
			m.lParam = cast(LPARAM)&tis;
			tview.prevWndProc(m);
			assert(m.result);
			node.hnode = cast(HTREEITEM)m.result;
			
			node.tchildren.doNodes();
		}
	}
	
	
	package void populateChildNode(inout TV_ITEMA dest, TreeNode node)
	{
		with(dest)
		{
			mask = /+ TVIF_CHILDREN | +/ TVIF_PARAM | TVIF_TEXT;
			/+ cChildren = I_CHILDRENCALLBACK; +/
			lParam = cast(LPARAM)cast(void*)node;
			pszText = toStringz(node.text);
			//cchTextMax = node.text.length; // ?
		}
	}
}


class TreeView: ControlSuperClass // docmain
{
	this()
	{
		_initTreeview();
		
		backc = defaultBackColor;
		forec = defaultForeColor;
		
		wstyle |= WS_TABSTOP | TVS_HASBUTTONS | TVS_LINESATROOT | TVS_HASLINES;
		wexstyle |= WS_EX_CLIENTEDGE;
		ctrlStyle |= ControlStyles.SELECTABLE;
		wclassStyle = treeviewClassStyle;
		
		tchildren = new TreeNodeCollection(this, null);
	}
	
	
	/+
	~this()
	{
		/+
		if(tchildren)
			tchildren._dtorReset();
		+/
	}
	+/
	
	
	static Color defaultBackColor() // getter
	{
		return SystemColors.window;
	}
	
	
	static Color defaultForeColor() //getter
	{
		return SystemColors.windowText;
	}
	
	
	final void borderStyle(BorderStyle bs) // setter
	{
		switch(bs)
		{
			case BorderStyle.FIXED_3D:
				_style(_style() & ~WS_BORDER);
				_exStyle(_exStyle() | WS_EX_CLIENTEDGE);
				break;
				
			case BorderStyle.FIXED_SINGLE:
				_exStyle(_exStyle() & ~WS_EX_CLIENTEDGE);
				_style(_style() | WS_BORDER);
				break;
				
			case BorderStyle.NONE:
				_style(_style() & ~WS_BORDER);
				_exStyle(_exStyle() & ~WS_EX_CLIENTEDGE);
				break;
		}
		
		if(created)
		{
			redrawEntire();
		}
	}
	
	
	final BorderStyle borderStyle() // getter
	{
		if(_exStyle() & WS_EX_CLIENTEDGE)
			return BorderStyle.FIXED_3D;
		else if(_style() & WS_BORDER)
			return BorderStyle.FIXED_SINGLE;
		return BorderStyle.NONE;
	}
	
	
	/+
	// Cannot be set after the control has been created.
	final void checkBoxes(bool byes) // setter
	{
		if(byes)
			_style(_style() | TVS_CHECKBOXES);
		else
			_style(_style() & ~TVS_CHECKBOXES);
	}
	
	
	final bool checkBoxes() // getter
	{
		return (_style() & TVS_CHECKBOXES) != 0;
	}
	+/
	
	
	final void fullRowSelect(bool byes) // setter
	{
		if(byes)
			_style(_style() | TVS_FULLROWSELECT);
		else
			_style(_style() & ~TVS_FULLROWSELECT);
	}
	
	
	final bool fullRowSelect() // getter
	{
		return (_style() & TVS_FULLROWSELECT) != 0;
	}
	
	
	final void hideSelection(bool byes) // setter
	{
		if(byes)
			_style(_style() & ~TVS_SHOWSELALWAYS);
		else
			_style(_style() | TVS_SHOWSELALWAYS);
	}
	
	
	final bool hideSelection() // getter
	{
		return (_style() & TVS_SHOWSELALWAYS) == 0;
	}
	
	
	final void hotTracking(bool byes) // setter
	{
		if(byes)
			_style(_style() | TVS_TRACKSELECT);
		else
			_style(_style() & ~TVS_TRACKSELECT);
	}
	
	
	final bool hotTracking() // getter
	{
		return (_style() & TVS_TRACKSELECT) != 0;
	}
	
	
	final void indent(int newIndent) // setter
	{
		if(newIndent < 0)
			newIndent = 0;
		else if(newIndent > 32_000)
			newIndent = 32_000;
		
		ind = newIndent;
		
		if(created)
			SendMessageA(hwnd, TVM_SETINDENT, ind, 0);
	}
	
	
	final int indent() // getter
	{
		if(created)
			ind = cast(int)SendMessageA(hwnd, TVM_GETINDENT, 0, 0);
		return ind;
	}
	
	
	final void itemHeight(int h) // setter
	{
		if(h < 0)
			h = 0;
		
		iheight = h;
		
		if(created)
			SendMessageA(hwnd, TVM_SETITEMHEIGHT, iheight, 0);
	}
	
	
	final int itemHeight() // getter
	{
		if(created)
			iheight = cast(int)SendMessageA(hwnd, TVM_GETITEMHEIGHT, 0, 0);
		return iheight;
	}
	
	
	final void labelEdit(bool byes) // setter
	{
		if(byes)
			_style(_style() | TVS_EDITLABELS);
		else
			_style(_style() & ~TVS_EDITLABELS);
	}
	
	
	final bool labelEdit() // getter
	{
		return (_style() & TVS_EDITLABELS) != 0;
	}
	
	
	final TreeNodeCollection nodes() // getter
	{
		return tchildren;
	}
	
	
	final void pathSeparator(char sep) // setter
	{
		pathsep = sep;
	}
	
	
	final char pathSeparator() // getter
	{
		return pathsep;
	}
	
	
	final void scrollable(bool byes) // setter
	{
		if(byes)
			_style(_style() & ~TVS_NOSCROLL);
		else
			_style(_style() | TVS_NOSCROLL);
	}
	
	
	final bool scrollable() // getter
	{
		return (_style & TVS_NOSCROLL) == 0;
	}
	
	
	final void selectedNode(TreeNode node) // setter
	{
		if(created)
		{
			if(node)
			{
				SendMessageA(hwnd, TVM_SELECTITEM, TVGN_CARET, cast(LPARAM)node.handle);
			}
			else
			{
				// Should the selection be cleared if -node- is null?
				//SendMessageA(hwnd, TVM_SELECTITEM, TVGN_CARET, cast(LPARAM)null);
			}
		}
	}
	
	
	final TreeNode selectedNode() // getter
	{
		if(created)
		{
			HTREEITEM hnode;
			hnode = cast(HTREEITEM)SendMessageA(hwnd, TVM_GETNEXTITEM, TVGN_CARET, cast(LPARAM)null);
			if(hnode)
				return treeNodeFromHandle(hnode);
		}
		return null;
	}
	
	
	final void showLines(bool byes) // setter
	{
		if(byes)
			_style(_style() | TVS_HASLINES);
		else
			_style(_style() & ~TVS_HASLINES);
	}
	
	
	final bool showLines() // getter
	{
		return (_style() & TVS_HASLINES) != 0;
	}
	
	
	final void showPlusMinus(bool byes) // setter
	{
		if(byes)
			_style(_style() | TVS_HASBUTTONS);
		else
			_style(_style() & ~TVS_HASBUTTONS);
	}
	
	
	final bool showPlusMinus() // getter
	{
		return (_style() & TVS_HASBUTTONS) != 0;
	}
	
	
	// -showPlusMinus- should be false.
	final void singleExpand(bool byes) // setter
	{
		if(byes)
			_style(_style() | TVS_SINGLEEXPAND);
		else
			_style(_style() & ~TVS_SINGLEEXPAND);
	}
	
	
	final bool singleExpand() // getter
	{
		return (_style & TVS_SINGLEEXPAND) != 0;
	}
	
	
	final void showRootLines(bool byes) // setter
	{
		if(byes)
			_style(_style() | TVS_LINESATROOT);
		else
			_style(_style() & ~TVS_LINESATROOT);
	}
	
	
	final bool showRootLines() // getter
	{
		return (_style() & TVS_LINESATROOT) != 0;
	}
	
	
	final void sorted(bool byes) // setter
	{
		_sort = byes;
	}
	
	
	final bool sorted() // getter
	{
		return _sort;
	}
	
	
	// First visible node, based on the scrolled position.
	final TreeNode topNode() // getter
	{
		if(created)
		{
			HTREEITEM hnode;
			hnode = cast(HTREEITEM)SendMessageA(hwnd, TVM_GETNEXTITEM,
				TVGN_FIRSTVISIBLE, cast(LPARAM)null);
			if(hnode)
				return treeNodeFromHandle(hnode);
		}
		return null;
	}
	
	
	// Number of visible nodes, including partially visible.
	final int visibleCount() // getter
	{
		if(!created)
			return 0;
		return cast(int)SendMessageA(hwnd, TVM_GETVISIBLECOUNT, 0, 0);
	}
	
	
	final void beginUpdate()
	{
		SendMessageA(handle, WM_SETREDRAW, false, 0);
	}
	
	
	final void endUpdate()
	{
		SendMessageA(handle, WM_SETREDRAW, true, 0);
		invalidate(true); // Show updates.
	}
	
	
	final void collapseAll()
	{
		if(created)
		{
			void collapsing(TreeNodeCollection tchildren)
			{
				foreach(TreeNode node; tchildren._nodes)
				{
					SendMessageA(hwnd, TVM_EXPAND, TVE_COLLAPSE, cast(LPARAM)node.hnode);
					collapsing(node.tchildren);
				}
			}
			
			
			collapsing(tchildren);
		}
	}
	
	
	final void expandAll()
	{
		if(created)
		{
			void expanding(TreeNodeCollection tchildren)
			{
				foreach(TreeNode node; tchildren._nodes)
				{
					SendMessageA(hwnd, TVM_EXPAND, TVE_EXPAND, cast(LPARAM)node.hnode);
					expanding(node.tchildren);
				}
			}
			
			
			expanding(tchildren);
		}
	}
	
	
	final TreeNode getNodeAt(int x, int y)
	{
		if(created)
		{
			TVHITTESTINFO thi;
			HTREEITEM hti;
			thi.pt.x = x;
			thi.pt.y = y;
			hti = cast(HTREEITEM)SendMessageA(hwnd, TVM_HITTEST, 0, cast(LPARAM)&thi);
			if(hti)
			{
				TreeNode result;
				result = treeNodeFromHandle(hti);
				if(result)
				{
					assert(result.tview is this);
					return result;
				}
			}
		}
		return null;
	}
	
	
	final TreeNode getNodeAt(Point pt)
	{
		return getNodeAt(pt.x, pt.y);
	}
	
	
	/+
	// TODO: finish.
	final int getNodeCount(bool includeSubNodes)
	{
		int result;
		result = tchildren.count();
		
		if(includeSubNodes)
		{
			// ...
		}
		
		return result;
	}
	+/
	
	
	protected override Size defaultSize() // getter
	{
		return Size(120, 100);
	}
	
	
	/+
	override void createHandle()
	{
		if(isHandleCreated)
			return;
		
		createClassHandle(TREEVIEW_CLASSNAME);
		
		onHandleCreated(EventArgs.empty);
	}
	+/
	
	
	protected override void createParams(inout CreateParams cp)
	{
		super.createParams(cp);
		
		cp.className = TREEVIEW_CLASSNAME;
	}
	
	
	protected override void onHandleCreated(EventArgs ea)
	{
		super.onHandleCreated(ea);
		
		SendMessageA(hwnd, TVM_SETINDENT, ind, 0);
		
		SendMessageA(hwnd, TVM_SETITEMHEIGHT, iheight, 0);
		
		tchildren.doNodes();
	}
	
	
	protected override void onHandleDestroyed(EventArgs ea)
	{
		tchildren._resetHandles();
		
		super.onHandleDestroyed(ea);
	}
	
	
	protected override void wndProc(inout Message m)
	{
		// TODO: support these messages.
		switch(m.msg)
		{
			case TVM_INSERTITEMA:
			case TVM_INSERTITEMW:
				m.result = cast(LRESULT)null;
				return;
			
			case TVM_SETITEMA:
			case TVM_SETITEMW:
				m.result = cast(LRESULT)-1;
				return;
			
			case TVM_DELETEITEM:
				m.result = FALSE;
				return;
			
			default: ;
		}
		
		super.wndProc(m);
	}
	
	
	protected override void prevWndProc(inout Message msg)
	{
		msg.result = CallWindowProcA(treeviewPrevWndProc, msg.hWnd, msg.msg, msg.wParam, msg.lParam);
	}
	
	
	TreeViewEventHandler afterCollapse;
	TreeViewEventHandler afterExpand;
	TreeViewEventHandler afterSelect;
	NodeLabelEditEventHandler afterLabelEdit;
	TreeViewCancelEventHandler beforeCollapse;
	TreeViewCancelEventHandler beforeExpand;
	TreeViewCancelEventHandler beforeSelect;
	NodeLabelEditEventHandler beforeLabelEdit;
	
	
	protected void onAfterCollapse(TreeViewEventArgs ea)
	{
		afterCollapse(this, ea);
	}
	
	
	protected void onAfterExpand(TreeViewEventArgs ea)
	{
		afterExpand(this, ea);
	}
	
	
	protected void onAfterSelect(TreeViewEventArgs ea)
	{
		afterSelect(this, ea);
	}
	
	
	protected void onAfterLabelEdit(NodeLabelEditEventArgs ea)
	{
		afterLabelEdit(this, ea);
	}
	
	
	protected void onBeforeCollapse(TreeViewCancelEventArgs ea)
	{
		beforeCollapse(this, ea);
	}
	
	
	protected void onBeforeExpand(TreeViewCancelEventArgs ea)
	{
		beforeExpand(this, ea);
	}
	
	
	protected void onBeforeSelect(TreeViewCancelEventArgs ea)
	{
		beforeSelect(this, ea);
	}
	
	
	protected void onBeforeLabelEdit(NodeLabelEditEventArgs ea)
	{
		beforeLabelEdit(this, ea);
	}
	
	
	/+package+/ LRESULT customMsg(inout CustomMsg msg) // package
	{
		LRESULT result;
		
		switch(msg.message)
		{
			case WM_NOTIFY:
				{
					NMHDR* nmh;
					NM_TREEVIEW* nmtv;
					TreeViewCancelEventArgs cea;
					
					nmh = cast(NMHDR*)msg.lParam;
					assert(nmh.hwndFrom == hwnd);
					
					switch(nmh.code)
					{
						/+
						case TVN_GETDISPINFOA:
							
							break;
						+/
						
						case TVN_SELCHANGINGA:
							nmtv = cast(NM_TREEVIEW*)nmh;
							switch(nmtv.action)
							{
								case TVC_BYMOUSE:
									cea = new TreeViewCancelEventArgs(cast(TreeNode)cast(void*)nmtv.itemNew.lParam,
										false, TreeViewAction.BY_MOUSE);
									onBeforeSelect(cea);
									result = cea.cancel;
									break;
								
								case TVC_BYKEYBOARD:
									cea = new TreeViewCancelEventArgs(cast(TreeNode)cast(void*)nmtv.itemNew.lParam,
										false, TreeViewAction.BY_KEYBOARD);
									onBeforeSelect(cea);
									result = cea.cancel;
									break;
								
								//case TVC_UNKNOWN:
								default:
									cea = new TreeViewCancelEventArgs(cast(TreeNode)cast(void*)nmtv.itemNew.lParam,
										false, TreeViewAction.UNKNOWN);
									onBeforeSelect(cea);
									result = cea.cancel;
							}
							break;
						
						case TVN_SELCHANGEDA:
							nmtv = cast(NM_TREEVIEW*)nmh;
							switch(nmtv.action)
							{
								case TVC_BYMOUSE:
									onAfterSelect(new TreeViewEventArgs(cast(TreeNode)cast(void*)nmtv.itemNew.lParam,
										TreeViewAction.BY_MOUSE));
									break;
								
								case TVC_BYKEYBOARD:
									onAfterSelect(new TreeViewEventArgs(cast(TreeNode)cast(void*)nmtv.itemNew.lParam,
										TreeViewAction.BY_KEYBOARD));
									break;
								
								//case TVC_UNKNOWN:
								default:
									onAfterSelect(new TreeViewEventArgs(cast(TreeNode)cast(void*)nmtv.itemNew.lParam,
										TreeViewAction.UNKNOWN));
							}
							break;
						
						case TVN_ITEMEXPANDINGA:
							nmtv = cast(NM_TREEVIEW*)nmh;
							switch(nmtv.action)
							{
								case TVE_COLLAPSE:
									cea = new TreeViewCancelEventArgs(cast(TreeNode)cast(void*)nmtv.itemNew.lParam,
										false, TreeViewAction.COLLAPSE);
									onBeforeCollapse(cea);
									result = cea.cancel;
									break;
								
								case TVE_EXPAND:
									cea = new TreeViewCancelEventArgs(cast(TreeNode)cast(void*)nmtv.itemNew.lParam,
										false, TreeViewAction.EXPAND);
									onBeforeExpand(cea);
									result = cea.cancel;
									break;
								
								default: ;
							}
							break;
						
						case TVN_ITEMEXPANDEDA:
							nmtv = cast(NM_TREEVIEW*)nmh;
							switch(nmtv.action)
							{
								case TVE_COLLAPSE:
									onAfterCollapse(new TreeViewEventArgs(cast(TreeNode)cast(void*)nmtv.itemNew.lParam,
										TreeViewAction.COLLAPSE));
									break;
								
								case TVE_EXPAND:
									onAfterExpand(new TreeViewEventArgs(cast(TreeNode)cast(void*)nmtv.itemNew.lParam,
										TreeViewAction.EXPAND));
									break;
								
								default: ;
							}
							break;
						
						case TVN_BEGINLABELEDITA:
							{
								TV_DISPINFOA* nmdi;
								nmdi = cast(TV_DISPINFOA*)nmh;
								NodeLabelEditEventArgs nleea;
								TreeNode node;
								node = cast(TreeNode)cast(void*)nmdi.item.lParam;
								// The nleea.label should contain the NEW text, which
								// there is no new text before it's edited... ?
								// TODO: check if correct implementation.
								nleea = new NodeLabelEditEventArgs(node /+ , node.text +/);
								onBeforeLabelEdit(nleea);
								result = nleea.cancelEdit;
							}
							break;
						
						case TVN_ENDLABELEDITA:
							{
								char[] label;
								TV_DISPINFOA* nmdi;
								nmdi = cast(TV_DISPINFOA*)nmh;
								NodeLabelEditEventArgs nleea;
								TreeNode node;
								// TODO: check if correct implementation.
								node = cast(TreeNode)cast(void*)nmdi.item.lParam;
								if(nmdi.item.pszText)
									label = std.string.toString(nmdi.item.pszText).dup;
								nleea = new NodeLabelEditEventArgs(node, label);
								if(!nmdi.item.pszText)
									nleea.cancelEdit = true;
								onAfterLabelEdit(nleea);
								if(nleea.cancelEdit)
								{
									result = FALSE;
								}
								else
								{
									// TODO: check if correct implementation.
									// Update the node's cached text..
									node.ttext = label;
									
									result = TRUE;
								}
							}
							break;
						
						default: ;
					}
				}
				break;
			
			default: ;
		}
		
		return result;
	}
	
	
	private:
	TreeNodeCollection tchildren;
	int ind = 19; // Indent.
	char pathsep = '\\';
	bool _sort = false;
	int iheight = 16;
	
	
	TreeNode treeNodeFromHandle(HTREEITEM hnode)
	{
		TV_ITEMA ti;
		ti.mask = TVIF_HANDLE | TVIF_PARAM;
		ti.hItem = hnode;
		if(SendMessageA(hwnd, TVM_GETITEMA, 0, cast(LPARAM)&ti))
		{
			return cast(TreeNode)cast(void*)ti.lParam;
		}
		return null;
	}
}

